/*
 * TapestryHowto.java
 *
 * Copyright (c) 2001 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.apps.test;
import java.io.FileInputStream;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkException;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.api.StagesInitializedSignal;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import ostore.security.QSPublicKey;
import ostore.tapestry.api.TapestryRouteMsg;
import ostore.util.ByteUtils;
import ostore.util.Debug;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;
import ostore.util.NodeId;
import ostore.util.QSIO;
import ostore.util.QuickSerializable;
import ostore.util.SHA1Hash;
import ostore.util.SecureHash;
import ostore.util.OutputBuffer;
import ostore.util.TypeTable;

/**
 * A simple example of how to write your own Tapestry client and message
 * types.  The example is as follows.  We bring up two machines, a sender
 * and a receiver.  The sender sends a Ping message to the receiver, which
 * the receiver acknoledges with a Pong.  This happens a number of times,
 * and then the test exits.  It's simple, but it covers most of the things
 * you need to know to do more complicated things.  Please just read the
 * source, it's well commented.
 *
 * <p>First off, this class is going to be a SEDA/Sandstorm stage, so it
 * has to inherit from {@link seda.sandStorm.api.EventHandlerIF}.
 *
 * @author Sean C. Rhea
 * @version $Id: TapestryHowto.java,v 1.12 2003/04/02 00:45:05 perlegos Exp $
 */
public class TapestryHowto implements EventHandlerIF {

    /** 
     * We'll use this to tag debugging messages.
     */
    private String class_tag = "TapestryHowto";

    /**
     * Is debugging on or off?
     */
    private static final boolean DEBUG = true;

    /**
     * Initializes the stage once Sandstorm is up and running.
     * 
     * <p>You'll notice that most stages won't have constructors.  The
     * reason is that all initialization can take place here, where you
     * have the information from your stage's config file to help you.
     * I'll detail the important things you want to do in this function
     * below.
     */
    public void init (ConfigDataIF config) throws Exception {

	// First, we find our NodeId (the IP address of this machine, and
	// the port we're listening on) and look up the appropriate
	// instance of the Classifier with it.  The Classifier is the piece
	// of code that handles the publish/subscribe mechanism used for
	// event dispatch between stages.

	NodeId self_node_id = new NodeId (config.getString ("node_id"));
	_classifier = Classifier.getClassifier (self_node_id);

	// Next, I look up the name of this stage, and change my class_tag
	// variable to it.  This is because if we're running more than one
	// instance of this stage in a given Sandstorm instance, they'll
	// have to have different names, and knowing which is which in
	// debug messages will be useful later on.

	class_tag = config.getStage ().getName ();

	// In addition to the class_tag variable, I like to have a final
	// variable called tag in each function, which is just the class
	// tag with the function name appended.  I don't create it if
	// debugging isn't on, though, since string concatination is
	// expensive in Java, and I don't want it happening every time a
	// function is entered.

	final String tag = (DEBUG ? (class_tag + ".init") : null);

	// This turnOn nonsense is part of an old interface.  Read the
	// class comments in ostore.util.Debug if you're interested, but
	// otherwise just do like this.

	if (DEBUG) Debug.turnOn (class_tag);

	// Look up some variables from the config file.  "sender" will
	// specify my role in this test: sender or receiver, and
	// "msg_count" will specify how many ping/pong pairs will be sent.

	_sender = config.getBoolean ("sender");
	_msg_count = config.getInt ("msg_count");
	_nodeName = config.getInt ("nodeName");
	//System.out.println ("Test1");

	// The GUID of an OceanStore node is the SHA-1 hash of it's public
	// key, so to look up my peer's name, I just load in it's public
	// key and hash it.

	String peer_pkey = config.getString ("peer_pkey");
	try {
	    FileInputStream keyfile = new FileInputStream (peer_pkey);
	    QSPublicKey pkey = (QSPublicKey)  QSIO.read (keyfile);
	    _peer = new SHA1Hash (pkey);
	    keyfile.close ();
	}
	catch (Exception e) {
	    BUG (tag + ": caught exception " + e +
		    " while trying to read " + peer_pkey + " from disk.");
	}

	// Now it's time to register for the events I want to receive.  I
	// put plain old events in this array, and the for loop below
	// registers for me to receive them.

        String [] event_types = {
            "seda.sandStorm.api.StagesInitializedSignal"
        };

	SinkIF sink = config.getStage ().getSink ();
        for (int i = 0; i < event_types.length; ++i) {
            Filter filter = new Filter ();
            if (! filter.requireType (Class.forName (event_types [i])))
                BUG (tag + ": could not require type " + event_types [i]);
            Classifier.getClassifier(self_node_id).subscribe (filter, sink);
        }

	// For messages I expect to receive through Tapestry, however, I
	// need to be a little more careful.  (See below.)

	String [] inb_msg_types = {
	    "ostore.apps.test.TapestryHowto$Ping",
	    "ostore.apps.test.TapestryHowto$Pong"
	};

	for (int i = 0; i < inb_msg_types.length; ++i) {
	    
	    // First, I need to make sure the types of messages I intend to
	    // receive are registered with the TypeTable, otherwise
	    // Tapestry won't know how to reconstruct them off the wire.

	    TypeTable.register_type (inb_msg_types [i]);

	    Filter filter = new Filter ();
	    if (! filter.requireType (Class.forName (inb_msg_types [i])))
		BUG (tag + ": could not require type " + inb_msg_types [i]);

	    // Second, in addition to requiring that the events be of a
	    // given type (above), I require that their inbound field is
	    // set to true.  Otherwise, I'll see the messages that I send
	    // as well (in which inbound would be set to false).

	    if (! filter.requireValue ("inbound", new Boolean (true)))
		BUG (tag + ": could not require inbound = true for " 
			+ inb_msg_types [i]);
	    _classifier.subscribe (filter, sink);
	}
    }

    /**
     * This is the function that will be called when a stage dispatches a
     * message that we've subscribed to.  
     */
    public void handleEvent(QueueElementIF item) 
    throws EventHandlerException {
	final String tag = (DEBUG ? (class_tag + ".handleEvent") : null);

	// Aside from being an old interface, using the Debug.printtagln
	// function is a good idea.  It will print something like 
	// "tag: msg", and it could be redirected to a file (instead of
	// STDOUT) or the like.  

	if (DEBUG) Debug.printtagln (tag, "got event " + item);

	// If I get a StagesInitializedSignal, I start up the test; if I'm
	// the sender, send_ping will dispatch a Ping to the receiver.
	if (item instanceof StagesInitializedSignal) {
	    send_ping ();
	}

	// If I get a Ping, I send back a Pong with the same sequence
	// number.
	else if (item instanceof Ping) {
	    if (_sender)
		BUG (tag + ": sender got a ping.");
	    send_pong (((Ping) item).seq);
	}

	// If I get a Pong, I chech the sequence number and send another
	// Ping, unless I've already sent all the Pings I intend to send,
	// in which case I exit the test successfully.
	else if (item instanceof Pong) {
	    if (! _sender)
		BUG (tag + ": receiver got a pong.");
	    send_ping ();
	    if (((Pong) item).seq == 1) {
		System.out.println ("Got last msg.  Test Successful.");
		System.exit (0);
	    }
	}

	// I should only get the event types that I've subscribed for in my
	// init function, above.
	else {
	    BUG (tag + ": got unknown msg type " + 
		    item.getClass ().getName ());
	}
    }

    /**
     * If Sandstorm has more than one event waiting for this stage, it will
     * call this function instead of handleEvent.  If you wanted to get
     * fancy, you could do something like sorting these events based on
     * priority before handling them.
     */
    public void handleEvents(QueueElementIF items[]) 
    throws EventHandlerException {
	for (int i=0; i < items.length; i++)
	    handleEvent (items[i]);
    }

    /**
     * Sandstorm calls this function when it unloads your stage, which to
     * my knowledge, it never does.  I think it's safe to ignore it
     * completely.
     */
    public void destroy() throws Exception {
	// do nothing
    }

    /**
     * A helper function; it's possible that dispatching an event could
     * fail--for instance, if the receiver's queue was full--I don't expect
     * that to happen in this simple test.
     */
    private final void dispatch (QueueElementIF item) {
	final String tag = (DEBUG ? (class_tag + ".dispatch") : null);
	if (DEBUG) Debug.printtagln (tag, "dispatching " +
		item.getClass ().getName () + " event: " + item + ".");
	try {
	    _classifier.dispatch (item);
	} catch (SinkException e) {
	    BUG (class_tag + ".dispatch: could not dispatch " + item);
	}
    }

    /**
     * Prints out the error condition and exits.
     */
    private static final void BUG (String msg) {
	System.err.println (msg + "  This is a bug in " +
		"ostore.apps.test.TapestryHowto.  " +
		"Please let Sean know about it ASAP.");
	System.exit (1);
    }

    /**
     * One of two Tapestry messages used in this test.  I inherit from
     * TapestryRouteMsg because I'm talking to a specific other node in
     * OceanStore.  If I wanted to talk to any node holding an object with
     * GUID g, I would use a TapestryLocateMsg instead.  Other interesting
     * Tapestry message types are:
     *
     * <ul>
     * <li>TapestryMacRouteMsg -- cryptographically guarentees that the
     *     message is from the node it says it's from
     * <li>TapestryPrefixRouteMsg -- routes to the node with the GUID which
     *     most closely matches a given GUID; used by the OceanStore
     *     archive to find storage candidates for erasure coded fragments
     * <li>TapestryPublishMsg -- advertise that your own node is storing a
     *     given object
     * <li>TapestryUnpublishMsg -- stop advertising an object
     * </ul>
     *
     * <p>Refer to the Tapestry javadocs for more details about any of the
     * above.
     *
     * <p>Since this class inherits from TapestryRouteMsg, it indirectly
     * inherits from QuickSerializable and QuickSerializable, meaning that we have
     * to define three main functions:
     * 
     * <ul>
     * <li>A serialization function, to_bytes, which takes a byte array and
     *     an offset and returns the resulting offset (number of bytes
     *     written plus the original offset)
     * <li>A deserialization constructor, which takes a byte array and an
     *     offset and returns the resulting offset (number of bytes read
     *     plus the original offset)
     * <li>A digest function, add_to_digest, which works like the
     *     serialization function, but rather than writing everything into
     *     a new array, just passes it on to a digest object.  This is used
     *     to compute the hash of messages for use inside Tapestry.  
     * <li>Finally, there is a legacy function, type_code, which must be
     *     defined but need not be implemented.
     * </ul>
     *
     * <p>See below for details.
     */
    public static class Ping 
    extends TapestryRouteMsg implements QuickSerializable { 
    
	public int seq;

	/**
	 * A constructor for messages I'm going to send to someone else.
	 * Sets the destination node and sequence number.
	 */
	public Ping (SecureHash dest, int seq) {
	    super (dest);
	    this.seq = seq;
	}

	/**
	 * A deserialization constructor.  Used by Tapestry on the
	 * receiving node to reproduce the message from its "on-the-wire"
	 * representation.  
	 */
	public Ping (InputBuffer buffer) { 
	    seq = buffer.nextInt ();
	}

	/**
	 * A serialization function.  Produces the "on-the-wire"
	 * representation of this object.  Must be the inverse of the
	 * deserialization constructor, above, or bad things will happen.
	 */
	public void serialize (OutputBuffer buffer) {
	    buffer.add (seq);
	}



	/**
	 * It's always nice for debugging to define one of these.
	 */
	public String toString () {
	    return "(Ping " + peer + " " + inbound + " " + seq + ")";
	}
    }

    /**
     * This class is identical in all by type from the Ping class above, so
     * I'm not going to comment it further.
     */
    public static class Pong 
    extends TapestryRouteMsg implements QuickSerializable { 

	public int seq;

	public Pong (SecureHash dest, int seq) {
	    super (dest);
	    this.seq = seq;
	}

	public Pong (InputBuffer buffer) { 
	    seq = buffer.nextInt ();
	}

	public void serialize (OutputBuffer buffer) {
	    buffer.add (seq);
	}



	public String toString () {
	    return "(Pong " + peer + " " + inbound + " " + seq + ")";
	}
    }

    /**
     * If we're the sender and we haven't sent all of our Ping messages
     * yet, send a new one with the next sequence number.
     */
    private void send_ping () {
	if (_sender && (_msg_count > 0)) {
	    Ping outb = new Ping (_peer, _msg_count);
        System.out.println ("*Node " + _nodeName + " sending " + outb + "*");
	    dispatch (outb);
	    --_msg_count;
	}
    }

    /**
     * Send a Pong reply to the Ping with the given sequence number.
     */
    private void send_pong (int seq) {
	final String tag = (DEBUG ? (class_tag + ".send_pong") : null);

	if (seq != _msg_count)
	    BUG (tag + ": seq != _msg_count.");
	--_msg_count;
	Pong outb = new Pong (_peer, seq);
	dispatch (outb);
    }

    private Classifier _classifier;
    private SecureHash _peer;
    private boolean _sender;
    private int _msg_count;
    private int _nodeName;
} 














