/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.bloom;

import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.SingleThreadedEventHandlerIF;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.api.SinkException;
import seda.sandStorm.api.StagesInitializedSignal;
import ostore.util.ByteUtils;
import ostore.util.Carp;
import ostore.util.Clock;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.util.NodeId;
import ostore.util.PriorityQueue;
import ostore.util.QuickSerializable;
import ostore.util.SecureHash;
import ostore.util.SHA1Hash;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import ostore.tapestry.api.TapestryPublishMsg;
import ostore.tapestry.api.TapestryUnpublishMsg;
import ostore.tapestry.impl.StaticAllNodesMsg;
import ostore.tapestry.impl.TapestryReadyMsg;
import ostore.tapestry.impl.StaticPingMsg;
import ostore.tapestry.impl.StaticPongMsg;
import ostore.tapestry.impl.StaticPangMsg;
import ostore.tapestry.impl.StaticBeginMsg;
import ostore.tapestry.impl.BloomWrapper;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import java.util.ListIterator;

/**
 * The probabilistic location stage.
 *
 * @author     Sean C. Rhea
 * @version    $Id: BloomStage.java,v 1.21 2003/02/19 10:32:30 hweather Exp $
 */
public class BloomStage implements EventHandlerIF, SingleThreadedEventHandlerIF {

    private static final boolean _new_way = false;

    public static BloomStage instance (NodeId node_id) {
	synchronized (_instances) {
	    return (BloomStage) _instances.get (node_id);
	}
    }

    private static Map _instances;
    static {
	_instances = new HashMap ();
    }

    private String class_tag = "BloomStage";
    private static final boolean DEBUG = DebugFlags.BLOOM;

    private void dispatch (QueueElementIF item) {
	final String tag = class_tag + ".dispatch";
	if (DEBUG) Debug.printtagln (tag, "dispatching " + item + ".");
	try {
	    _classifier.dispatch(item);
	} catch (SinkException e) {
	    BUG (class_tag + ".dispatch: could not dispatch " + item);
	}
    }

    private static final void BUG (String msg) {
	Carp.die (msg + "  This is a bug in ostore.bloom.BloomStage.  " +
		"Please let Sean know about it ASAP.");
    }

    public void init (ConfigDataIF config) throws Exception {
	final String tag = class_tag + ".init";

	ostore.util.Types.registerAllTypes ();

	_simulator_running = config.getBoolean ("simulator_running");
	_fed_node_id = new NodeId (config.getString ("federation"));
	_self_node_id = new NodeId (config.getString ("node_id"));
	synchronized (_instances) {
	    _instances.put (_self_node_id, this);
	}

	if (DEBUG) Debug.turnOn (class_tag);
	class_tag = class_tag + ".[" + _self_node_id+ "]";

	_classifier = Classifier.getClassifier (_self_node_id);
	_sink = config.getStage().getSink();
	_src_filtering = config.getBoolean ("src_filtering");
	_dest_filtering = config.getBoolean ("dest_filtering");
	int filter_size = config.getInt ("filter_size");
	_bloom_limit = config.getInt ("bloom_limit");
	_filter_depth = config.getInt ("filter_depth");
	_bloom_hashes = config.getInt ("bloom_hashes");

	if (_new_way) 
	    _local_filter = new BloomFilter (1, filter_size);
	else
	    _local_filter = new BloomFilter (1, filter_size/_filter_depth);

	_filter_width = _local_filter.width ();

	_local_filter_counts = new int [_filter_depth] [];
	for (int i = 0; i < _filter_depth; ++i)
	    _local_filter_counts [i] = new int [_filter_width];

	_last_sourced_update_seq = -1;
	_last_query_id = -1;

	_source_neighbors = new HashMap ();
	_source_distances = new HashMap ();
	_update_info_map = new HashMap ();
	_ack_sets = new HashMap ();
	_neighbors_docs = new HashMap ();
	_last_update_docs = new HashMap ();
	_recent_updates = new HashSet ();
	_recent_updates_pq = new PriorityQueue (10);
	_shortest_paths = new HashMap ();
	_published_guids = new HashSet ();
	_pending_req_queue = new LinkedList ();
	_guids_waiting_to_be_published = new HashSet ();
	_guids_waiting_to_be_unpublished = new HashSet ();

	// TODO: for now, we fix the overlay graph at creation
        // SEC hacked to get a little better way of putting in neighbors
        HashSet my_neighbors = new HashSet();
          
        if (config.getString("num_adjacencies") != null) {
	    // DEBUG
	    System.err.println("### USING ADJACENCY LISTS");
	    ////////

          int num_adjacencies = config.getInt("num_adjacencies");
          String my_node_id = config.getString("node_id");
          
          for (int i = 0; i < num_adjacencies; i++) {
            String nodes[] = config.getStringList("adjacency-" + i);

            boolean in_list = false;
            
	    NodeId[] node_ids = new NodeId[nodes.length];
	    for (int j = 0; j < nodes.length; j++)
		node_ids[j] = new NodeId(nodes[j]);

            for (int j = 0; (j < node_ids.length) && !in_list; j++)
              in_list |= (node_ids[j].equals(_self_node_id));

            if (in_list)
              for (int j = 0; j < node_ids.length; j++)
                if (!node_ids[j].equals(_self_node_id))
                  my_neighbors.add(node_ids[j]);
          }

	  // DEBUG
	  System.err.println("### Me: " + config.getString("node_id"));
	  for (Iterator i=my_neighbors.iterator(); i.hasNext(); ) {
	      System.err.println("### Neighbor: " + i.next());
	  }
	  ////////
        }
        else { // SEC the old way of doing it

	    // DEBUG
	    System.err.println("### USING NEIGHBOR LIST");
	    ////////

          int num_neighbors = config.getInt ("num_neighbors");
          for (int i = 0; i < num_neighbors; ++i)
            my_neighbors.add(new NodeId (config.getString ("neighbor-" + i)));
        }

        Iterator neighbor_iter = my_neighbors.iterator();

        while (neighbor_iter.hasNext()) {          
          NodeId n = (NodeId) neighbor_iter.next();
          
          _update_info_map.put (n, new NeighborUpdateInfo ());
          if (_new_way) {
            _last_update_docs.put (n, new BloomFilter (1, filter_size));
          }
          else {
            _last_update_docs.put (n,
                                   new BloomFilter(_filter_depth,filter_size));
          }
          
          // Direct neighbors are always the next hop on the shortest
          // path to themselves.
          _shortest_paths.put (n, n);
          
          // TODO: eventually, we don't want to assume that our neighbors
          // have the same depth filters that we do...
          if (_new_way) {
            _neighbors_docs.put (n, new BloomFilter (1, filter_size));
          }
          else {
            _neighbors_docs.put (n, 
                                 new BloomFilter (_filter_depth, filter_size));
          }
          
          Integer one = new Integer (1);
          _source_distances.put (n, one);
            
          System.out.println ("Node " + _self_node_id+ 
                              " accepting updates from node " + 
                              n+ ", 1 hops away through neighbor " + 
                              n+ ".");
        }
        
	String [] event_types = {
	    "seda.sandStorm.api.StagesInitializedSignal",
	    "ostore.tapestry.impl.TapestryReadyMsg",
	    "ostore.tapestry.api.TapestryPublishMsg",
	    "ostore.tapestry.api.TapestryUnpublishMsg",
	    "ostore.bloom.BloomStage$SendUpdateMsgTimeout"
	};

	for (int i = 0; i < event_types.length; ++i) {
	    Filter filter = new Filter ();
	    if (! filter.requireType (Class.forName (event_types [i])))
		BUG (tag + ": could not require type " + event_types [i]);
	    _classifier.subscribe (filter, _sink);
	}

	String [] msg_types = {
	    "ostore.bloom.UpdateMessage",
	    "ostore.bloom.TopologyDiscoverQuery",
	    "ostore.bloom.TopologyDiscoverResponse",
	    "ostore.bloom.QueryMessage",
	    "ostore.tapestry.impl.StaticAllNodesMsg",
	    "ostore.tapestry.impl.StaticPingMsg",
	    "ostore.tapestry.impl.StaticPongMsg",
	    "ostore.tapestry.impl.StaticPangMsg",
	    "ostore.tapestry.impl.StaticBeginMsg"
	    };

	for (int i = 0; i < msg_types.length; ++i) {
	    Filter filter = new Filter ();
	    if (! filter.requireType (Class.forName (msg_types [i])))
		BUG (tag + ": could not require type " + msg_types [i]);
	    if (! filter.requireValue ("inbound", new Boolean (true)))
		BUG (tag + ": could not require inbound = true for " 
			+ msg_types [i]);
	    _classifier.subscribe (filter, _sink);
	}

	String [] outbound_types = {
	    "ostore.bloom.BloomQuery"
	};

	for (int i = 0; i < outbound_types.length; ++i) {
	    Filter filter = new Filter ();
	    if (! filter.requireType (Class.forName (outbound_types [i])))
		BUG (tag + ": could not require type " + outbound_types [i]);
	    if (! filter.requireValue ("inbound", new Boolean (false)))
		BUG (tag + ": could not require inbound = true for " 
			+ outbound_types [i]);
	    _classifier.subscribe (filter, _sink);
	}
    }

    public void destroy() throws Exception {
	// TODO
    }

    public void handleEvent(QueueElementIF item) 
    throws EventHandlerException {
	final String tag = class_tag + ".handleEvent";

	if (DEBUG) Debug.printtagln (tag, "got " + item + ".");

	if (_pending_req_queue != null) {
	    // The neighbor graph is not fully built yet.  Only handle the
	    // following msgs.

	    if (item instanceof StagesInitializedSignal) {
		handle_init ((StagesInitializedSignal) item);
	    }
	    else if (item instanceof TapestryReadyMsg) {
		handle_tapestry_ready ((TapestryReadyMsg) item);
	    }
	    else if (item instanceof StaticAllNodesMsg) {
		handle_static_all_nodes_msg ((StaticAllNodesMsg) item);
	    }
	    else if (item instanceof StaticPingMsg) {
		handle_static_ping_msg ((StaticPingMsg) item);
	    }
	    else if (item instanceof StaticPongMsg) {
		handle_static_pong_msg ((StaticPongMsg) item);
	    }
	    else if (item instanceof StaticPangMsg) {
		handle_static_pang_msg ((StaticPangMsg) item);
	    }
	    else if (item instanceof StaticBeginMsg) {
		handle_static_begin_msg ((StaticBeginMsg) item);
	    }
	    else
		_pending_req_queue.addLast (item);
	}
	else {

	    if (item instanceof UpdateMessage) {
		UpdateMessage um = (UpdateMessage) item;
		NeighborUpdateInfo sender_info = 
		    (NeighborUpdateInfo) _update_info_map.get (um.peer);
		handle_um (um, sender_info);
	    }
	    else if (item instanceof TopologyDiscoverQuery) {
		handle_tdq ((TopologyDiscoverQuery) item);
	    }
	    else if (item instanceof TopologyDiscoverResponse) {
		handle_tdr ((TopologyDiscoverResponse) item);
	    }
	    else if (item instanceof TapestryPublishMsg) {
		handle_publish_msg ((TapestryPublishMsg) item);
	    }
	    else if (item instanceof TapestryUnpublishMsg) {
		handle_unpublish_msg ((TapestryUnpublishMsg) item);
	    }
	    else if (item instanceof BloomQuery) {
		handle_bloom_query ((BloomQuery) item);
	    }
	    else if (item instanceof QueryMessage) {
		handle_query_message ((QueryMessage) item);
	    }
	    else if (item instanceof SendUpdateMsgTimeout) {
		if ((! _guids_waiting_to_be_published.isEmpty ())
			|| (! _guids_waiting_to_be_unpublished.isEmpty ())) {
		    Set pub_guids = _guids_waiting_to_be_published;
		    _guids_waiting_to_be_published = new HashSet ();
		    Set unpub_guids = _guids_waiting_to_be_unpublished;
		    _guids_waiting_to_be_unpublished = new HashSet ();
		    publish_guids (pub_guids, unpub_guids);
		}
	    }
	    else {
		BUG (tag + ": caught unexpected event " + item + ".");
	    }
	}
    }

    public void handleEvents(QueueElementIF items[]) 
    throws EventHandlerException {
	for(int i=0; i<items.length; i++)
	    handleEvent(items[i]);
    }

    private void handle_init (StagesInitializedSignal signal) {
	final String tag = class_tag + ".handle_init";
	/*
	_self_guid = ((TapestryReadyMsg)item).node_guid;
	_guids_waiting_to_be_published.add (_self_guid);
	if (_guids_waiting_to_be_published.size () == 1)
	    _classifier.dispatch_later (new SendUpdateMsgTimeout (), 1);
	*/
    }

    private void handle_tapestry_ready (TapestryReadyMsg msg) {
	final String tag = class_tag + ".handle_tapestry_ready";
	_self_guid = msg.node_guid;
	_guids_waiting_to_be_published.add (_self_guid);
	if (_guids_waiting_to_be_published.size () == 1)
	    _classifier.dispatch_later (new SendUpdateMsgTimeout (), 1);
    }

    private void handle_bloom_query (BloomQuery query) {
	if (_simulator_running) {
	    if (!(query instanceof BloomWrapper)) Carp.die("BUG8");
	    BloomWrapper wrapper = (BloomWrapper) query;
	    if (wrapper.user_data instanceof SimulatedReadResp) {
		SimulatedReadResp resp = (SimulatedReadResp) wrapper.user_data;
		SecureHash guid = resp.guid;
		SimulationLog.ReadTransactionInfo xact = 
		    SimulationLog.instance (_self_node_id).xact (
			    query.dest, guid);
		xact.started_routing_as_bloom = true;
		xact.finished_routing_as_bloom = true;
	    }
	    else if (wrapper.user_data instanceof SimulatedRead) {
		SecureHash guid = wrapper.dest;
		SimulationLog.ReadTransactionInfo xact = 
		    SimulationLog.instance (_self_node_id).xact (
			    _self_node_id, guid);
		xact.started_location_as_bloom = true;
		xact.finished_location_as_bloom = true;
	    }
	}

	QueryMessage qm = new QueryMessage (null /* set later */,
		++_last_query_id, _self_guid, query.dest, query.query, 
		query.query_state, query.id, query);

	handle_query_message (qm);
    }

    private void handle_query_message (QueryMessage qm) {
	final String tag = class_tag + ".handle_query_message";

	if (_published_guids.contains (qm.dest)) {

	    if (qm.peer != null)
		qm.add_to_path (qm.peer);
	    if (qm.path_length () > 0) {
		//System.out.print ("Success: " + qm.user_data);
		BloomWrapper wrapper = (BloomWrapper) qm.user_data;
		if (wrapper.user_data instanceof SimulatedRead) {
		    SimulationLog.ReadTransactionInfo xact = 
			SimulationLog.instance (_self_node_id).xact (
				qm.src, qm.dest);
		    xact.bloom_obj_hops = qm.path_length ();
		    //System.out.println (", hops = " + qm.path_length ());
		    //System.out.println (xact);
		}
		else {
		    //System.out.println (", but not a read.");
		}
	    }
	    BloomQuery inb = (BloomQuery) ByteUtils.qs_clone (qm.user_data);
	    inb.inbound = true;
	    inb.success = true;
	    inb.src     = qm.src;
	    inb.dest    = qm.dest;
	    inb.query   = qm.query;
	    inb.query_state = qm.query_state;
	    inb.id      = qm.user_id;
	    dispatch (inb);
	    return;
	}

	int [] hashes = guid_hash (qm.dest);

	// Clone the thing so we can start changing it.
	try {
	    qm = (QueryMessage) qm.clone ();
	}
	catch (CloneNotSupportedException e) {
	    Carp.die( "BUG9" );
	}

	// Add the previous sender to the list of visited nodes.
	if (qm.peer != null)
	    qm.add_to_path (qm.peer);

	// Find someone to forward it to.
	int closest_hops = _filter_depth - qm.path_length ();
	if (_new_way)
	    closest_hops = 1;

	NodeId closest = null;
	if (DEBUG) Debug.printtagln (tag, "path length is " + 
		qm.path_length () + ", filter depth is " + _filter_depth +
		", searching to row " + (closest_hops - 1) + ".");

	for (int row = 0; row < closest_hops; ++row) {
	    ListIterator i = _neighbor_distances.listIterator ();
	    while (i.hasNext ()) {
		NodeId node_id = ((NeighborDistanceInfo) i.next ()).neighbor;
		if (qm.on_path (node_id)) {
		    if (DEBUG) Debug.printtagln (tag, node_id + 
			    " already on path.");
		    continue;
		}
		
		BloomFilter bf = (BloomFilter) _neighbors_docs.get (node_id);
		int j = 0;
		while (j < hashes.length) {
		    int bit = hashes [j++];
		    if (bf.bit (row, bit) == false) {
			if (DEBUG) Debug.printtagln (tag, node_id + 
				" does not match on bit " + bit + ", row " +
				row + ".");
			break;
		    }
		}
		if (j == hashes.length) {
		    closest_hops = row;
		    closest = node_id;
		    break;
		}
	    }
	}

	if (closest != null) {
	    qm.peer = closest;
	    qm.inbound = false;
	    dispatch (qm);
	}
	else {
	    if (_simulator_running) {
		if (!(qm.user_data instanceof BloomWrapper)) Carp.die("BUG10");
		BloomWrapper wrapper = (BloomWrapper) qm.user_data;
		if (wrapper.user_data instanceof SimulatedReadResp) {
		    SimulatedReadResp resp = 
			(SimulatedReadResp) wrapper.user_data;
		    SecureHash guid = resp.guid;
		    SimulationLog.ReadTransactionInfo xact = 
			SimulationLog.instance (_self_node_id).xact (
				qm.dest, guid);
		    if (xact == null) Carp.die("ack!");
		    if (qm.src.equals (_self_guid))
			xact.started_routing_as_bloom = false;
		    xact.finished_routing_as_bloom = false;
		}
		else if (wrapper.user_data instanceof SimulatedRead) {
		    SecureHash guid = wrapper.dest;
		    SimulationLog.ReadTransactionInfo xact = 
			SimulationLog.instance (_self_node_id).xact (
				qm.src, guid);
		    if (xact == null) Carp.die("thpth!");
		    if (qm.src.equals (_self_guid))
			xact.started_location_as_bloom = false;
		    xact.finished_location_as_bloom = false;
		    if (qm.path_length () > 0) {
			xact.bloom_obj_hops = qm.path_length ();
		    }
		}
	    }

	    BloomQuery inb = (BloomQuery) ByteUtils.qs_clone (qm.user_data);
	    inb.inbound = true;
	    inb.success = false;
	    inb.src     = qm.src;
	    inb.dest    = qm.dest;
	    inb.query   = qm.query;
	    inb.query_state = qm.query_state;
	    inb.id      = qm.user_id;
	    dispatch (inb);
	}
    }

     /** 
      * This function does nothing more than make sure we get all the
      * messages from each of our Bloom neighbors in order.
      * <code>handle_one_um</code>, below, actually implements the Bloom
      * update algorithm.
      */
    private void handle_um (UpdateMessage um, NeighborUpdateInfo sender_info) {
	final String tag = class_tag + ".handle_um";

	if (DEBUG) Debug.printtagln (tag, "got " + um +
		", need seq num " + 
		(sender_info.last_recv_seq + 1) + ".");

        if (sender_info == null)
        {
            Carp.die(tag + ": null sender info from " + um.peer);
        }

	ListIterator i = sender_info.queued_updates.listIterator ();
	while (i.hasNext ()) {
	    UpdateMessage x = (UpdateMessage) i.next ();
	    if (um.local_seq <= x.local_seq) {
		i.previous ();
		break;
	    }
	}

	i.add (um); 

	um = (UpdateMessage) sender_info.queued_updates.listIterator ().next ();
	
	while ((um.local_seq == sender_info.last_recv_seq + 1) &&
		handle_one_um (um, sender_info)) {

	    sender_info.queued_updates.removeFirst ();
	    sender_info.last_recv_seq += 1;

	    if (sender_info.queued_updates.isEmpty ())
		break;

	    um = (UpdateMessage) 
		sender_info.queued_updates.listIterator ().next ();
	}
    }

    private boolean handle_one_um (UpdateMessage um, 
	    NeighborUpdateInfo sender_info) {
	final String tag = class_tag + ".handle_one_um";

	if (DEBUG) Debug.printtagln (tag, "processing " + um.id + ".");

	if (um.local_seq != sender_info.last_recv_seq + 1)
	    Carp.die( "handle_um broke");

	// First off, who is this message from?

	NodeId nsrc = um.id.node_id;

	// Bits being set are being set on account of a new replica on the
	// node which originated this update.  If we have already seen the
	// results of that add in another message, we may thus ignore them.
	// Unfortuately, bits being cleared may or may not be due a change
	// on that replica, so we must pay attention to them just in case.

	Map ack_set = null;
	if (_src_filtering) {

	    // Now, do we know, for each of our neighbors, whether or not
	    // they are accepting updates from this neighbor?

	    ack_set = (Map) _ack_sets.get (nsrc);
	    if (ack_set == null) {

		// This is the first time we've gotten an update from this
		// source.

		// If we don't know that we're either on or not on the
		// shortest path between this source and each of our
		// neighbors, we'll hold off on sending it to any of them.

		ack_set = new HashMap ();
		_ack_sets.put (nsrc, ack_set);

		Integer sdst = (Integer) _source_distances.get (nsrc);
		/*Assert.croak (nsrc.equals (_self_node_id) || (sdst != null),
			tag + ": no source distance for source " +
			nsrc+ " through " + um.peer);*/

		if (sdst == null) {
		    int i = 0;
		    for (; i < um.num_bits (); ++i) {
			if (! um.clear (i))
			    break;
		    }
		    if (i != um.num_bits ()) {
			if( !nsrc.equals (_self_node_id))
			    Carp.die( tag + ": no src dist for " + nsrc + 
				      " but some bits set.");
		    }
		}

		if (nsrc.equals (_self_node_id)) {

		    if (DEBUG) Debug.printtagln (tag, "alpha");

		    // TODO: think about this
		    // Assert.croak (false, "unreachable");

		    // If we're the source (the update has looped), we need
		    // to forward on the clear bits, but never the set
		    // bits.

		    Iterator i = _neighbors_docs.keySet ().iterator ();
		    while (i.hasNext ()) {
			NodeId n = (NodeId) i.next ();
			ack_set.put (n, new Boolean (false));
			if (DEBUG) System.out.println (tag + ": - 1 - neighbor " +
				n + " not accepting updates from "
				+ "source " + nsrc);
		    }
		}
		else { 

		    Iterator i = _neighbors_docs.keySet ().iterator ();
		    while (i.hasNext ()) {
			NodeId n = (NodeId) i.next ();
			if (n.equals (nsrc)) {

			    // If the source is a neighbor of ours, he
			    // certainly doesn't want his own updates sent
			    // back to him, so add him to the list
			    // automatically.

			    ack_set.put (nsrc, new Boolean (false));
			    if (DEBUG) System.out.println (tag + ": neighbor " +
				    n + " not accepting updates from "
				    + "itself.");
			    continue;
			}
			else if (n.equals (um.peer)) {

			    // Same goes for the person who sent us this
			    // update (if they are different from the
			    // source).

			    ack_set.put (um.peer, new Boolean (false));
			    if (DEBUG) System.out.println (tag + ": - 2 - neighbor " +
				    n + " not accepting updates from "
				    + "source " + nsrc);
			    continue;
			}

			dispatch (new TopologyDiscoverQuery (n, nsrc,
				    sdst == null ? -1 : (sdst.intValue () + 1)));
		    }

		    if (ack_set.size () != _neighbors_docs.size ()) {
			if (DEBUG) Debug.printtagln (tag, "A need " + 
				(_neighbors_docs.size () - ack_set.size ()) +
				" more TDRs to process " + um.id + ".");
			return false;
		    }
		}
	    }
	    else if (ack_set.size () != _neighbors_docs.size ()) {

		// We have not yet received all the responses to our
		// TopologyDiscoverQuery for this source.

		if (DEBUG) Debug.printtagln (tag, "B need " + 
			(_neighbors_docs.size () - ack_set.size ()) +
			" more TDRs to process " + um.id + ".");
		return false;
	    }
	}

	{
	    long now = current_time ().getTime ();
	    while ((! _recent_updates_pq.empty ()) && 
	           (now > _recent_updates_pq.topPriority (false) + 10000)) {
		UpdateId id = (UpdateId) _recent_updates_pq.extract (false);
		if (! _recent_updates.remove (id))
		    Carp.die( "but with recent updates.");
	    }
	}

	boolean ignore_set_bits = false;
	if (_recent_updates.contains (um.id)) {
	    if (DEBUG) Debug.printtagln (tag, "got UM " + um + 
		    " again (via " + um.peer + 
		    ").  Ignoring set bits.");
	    ignore_set_bits = true;
	}
	else {
	    // Only record this the first time we see it with set bits.

	    int i = 0;
	    for (; i < um.num_bits (); ++i) {
		if (! um.clear (i))
		    break;
	    }
	    if (i != um.num_bits ()) {
		_recent_updates.add (um.id);
		_recent_updates_pq.insert (
			um.id, current_time ().getTime (), false);

		// Okay, we are going to propogate it.

		if (DEBUG) Debug.printtagln (tag, "got UM "  + um + 
			" from " + um.peer + ".");
	    }
	}

	if (! _dest_filtering) {
	    if (_src_filtering && ignore_set_bits) {
		for (int i = 0; i < um.num_bits (); ++i) {
		    if (! um.clear (i)) {
			BUG (tag + ": not ignoring "  + um + " from " + 
				um.peer + ".");
		    }
		}
	    }

	    ignore_set_bits = false;
	}

	// The rest of this function follows the pseudo-code in the paper.
	// The line numbers in the comments are w.r.t. that.

	{
	    BloomFilter nf = (BloomFilter) _neighbors_docs.get (um.peer);
	    if (nf == null) Carp.die("BUG11");

	    // 3  foreach (r, c, v) \in U do
	    for (int i = 0; i < um.num_bits (); ++i) {

		int um_row = _new_way ? 0 : um.row (i);

		// 4  F_rc <-- v
		if (um.clear (i)) {
		    // If I've ignored a recent update, and he doesn't know
		    // about it, than he may tell me to clear bits which I
		    // don't currently have set.
		    if (! _dest_filtering) 
			if(!nf.bit (um_row, um.col (i))) Carp.die("BUG1");
		    nf.clear_bit (um_row, um.col (i));
		}
		else if (! ignore_set_bits) {
		    if(nf.bit (um_row, um.col (i))) Carp.die("BUG2");
		    nf.set_bit (um_row, um.col (i));
		}
	    }
	    // 5  endfor
	}

	// 7   foreach n' \in N \ n do
	for (Iterator j = _last_update_docs.keySet ().iterator ();
		j.hasNext (); ) {

	    NodeId dest = (NodeId) j.next ();
	    if (dest.equals (um.peer)) continue;

	    NeighborUpdateInfo info = 
		(NeighborUpdateInfo) _update_info_map.get (dest);
	    info.last_sent_seq++;

	    // 8   U' <-- 0

	    UpdateMessage outbound = new UpdateMessage (dest, um.id, 
		    info.last_sent_seq, _filter_width, _filter_depth);

	    // 9   L <-- LastUpdateFilter (n')
	    BloomFilter dest_last_update_docs = 
		(BloomFilter) _last_update_docs.get (dest);

	    // 10  foreach (r, c, v) \in U do
	    for (int i = 0; i < um.num_bits (); ++i) {

		int row = um.row (i); 
		int col = um.col (i);
		int val = um.clear (i) ? 0 : 1;

		// 11  r' <-- r + 1
		int newrow = row + 1;

		if (newrow >= _filter_depth)
		    continue;

		// 14  if v = 0 and L_{r'c} = 1
		if ((val == 0) && dest_last_update_docs.bit (
			    _new_way ? 0 : newrow, col)) {

		    // 15  c <-- true
		    boolean clear = true;

		    // 16  foreach n'' \in N \ n' do
		    for (Iterator k = _neighbors_docs.keySet ().iterator ();
			    k.hasNext (); ) {

			NodeId kdest = (NodeId) k.next ();
			if (kdest.equals (dest)) continue; // TODO: was break

			// 17  F' <-- NeighborFilter (n'')
			BloomFilter fp = 
			    (BloomFilter) _neighbors_docs.get (kdest);

			// 18  if F'_{r'c} = 1 then c <-- false endif
			if (fp.bit (_new_way ? 0 : newrow, col)) {
			    clear = false;
			    break;
			}
		    }
		    // 19  endfor

		    // 20  if c then
		    if (clear == true) {

			// 21  L_{r'c} <-- 0
			dest_last_update_docs.clear_bit (
				_new_way ? 0 : newrow, col);

			// 22  U' <-- U' union {(r', c, 0)}
			outbound.add_bit (newrow, col, true);
		    }
		}
		// 26  elsif v = 1 and L_{r'c} = 0
		else if ((val == 1) && 
			(! dest_last_update_docs.bit (
			_new_way ? 0 : newrow, col))) {

		    if ((! ignore_set_bits) &&
			    ((! _dest_filtering) || (! dest.equals (nsrc)))) {

			boolean propogate_set_bits = true;

			if (_src_filtering) {
			    Boolean b = (Boolean) ack_set.get (dest);
			    if (b == null) Carp.die("BUG3");
			    propogate_set_bits = b.booleanValue ();
			}

			// 27  if not IgnoringSource (n', s) then
			if (propogate_set_bits) {

			    if(outbound.has_bit (newrow, col, true))
				Carp.die("BUG4");

			    // 28  L_{r'c} <-- 1
			    dest_last_update_docs.set_bit (
			    _new_way ? 0 : newrow, col);

			    // 29  U' <-- U' union {(r', c, 1)}
			    outbound.add_bit (newrow, col, false /* set */);
			}
			else if (DEBUG) {
			    Debug.printtagln (tag, "no set bits " +
				    "propagated to neighbor " + 
				    dest + " b/c of src filtering.");
			}
			// 30  endif
		    }
		}
		// 31  endif
	    }
	    // 32  endfor

	    // Okay, this update message is ready.  Dispatch it.

	    // 34  if U' /= 0
	    if (outbound.num_bits () > 0) {
		dispatch (outbound);
	    }
	    else {

		if (DEBUG) Debug.printtagln (tag, ": no bits changed " +
			"for neighbor " + dest + ".");

		// drop outbound
		info.last_sent_seq--;
	    }

	    // Go on to the next neighbor.
	}

	return true;  // success
    }

    private void handle_tdq (TopologyDiscoverQuery query) {
	final String tag = class_tag + ".handle_tdq";
	if(!_src_filtering) Carp.die("BUG14");
	if(query.update_source.equals (_self_node_id))
	    Carp.die("BUG15");

	boolean on_shortest_path = false;

	NodeId shortest_path_succ = (NodeId) 
	    _shortest_paths.get (query.update_source);

	if (shortest_path_succ == null) {
	    if ((query.hops != -1 ) && (query.hops <= _filter_depth)) {
		// We aren't already accepting updates from this source
		// through another neighbor.  But are we accepting too many
		// sources at this depth from this neighbor?

		Set [] depths = (Set []) _source_neighbors.get (query.peer);
		if (depths == null) {
		    depths = new Set [_filter_depth+1];
		    _source_neighbors.put (query.peer, depths);
		}
		if (depths [query.hops] == null)
		    depths [query.hops] = new HashSet (_bloom_limit);

		if (depths [query.hops].size () < _bloom_limit) {

		    depths [query.hops].add (query.update_source);
		    _shortest_paths.put (query.update_source, query.peer);
		    on_shortest_path = true;
		    System.out.println ("Node " + _self_node_id + 
			    " accepting updates from node " + 
			    query.update_source + ", " + query.hops + 
			    " hops away through neighbor " + 
			    query.peer + ".");

		    if(_source_distances.containsKey (query.update_source))
			Carp.die("already have src dist!");

		    _source_distances.put (query.update_source, 
			    new Integer (query.hops));
		}
	    }
	}
	else {
	    if (shortest_path_succ.equals (query.peer)) {
		Carp.die (tag + ": got second TDQ from " + query.peer
			+ " for source " + query.update_source + ".");
	    }
	}
	
	dispatch (new TopologyDiscoverResponse (query.peer,
		    query.update_source, on_shortest_path));
    }

    private void handle_tdr (TopologyDiscoverResponse resp) {
	final String tag = class_tag + ".handle_tdr";
	if(!_src_filtering) Carp.die("BUG16");

	NodeId responder = resp.peer;

	// source -> NULL
	// source -> neighbor -> true/false

	Map ack_set = (Map) _ack_sets.get (resp.update_source);
	if(ack_set == null) Carp.die("BUG17");

	if(ack_set.containsKey (responder)) Carp.die("BUG18");

	ack_set.put (responder, new Boolean (resp.on_shortest_path));
	if (DEBUG) Debug.printtagln (tag, "neighbor " +
		responder + (resp.on_shortest_path ? " is" : " not") +
		" accepting updates from " + "source " + 
		resp.update_source);

	if (DEBUG) Debug.printtagln (tag, "Have " + ack_set.size () + 
		" responses, for source " + resp.update_source + 
		"; want " + _neighbors_docs.size () + ".");

	if (ack_set.size () == _neighbors_docs.size ()) {

	    // TODO: do this better

	    Iterator k = _update_info_map.keySet ().iterator ();
	    while (k.hasNext ()) {
		NodeId sender = (NodeId) k.next ();
		NeighborUpdateInfo sender_info = 
		    (NeighborUpdateInfo) _update_info_map.get (sender);

		if (sender_info.queued_updates.size () > 0) {
		    UpdateMessage um = (UpdateMessage)
			sender_info.queued_updates.removeFirst ();
		    handle_um (um, sender_info);
		}
	    }
	}
    }

    public static class SendUpdateMsgTimeout implements QueueElementIF {
	public String toString () { return "SendUpdateMsgTimeout"; }
    }

    private void handle_publish_msg (TapestryPublishMsg msg) {
	final String tag = class_tag + ".handle_publish_msg";
	if (DEBUG) Debug.printtagln(tag, ": Got publish for " + msg.guid);
	if ((msg.tag != null) && 
	    /* TODO: hack for Patrick */
	    ! (msg.tag instanceof ostore.replica.ReplicaTag))
	    return;
	if (! _guids_waiting_to_be_unpublished.remove (msg.guid)) {
	    _guids_waiting_to_be_published.add (msg.guid);
	    if (_guids_waiting_to_be_published.size () == 1)
		_classifier.dispatch_later (new SendUpdateMsgTimeout (), 1);
	}
    }

    private void handle_unpublish_msg (TapestryUnpublishMsg msg) {
	final String tag = class_tag + ".handle_publish_msg";
	if (DEBUG) Debug.printtagln(tag, ": Got unpublish for " + msg.guid);
	if ((msg.tag != null) && 
	    /* TODO: hack for Patrick */
	    ! (msg.tag instanceof ostore.replica.ReplicaTag))
	    return;
	if (! _guids_waiting_to_be_published.remove (msg.guid)) {
	    _guids_waiting_to_be_unpublished.add (msg.guid);
	    if (_guids_waiting_to_be_unpublished.size () == 1)
		_classifier.dispatch_later (new SendUpdateMsgTimeout (), 1);
	}
    }

    private void handle_static_all_nodes_msg (StaticAllNodesMsg msg) {
	final String tag = class_tag + ".handle_static_all_nodes_msg";

	if (DEBUG) Debug.printtagln (tag, "processing StaticAllNodesMsg" );
	if (_pending_req_queue == null) {
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS) Carp.out (tag + ": " + 
		    "got duplicate StaticAllNodesMsg: " +
		    msg + ".  Dropping it.");
	    return;
	}

	if (! msg.peer.equals (_fed_node_id)) {
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS) Carp.out (tag + ": " + 
		    "got StaticAllNodesMsg: " + msg + 
		    " from someone other than the federation.  " +
		    "Dropping it.");
	    return;
	}

	_out_times = new HashMap ();
	_timing_map = new PriorityQueue (msg.nodes ().size ());

	for (Iterator i = msg.nodes ().keySet ().iterator (); 
		i.hasNext (); ) {
	    NodeId n = (NodeId) i.next ();
	    if (n.less_than (_self_node_id))
		_out_times.put (n, current_time ());
	}
    }

    protected void handle_static_ping_msg (StaticPingMsg msg) {
	final String tag = class_tag + ".handle_static_ping_msg";
	if (DEBUG) Debug.printtagln (tag, "processing StaticPingMsg" );
	long now = Clock.current_date(_self_node_id).getTime();
	_out_times.put (msg.peer, current_time ());
    }

    private void handle_static_pong_msg (StaticPongMsg msg) 
    throws EventHandlerException {
	final String tag = class_tag + ".handle_static_pong_msg";
	if (DEBUG) Debug.printtagln (tag, "processing StaticPongMsg" );

	if ((_out_times == null) || (_out_times.isEmpty ())) {
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS) Carp.out (tag + ": " + 
		    "got duplicate StaticPongMsg: " +
		    msg + ".  Dropping it.");
	    return;
	}

	long recv_time = current_time ().getTime ();

	if (! _out_times.containsKey (msg.peer)) {
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS) Carp.out (tag + ": " + 
		    "got StaticPongMsg: " + msg + 
		    " from someone unexpected.  " +
		    "Dropping it.");
	    return;
	}

	round_trip_done (recv_time, msg.peer);
    }

    private void handle_static_pang_msg (StaticPangMsg msg) 
    throws EventHandlerException {
	final String tag = class_tag + ".handle_static_pang_msg";
	if (DEBUG) Debug.printtagln (tag, "processing StaticPangMsg" );

	if ((_out_times == null) || (_out_times.isEmpty ())) {
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS) Carp.out (tag + ": " + 
		    "got duplicate StaticPangMsg: " +
		    msg + ".  Dropping it.");
	    return;
	}

	long recv_time = current_time ().getTime ();

	if (! _out_times.containsKey (msg.peer)) {
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS) Carp.out (tag + ": " + 
		    "got StaticPangMsg: " + msg + 
		    " from someone unexpected.  " +
		    "Dropping it.");
	    return;
	}

	round_trip_done (recv_time, msg.peer);
    }

    protected void round_trip_done (long recv_time, NodeId peer) {
	final String tag = class_tag + ".round_trip_done ";

	long send_time = 
	    ((Date) _out_times.remove (peer)).getTime ();
	long delta = recv_time - send_time;

	_timing_map.insert (peer, delta, false);

	if (DEBUG) Debug.printtagln (tag, 
		"_out_times.size () = " + _out_times.size ());
    }

    protected void handle_static_begin_msg (StaticBeginMsg msg) 
    throws EventHandlerException {
	final String tag = class_tag + ".handle_static_begin_msg";
	if ( ! _out_times.isEmpty ()) Carp.die("Ack!");

	// The timing map is complete.

	_out_times = null;  // free up unused memory

	_neighbor_distances = new LinkedList ();

	while (! _timing_map.empty ()) {

	    long alpha_millis = _timing_map.topPriority (false);
	    double alpha_secs = (double) alpha_millis * 1000.0;
	    NodeId n = (NodeId) _timing_map.extract (false);

	    if (_neighbors_docs.containsKey (n)) {

		if (DEBUG) Debug.printtagln (tag, "Bloom neighbor " + 
			n + " is at distance " + 
			alpha_millis + " ms.");

		// Already sorted by virtue of _timing_map.
		_neighbor_distances.addLast (
			new NeighborDistanceInfo (n, alpha_secs));

		if (_neighbor_distances.size () == _neighbors_docs.size ())
		    break;
	    }
	}

	_timing_map = null;  // free up unused memory

	// Handle all the pending requests.

	LinkedList p = _pending_req_queue;
	_pending_req_queue = null;  
	while (! p.isEmpty ())
	    handleEvent ((QueueElementIF) p.removeFirst ());
    }

    private void publish_guids (Set pub_guids, Set unpub_guids) {
	Set cols = new HashSet ();

	for (Iterator i = pub_guids.iterator (); i.hasNext (); ) {
	    SecureHash guid = (SecureHash) i.next ();
	    if (! _published_guids.add (guid)) 
		Carp.die("BUG12");
	    update_local_bf (guid, false, cols);
	}

	for (Iterator i = unpub_guids.iterator (); i.hasNext (); ) {
	    SecureHash guid = (SecureHash) i.next ();
	    if (! _published_guids.remove (guid)) 
		Carp.die("BUG13: could not remove " + guid);
	    update_local_bf (guid, true, cols);
	}

	// TODO: combine with removed guids...

	// It's possible that whatever cols were cleared in removing the
	// Guids in removed were set again by adding Guid g.  If so,
	// there's no need to send out any updates.
	if (cols.isEmpty ())
	    return;

	UpdateId update_id = 
	    new UpdateId (_self_node_id, ++_last_sourced_update_seq);
	Iterator i = _last_update_docs.keySet ().iterator ();
	while (i.hasNext ()) {

	    NodeId neighbor = (NodeId) i.next ();
	    BloomFilter bf = (BloomFilter) _last_update_docs.get (neighbor);
	    NeighborUpdateInfo info = 
		(NeighborUpdateInfo ) _update_info_map.get (neighbor);
	    UpdateMessage um = new UpdateMessage (neighbor, update_id,
		    ++info.last_sent_seq, _filter_width, _filter_depth);
	    
	    Iterator j = cols.iterator ();
	    while (j.hasNext ()) {
		int col = ((Integer) j.next ()).intValue ();
		if(col >= _filter_width) Carp.die("BUG19");
		if (_new_way) {
		    boolean val = _local_filter.bit (0, col);
		    if (val) {
			// The bit has been set; ie. it was previously
			// clear.  Make sure it is now set for this
			// neighbor.

			if (! bf.bit (0, col)) {
			    bf.set_bit (0, col);
			    um.add_bit (0, col, false /* set */);
			}
		    }
		    else {
			// The bit was cleared; ie. it was previously set.
			// See if it is set anywhere else.
			boolean clear = true;
			for (Iterator k = _neighbors_docs.keySet ().iterator ();
				k.hasNext (); ) {

			    NodeId kdest = (NodeId) k.next ();
			    if (kdest.equals (neighbor)) continue; 

			    BloomFilter fp = 
				(BloomFilter) _neighbors_docs.get (kdest);

			    if (fp.bit (0, col)) {
				clear = false;
				break;
			    }
			}

			if (clear == true) {
			    bf.clear_bit (0, col);
			    um.add_bit (0, col, true /* clear */);
			}
		    }
		}
		else {
		    if(bf.bit (0, col) == _local_filter.bit (0, col))
			Carp.die("BUG20");
		// if the local filter is not set, we want to clear this col
		um.add_bit (0, col, ! _local_filter.bit (0, col));
		if (_local_filter.bit (0, col)) {
		    bf.set_bit (0, col);
		}
		else {
		    bf.clear_bit (0, col);
		}
		if(bf.bit (0, col) != _local_filter.bit (0, col))
		    Carp.die("BUG21");
		}
	    }

	    dispatch (um);
	}
    }

    private void update_local_bf (SecureHash g, boolean clear, Set bits) {
	int [] hashes = guid_hash (g);

	for (int j = 0; j < hashes.length; ++j) {
	    int bit = hashes [j];

	    boolean changed = false;
	    if (clear) {
		if ((--_local_filter_counts [0][bit]) == 0) {
		    _local_filter.clear_bit (0, bit);
		    changed = true;
		}
	    }
	    else {
		if ((_local_filter_counts [0][bit]++) == 0) {
		    _local_filter.set_bit (0, bit);
		    changed = true;
		}
	    }

	    if (changed) {
		if (! bits.remove (new Integer (bit)))
		    bits.add (new Integer (bit));
	    }
	}
    }

    private int [] guid_hash (SecureHash g) {
	byte [] data = ByteUtils.qs_to_bytes (g);
	if(data.length / 4 < _bloom_hashes) Carp.die("BUG5");
	int [] result = new int [_bloom_hashes];
	int [] offset = {0};
	for (int i = 0; i < _bloom_hashes; ++i) {
	    int n = ByteUtils.bytesToInt (data, offset);
	    n &= 0x7fffffff;
	    result [i] = n % _filter_width;
	}
	return result;
    }

    private Date current_time () {
	return Clock.current_date (_self_node_id);
    }

    private static class NeighborUpdateInfo {
	public int last_sent_seq = -1, last_recv_seq = -1;
	public LinkedList /* of UpdateMessage */ queued_updates = 
	    new LinkedList ();
    }

    private static class NeighborDistanceInfo {
	public NeighborDistanceInfo (NodeId n, double a) {
	    neighbor = n;
	    alpha = a;
	}

	public NodeId neighbor;
	public double alpha;
    }

    private SinkIF _sink;
    private NodeId _self_node_id;
    private SecureHash _self_guid;
    private Classifier _classifier;

    private boolean _src_filtering, _dest_filtering;
    private int _filter_width, _filter_depth, _bloom_hashes, _bloom_limit;

    private BloomFilter _local_filter;
    private int [] [] _local_filter_counts;

    private long _last_sourced_update_seq, _last_query_id;

    /**
     * Stores an NeighborUpdateInfo for each neighbor.
     */
    private Map /* NodeId -> NeighborUpdateInfo */ _update_info_map;

    /**
     * Is the first node accepting updates from the second?
     */
    private Map /* NodeId -> NodeId -> Boolean */ _ack_sets;
    
    private Map /* NodeId -> BloomFilter */ _neighbors_docs;
    private Map /* NodeId -> BloomFilter */ _last_update_docs;

    private Set /* UpdateId */ _recent_updates;
    private PriorityQueue /* of UpdateId */ _recent_updates_pq;

    /**
     * Which of our neighbors is on the shortest path to the given update
     * source?
     */
    private Map /* NodeId -> NodeId */ _shortest_paths;

    private Map /* NodeId -> Integer */ _source_distances;
    private Map /* NodeId -> Set of NodeId */ _source_neighbors;

    private Set _published_guids;

    private LinkedList /* of NeighborDistanceInfo */ _neighbor_distances;

    private Set _guids_waiting_to_be_published;
    private Set _guids_waiting_to_be_unpublished;
    private Map _out_times;
    private NodeId _fed_node_id;
    private LinkedList _pending_req_queue;
    private PriorityQueue _timing_map;

    private boolean _simulator_running;
}
