/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.bloom;
import java.util.Iterator;
import java.util.Vector;
import ostore.network.NetworkMessage;
import ostore.util.AssertionViolatedException;
import ostore.util.Carp;
import ostore.util.ByteUtils;
import ostore.util.DebugFlags;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;
import ostore.util.NodeId;
import ostore.util.OutputBuffer;
import ostore.util.QSException;
import ostore.util.QuickSerializable;

/**
 * Update message.
 *
 * @author     Sean C. Rhea
 * @version    $Id: UpdateMessage.java,v 1.15 2002/10/31 07:53:37 jeffpang Exp $
 */
public class UpdateMessage 
extends NetworkMessage implements QuickSerializable {

    private static final boolean DEBUG = DebugFlags.BLOOM;

    public UpdateId id;
    public long local_seq;
    public int filter_width, filter_depth;

    public UpdateMessage (NodeId peer, UpdateId id, long local_seq,
	    int filter_width, int filter_depth) {
	super (peer, false /* outbound */);
	this.id = id;
	this.local_seq = local_seq;
	this.filter_width = filter_width;
	this.filter_depth = filter_depth;
	this._bits = new Vector (10, 10);
    }

    public UpdateMessage (InputBuffer buffer) throws QSException {
	super (buffer);
	id = (UpdateId) buffer.nextObject ();
	local_seq = buffer.nextLong ();
	filter_width = buffer.nextInt ();
	filter_depth = buffer.nextInt ();

	if (filter_width >= 2*1024*1024)
	    throw new AssertionViolatedException ("TODO: filters to wide");

	int size = buffer.nextInt ();
	this._bits = new Vector (size, 10);

	for (int i = 0; i < size; ++i) {
	    int bit = buffer.nextInt ();
	    _bits.add (_bits.size (), new Integer (bit));
	}
    }


    public void serialize (OutputBuffer buffer) {
	super.serialize (buffer);
	buffer.add (id);
	buffer.add (local_seq);
	buffer.add (filter_width);
	buffer.add (filter_depth);

	if (filter_width >= 2*1024*1024)
	    throw new AssertionViolatedException ("TODO: filters to wide");

	buffer.add (_bits.size ());
	Iterator i = _bits.iterator ();
	while (i.hasNext ()) {
	    Integer bit = (Integer) i.next ();
	    buffer.add (bit.intValue ());
	}
    }

    public void add_bit (int row, int col, boolean clear) {
	if (DEBUG && (row < 0)) Carp.die("BUG");
	if (DEBUG && (col < 0)) Carp.die("BUG");
	if (DEBUG && (row >= filter_depth)) Carp.die("BUG");
	if (DEBUG && (col >= filter_width)) Carp.die("BUG");
	if (DEBUG && (has_bit (row, col, true))) Carp.die("BUG");
	if (DEBUG && (has_bit (row, col, false))) Carp.die("BUG");
	int bit = row * filter_width + col;
	if (! clear)
	    bit |= 0x80000000;
	_bits.add (_bits.size (), new Integer (bit));

    }

    public void rem_bit (int row, int col, boolean clear) {
	if (DEBUG && (row >= filter_depth)) Carp.die("BUG");
	if (DEBUG && (col >= filter_width)) Carp.die("BUG");
	int bit = row * filter_width + col;
	if (clear)
	    bit |= 0x80000000;

	int i = 0; 
	while (((Integer) _bits.get (i)).intValue () != bit)
	    if (++i >= _bits.size ())
		if (DEBUG) Carp.die("BUG");
	if (DEBUG &&
	    ((Integer) _bits.get (i)).intValue () != bit) Carp.die("BUG");

	while (i < _bits.size ()) {
	    _bits.set (i, _bits.get (i+1));
	    ++i;
	}

	_bits.remove (_bits.size () - 1);
    }

    public boolean has_bit (int row, int col, boolean clear) {
	if (DEBUG && (row >= filter_depth)) Carp.die("BUG");
	if (DEBUG && (col >= filter_width)) Carp.die("BUG");
	int bit = row * filter_width + col;
	if (clear)
	    bit |= 0x80000000;
	for (int i = 0; i < _bits.size (); ++i) {
	    if (((Integer) _bits.get (i)).intValue () == bit)
		return true;
	}
	return false;
    }

    public int num_bits () {
	return _bits.size ();
    }

    public int row (int b) {
	if (DEBUG && (b >= num_bits ())) Carp.die("BUG");
	return nosign (b) / filter_width;
    }

    public int col (int b) {
	if (DEBUG && (b >= num_bits ())) Carp.die("BUG");
	return nosign (b) % filter_width;
    }

    public boolean clear (int b) {
	if (DEBUG && (b >= num_bits ())) Carp.die("BUG");
	return ((((Integer) _bits.get (b)).intValue ()) & 0x80000000) == 0;
    }

    public Object clone () throws CloneNotSupportedException {
	UpdateMessage result = (UpdateMessage) super.clone ();
	result.id = id;
	result.local_seq = local_seq;
	return result;
    }

    public String toString () {
	String result = "(UpdateMessage " + id + " " + 
	    (inbound ? "from " : "to ") + peer + " " + local_seq;
	for (int i = 0; i < num_bits (); ++i) 
	    result += " " + (clear (i) ? "-" : "+") + "(" + row (i) + "," +
		col (i) + ")";
	result += ")";
	return result;
    }

    private int nosign (int b) {
	return (((Integer) _bits.get (b)).intValue ()) & 0x7fffffff;
    }

    private Vector _bits;
}

