/**
 * Classifier.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.dispatch;

import java.util.Vector;
import java.util.Set;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;
import java.lang.reflect.Field;
import ostore.network.NetworkMessage;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.util.Types;
import ostore.util.Carp;
import ostore.util.NodeId;
import seda.sandStorm.main.Sandstorm;
import seda.sandStorm.main.stageDescr;
import seda.sandStorm.api.ManagerIF;
import seda.sandStorm.api.SignalMgrIF;
import seda.sandStorm.api.StagesInitializedSignal;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.api.SinkException;
import seda.sandStorm.api.SinkFullException;
import seda.sandStorm.api.SinkClosedException;
import seda.sandStorm.api.NoSuchStageException;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.core.ssTimer;
import seda.sandStorm.core.SimpleSink;

/**
 * The event classification engine for an Oceanstore replica.
 * <p>
 * Work modules register a queue satisfying the <code>SinkIF</code>
 * interface, which will receive all <code>QueueElementIF</code> objects that
 * satisfy the specified <code>Filter</code>.  A single <code>SinkIF</code>
 * may subscribe with several <code>Filter</code>s, or with several
 * priorities for a single <code>Filter</code>.  Each <code>SinkIF</code>
 * will receive exactly one copy of each <code>QueueElementIF</code> which
 * satisfies any of its <code>Filter</code>s, at the highest of its priorities.
 * <p>
 * By default, all <code>Filter</code>s are subscribed with
 * <code>PRIORITY_NORMAL</code>.  The <code>Classifier</code> will throw an
 * exception if it cannot properly enqueue an event on the corresponding
 * <code>SinkIF</code>.  Modules may register with lower priorities, such
 * as <code>PRIORITY_LOW</code>.  Any exceptions thrown by queues registered
 * at this lower priority will be ignored.
 * <p>
 * The <code>Classifier</code> stores a trie of filter requirements for
 * each type of event.  A node in the trie represents the set of
 * <code>QueueElementIF</code>s that pass all requirements on the path to
 * it from the trie root.  Each node stores the <code>SinkIF</code>s
 * requesting all <code>QueueElementIF</code>s from that set.  The tries
 * are stored in an unsynchronized <code>HashMap</code>, indexed by its
 * event type.  
 * <p>
 * On dispatch, the <code>Classifier</code> uses the event type to search
 * the <code>HashMap</code> for the proper trie.  It then passes down the
 * trie, checking field requirements and collecting the
 * <code>SinkIF</code>s for nodes which it passes.  Finally, a single copy
 * of the <code>QueueElementIF</code> is placed into each of these
 * <code>SinkIF</code>s.
 * <p>
 * The <code>HashMap</code> of tries is unsynchronized, to maximize
 * parallelism.  When it must be modified, to add a new event type, a new
 * <code>HashMap</code> is constructed and atomically replaces the old
 * one.  All calls to <code>dispatch</code> can proceed without blocking.
 *
 * @author	Dennis Geels
 * @version $Id: Classifier.java,v 1.48 2003/06/18 21:07:21 srhea Exp $
 *
 * @see Filter
 */

public class Classifier extends SimpleSink
{
    ////////////////////       CONSTANTS	////////////////////

    private static final String tag = "Classifier";    

    private static final boolean DEBUG = false;
    
    // A few <code>SinkIF</code> priorities / error masks
    private static final Integer PRIORITY_INFINITE	= new Integer(11);
    public static final Integer PRIORITY_HIGH		= new Integer(10);
    public static final Integer PRIORITY_NORMAL		= PRIORITY_HIGH;
    public static final Integer PRIORITY_LOW 		= new Integer(0);

    ////////////////////    STATIC VARIABLES 	////////////////////

    /**
     * A <code>NodeId->Classifier</code> map of classifiers.
     * Each virtual machine has a single <code>Classifier</code>.
     */
    private static HashMap _classifiers;
    
    ////////////////////       VARIABLES	////////////////////

    /**
     * A map of all the type nodes, indexed by <code>Class</code> objects.
     */
    protected HashMap type_nodes;

    /**
     * The <code>Class</code> object for the root of the type hierarchy.
     */
    protected Class root_class;
    
    /**
     * A timer used to delay event delivery
     */
    protected TimerCB timer_cb;

    /**
     * The <code>NodeId</code> that is associated with this classifier.
     */
    protected NodeId _node_id;

    /**
     * <code>Vector</code> of <code>TSink</code>s
     */
    protected Vector tsinks;

    /**
     * <code>Hashtable</code> contains all types that have been subscribed to
     */
    private Hashtable subscribed_types;

    /**
     * a <code>SlowLink</code> field describing a single stage-to-stage
     * link whose events we should delay, or <code>null</code>.
     */
    private SlowLink slow_link;

    ////////////////////  STATIC CONSTRUCTORS	////////////////////

    static {
	if( DEBUG ) Debug.turnOn( tag );
	
	// Initialize the map of classifiers.
	_classifiers = new HashMap ();

	// Make sure all the QuickSerializable codes have been registered.
	Types.registerAllTypes();
    }

    ////////////////////      CONSTRUCTORS	////////////////////

    /**
     * Create a new Classifier object.
     */
    public Classifier(NodeId node_id)
    {
	_node_id = node_id; // DO NOT TAKE THIS OUT --srhea

	// Use a small map with a low load factor.
	type_nodes = new HashMap( 20, 0.3f );

	try {
	    // All events implement the interface for SandStorm messages:
	    root_class = Class.forName("seda.sandStorm.api.QueueElementIF");
	} catch ( ClassNotFoundException cnfe ) {
	    Carp.die( cnfe );
	}
	TrieRoot root = new TrieRoot();
	type_nodes.put( root_class, root );

	// Initialize an internal timer for <code>dispatch_later</code>.
	timer_cb = new DefaultTimerCB (this);
	
	// Register with SandStorm for its signals
	Sandstorm ss = null;
	try {
	    ss = Sandstorm.getSandstorm();
	}
	catch (NoClassDefFoundError e) {
	}

	if( ss == null ) {
	    if( DEBUG ) Debug.printtagln( tag, "No SandStorm running" );
	} else {
	    SignalMgrIF ss_signal = ss.getManager().getSignalMgr();
	    ss_signal.register( new StagesInitializedSignal(), this );

	    if( DebugFlags.ADD_ARTIFICIAL_DELAY ) {
		String link = 
		    ss.getManager().getConfig().getString("global.initargs.delay");
		if( link != null ) {
		    try {
			StringTokenizer st = new StringTokenizer( link, ":" );
			String sender = st.nextToken();
			String receiver = st.nextToken();
			long delay = Long.parseLong( st.nextToken() );
			Carp.out( tag + " delaying " + sender + " -> " +
				  receiver + " by " + delay + " milliseconds." );
			slow_link = new SlowLink( sender, receiver, delay );

		    } catch( Exception e ) {
			Carp.carp( e );
			Carp.die( "invalid link specification: " + link );
		    }
		} else {
		    if( DEBUG ) Debug.printtagln( tag, "No artificial delays" );
		}
	    }
	}

	tsinks = new Vector();
	subscribed_types = new Hashtable();

    }

    ////////////////////     STATIC METHODS 	////////////////////

    /**
     * Returns the <code>Classifier</code> instance corresponding to a
     * <code>NodeId</code>. 
     *
     * @param node_id	The <code>NodeId</code> which parameterizes a
     * virtual SandStorm instance.
     * @return the (possibly new) <code>Classifier</code> instance
     * corresponding to <code>node_id</code>.
     */
    public static Classifier getClassifier( NodeId node_id )
    {
	Classifier result = null;
	synchronized (_classifiers) {
	    result = (Classifier) _classifiers.get (node_id);
	    if (result == null) {
		result = new Classifier(node_id);
		_classifiers.put (node_id, result);
	    }
	}
	return result;
    }

    /**
     * Deletes the <code>Classifier</code> instance corresponding to a
     * <code>NodeId</code>. Should be called only when a node is going down
     * and going to come up again with a new node id.  Use very carefully!
     *
     * @param node_id	The <code>NodeId</code> which parameterizes a
     * virtual SandStorm instance.
     */
    public static void removeClassifier( NodeId node_id )
    {
	synchronized (_classifiers) {
	    _classifiers.remove (node_id);
	}
    }

    /**
     * Returns a <code>Set</code> containing the <code>NodeId</code>s for
     * each <code>Classifier</code> in this jvm.  The <code>Set</code> is
     * backed by the internal map; do not attempt to change it.
     * <p>
     * <b>WARNING:</b> Used for shutdown control only. Do not call this method.
     */
    public static Set getNodeIds()
    {
	Set result = null;
	synchronized( _classifiers ) {
	    result = _classifiers.keySet();
	}
	return result;
    }

    //////////////////// 	METHODS		////////////////////

    /**
     * The <code>NodeId</code> that is associated with this classifier.
     */
    public NodeId node_id () {
	return _node_id;
    }
    
    /**
     * Subscribes a <code>SinkIF</code> to an event stream defined by a
     * <code>Filter</code>.
     * <p>
     * Each event class corresponds to a trie of field restrictions.  This
     * method registers the sink at the appropriate leaf of each trie
     * corresponding to the required event type or any of its
     * subclasses.
     * <p>
     * Currently each trie is modified atomically, but the whole set of
     * tries is not.  The <code>Classifier</code> may dispatch events of
     * some subclasses before this <code>Filter</code> has been registered
     * everywhere.  All tries will be ready when this method returns.  If
     * this level of consistency is insufficient, use
     * {@link seda.sandStorm.api.SinkIF#setEnqueuePredicate} to disable the
     * sink completely until after subscription is complete.
     * <p>
     * If the <code>HashMap</code> of tries requires modification (e.g. for
     * a new type of event), a new <code>HashMap</code> is constructed and
     * moved into place atomically.
     * <p>
     * The type hierarchy must be a tree.  We disallow subscriptions which
     * require a type with multiple independent parents in the hierarchy.
     *
     * @param filter 	defines the set of <code>QueueElementIF</code>s which
     * 			the caller requests.
     * @param sink	the <code>SinkIF</code> into which we place satisfying
     * 			<code>QueueElementIF</code>s.
     * @param priority	a number between 0 (lowest priority) and 10
     (highest). Only the <code>PRIORITY_*</code> constants are meaningful.
     *
     * @return true iff subscription succeeded.
     */

    public synchronized boolean subscribe( Filter filter, SinkIF sink,
					   Integer priority )
    {
	if( DEBUG ) Debug.printtagln( tag+".subscribe", "called: " +
				      filter + ", sink, " + priority);
	if( (priority.compareTo(PRIORITY_LOW) < 0) ||
	    (priority.compareTo(PRIORITY_HIGH) > 0) ) {
	    if( DEBUG ) Debug.printtagln( tag+".subscribe",
					  "Invalid priority: " + priority );
	    return false;
	}
	// First, find the TrieRoot from which to start the filter.
	Class required_type = filter.required_type;
	if( DEBUG ) Debug.printtagln( tag+".subscribe", "looking up " +
				      required_type.getName() );
	TrieRoot root = (TrieRoot)type_nodes.get( required_type );

	if( root == null ) {
	    // This type doesn't exist in our trie yet; create a new one.
	    root = add_root( required_type );
	    if( root == null ) {
		if( DEBUG ) Debug.printtagln( tag+".subscribe", "bad type: "
					      + required_type.getName() );
		return false;
	    }
	    // It is possible that the Classifier could dispatch an event
	    // through the new trie before we add the new Filter below.
	    // That dispatch would behave the same as if the new node had
	    // not yet been added, because it is a duplicate of its
	    // superclass_node.  
	} else {
	    if( DEBUG ) Debug.printtagln( tag+".subscribe",
					  "found type node " + root );
	}


	// Notify all TSinks of the SinkIF subscribing
	if (subscribed_types.containsKey(required_type)) {
	    subscribed_types.put(required_type, Boolean.TRUE);
	} else {
	    subscribed_types.put(required_type, sink);
	}
	for(int i = 0; i < tsinks.size(); i++) {
	    TSink t = ((TSink)tsinks.get(i));
	    if (sink != t.mySinkIF)
		t.subscribeNotify(filter, sink, priority);
	}

	// Now add the rest of the filter to the trie rooted at this node.
	// (And the tries for subclasses)
	add_sink( sink, filter, priority, root );

	if( DEBUG ) Debug.printtagln( tag+".subscribe", "done" );
	return true;
    }

    /**
     * Subscribe a normal priority <code>SinkIF</code> to an event stream
     * defined by a <code>Filter</code>.
     *
     * @param filter 	defines the set of <code>QueueElementIF</code>s which
     * 			the caller requests.
     * @param sink	the <code>SinkIF</code> into which we place satisfying
     * 			<code>QueueElementIF</code>s.
     * @return true iff subscription succeeded.
     */

    public boolean subscribe( Filter filter, SinkIF sink )
    {
	return subscribe( filter, sink, PRIORITY_NORMAL );
    }

    /**
     * Search through the event class hierarchy, finding the closest
     * superclass of the specified class.
     *
     * @param orphan	the class whose ancestor to find.
     * @return the most immediate superclass of <code>orphan</code, or
     * <code>null</code> if no superclasses are found (must be a malformed
     * <code>Filter</code>).
     */

    private Class findNearestAncestor( Class orphan )
    {
	if( DEBUG ) Debug.printtagln( tag+".findNearestAncestor",
				      "called: " + orphan.getName() );

	Class superclass = root_class;
	if( ! superclass.isAssignableFrom( orphan ) ) {
	    // <code>orphan</code> is not a valid event
	    return null;
	}

	// Descend the class tree - assume a single correct path.
	looking: while( true ) {
	    // Block on the superclass, because
	    // <code>adjustLinks</code> might be restructuring the
	    // <code>subclasses</code> vector.
	    if( DEBUG ) Debug.printtagln( tag+".findNearestAncestor",
					  "checking: " +superclass.getName());
	    synchronized( superclass ) {
		TrieRoot superclass_node = (TrieRoot)type_nodes.get(superclass);
		
		for( int i=0; i<superclass_node.subclasses.size(); i++ ) {
		    Class nominee = (Class)superclass_node.subclasses.get(i);
		    if( nominee.isAssignableFrom( orphan ) ) {
			// Found a closer ancestor. Ignore other siblings.
			// Make this class the superclass and descend.
			if( DEBUG ) {
			    Debug.printtagln( tag+".findNearestAncestor",
					      "yes: " +nominee.getName() );
			}
			superclass = nominee;
			continue looking;
		    } else {
			if( DEBUG ) {
			    Debug.printtagln( tag+".findNearestAncestor",
					      "no: " +nominee.getName() );
			}
		    }
		}
		break looking;
	    }
	}
	
	if( DEBUG ) Debug.printtagln( tag+".findNearestAncestor",
				      "done: " + superclass.getName() );
	return superclass;
    }

    /**
     * Adjusts <code>superclass</code> and <code>subclasses</code> fields
     * when a new type is inserted into the event class hierarchy.
     * The subclasses of the <code>superclass</code> node are examined and
     * reassigned to <code>node</code> if appropriate.
     * <p>
     * This method is synchronized for clarity.  It is called only from
     * {@link #add_root}, so it would be de facto synchronized anyway.
     *
     * @param superclass	the <code>Class</code> restricted by the
     *  new node's nearest ancestor in the type hierarchy.
     * @param new_type	the <code>Class</code> for the new node.
     * @param node	the new node.
     */
    private synchronized void adjustLinks( Class superclass,
					   Class new_type, TrieRoot node )
    {
	if( DEBUG ) Debug.printtagln( tag+".adjustLinks", "called: "
				      + superclass.getName() + ", " +
				      new_type.getName() );

	// First reset the links in node, which were cloned.
	node.superclass = superclass;
	node.subclasses = new Vector();
	
	TrieRoot superclass_node = (TrieRoot)type_nodes.get(superclass);
	Class subclass;
	// Find all the subclasses to be reassigned.
	// Iterate backwards, so that remove() doesn't mess us up.
	for( int i=(superclass_node.subclasses.size()-1); i>=0; i-- ) {
	    subclass = (Class)superclass_node.subclasses.get(i);
	    if( new_type.isAssignableFrom(subclass) ) {
		if( DEBUG ) Debug.printtagln( tag+".adjustLinks",
					      "reassigning: " + subclass );
		// Move the subclass link
		node.subclasses.add( subclass );
		superclass_node.subclasses.remove(i);
		// Move the superclass link
		((TrieRoot)type_nodes.get(subclass)).superclass = new_type;
	    } else {
		if( DEBUG ) Debug.printtagln( tag+".adjustLinks",
					      "skipping: " + subclass );
	    }

	}

	// Finally, add the new type to the superclass' list.
	superclass_node.subclasses.add(new_type);

	if( DEBUG ) Debug.printtagln( tag+".adjustLinks", "done" );
    }

    /**
     * Adds a new <code>TrieRoot</code> to type_nodes, updating the class
     * tree links.  The new trie will be a clone of its immediate
     * superclass' trie, with all appropriate <code>superclass</code> and
     * <code>subclasses</code> fields updated.  Because the new trie
     * contains the same filters as its parent, no observable difference in
     * dispatch will occur until one of them is modified.
     * <p>
     * This method is synchronized for clarity.  It is called only from
     * {@link #subscribe}, so it would be de facto synchronized anyway.
     *
     * @param new_type	The <code>Class</code> for the event type for which
     * 			we want to create a new trie.
     * @return the <code>TrieRoot</code> for the newly added type.
     */
    private synchronized TrieRoot add_root( Class new_type )
    {
	// This type doesn't exist in our trie yet; create a new one.
	if( DEBUG ) Debug.printtagln( tag+".add_root", "called" );
	
	TrieRoot new_root;
	// Start by finding this type's closest ancestor in the trie.
	Class superclass = findNearestAncestor( new_type );
	if( superclass == null ) {
	    Carp.error( "Error: " + new_type + " has no " +
			"superclass in QueueElementIF hierarchy." );
	    return null;
	}
	
	// All filters and sinks for the superclass apply to the new root.
	TrieRoot superclass_node = (TrieRoot)type_nodes.get( superclass );
	new_root = (TrieRoot)superclass_node.clone();
	if( DEBUG ) Debug.printtagln( tag+".add_root",
				      "new root: " + new_root );

	// Now construct a new map and add the new node.
	HashMap new_type_nodes = (HashMap)type_nodes.clone();
	new_type_nodes.put( new_type, new_root );
	
	synchronized( superclass ) {
	    // Now adjust the superclass and subclass links for all applicable
	    // nodes and swap the hashtables.
	    // The links are used in  <code>FindNearestAncestor</code>,
	    // so we need to adjust the links and swap in the new
	    // <code>type_nodes</code> atomically.  Otherwise, the
	    // <code>subclasses</code> field of the superclass could reference
	    // a root (<code>new_root</code>) which isn't in
	    // <code>type_nodes</code>.
	    // Since <code>FindNearestAncestor</code> descends the class
	    // hierarchy from the root, synchronizing the superclass suffices.
	    
	    adjustLinks( superclass, new_type, new_root );
	    type_nodes = new_type_nodes;
	}

	if( DEBUG ) Debug.printtagln( tag+".add_root", "done" );
	return new_root;
    }
    
    
    /**
     * Adds the specified sink to the trie rooted at <code>root</code>.
     * This method descends the trie, adding new nodes when necessary,
     * finally adding the sink to the appropriate <code>TrieNode</code>.
     * This method also recurses down the class hierarchy, adding the sink
     * to tries for all subclasses (whose tries should always contain a
     * superset of the superclass trie).
     * <p>
     * This method is synchronized for clarity.  It is called only from
     * {@link #subscribe}, so it would be de facto synchronized anyway.
     *
     * @param sink	The <code>SinkIF</code> to add
     * @param filter	The <code>Filter</code> to use.
     * @param priority	The <code>Filter</code>'s priority
     * @param root	The trie to descend, and on whose subclasses to
     *			recurse.
     */
    private synchronized void add_sink( SinkIF sink, Filter filter,
					Integer priority, TrieRoot root )
    {
	if( DEBUG ) Debug.printtagln( tag+".add_sink", "called" );

	synchronized( root ) {
	    // Step 1: Register the filter on this trie.

  	    TrieNode node = root;
	    // Use an array to track the set of remaining restrictions.
	    // Use <code>ordered_restrictions</code> for determinism.
  	    Object[] remaining = filter.ordered_restrictions.toArray();

  	    // Iterate across the restrictions already mapped in the node.
  	    // Descend the trie along a current mapping, if possible.
	    // If no restriction is shared, create a new node.
  	    // TODO: This approach is provably suboptimal if we ever see
  	    // complicated, overlapping Filters.  We will then need to
  	    // optimize the Trie properly.

  	    for( int last=remaining.length-1; last>=0; last-- ) {
  		// Drop the last restriction on each iteration
  		Field found = null;
  		HashMap map = null;
		
  		// Look for a current mapping in this node
  	    field_search: {
  		    for( int i=0; i<node.fields.size(); i++ ) {
  			found = (Field)node.fields.get(i);
  			for( int j=0; j<=last; j++ ) {
  			    if( found.equals( remaining[j] ) ) {
  				// Found an existing map for this field.
  				map = (HashMap)node.value_maps.get(i);
  				// Mark this restriction complete.
  				remaining[j] = remaining[last];
  				if( DEBUG ) Debug.printtagln( tag+".add_sink",
  							      "found field: " +
  							      found.getName() );
  				break field_search;
  			    }
  			}
  		    }
  		    // We didn't find any map for any of our remaining Fields.
  		    // Just choose the last one and make a new map.
  		    found = (Field)remaining[last];
  		    node.fields.add( found );
  		    map = new HashMap();
  		    node.value_maps.add( map );
  		    if( DEBUG ) Debug.printtagln( tag+".add_sink",
  						  "chose field: " +
  						  found.getName() );
  		} // end field_search
		
		// Now find the value and child node.
		Object value = filter.restrictions.get(found);
		TrieNode child = (TrieNode)map.get( value );
		if( DEBUG ) Debug.printtagln( tag+".add_sink",
					      "value: " + value );
		// Splice in a new child, if necessary.
		if( child == null ) {
		    if( DEBUG ) Debug.printtagln( tag+".add_sink",
						  "creating new subtrie" );
		    child = new TrieNode();
		    map.put( value, child );
		}
		// Now descend the Trie and continue
		if( DEBUG ) Debug.printtagln( tag+".add_sink",
					      "descending into " + child );
		node = child;
	    }
	    
	    // All restrictions have been processed.  The current TrieNode is
	    // therefore the node satisfying the Filter.  Insert the SinkIF here.
	    if( DEBUG ) Debug.printtagln( tag+".add_sink", "adding sink" );
	    node.addSink( sink, priority, filter.acceptance_rate );


	    // Step 2: Recurse down the subclasses, who also need to see
	    // this <code>Filter</code>.
	    Class subclass;
	    TrieRoot subclass_root;
	    for( int i=0; i<root.subclasses.size(); i++ ) {
		subclass = (Class)root.subclasses.get(i);
		subclass_root = (TrieRoot)type_nodes.get(subclass);
		if( DEBUG ) {
		    Debug.printtagln( tag+".add_sink",
				      "duplicating filter for event type "
				      + subclass.getName() );
		}
		add_sink( sink, filter, priority, subclass_root );
	    }
	    
	} // End synchronized( root );
	
	if( DEBUG ) Debug.printtagln( tag+".add_sink", "done" );
    }

    /**
     * Unsubscribe a <code>Filter</code>.  This method will not remove a
     * <code>SinkIF</code> from the trie completely if other 
     * <code>Filter</code>s for that <code>SinkIF</code> remain.
     * @param filter	The <code>Filter</code> to be removed.
     * @return true iff unsubscription succeeded.
     */

    public boolean unsubscribe( Filter filter )
    {
	Carp.error( "This method not yet implemented." );
	// TODO: implement.
	// Be very careful about synchronization issues.
	return false;
    }

    /**
     * Unsubscribe a <code>SinkIF</code>, removing its <code>Filter</code>s
     * from the trie.
     * @param sink	The <code>SinkIF</code> to be removed.
     * @return true iff unsubscription succeeded.
     */

    public boolean unsubscribe( SinkIF sink )
    {
	Carp.error( "This method not yet implemented." );
	// TODO: implement.
	// Be very careful about synchronization issues.
	return false;
    }

    /**
     * Enqueues the event into each <code>SinkIF</code> subscribed with a 
     * <code>Filter</code> that the given <code>QueueElementIF</code>
     * satisfies.
     * Uses the default error mask (<code>PRIORITY_NORMAL</code>).
     *
     * @param event	The <code>QueueElementIF</code> to dispatch.
     * @return the number of <code>SinkIF</code>s into which the event was
     * 			successfully enqueued.
     * @exception 	SinkFullException indicates that at least one
     * 			receiving sink of high priority was temporarily full.
     */

    public int dispatch( QueueElementIF event ) throws SinkFullException
    {
	return dispatch( event, PRIORITY_NORMAL );
    }
    
    /**
     * Enqueues the event into each <code>SinkIF</code> subscribed with a 
     * <code>Filter</code> the given <code>QueueElementIF</code> satisfies.
     *
     * This method is lossy in that it silently fails to enqueue the event
     * into any <code>SinkIF</code> which returns an <code>Exception</code>.
     * This dispatch variant is meant for "low priority" events which can be
     * safely dropped.  
     *
     * @param event	The <code>QueueElementIF</code> to dispatch.
     * @return the number of <code>SinkIF</code>s into which the event was
     * 			successfully enqueued.
     */

    public int dispatch_lossy( QueueElementIF event )
    {
	// Just use the normal dispatch() with an infinitely high error mask.
	int num_dispatched = 0;
	try {
	    num_dispatched = dispatch( event, PRIORITY_INFINITE );
	} catch( SinkFullException sfe ) {
	    // Shouldn't ever happen.
	    Carp.die( sfe );
	}
	return num_dispatched;
    }
	       
    /**
     * Enqueues the event into each <code>SinkIF</code> subscribed with a
     * <code>Filter</code> that the given <code>QueueElementIF</code>
     * satisfies.  All <code>SinkIF</code>s registered at or above the
     * specified priority are enqueued atomically.  If any throws a
     * <code>SinkFullException</code>, that exception is thrown and no
     * <code>SinkIF</code> receives the the event.  <code>SinkIF</code>s
     * which were registered below the specified priority are processed
     * second; Any <code>SinkFullException</code>s they throw are ignored.
     *
     * @param event	The <code>QueueElementIF</code> to dispatch.
     * @param error_mask	The lowest priority for <code>SinkIF</code>s
     * 				whose errors we want to see.
     * @return the number of <code>SinkIF</code>s into which the event was
     * 			successfully enqueued.
     * @exception 	SinkFullException indicates that at least one
     * 			receiving sink of high priority was temporarily full.
     */
    public int dispatch( QueueElementIF event, Integer error_mask )
	throws SinkFullException
    {
	if( DEBUG ) Debug.printtagln( tag+".dispatch", "called: " +
				      event + ", " + error_mask );
	if( event == null ) {
	    if( DEBUG ) Debug.printtagln( tag+".dispatch",
					  "dropping null event" );
	    return 0;
	}

	// THIS IS A HACK.  We tag dispatched NetworkMessages with the
	// NodeId of the Classifier they are dispatched through in order to
	// allow the MultiplexedNetwork stage to correctly set the peer
	// field on an outgoing message.

	if (event instanceof NetworkMessage)
	    ((NetworkMessage) event).sender = _node_id;

	// END HACK.
	
	int num_dispatched = 0;

	// Find the SinkIF's for this event.
	HashMap sink_map = collect_sinks( event );

	if( DebugFlags.ADD_ARTIFICIAL_DELAY ) {
	    if( (slow_link != null) && (slow_link.sink != null) &&
		sink_map.containsKey( slow_link.sink ) ) {
		if( slow_link.thread == null ) {
		    // We don't have the sender's thread object
		    String name = Thread.currentThread().getName();
		    if( name.indexOf( slow_link.sender ) >= 0 ) {
			// The thread name contains the stage name.
			if( DEBUG ) Debug.printtagln( tag+".dispatch", 
						      "found slow sender");
			slow_link.thread = Thread.currentThread();
		    }
		}
		if( (slow_link.thread != null) &&
		    (slow_link.thread.equals( Thread.currentThread() )) ) {
		    if( DEBUG ) Debug.printtagln( tag+".dispatch", 
						  "delaying message to " +
						  slow_link.receiver );
		    sink_map.remove( slow_link.sink );
		    slow_link.enqueue( event );
		}
	    }
	}

	
	Object[] sinks = sink_map.keySet().toArray();

        // Optimize for the common cases.
        if (sinks.length == 0) {
            // do nothing
        }
        else if (sinks.length == 1) {
            try {
                ((SinkIF)sinks[0]).enqueue (event);
                num_dispatched++;
            } catch( SinkClosedException sce ) {
                // We found dead sink; remove it.
                // TODO: uncomment these when unsubscribe is implemented
                //Carp.error( sce );
                //unsubscribe( (SinkIF)sinks[i] );
                Carp.out( "Attempted to dispatch " + event + 
                        " to a closed queue." );
            } catch( SinkFullException sfe ) {
                if( error_mask.compareTo(sink_map.get(sinks[0])) <= 0 ) {
                    // High priority; Stop here, and throw the exception.
                    throw sfe;
                } else if( DEBUG ) {
                    // Ignore low priority SinkIF's.
                    Debug.printtagln( tag+".dispatch",
                            "Masking SinkFullException from " +
                            "low-priority SinkIF: " + sinks[0] );
                }
            } catch( SinkException se ) {
                // Shouldn't ever happen.
                Carp.die( se );
            }
        }
        else {

            Object[] tokens = new Object[sink_map.size()];
            QueueElementIF[] singleton = {event};

            // Tentatively enqueue the event onto each sink.
            for( int i=0; i<sinks.length; i++ ) {
                try {
                    tokens[i] = ((SinkIF)sinks[i]).enqueue_prepare( singleton );
                    num_dispatched++;
                } catch( SinkClosedException sce ) {
                    // We found dead sink; remove it.
                    // TODO: uncomment these when unsubscribe is implemented
                    //Carp.error( sce );
                    //unsubscribe( (SinkIF)sinks[i] );
                    Carp.out( "Attempted to dispatch " + event + 
                            " to a closed queue." );
                } catch( SinkFullException sfe ) {
                    if( error_mask.compareTo(sink_map.get(sinks[i])) <= 0 ) {
                        // High priority; Stop here, and throw the exception.
                        for( int j=0; j<i; j++ ) {
                            if( tokens[i] != null ) 
                                ((SinkIF)sinks[i]).enqueue_abort( tokens[i] );
                        }
                        throw sfe;
                    } else if( DEBUG ) {
                        // Ignore low priority SinkIF's.
                        Debug.printtagln( tag+".dispatch",
                                "Masking SinkFullException from " +
                                "low-priority SinkIF: " + sinks[i] );
                    }
                } catch( SinkException se ) {
                    // Shouldn't ever happen.
                    Carp.die( se );
                }
            }

            if( DEBUG ) {
                Debug.printtagln( tag+".dispatch", "enqueued " + event +
                        " onto " + num_dispatched +
                        (num_dispatched==1?" SinkIF":" SinkIFs") );
            }

            // Now commit the transactions.  Skip sinks which failed silently.
            for( int i=0; i<sinks.length; i++ ) {
                if( tokens[i] != null ) {
                    try {
                        ((SinkIF)sinks[i]).enqueue_commit( tokens[i] );
                    }
                    catch (Exception e) {
                        System.err.println ("Classifer.dispatch: caught " + e);
                        System.err.println ("event = " + event);
                        System.err.println ("sinks[i] = " + sinks[i]);
                        System.err.println ("tokens[i] = " + tokens[i]);
                        e.printStackTrace (System.err);
                    }
                }
            }
        }

	return num_dispatched;
    }

    /**
     * Enqueues the event at a later time into each <code>SinkIF</code>
     * subscribed with a <code>Filter</code> that the given
     * <code>QueueElementIF</code> satisfies.  

     * An internal timer holds the event for the specified time, and then
     * <code>dispatch</code>es it.  The event is classified at dispatch
     * time, not when this method is called.  Therefore the set of sinks
     * which receive the event may change in the meantime.
     * NOTE: please <code>cancel</code> any unneeded events.  Each pending
     * event increases the resources consumed by the internal timer.
     *
     * @see #cancel
     *
     * @param event	The <code>QueueElementIF</code> to dispatch.
     * @param millis	The number of milliseconds from now to dispatch the
     * 			event.
     * @return a token to pass to <code>cancel</code> in order to cancel
     * the event before it is dispatched.
     */

    public Object dispatch_later( QueueElementIF event, long millis )
    {
	if( DEBUG ) Debug.printtagln( tag+".dispatch_later",
				      "called: " + event );
	Object token = timer_cb.schedule ( millis, event );
	
	if( DEBUG ) Debug.printtagln( tag+".dispatch_later",
				      "done. token: " + token );
	return token;
    }
    
    /**
     * Cancels an event previously scheduled by <code>dispatch_later</code>.
     * If the event has already been dispatched, this method has no effect.
     *
     * @see #dispatch_later
     *
     * @param token	The <code>Object</code> returned by
     * 			<code>dispatch_later</code>.
     * @exception IllegalArgumentException if <code>token</code> is not a
     * token returned by the internal timer.
     */

    public void cancel( Object token )
    {
	if( DEBUG ) Debug.printtagln( tag+".cancel",
				      "called: " + token  );
	timer_cb.cancel (token);
	if( DEBUG ) Debug.printtagln( tag+".cancel", "done" );
    }
    
    /**
     * Gets the <code>TSink</code> for the stage given as an argument.
     * When the requesting stage has the <code>TSink</code> for the 
     * given stage, it will be able to pass events directly instead of
     * invoking this <code>Classifier</code>, which should be more 
     * efficient.
     *
     * @param stagename   String name of the stage whose TSink is requested
     * @return TSink    A TSink in which events can be enqueued.
     * @exception NoSuchStageException if there is no stage named stagename.
     */
    public TSink getTSink( String stagename) throws NoSuchStageException {
	TSink rv = new TSink(stagename, this);
	tsinks.addElement(rv);
	return rv;
    }

    /**
     * Returns Boolean.TRUE if type has been subscribed to.
     *
     * @param event_type  The type of event
     * @return Boolean  true if event_type has been subscribed to
     */
    public Boolean subscribedTo( Class event_type ) 
    {
	Object o = subscribed_types.get(event_type);
	if (o != null) {
	    return Boolean.TRUE;
	} else {
	    return Boolean.FALSE;
	}
    }

    /**
     * Called by TSink.enqueue(), adds ts to tsinks.
     *
     * @param ts  The TSink calling this method
     * @param event_type  The type which ts is enqueueing
     */
    public Boolean addTSinkType( TSink ts, Class event_type ) {
	Class current_type = event_type;

	while (current_type != null) {
	    Object o = subscribed_types.get(current_type);
	    if (o != null) {
		if (o instanceof SinkIF && ts.mySinkIF.equals((SinkIF)o)) {
		    //return Boolean.FALSE;
		} else {
		    return Boolean.TRUE;
		} 
	    } else {
		//return Boolean.FALSE;
	    }
	    current_type = current_type.getSuperclass();
	}
	return Boolean.FALSE;
    }

    /**
     * Find the set of all <code>SinkIF</code>'s which have registered
     * <code>Filter</code>s which the specified event satisfies.
     *
     * @param event	The <code>QueueElementIF</code> to classify.
     */
    private HashMap collect_sinks( QueueElementIF event )
    {
    	// First, find the correct type node.
	Class event_type = event.getClass();

	// Don't block on hashtable.
	TrieRoot node = (TrieRoot)type_nodes.get( event_type );
	if( node == null ) {
	    // Look for a superclass
	    Class ancestor = findNearestAncestor( event_type );
	    node = (TrieRoot)type_nodes.get( ancestor );
	    if( DEBUG ) Debug.printtagln( tag+".collect_sinks",
					  "dispatching " + event +
					  " as class " + ancestor );
	}

	// Next, pass this QueueElementIF down the trie of this type node
	HashMap sink_map = new HashMap();
	synchronized( node ) {
	    if( DEBUG ) Debug.printtagln( tag+".collect_sinks",
					  "collecting at " + node );
	    node.collectSinks( event, sink_map );
	}
	return sink_map;
    }

    /* From ostore.util.QuickSerializable interface */
    /**
     * Dispatches an event, just like <code>dispatch</code>.
     * <p>
     * <b> DO NOT USE THIS METHOD </b>. 
     * This method exists only to satisfy the <code>SinkIF</code> interface.
     * We implement that interface (by extending the
     * <code>SimpleSink</code> class) to allow signals and network messages to
     * travel through the Classifier.
     *
     * @see #dispatch
     * @param element	The event to dispatch.
     */
    public void enqueue( QueueElementIF element ) throws SinkFullException
    {
	if( DEBUG ) Debug.printtagln( tag+".enqueue", "called: " + element );
	dispatch( element, PRIORITY_NORMAL );
	if( DEBUG ) Debug.printtagln( tag+".enqueue", "done");
    }

    /**
     * Allows the timer callback to be changed; this is used by the
     * simulator to install a timer using simulated time rather than
     * operating system timers based on real time.
     */
    public void set_timer_cb (TimerCB callback) {
	timer_cb = callback;
    }

    ////////////////////     INNER CLASSES	////////////////////

    /**
     * Records information used to delay events sent from one named stage
     * to another.
     */
    public class SlowLink
    {
	////////////////////     CONSTANTS	////////////////////
	////////////////////  STATIC VARIABLES 	////////////////////
	////////////////////     VARIABLES	////////////////////

	/**
	 * The name of the sending stage
	 */
	public String sender;
	
	/**
	 * The name of the receiving stage
	 */
	public String receiver;

	/**
	 * The number of milliseconds to delay messages on this link.
	 */
	public long delay;

	/**
	 * The <code>Thread</code> object for the sender
	 * TODO: allow more than one thread per stage?
	 */
	public Thread thread;
	
	/**
	 * The <code>SinkIF</code> for the receiver
	 * TODO: allow more than one sink?
	 */
	public SinkIF sink;

	/**
	 * A timer interface for <code>sink</code>.
	 */
	public TimerCB timer;
	
	////////////////////    CONSTRUCTORS	////////////////////

	/**
	 * Creates a new <code>SlowLink</code>.
	 */
	public SlowLink( String sender, String receiver, long delay )
	    throws NoSuchStageException
	{
	    this.sender = sender;
	    this.receiver = receiver;
	    this.delay = delay;
	    ManagerIF mgr = Sandstorm.getSandstorm().getManager();
	    Enumeration stages = mgr.getConfig().getStages();
	    while( stages.hasMoreElements() ) {
		stageDescr stage = (stageDescr)stages.nextElement();
		if( stage.stageName.indexOf( receiver ) >= 0 ) {
		    if( DEBUG ) Debug.printtagln( tag+".init", 
						  "found receiver " +
						  stage.stageName );
		    this.sink = mgr.getStage(stage.stageName).getSink();
		}
	    }
	    if( this.sink == null ) {
		Carp.die( "Could not find receiving stage" );
	    }
	    this.timer = new DefaultTimerCB( sink );
		// TODO: Allow other TimerCB
	}
	
	////////////////////   STATIC METHODS 	////////////////////
	//////////////////// 	  METHODS	////////////////////

	/**
	 * Place the event on the receiver's sink in <code>delay</code>
	 * milliseconds. 
	 */
	public void enqueue( QueueElementIF event )
	{
	    timer.schedule( delay, event );
	}
    }
    
    /**
     * A Simple callback interface for timers.
     */
    public interface TimerCB
    {
	////////////////////     CONSTANTS	////////////////////
	////////////////////  STATIC VARIABLES 	////////////////////
	////////////////////     VARIABLES	////////////////////
	////////////////////    CONSTRUCTORS	////////////////////
	////////////////////   STATIC METHODS 	////////////////////
	//////////////////// 	  METHODS	////////////////////
	/**
	 * dispatches a specified event some time in the future.
	 *
	 * @see #cancel
	 * 
	 * @param millis	The number of milliseconds from now to
	 * 			dispatch the event.
	 * @param event		The <code>QueueElementIF</code> to dispatch.
	 *
	 * @return a token to pass to <code>cancel</code> in order to cancel
	 * the event before it is dispatched.
	 */
	public Object schedule (long millis, QueueElementIF event);
	
	/**
	 * Cancels an event previously <code>schedule</code>d.
	 * If the event has already been dispatched, this method has no effect.
	 *
	 * @see #schedule
	 *
	 * @param token	The <code>Object</code> returned by
	 * 			<code>schedule</code>.
	 * @exception IllegalArgumentException if <code>token</code> is not a
	 * token returned by the internal timer.
	 */
	public void cancel (Object token);
    }	// end interface TimerCB
    
    /**
     * The default implementation of <code>TimerCB</code>, which uses
     * <code>seda.sandStorm.core.ssTimer</code>s to store and trigger events.
     */
    protected class DefaultTimerCB implements TimerCB
    {
	////////////////////     CONSTANTS	////////////////////
	////////////////////  STATIC VARIABLES 	////////////////////
	////////////////////     VARIABLES	////////////////////

	/**
	 * The <code>ssTimer</code> which stores and triggers events.
	 */
	protected ssTimer _timer;

	/**
	 * A <code>SinkIF</code> onto which the events will be enqueued.
	 * Presumably a <code>Classifier</code>.
	 */
	protected SinkIF _sink;
	
	////////////////////    CONSTRUCTORS	////////////////////
	////////////////////   STATIC METHODS 	////////////////////
	//////////////////// 	  METHODS	////////////////////
	
	/**
	 * Creates a new <code>DefaultTimerCB</code> which will enqueue
	 * events onto a specified sink
	 *
	 * @param sink	The <code>SinkIF</code> onto which the events will
	 * be enqueued.  Presumably a <code>Classifier</code>.
	 */
	public DefaultTimerCB (SinkIF sink) {
	    _sink = sink;
	}

	/* Specified by TimerCB interface */
	public Object schedule (long millis, QueueElementIF event) {
	    if (_timer == null) 
		_timer = new ssTimer();
	    return _timer.registerEvent (millis, event, _sink);
	}

	/* Specified by TimerCB interface */
	public void cancel (Object token) {
	    if (_timer == null) 
		_timer = new ssTimer();
	    if( token instanceof ssTimer.ssTimerEvent ) {
		_timer.cancelEvent( (ssTimer.ssTimerEvent)token );
	    } else {
		throw new IllegalArgumentException( "Not a timer token: " +
			token );
	    }
	}
    }	// end class DefaultTimerCB

}

