/**
 * TrieNode.java
 *	
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.dispatch;

import java.lang.reflect.Field;
import java.util.Vector;
import java.util.HashMap;
import java.util.Iterator;
import ostore.util.Carp;
import ostore.util.Debug;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.api.QueueElementIF;

/**
 * A node of a trie for use with a <code>Classifier</code> object.
 * <p>
 * A <code>TrieNode</code> stores the <code>SinkIF</code>s associated with
 * the <code>Filter</code> defined by the path from the root of the trie.  It
 * also stores enough information about its children to recursively classify
 * <code>QueueElementIF</code>s.  Information about the requirement represented by
 * this <code>TrieNode</code> is stored on the edge to this node from its
 * parent, not internally to this data structure.
 *
 * @author	Dennis Geels
 * @version $Id: TrieNode.java,v 1.16 2002/07/22 20:54:53 srhea Exp $
 */
// TODO: optimization: use hashtables vs. arrays vs. singleton depending on
// number of children?

public class TrieNode
{
    ////////////////////       CONSTANTS	////////////////////

    private static final String tag = "TrieNode";
    
    private static final boolean DEBUG =
	ostore.util.DebugFlags.DISPATCH_TRIENODE;

    ////////////////////    STATIC VARIABLES 	////////////////////
    ////////////////////       VARIABLES	////////////////////

    /**
     * Contains all of the <code>SinkIF</code>s to receive events which
     * satisfy this node's requirement.
     */
    protected Vector sinks;

    /**
     * Contains the priorities for the <code>SinkIF</code>s.  This
     * <code>Vector</code> is parallel to <code>sinks</code>.
     */
    protected Vector priorities;

    /**
     * Contains the pass rates for the <code>SinkIF</code>s.  This
     * <code>Vector</code> is parallel to <code>sinks</code>.
     */
    protected Vector rates;

    /**
     * Contains the set of <code>Field</code>s checked by this node's
     * children. 
     */
    protected Vector fields;

    /**
     * Contains the set of value->child mappings for each <code>Field</code>.
     * This <code>Vector</code> is parallel to <code>fields</code>.
     */
    protected Vector value_maps;

    ////////////////////  STATIC CONSTRUCTORS	////////////////////

    static {
	if( DEBUG ) Debug.turnOn( tag );
    }
    ////////////////////      CONSTRUCTORS	////////////////////

    /**
     * Creates a new <code>TrieNode</code> with no sinks or children.
     */
    protected TrieNode() {
	sinks = new Vector();
	priorities = new Vector();
	rates = new Vector();
	fields = new Vector();
	value_maps = new Vector();
    }
    
    /**
     * Creates a new <code>TrieNode</code> with the specified fields.  This
     * constructor is used within <code>clone</code>.
     */
    private TrieNode( Vector sinks, Vector priorities, Vector rates,
		      Vector fields, Vector value_maps )
    {
	this.sinks = sinks;
	this.priorities = priorities;
	this.rates = rates;
	this.fields = fields;
	this.value_maps = value_maps;
    }
    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////

    /**
     * Recursively find all <code>TrieNode</code>s which pass the specified
     * event, collecting all the <code>SinkIF</code>s subscribed at each node.
     *
     * @param event	The <code>QueueElementIF</code> to evaluate.
     * @param found	The set of <code>SinkIF</code>s collected. Each
     * 			<code>SinkIF</code> maps to its priority.
     */
    
    public void collectSinks( QueueElementIF event, HashMap found )
    {
	if( DEBUG ) Debug.printtagln( tag+".collectSinks", name() + 
					  " called" );
	// Assume that the event satisfies this node.
	// Collect all the sinks for this node:
	Object priority;
	Object old_priority;
	Object sink;
	Double rate;

	// If any sinks are just sampling the event stream, we need to
	// decide whether they accept this event or drop it.
	double rand = 0; // Math.random();
	
	for( int i=0; i<sinks.size(); i++ ) {
	    rate = (Double)rates.get(i);
	    if( rate.doubleValue() < rand ) {
		continue;
	    }
	    sink = sinks.get(i);
	    priority = priorities.get(i);
	    if( DEBUG ) Debug.printtagln( tag+".collectSinks", name() +
					  " found sink: " +
					  sink + ", " + priority );

	    // Store this sink at its highest priority seen so far.
	    old_priority = found.put( sink, priority );
	    if( (old_priority != null) &&
		(((Integer)old_priority).compareTo(priority) > 0) ) {
		found.put( sink, old_priority );
	    }
	}

	// Check all children:
	Field field;
	HashMap map;
	Object value = null;
	TrieNode child;
	for( int i=0; i<fields.size(); i++ ) {
	    field = (Field)fields.get(i);
	    map = (HashMap)value_maps.get(i);
	    try {
		value = field.get(event);
	    } catch( Exception e ) {
		// Really shouldn't happen.
		Carp.die( e );
	    }
	    child = (TrieNode)map.get(value);
	    if( DEBUG ) Debug.printtagln( tag+".collectSinks", name() + 
					  " checking field: " +
					  field.getName() +" => "+ value );
	    if( child == null ) {
		if( DEBUG ) Debug.printtagln( tag+".collectSinks", name() +
					      " no child" );
	    } else {
		// A sink is subscribed for this value of this field.
		if( DEBUG ) Debug.printtagln( tag+".collectSinks", name() + 
					      " found child: " + child );
		child.collectSinks( event, found );
	    }
	}
	
	return;
    }

    /**
     * Add a child node with a specified field-value restriction.  Any
     * <code>QueueElementIF</code> which satisfies this node's restriction
     * should test the new child's as well.
     *
     * @param child	The node to add.
     * @param field	The <code>Field</code> restricted by the child.
     * @param value	The value that <code>Field</code> should have.
     */
    public void addChild( TrieNode child, Field field, Object value )
    {
	if( DEBUG ) Debug.printtagln( tag+".addChild", name() + " called" );
	
	// Find the map for that Field
	HashMap map;
	for( int i=0; i<fields.size(); i++ ) {
	    if( fields.get(i).equals(field) ) {
		map = (HashMap)value_maps.get(i);
		map.put( value, child );
		return;
	    }
	}
	// No map; make a new one.
	fields.add(field);
	map = new HashMap();
	value_maps.add(map);
	map.put( value, child );
	return;
    }

    /**
     * Add a new <code>SinkIF</code> to receive for all
     * <code>QueueElementIF</code> satisfying this node in the trie. 
     *
     * @param sink	the <code>SinkIF</code>.
     * @param priority	a number between 0 (lowest priority) and 10 (highest).
     * @param acceptance_rate	the fraction of qualifying events to pass on
     * 			to the sink.  Useful for sampling the input stream.
     */
    
    public void addSink( SinkIF sink, Integer priority,
			    double acceptance_rate )
    {
	if( DEBUG ) Debug.printtagln( tag+".addSink", name() + 
				      " adding sink." );
	// We don't check for duplicates.  We want to leave them in, so that
	// one could be removed independently.
	sinks.add( sink );
	priorities.add( priority );
	rates.add( new Double(acceptance_rate) );
    }

    /**
     * Returns a short representation of this <code>TrieNode</code>
     * I don't override <code>toString()</code> because this version seems
     * more expensive computationally, and isn't significantly more useful.
     *
     * @return  <code>TrieNode@&lt;hashCode&gt;</code>
     */
    private String name()
    {
	return ("TrieNode");
    }

    /**
     * Performs a deeper copy than <code>Object.clone</code> would.
     * The entire trie is cloned, and may be modified independently of the
     * original.  The <code>Field</code>s and <code>SinkIF</code>s
     * referenced by the <code>TrieNode</code>s are not cloned.
     *
     * @return A new <code>TrieNode</code> with <code>cloned</code>
     * internal data structures.
     */
    public Object clone() {
	// Each <code>Vector</code> can be cloned, except for
	// <code>value_maps</code>.  Those contain the actual links to
	// children, which we need to copy explicity.
	
	Vector new_maps = new Vector( value_maps.size() );
	// Copy each map
	for( int i=0; i<value_maps.size(); i++ ) {
	    HashMap map = (HashMap)value_maps.get(i);
	    HashMap new_map = new HashMap( map.size() );
	    // Copy each mapping - be sure to recursively clone children!
	    Iterator values = map.keySet().iterator();
	    while( values.hasNext() ) {
		Object value = values.next();
		TrieNode child = (TrieNode)map.get( value );
		new_map.put( value, child.clone() );
	    }
	    new_maps.add( new_map );
	}
	
	TrieNode cloned = new TrieNode( (Vector)sinks.clone(),
					(Vector)priorities.clone(),
					(Vector)rates.clone(),
					(Vector)fields.clone(),
					new_maps );
	return cloned;
    }

    public String toString () {
	return "ostore.dispatch.TrieNode";
    }

}
