/**
 * HMACDHHandler.
 *
 * @author    Gifford Cheung
 * @version    $Id: HMACDHHandler.java,v 1.35 2002/09/09 05:04:15 srhea Exp $
 */  

/**
 * Future TODO: implement message/event logging....?
 */

package ostore.mac;
import java.util.*;
import java.io.*;
import ostore.util.*;
import ostore.dispatch.*;
import ostore.security.*;
import seda.sandStorm.api.*;
import seda.sandStorm.core.ssTimer;
import seda.sandStorm.internal.SinkProxy;

// Security stuff
import java.security.*;
import java.security.spec.*;
import javax.crypto.*;
import javax.crypto.spec.*;
import javax.crypto.interfaces.*;
import cryptix.jce.provider.*;
import cryptix.jce.provider.random.*;
import cryptix.jce.provider.key.*;

/**
 * MacHandler that uses HMAC for the message authentication and  
 * Diffie-Hellman Key Generation.  <br></br><br></br>
 * <b><i>Station-To-Station Protocol Implentation:</i></b> <br></br><br></br>
 * <b>Basic Protocol:</b><br></br>
 * There are four states: (1)UNINITIALIZED, (2)WAITFORRESPONSE, (3)WAITFORFINAL,
 * (4)ESTABLISHED.<br></br>
 * There are three messages (across the network): (1)INITIALIZE, (2)RESPOND,
 * (3)FINALIZE. <br></br>
 * There are three events (internal): (1)INIT (request to send out message 1), 
 * (2) STS_TIMEOUT, (3)DH_TIMEOUT <br></br>
 * State table<br></br>
 * <table border = 1><tr><td><b>State</b></td>
 * <td><b>Event</b></td><td><b>Next Stage</b>
 * </td><td><b>Action</b></td></tr>
 * <tr><td>UNINIT</td><td>Tapestry sends (1)INIT event</td>
 *  <td>WAITFORRESPONSE</td><td>Send message (1)</td></tr>
 * <tr><td>UNINIT</td><td>Receive message (1)</td>
 *  <td>WAITFORFINAL</td><td>Compute secret, Send message (2)</td></tr>
 * <tr><td>WAITFORRESPONSE</td><td>Receive message (2)</td>
 *  <td>ESTABLISHED</td><td>Confirm signature and Compute secret, 
 *  Send message (3)</td></tr>
 * <tr><td>WAITFORRESPONSE</td><td>Receive message (1)</td>
 *  <td>WAITFORRESPONSE or <br></br> WAITFORFINAL</td><td>If my guid is 
 *    greater that their guid, I become WAITFORFINAL, else ignore message.
 *  </td></tr>
 * <tr><td>WAITFORFINAL</td><td>Receive message (3)</td>
 *  <td>ESTABLISHED</td><td>Confirm signature</td></tr></table>
 * <br></br><br></br>
 * 
 * <b>establishedParties & pendingParties:</b><br></br>
 * Needed: To be able to go through a false or faulty handshake, 
* one that is cancelled before completion, and not lose the existing
* connection. <br></br>
* Solution: Keep two Hashtables (indexed by guid) one for established
* connections (<tt>establishedParties</tt>) and another for in-progress
* connections (<tt>pendingParties</tt>). If a <tt>pendingParties</tt> 
* connection successfully finishes a handshake, that entry replaces the
* one in <tt>establishedParties</tt>. <br></br>
* Functional access to this MacHandler only accesses 
* <tt>establishedParties</tt><br></br><br></br>
* <b>Timeouts:</b><br></br>
    * The HMACDHHandler handles timeouts (both session and Diffie-Hellman key
	    * expirations) by sending itself delayed messages. The time (in milliseconds)
    * of these are determined by config variables: <tt>timeoutlength</tt> and 
    * <tt>dhlifespan</tt><br></br><br></br>
    * TODO describe sts timeout and dh timeout<br></br>
    * <b>Synchronicity:</b><br></br>
    * Internally: <tt>establishedParties</tt>I
    * This stage is not designed to be replicated. Replication may introduce
    * race conditions that have not been thoroughly investigated.
    */

public class HMACDHHandler extends MacHandler {

    public static class MacBuffer extends OutputBufferImpl {

	public Mac engine;

	public MacBuffer (Mac engine) {
	    this.engine = engine;
	}

	public void add (byte input ) {
	    try { 
		engine.update (input);
	    }
	    catch (IllegalStateException e) {
		throw new AssertionViolatedException (
			"Mac object should be initialized.");
	    }
	}

	public void add (byte[] input, int offset, int length) {
	    try { 
		engine.update (input, offset, length);
	    }
	    catch (IllegalStateException e) {
		throw new AssertionViolatedException (
			"Mac object should be initialized.");
	    }
	}
    }

    public static final String tag = "HMACDHHandler";
    public static final boolean SEAN_DEBUG = false;
    public static final boolean DEBUG = false;
    public static final boolean ANNOUNCE = false;
    protected static final int _keysize       = 1024; //4096;
    protected static Hashtable instances = new Hashtable ();

    protected Hashtable establishedParties = new Hashtable ();
    protected Hashtable pendingParties = new Hashtable ();
    protected KeyPairGenerator _kpgen;
    protected KeyAgreement _kagree;
    protected SecureRandom _r;
    protected Signature _signEngine;
    protected String _pkeyfile;
    protected String _skeyfile;
    protected QSPublicKey _sigPub;
    protected QSPrivateKey _sigPri;
    protected Mac _hmac;
    protected SecureHash _myGuid;
    protected SinkIF _this_sink;
    protected int _stsTime;
    protected int _retryLimit;
    protected int _dhlifespan;
    protected boolean _selfTest;
    protected Classifier _classifier;

    synchronized protected void handle_init (TellInit tell) {
	if (DEBUG) {
	    Debug.printtagln (tag, "HMDHH " + _myGuid + 
		    ": TellSelf.init recv'd");
	    Debug.printtagln (tag, "HMDHH " + _myGuid + 
		    ": printing pending");
	    for (Enumeration e = pendingParties.elements (); 
		    e.hasMoreElements (); ) {
		((PartyInfo)e.nextElement ()).print ();
		Debug.printtagln (tag, "---------------------");
	    }
	    Debug.printtagln (tag, 
		    "***************************");
	    Debug.printtagln (tag, "HMDHH " + _myGuid + 
		    ": printing established");
	    for (Enumeration e = establishedParties.keys (); 
		    e.hasMoreElements (); ) {
		Debug.printtagln (tag, _myGuid + 
			e.nextElement ().toString ());
		Debug.printtagln (tag, "---------------------");
	    }
	    for (Enumeration e = establishedParties.elements (); 
		    e.hasMoreElements (); ) {
		if (DEBUG)
		    ((PartyInfo)e.nextElement ()).print ();
		Debug.printtagln (tag, "---------------------");
	    }
	}
	// if establishedParties (guid) gone then nothing
	// establishED
	// if pendingParties (guid) gone then nothing
	// establishING
	// if nothing is establishED or establishING for
	// this guid, then get something started.
	// if we are already establishED: reannounce and
	// return.

	if (establishedParties.containsKey (tell.guid)) {
	    dispatch (new KeyEstablished (tell.guid));
	    if (DEBUG) Debug.printtagln (tag, 
		    "HMAC" + _myGuid + ": REANNOUNCING.");
	    return;
	}
	// if we are establishING: return and do nothing
	if (pendingParties.containsKey (tell.guid) ) 
	    return; 

	// create a new PartyInfo and place into pendingParties 
	PartyInfo freshEntry = new PartyInfo ();
	pendingParties.put (tell.guid, freshEntry);

	if (DEBUG) Debug.printtagln (tag, "adding " + freshEntry +
		" to pendingParties with guid " + tell.guid);

	// generate new DH pair 
	// this ISN'T synchronized, even if it alters a
	// pendingParties
	// element because  (1) this is a slow operation
	//                  (2) this particular PartyInfo
	//                      will NOT be next altered
	//                      UNTIL a new DHMessage comes
	//                      in, referring to it and
	//                      that won't come in until I
	//                      send out the 1st message, I
	//                      haven't till fifteen lines
	//                      from now.
	makeNewDHKeys (freshEntry);

	// make INITIALIZE message
	DHMessage out = new DHMessage (tell.guid,
		freshEntry.dhpub,
		null,
		new QSHMACMD5(),
		null, null,
		DHMessage.INITIALIZE);
	out.inbound = false;

	// change state to WAITFORRESPONSE
	freshEntry.state = freshEntry.WAITFORRESPONSE;

	// send message INITIALIZE and
	// send out TIMER for session timeout
	TellStsTimeout st = new TellStsTimeout (tell.guid);
	if (DEBUG) Debug.printtagln (tag, "HMDHH" + _myGuid + 
		": Sending init msg to " + tell.guid);
	dispatch (out);
	freshEntry.timerToken = _classifier.dispatch_later (st, _stsTime);
    }

    synchronized protected void handle_sts_timeout (TellStsTimeout tell) {
	if (ANNOUNCE) 
	    Debug.printtagln (tag, "HMDHH" + _myGuid +
		    ":TellSelf.STS_TIMEOUT rcv'd"
		    + "against: " + tell.guid);
	QueueElementIF announce;

	// *******************************************
	// IT IS STILL POSSIBLE FOR A TIMEOUT TO QUEUE  
	// UP EVEN THOUGH WE ARE ESTABLISHED
	// ******************************************

	// a session has timed out
	// (1) retrieve the session
	PartyInfo info = (PartyInfo) pendingParties.get (tell.guid);
	// (2) do nothing if we retrieved nothing
	//     reason for this typically is
	//     a timeout was placed in the queue before
	//     a successful STS run. 
	//     If this isn't the reason, we still would be
	//     able to do anything
	if (info == null) {
	    // do nothing
	    // this statement skips steps 3/4  
	    if (ANNOUNCE) Debug.printtagln (tag,"HMACDHH" + 
		    _myGuid + ": " +
		    "TIMEOUT received, but no info. " +
		    "may have already established.");
	}
	else {

	    // (3) increment the retry counter
	    info.sessionRetries += 1;
	    // (4) if retry limit not reached start all over again. 
	    if (info.sessionRetries >= _retryLimit) {
		announce = new HandshakeFailed (tell.guid);
	    }
	    else {
		announce = new DHMessage (tell.guid,
			info.dhpub,
			null,
			new QSHMACMD5(),
			null, null,
			DHMessage.INITIALIZE);
		((DHMessage) announce).inbound = false;

		info.state = info.WAITFORRESPONSE;
		TellStsTimeout st = 
		    new TellStsTimeout (tell.guid);
		info.timerToken = 
		    _classifier.dispatch_later (st, _stsTime);
	    }

	    dispatch (announce);
	}
    }

    synchronized protected void handle_dh_timeout (TellDhTimeout tell) {
	if (ANNOUNCE) Debug.printtagln (tag, "DH TIMEOUT:" + _myGuid);
	// (1)snag lock 
	/* NOTE on establishedParties
	 * establishingParties is accessed both by handleEvent and 
	 * functional access to this stage, thus the need for
	 * "synchronized" access to it.
	 */
	synchronized (establishedParties) {
	    // (2)wipe all keys and timers
	    for (Enumeration e = pendingParties.elements (); 
		    e.hasMoreElements () ;) {
		PartyInfo i = (PartyInfo) e.nextElement ();
		if (i.timerToken != null) {
		    _classifier.cancel (i.timerToken);
		}
	    }
	    establishedParties.clear ();
	    // (3) usedKeys cleared out by virtue of emptying 
	    // establishedParties. 
	}
	_classifier.dispatch_later (tell,_dhlifespan);
    }

    synchronized protected void handle_state_uninit (
	    DHMessage dhmsg, int msgType, SecureHash otherguid, 
	    PartyInfo info) {

	if (msgType != DHMessage.INITIALIZE) {
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS)
		Carp.out (tag + ": not in init state.  dhmsg=" + dhmsg +
		", msgType=" + msgType + ", otherguid=" + otherguid + 
		", info=" + info + ".  Dropping it.");
	    return;
	}

	//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
	// Conditions:
	// Condition (4.abc)
	// an INITIALIZE message needs a non-null entry
	// for (a)QSkey and nothing in (b)sigkey or (c)verify
	// if we catch something wrong here, we drop the message
	// and reset the pendingParties 
	// TODO: this still allows for Denial of Service attacks...
	if (dhmsg.QSkey == null || 
		dhmsg.sigkey != null ||    
		dhmsg.mac != null) {
	    // ignore (TODO write error to log file)
	    if (DEBUG) Debug.printtagln (tag, "removing " + otherguid +
		    " from pendingParties");

	    pendingParties.remove (otherguid);
	    return;
	}
	//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

	// (1)
	// TODO a check for a bad key submission
	//        one that takes advantage of the 
	//        math involved such as a QS public key
	//        of value 0 and for replay attacks (keysused)
	if (establishedParties.containsKey (otherguid)) {
	    PartyInfo check = 
		(PartyInfo) establishedParties.get (otherguid);
	    if (check.usedKeys.contains (dhmsg.QSkey)) {
		// then we've already seen this DH key
		// TODO write event to log file and ignore msg
		return;
	    }
	}
	// (1.a) store their DHKey. Also checks out we 'save it'.
	info.dhpubOther = (DHPublicKey) dhmsg.QSkey.key ();
	info.usedKeys.add (dhmsg.QSkey);

	// (2) do we need to compute a new DH?
	if (establishedParties.containsKey (otherguid)) {
	    PartyInfo e = 
		(PartyInfo) establishedParties.get (otherguid);
	    info.dhpub = e.dhpub;
	    info.dhpri = e.dhpri;
	}
	else {
	    makeNewDHKeys (info);
	}
	if (DEBUG) Debug.printtagln (tag, "HMACDHH" + _myGuid + 
		": computed DH");

	// (3) compute secret
	try {
	    _kagree.init (info.dhpri);
	    _kagree.doPhase (dhmsg.QSkey.key (), true);
	}
	catch (InvalidKeyException e) {
	    // key issue reset the stage and ignore message 
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS)
		Carp.out (tag + ": InvalidKeyException.");
	    info.state = PartyInfo.UNINIT;
	    info.secret = null;
	    return;
	}

	byte[] seed = _kagree.generateSecret ();
	info.secret = new RawSecretKey ("DH seed", seed);
	if (DEBUG) Debug.printtagln (tag, "HMACDHH" + _myGuid + 
		": computed secret");

	// (4) make and send DHMESSAGE RESPOND
	DHMessage out = new DHMessage (otherguid,
		info.dhpub,
		(QSRsaPublicKey)_sigPub,
		new QSHMACMD5(),
		null, null, // makeDHVerify later
		DHMessage.RESPOND);
	out.inbound = false;

	// compute verification info and place into DHVerify
	//  MAC (Signature (Hash (both DH keys and Mac type)))
	try {
	    makeDHVerify (new QSDHPublicKey (info.dhpub), 
		    dhmsg.QSkey, info.secret, out);
	}
	catch (InvalidKeyException e) {
	    // key issue reset the stage and ignore message 
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS)
		Carp.out (tag + ": InvalidKeyException.");
	    info.state = PartyInfo.UNINIT;
	    info.secret = null;
	    return;
	}
	catch (SignatureException e) {
	    // unable to sign the message
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS)
		Carp.out (tag + ": signature exception.");
	    info.state = PartyInfo.UNINIT;
	    info.secret = null;
	    return;
	} 

	// send out 
	if (DEBUG) Debug.printtagln (tag, "HMACDHH: " + _myGuid +
		"sending out RESPOND  msg");
	dispatch (out);

	// (5) change state to WAITFORFINAL
	info.state = PartyInfo.WAITFORFINAL;

	// (6) TIMER send out session timer
	TellStsTimeout st = 
	    new TellStsTimeout (otherguid);
	info.timerToken =
	    _classifier.dispatch_later (st, _stsTime);
    }    

    synchronized protected void handle_state_waitforresp_resp (
	    DHMessage dhmsg, SecureHash otherguid, PartyInfo info) {

	//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
	// Conditions:
	// Condition (4.abc)
	if (dhmsg.QSkey == null ||
		dhmsg.sigkey == null ||
		!otherguid.verify (dhmsg.sigkey) ||
		dhmsg.mac == null) {
	    // ignore (TODO write error to log file)
	    if (DEBUG) Debug.printtagln (tag, _myGuid + 
		    "HMDHH: dhmsg failed " +
		    "Condition 4.abc check");
	    return;
	}

	//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

	// (1) Store the public signature key and QSkey
	info.sigpub = dhmsg.sigkey;
	info.dhpubOther = (DHPublicKey)dhmsg.QSkey.key ();

	// (2)
	// TODO check for a bad key submission
	//        one that takes advantage of the 
	//        math involved such as a QS public key
	//        of value 0 and for replay attacks (keysused)
	// THEN check against usedKeys
	if (establishedParties.containsKey (otherguid)) {
	    PartyInfo check = 
		(PartyInfo) establishedParties.get (otherguid);
	    if (check.usedKeys.contains (dhmsg.QSkey)) {
		// then we've already seen this DH key
		// TODO write event to log file and ignore msg
		return;
	    }
	}
	// (2.a) if their DHKey checks out we 'save it'.
	info.usedKeys.add (dhmsg.QSkey);

	// (3) compute secret
	try {
	    _kagree.init (info.dhpri); 
	    _kagree.doPhase (dhmsg.QSkey.key (), true); 
	}
	catch (InvalidKeyException e) { 
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS)
		Carp.out (tag + ": invalid key exception.");
	    info.state = PartyInfo.WAITFORRESPONSE;
	    info.secret = null;
	    return;
	}

	byte[] seed = _kagree.generateSecret ();
	info.secret = new RawSecretKey ("DH seed", seed);

	// (4) check dhmsg.verify
	// JUMP InvalidKeyException
	// JUMP SignatureException
	// JUMP TypeTable.NoSuchTypeCode
	try {
	    if (!checkDHVerify (
			dhmsg.user_data, dhmsg.mac,dhmsg,info)) {
		if (DebugFlags.LOG_BAD_NETWORK_EVENTS)
		    Carp.out (tag + ": didn't verify.");
		return;
	    }
	}
	catch (InvalidKeyException e) { 
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS)
		Carp.out (tag + ": invalid key exception.");
	    info.state = PartyInfo.WAITFORRESPONSE;
	    info.secret = null;
	    return;
	}
	catch (SignatureException e) {
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS)
		Carp.out (tag + ": signature exception.");
	    info.state = PartyInfo.WAITFORRESPONSE;
	    info.secret = null;
	    return;
	}
	catch (TypeTable.NoSuchTypeCode e) {
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS)
		Carp.out (tag + ":nosuchtypecode exception.");
	    info.state = PartyInfo.WAITFORRESPONSE;
	    info.secret = null;
	    return;
	}

	// (5) change state to ESTABLISHED
	info.state = PartyInfo.ESTABLISHED;

	if (ANNOUNCE) 
	    Debug.printtagln (tag, "HMACDHHandler: Xestablished by " +
		    _myGuid + "(me) with " +
		    otherguid + "(them)");

	synchronized (establishedParties) {

	    // (6) make and send FINALIZE
	    DHMessage out = new DHMessage (otherguid,
		    null,
		    (QSRsaPublicKey)_sigPub,
		    new QSHMACMD5(),
		    null, null, // DHVerify later
		    DHMessage.FINALIZE);
	    out.inbound = false;

	    // compute verification info and place into DHVerify
	    //  MAC (Signature (Hash (both DH keys and Mac type)))
	    try {
		makeDHVerify (new QSDHPublicKey (info.dhpub), 
			dhmsg.QSkey, info.secret, out);
	    }
	    catch (InvalidKeyException e) { 
		if (DebugFlags.LOG_BAD_NETWORK_EVENTS)
		    Carp.out (tag + ": invalid key exception.");
		info.state = PartyInfo.WAITFORRESPONSE;
		info.secret = null;
		return;
	    }
	    catch (SignatureException e) {
		if (DebugFlags.LOG_BAD_NETWORK_EVENTS)
		    Carp.out (tag + ": signature exception.");
		info.state = PartyInfo.WAITFORRESPONSE;
		info.secret = null;
		return;
	    }

	    // send out 
	    dispatch (out);

	    // (7) move info to establishedParties and
	    //       remove info from pendingParties 
	    //       cancel timer
	    if (info.timerToken != null)
		_classifier.cancel (info.timerToken);
	    info.timerToken = null;
	    if (establishedParties.contains (otherguid)) {
		PartyInfo old =
		    (PartyInfo) establishedParties.get (otherguid);
		info.usedKeys.addAll (old.usedKeys);
	    }

	    if (DEBUG) Debug.printtagln (tag, "removing " + otherguid +
		    " from pendingParties and changing state");

	    pendingParties.remove (otherguid);
	    establishedParties.put (otherguid, info);
	    sendEstablished (otherguid);
	}

	if (DEBUG) { 
	    Debug.printtagln (tag, "HMACDHHANDLER info for"+_myGuid);
	    info.print ();
	}
    }

    synchronized protected void handle_state_waitforresp_init (
	    DHMessage dhmsg, SecureHash otherguid, PartyInfo info) {
	//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
	// Conditions:
	// Condition (4.abc)
	if (dhmsg.QSkey == null || 
		dhmsg.sigkey != null || 
		dhmsg.mac != null) {
	    // ignore (TODO write error to log file)
	    return;
	}
	//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

	// Conditions are met, so this is a readable message,
	// Now, should we ignore this message? Because in this
	// case, both sides sent (1) and are stage WAITFORRESPONSE
	// If my guid is greater than other guid change my state.
	int compare = _myGuid.compareTo (otherguid);
	if (compare > 0) {
	    if (ANNOUNCE) {
		Debug.printtagln (tag,"HMACDHH: DUAL-INIT conflict" +
			" I, " + _myGuid + 
			" WILL BECOME UNINIT FOR " +
			otherguid + " AND WILL SEND" +
			" MYSELF THE MESSAGE AGAIN");
	    }
	    info.state = PartyInfo.UNINIT;
	    // send the message back to myself
	    dispatch (dhmsg);
	    //JUMP SinkFullException
	}
	if (compare == 0) 
	    Carp.die ("Attempting to handshake with self! me:"
		    + _myGuid + " them: " + otherguid);
	if (ANNOUNCE && compare <= 0) {
	    Debug.printtagln (tag,"HMACDHH: DUAL-INIT conflict" +
		    " I, " + _myGuid + 
		    " WILL IGNORE THIS MESSAGE FOR " +
		    otherguid);
	}
    }

    synchronized protected void handle_state_waitforfin (
	    DHMessage dhmsg, int msgType, SecureHash otherguid, 
	    PartyInfo info) {

	if (msgType != DHMessage.FINALIZE) {
	    // TODO
	}

	//System.out.println ("FINIALIZE while WAITFOR FIN");
	//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
	// Conditions:
	// Condition (4.abc)
	if (dhmsg.QSkey != null ||
		dhmsg.sigkey == null ||
		!otherguid.verify (dhmsg.sigkey) ||
		dhmsg.mac == null ) {
	    if (DEBUG) Debug.printtagln (tag, "HMACDHHandler: " +
		    "message ignored. :"  +
		    _myGuid);
	    // ignore. (TODO write error to log file)
	    return;    
	}        
	//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

	// (1) Store the public signature key
	info.sigpub = dhmsg.sigkey;

	// (2) check dhmsg.verify
	try {
	    if (!checkDHVerify (dhmsg.user_data, dhmsg.mac, dhmsg, info)){
		//JUMP InvalidKeyException
		//JUMP SignatureException
		//JUMP TypeTable.NoSuchTypeCode
		// ignore (TODO write error to log file)
		if (DEBUG) Debug.printtagln (tag, "HMACDHHandler: " +
			_myGuid +
			"dhverify failed.");
		if (DEBUG) {
		    Debug.printtagln (tag, _myGuid.toString ());
		    info.print ();
		    return;
		}
	    }
	}
	catch (InvalidKeyException e) {
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS)
		Carp.out (tag + ": invalid key exception.");
	    info.state = PartyInfo.WAITFORFINAL;
	    return;
	}
	catch (SignatureException e) {
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS)
		Carp.out (tag + ": signature exception.");
	    return;
	}
	catch (TypeTable.NoSuchTypeCode e) {
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS)
		Carp.out (tag + ":nosuchtypecode exception.");
	    return;
	}

	//System.out.println ("FINIALIZE checked out");

	// (3) change state to ESTABLISHED
	info.state = PartyInfo.ESTABLISHED;
	//System.out.println ("FINIALIZE established");

	if (ANNOUNCE) 
	    Debug.printtagln (tag,"HMACDHHandler: established by " +
		    _myGuid + "(me) with " +
		    otherguid + "(them)");

	// (4) move info to establishedParties and
	//       remove info from pendingParties 
	synchronized (establishedParties) {
	    _classifier.cancel (info.timerToken);
	    info.timerToken = null;
	    if (establishedParties.contains (otherguid)) {
		PartyInfo old =
		    (PartyInfo) establishedParties.get (otherguid);
		info.usedKeys.addAll (old.usedKeys);
	    }
	    pendingParties.remove (otherguid);
	    establishedParties.put (otherguid, info);
	    sendEstablished (otherguid);
	}
    }

    synchronized public void handle_dh_message (DHMessage dhmsg) {
	//=============================================
	// Prep dhmsg for Stage evaluation
	//=============================================
	if (SEAN_DEBUG) 
	    Debug.printtagln (tag, "got " + dhmsg);

	if (DEBUG) {
	    Debug.printtagln (tag, "HMDHH" + _myGuid);
	    dhmsg.print ();
	}

	// Condition (4.b <Prep>) and to send out message: extract guid
	SecureHash otherguid = dhmsg.peer;

	// Condition (1) record msgType
	int msgType = dhmsg.messageType;

	// Condition (2) grab and verify the MACType
	// drop the message if MACType not identified
	if (dhmsg.MACType == null) return;
	// drop the message if MACType incorrect (not MACTYPE_HMACMD5)
	if (TypeTable.type_code (dhmsg.MACType) != MACTYPE_HMACMD5) return;

	// Condition (3) get associated PartyInfo
	PartyInfo info;
	if (!pendingParties.containsKey (otherguid)) {
	    if (msgType == DHMessage.INITIALIZE) {
		// if the PartyInfo entry doesn't exist and we have an 
		// appropriate message, create it.
		PartyInfo newPending = new PartyInfo ();
		pendingParties.put (otherguid, newPending);
		info = newPending;
	    }
	    else return; // no partyinfo entry means nothing can happen
	}
	else {
	    info = (PartyInfo) pendingParties.get (otherguid);
	}
	if (DEBUG) {
	    Debug.printtagln (tag, "HMDHH" + _myGuid);
	    info.print ();
	}

	if (SEAN_DEBUG) 
	    Debug.printtagln (tag, "initial state=" + info);

	// Conditions satisfied at this point:
	// (1) msgType retreived 
	// (2) MACType retreived and == MACTYPE_HMACMD5  
	// (3) info is a non-null PartyInfo entry linked w/ the sender guid
	// Conditions needed to be satisfied:
	// (4) the rest of dhmsg need to be checked (at least non-null) 
	//        according to the state:
	//        (a)QSkey, (b)sigkey, (c)verify

	// State evaluation: get the state 
	int state = info.state;
	// =================================================
	if (state == PartyInfo.UNINIT) {
	    handle_state_uninit (dhmsg, msgType, otherguid, info);
	}
	else if (state == PartyInfo.WAITFORRESPONSE) {
	    if (msgType == DHMessage.RESPOND) 
		handle_state_waitforresp_resp (dhmsg, otherguid, info);
	    else if (msgType == DHMessage.INITIALIZE) 
		handle_state_waitforresp_init (dhmsg, otherguid, info);
	    else {
		// TODO
	    }
	}
	else if (state == PartyInfo.WAITFORFINAL) {
	    handle_state_waitforfin (dhmsg, msgType, otherguid, info);
	}

	if (SEAN_DEBUG) 
	    Debug.printtagln (tag, "final state=" + info);
    }

    synchronized public void handleEvent (QueueElementIF item)
    throws EventHandlerException {

	if (SEAN_DEBUG) Debug.printtagln (tag, "got event " + item + ".");

	if (item instanceof TellInit) {
	    handle_init ((TellInit) item);
	}
	else if (item instanceof TellStsTimeout) {
	    handle_sts_timeout ((TellStsTimeout) item);
	} 
	else if (item instanceof TellDhTimeout) {
	    handle_dh_timeout ((TellDhTimeout) item);
	}
	else if (item instanceof DHMessage) {
	    handle_dh_message ((DHMessage) item);
	}
	else if (item instanceof MacRequest) {
	    MacRequest req = (MacRequest) item;
	    QSByteArray mac = null;
	    try {
		mac = doMAC (req.data_to_mac, req.peer);
	    }
	    catch (MacHandler.GuidNotEstablished e) {
		if (SEAN_DEBUG) Debug.printtagln (tag, 
			"No secret key established with " +
			req.peer + " yet.  Enqueuing request.");

		boolean start_agree = false;
		LinkedList l = (LinkedList) _msgs_needing_macs.get (req.peer);
		if (l == null) {
		    if (SEAN_DEBUG) Debug.printtagln (tag, 
			    "Starting a key agreement with " + req.peer + ".");
		    l = new LinkedList ();
		    _msgs_needing_macs.put (req.peer, l);
		    start_agree = true;
		}
		else if (SEAN_DEBUG) Debug.printtagln (tag, 
			"Key agreement already started with "
			+ req.peer + ".");
		l.addLast (req);

		if (start_agree) 
		    establish (req.peer);

		return;
	    }

	    dispatch (new MacResponse (req.id, mac));
	}
	else {
	    BUG (tag + ": unknown event " + item);
	}
    }

    private void sendEstablished (SecureHash otherguid) {
	LinkedList l = (LinkedList) _msgs_needing_macs.remove (otherguid);
	if (l != null) {
	    while (! l.isEmpty ()) {
		MacRequest req = (MacRequest) l.removeFirst ();
		QSByteArray mac = null;
		try {
		    mac = doMAC (req.data_to_mac, otherguid);
		}
		catch (MacHandler.GuidNotEstablished e) {
		    Carp.die (tag + ": BUG");
		}
		dispatch (new MacResponse (req.id, mac));
	    }
	}
	dispatch  (new KeyEstablished (otherguid));
    }

    private Map _msgs_needing_macs = new HashMap ();

    synchronized public void handleEvents (QueueElementIF item[])
    throws EventHandlerException {
	for (int i = 0; i < item.length; ++i)
	    handleEvent (item[i]);
    }

    synchronized protected boolean checkDHVerify (QuickSerializable user_data,
	    QSByteArray mac, DHMessage dhmsg, PartyInfo info) 
    throws InvalidKeyException, SignatureException, 
    TypeTable.NoSuchTypeCode {

	boolean allzgood;
	SignedQS s = null;
	STSVerify recreatedSTS;
	byte[] recreated, extracted;

	// 1. check MAC
	_hmac.init (info.secret);
	MacBuffer buf = new MacBuffer (_hmac);
	buf.add (user_data);
	allzgood = ByteUtils.equals (_hmac.doFinal (), mac.bytes ());

	if (allzgood) {
	    // extract signature
	    s = (SignedQS) user_data;

	    // verify Signature
	    // next line: workaround cryptix bug
	    try {reInitSigEngine ();} catch (Exception e) {Carp.die (e);}
	    allzgood = s.verify (info.sigpub.key (), _signEngine);
	}
	if (DEBUG) Debug.printtagln (tag, "checkDH: sig. allzgood = " + 
		allzgood + " " + _myGuid);

	// 2. check the HASH
	if (allzgood) {
	    // recreate STS (as if they would have done, i.e.
	    // "my Public + other Public + MACType"
	    // becomes: 
	    // "other Public + my Public + MACType"
	    recreatedSTS = new STSVerify (new QSDHPublicKey (info.dhpubOther), 
		    new QSDHPublicKey (info.dhpub));

	    // verify the HASH is equivalent
	    allzgood = ((SHA1Hash)s.user_data ()).verify (recreatedSTS);
	}

	if (DEBUG) Debug.printtagln (tag, "checkDH: sig & hash allzgood = " + 
		allzgood + " " + _myGuid);
	return allzgood;

    }

    synchronized protected void makeDHVerify (QSDHPublicKey myPu, 
	    QSDHPublicKey otherPu, 
	    SecretKey macSecret,
	    DHMessage destination ) 
    throws SignatureException, InvalidKeyException {

	// (1) create STSVerify info my Public + other Public + MACType
	STSVerify sts = new STSVerify (myPu, otherPu);
	// (2) Hash the info
	SecureHash digested = new SHA1Hash (sts);
	// (3) Sign the Hash
	// next line: workaround cryptix bug
	try {reInitSigEngine ();} catch (Exception e) {Carp.die (e);}
	SignedQS sig = new SignedQS (digested, 
		new QSSha1RsaSignature (),
		_sigPri.key (),
		_r,
		_signEngine);
	// (4) MAC the Signature (the SHA1Hash)
	_hmac.reset ();
	_hmac.init (macSecret);
	MacBuffer buf = new MacBuffer (_hmac);
	buf.add (sig);
	QSByteArray mac = new QSByteArray (_hmac.doFinal ());
	// (5) insert into destination
	destination.user_data = sig;
	destination.mac = mac;
    }

    synchronized protected void makeNewDHKeys (PartyInfo info) {
	// TODO in the future, we could create a keygenerator with
	// functional access that we could simple grab a ready made
	// key. (might save some time, this can take up to 8 secs)

	KeyPair pair = _kpgen.generateKeyPair ();
	info.dhpub = (DHPublicKey) pair.getPublic ();
	info.dhpri = (DHPrivateKey) pair.getPrivate ();
    }

    synchronized public void init (ConfigDataIF config) throws Exception {

	if (DEBUG) 
	    Debug.printtagln (tag, "HMACDHHandler initializing.------------");

	//After the new type code, registerAllTypes () necessary anymore?
	//I'm referring to the switch from int type codes to long hash type code
	//Types.registerAllTypes (); 

	_classifier = Classifier.getClassifier (
		new NodeId (config.getString ("node_id")));


	String stageName = config.getStage ().getName ();
	try 
	{
	    if (config.getBoolean ("Profile"))
	    {
		config.getManager ().getProfiler ().add (
			"ostore.mac.HMACDHHandler", this);
		SinkIF sink =
		    config.getManager ().getStage (stageName).getSink ();
		_this_sink = new SinkProxy (sink,
			config.getManager (),
			config.getStage ().getWrapper ());
	    }
	    else
		_this_sink =
		    config.getManager ().getStage (stageName).getSink ();
	} 
	catch (NoSuchStageException e) 
	{
	    Carp.die (stageName + ".init: could not get own sink.");
	}


	// sts timeout length (milliseconds)
	_stsTime = config.getInt ("timeoutlength");
	if (ANNOUNCE) System.out.println ("timeout length= " + _stsTime);

	// dh key lifespan (milliseconds)
	_dhlifespan = config.getInt ("dhlifespan");

	// get sts retry limit
	_retryLimit = config.getInt ("sessionretries");

	// get my signature key filenames
	_pkeyfile = config.getString ("pkey");
	_skeyfile = config.getString ("skey");

	// are we running the selfTest? (warning this will system.exit
	// on completion.)
	_selfTest = config.getBoolean ("selftest");

	// Making sure that the type_codes for the types we require are
	// actually registered.
	TypeTable.register_type ("ostore.mac.DHMessage" );
	TypeTable.register_type ("ostore.mac.QSHMACMD5" );

	MACTYPE_HMACMD5 = TypeTable.type_code ("ostore.mac.QSHMACMD5");

	// subscribe ourselves to certain types of messages
	// DHMessages (inbound)
	Filter filter = new Filter ();
	if (! filter.requireType (Class.forName
		    ("ostore.mac.DHMessage")))
	    Carp.die ("HMACDHHandler: could not require type.");
	if (! filter.requireValue ("inbound", new Boolean (true)))
	    Carp.die ("HMACDHHandler: could not require inbound.");
	_classifier.subscribe (filter, _this_sink);

	filter = new Filter ();
	if (! filter.requireType (Class.forName (
			"ostore.mac.HMACDHHandler$TellInit")))
	    Carp.die ("HMACDHHandler: could not require type.");
	_classifier.subscribe (filter, _this_sink);

	filter = new Filter ();
	if (! filter.requireType (Class.forName (
			"ostore.mac.HMACDHHandler$TellStsTimeout")))
	    Carp.die ("HMACDHHandler: could not require type.");
	_classifier.subscribe (filter, _this_sink);

	filter = new Filter ();
	if (! filter.requireType (Class.forName (
			"ostore.mac.HMACDHHandler$TellDhTimeout")))
	    Carp.die ("HMACDHHandler: could not require type.");
	_classifier.subscribe (filter, _this_sink);

	filter = new Filter ();
	if (! filter.requireType (Class.forName (
			"ostore.mac.MacRequest")))
	    Carp.die ("HMACDHHandler: could not require type.");
	_classifier.subscribe (filter, _this_sink);

	try {
	    if (DEBUG || ANNOUNCE) Debug.turnOn (tag);
	    initProviders ();

	    // Init key gen engines
	    _kpgen = KeyPairGenerator.getInstance ("DH","CryptixCrypto");
	    _kpgen.initialize (_keysize, _r);
	    _kagree = KeyAgreement.getInstance ("DH", "CryptixCrypto");

	    initSigEngine ();

	    // Initalize the Mac Engine
	    _hmac = Mac.getInstance ("HMAC-MD5", "CryptixCrypto");

	    // TODO for some reason, skipping the selfTest breaks
	    // things...?? ....  2/1/2002 -- is this true anymore?...

	    //if (_selfTest) 
	    selfTest (); 	
	}
	catch (Exception e) {
	    e.printStackTrace ();
	    Carp.die (e);
	}

	// for getInstance
	instances.put (_myGuid, this);
	if (ANNOUNCE) Debug.printtagln (tag,"INSTANCE REGISTERED:" + _myGuid);

	// begin DH key life countdown
	_classifier.dispatch_later (new TellDhTimeout (), _dhlifespan);

	if (DEBUG) Debug.printtagln (tag, 
		"HMACDHHandler Initiated.---------------");

    }

    protected long MACTYPE_HMACMD5;


    /**
     * Return the HMACDHHandler associated with a particular guid. <br></br> 
     * 
     */
    synchronized public static HMACDHHandler getInstance (SecureHash guid) { 
	if (ANNOUNCE) Debug.printtagln (tag,"INSTANCE REQUESTED:" + guid);
	Object r = instances.get (guid);
	if (r != null)
	    return (HMACDHHandler) r;
	else
	    return null;
    }

    /**
     *  Establish a connection between the guid and this instance of sandstorm
     * <br></br> 
     * Once a connection is established, the MacHandler will broadcast
     * a {@link MacHandler.KeyEstablished} announcement.
     */
    synchronized public void establish (SecureHash guid) {
	// send a message back to myself through to the eventhandler
	// to initiate a DH handshake (STS) with this guid.
	dispatch (new TellInit (guid));
    }

    /**
     * Note: a failure may mean that a message was sent right before
     * the sender decided to change keys that that message arrived 
     * after a new key was established (though this is unlikely) it
     * is a possibility that needs to be accounted for.
     * @return <code>true</code> iff this message is authentic, i.e.
     * verified by the internal MAC
     */
    synchronized public boolean authenticate (QuickSerializable user_data, 
	    QSByteArray incoming_mac, SecureHash guid)
    throws GuidNotEstablished {

	SecretKey secret;
	int state;
	byte[] computed_mac;

	// get secret or throw GuidNotEstablished
	// since the get operation is synchronized, that's all we
	// need for synchronization here.
	PartyInfo info = (PartyInfo) establishedParties.get (guid);
	if (info == null) 
	    throw new GuidNotEstablished (guid.toString ());
	state = info.state;

	if (state != PartyInfo.ESTABLISHED) 
	    throw new GuidNotEstablished (guid.toString ());
	secret = info.secret;

	// init engine and do mac
	try {
	    _hmac.init (secret);
	    MacBuffer buf = new MacBuffer (_hmac);
	    buf.add (user_data);
	    computed_mac = _hmac.doFinal ();
	}
	catch (InvalidKeyException e) {
	    //TODO does this mean the key simply expired?
	    Carp.out ("STRANGE THINGS: HMACDHHandler: InvalidKeyException" +
		    "while authenticating");
	    return false;
	}

	// compare and return result
	return ByteUtils.equals (computed_mac, incoming_mac.bytes ());
    }

    /**
     * Computes the MAC code for the this message
     * @return <code>true</code> iff the MAC was computed successfully.
     */
    synchronized public QSByteArray doMAC (QuickSerializable user_data, SecureHash guid)
    throws GuidNotEstablished {

	SecretKey secret;
	int state;
	byte[] mac = null;

	// get secret or throw GuidNotEstablished
	PartyInfo info = (PartyInfo) establishedParties.get (guid);
	if (info == null) 
	    throw new GuidNotEstablished (guid.toString ());
	secret = info.secret;

	// init engine and do mac
	try {
	    _hmac.init (secret);
	    MacBuffer buf = new MacBuffer (_hmac);
	    buf.add (user_data);
	    mac = _hmac.doFinal ();
	}
	catch (InvalidKeyException e) {
	    //TODO does this mean the key simply expired?
	    Carp.out ("STRANGE THINGS: HMACDHHandler: InvalidKeyException" +
		    "while doingMAC");
	}

	return new QSByteArray (mac);
    }

    public String idString () { return "HMAC-DH Mac Handler"; }

    public SecureHash myGuid () { return _myGuid; }

    synchronized void initProviders ()
    throws NoSuchProviderException, NoSuchAlgorithmException {    

	if (DEBUG) Debug.printtagln (tag, "Adding Provider: CryptixJCE");
	Provider cryptixJCE = new cryptix.jce.provider.CryptixCrypto ();
	Security.addProvider (cryptixJCE);
	//cryptixJCE.list (System.out);

	if (DEBUG) Debug.printtagln (tag, "Adding Provider: CryptixRandom");
	Provider cryptixRandom = new cryptix.jce.provider.CryptixRandom ();
	Security.addProvider (cryptixRandom);
	//cryptixRandom.list (System.out);

	// Initialize

	// Initialize SecureRandom (Random number generator)
	boolean available = DevRandom.isAvailable ();
	if (DEBUG) 
	    Debug.printtagln (tag, "DevRandom.isAvailable = " + available);
	// Note that we are using DevRandom from CryptixRandom
	// which pulls their numbers from /usr/random
	// which, by the way, is butt fast.
	_r = SecureRandom.getInstance ("DevRandom","CryptixRandom");
    }

    synchronized void initSigEngine () 
    throws NoSuchProviderException, NoSuchAlgorithmException,
    InvalidKeyException, QSException {

	if (ostore.security.NativeRSAAlgorithm.available) {
	    Security.addProvider (new ostore.security.NativeRSAProvider ());
	    _signEngine = Signature.getInstance (
		    "SHA-1/RSA/PKCS#1", "NativeRSA");
	}
	else {
	    _signEngine = Signature.getInstance (
		    "SHA-1/RSA/PKCS#1", "CryptixCrypto");
	}

	//TODO should we be grabbing this from some global config for
	// this OceanStore instance? we ought to have a EncryptedQSIO

	// Grab keys from _pkeyfile and _skeyfile
	try {
	    _sigPub = (QSPublicKey) QSIO.read (_pkeyfile);
	    _sigPri = (QSPrivateKey) QSIO.read (_skeyfile);
	}
	catch (IOException e) {
	    Carp.die ("HMACDHHandler: IO Exception reading " +
		    _pkeyfile + " or " + _skeyfile );
	}
	_signEngine.initSign (_sigPri.key (), _r);

	// get my guid
	_myGuid = new SHA1Hash (_sigPub);

	if (ANNOUNCE) 
	    Debug.printtagln (tag,"HMACDH: My Guid is " + _myGuid);
    }

    synchronized void reInitSigEngine ()
    throws NoSuchProviderException, NoSuchAlgorithmException,
    InvalidKeyException {

	if (ostore.security.NativeRSAAlgorithm.available) {
	    Security.addProvider (new ostore.security.NativeRSAProvider ());
	    _signEngine = Signature.getInstance (
		    "SHA-1/RSA/PKCS#1", "NativeRSA");
	}
	else
	    _signEngine = Signature.getInstance (
		    "SHA-1/RSA/PKCS#1", "CryptixCrypto");
    }

    synchronized void selfTest () {

	if (DEBUG) 
	    Debug.printtagln (tag, "STARTING SELFTEST ---------------------");

	// testing timeoutlength ... and dhlifespan maybe 
	if (DEBUG) 
	    Debug.printtagln (tag, "testing timing-------------------------");
	// timeoutlength
	// TODO TODO TODO 



	PartyInfo AlfInfo = new PartyInfo ();
	PartyInfo BigFootInfo = new PartyInfo ();

	// create DHKeys for Alf 
	if (DEBUG) 
	    Debug.printtagln (tag, "(1) Alf: create a set of dh keys of size " + 
		    _keysize + " ...");
	makeNewDHKeys (AlfInfo);
	if (DEBUG) 
	    Debug.printtagln (tag, "done.");

	// create DHKeys for Bigfoot 
	if (DEBUG) 
	    Debug.printtagln (tag, "(2) Bigfoot: create a set of dh " +
		    "keys of size " + _keysize + " ...");
	makeNewDHKeys (BigFootInfo);
	if (DEBUG) 
	    Debug.printtagln (tag, "done.");

	// get signature keys from quick.pri and quick.pub
	// we did this already and got: _sigPub and _sigPri
	// testing w/ two DIFFERENT sig keys is done in StageTester.java

	// generate secret
	if (DEBUG) 
	    Debug.printtagln (tag, "(3) Alf: generate shared secret ...");
	try {
	    _kagree.init (AlfInfo.dhpri); //JUMP InvalidKeyException
	    _kagree.doPhase (BigFootInfo.dhpub, true);
	    byte[] seed = _kagree.generateSecret ();
	    AlfInfo.secret = new RawSecretKey ("DH seed", seed);
	}
	catch (InvalidKeyException e) {
	    Carp.die (e);
	}
	if (DEBUG) 
	    Debug.printtagln (tag, "done.");

	// create test DHMessage 
	if (DEBUG) Debug.printtagln (tag, 
		"(4) Alf: create test DHMessage (RESPOND) ...");
	DHMessage dhmTest = new DHMessage (new SHA1Hash (),     //toWhom
		AlfInfo.dhpub,     //DH Public key
		_sigPub,            //Sign Pub key
		new QSHMACMD5(),    //mac type
		null, null,         //verification
		DHMessage.RESPOND); // type 

	// **** makeDHVerify 
	try {
	    makeDHVerify (new QSDHPublicKey (AlfInfo.dhpub), 
		    new QSDHPublicKey (BigFootInfo.dhpub), 
		    AlfInfo.secret, 
		    dhmTest); 
	    // JUMP SignatureException 
	    // JUMP InvalidKeyException
	}
	catch (SignatureException e) { 
	    Carp.die (e);
	}
	catch (InvalidKeyException e) {
	    Carp.die (e);
	}
	if (DEBUG) 
	    Debug.printtagln (tag, "done.");
	// encode DHMessage
	if (DEBUG) 
	    Debug.printtagln (tag, "(5) Alf: qs_cloning DHMessage...");
	dhmTest = (DHMessage) ostore.util.QSClone.qs_clone (dhmTest);
	if (DEBUG) 
	    Debug.printtagln (tag, "done.");
	// BigFoot figures out secret key.
	if (DEBUG) 
	    Debug.printtagln (tag, "(7) BigFoot: figure out secret...");
	try {
	    _kagree.init (BigFootInfo.dhpri); // JUMP InvalidKeyException
	    _kagree.doPhase (dhmTest.QSkey.key (), true );
	    byte[] seed = _kagree.generateSecret ();
	    BigFootInfo.secret = new RawSecretKey ("DH seed", seed);
	}
	catch (InvalidKeyException e) {
	    Carp.die (e);
	}
	BigFootInfo.dhpubOther = (DHPublicKey)dhmTest.QSkey.key ();
	BigFootInfo.sigpub = dhmTest.sigkey;
	if (DEBUG) 
	    Debug.printtagln (tag, "done.");
	// **** checkDHVerify
	if (DEBUG) 
	    Debug.printtagln (tag, "(8) BigFoot: checkDHVerify...");
	try {
	    boolean check = checkDHVerify (dhmTest.user_data, dhmTest.mac, 
		    dhmTest, 
		    BigFootInfo); 
	    // JUMP InvalidKeyException
	    // JUMP SignatureException
	    // JUMP TypeTable.NoSuchTypeCode
	    if (DEBUG) 
		Debug.printtagln (tag, "[returns:" + check + "]...");
	    if (!check) {
		if (DEBUG) 
		    Debug.printtagln (tag, "checkDHVerify FAILED");
		Carp.die  ();
	    }
	}
	catch (InvalidKeyException e) {
	    Carp.die (e);
	}
	catch (SignatureException e) {
	    Carp.die (e);
	}
	catch (TypeTable.NoSuchTypeCode e) {
	    Carp.die (e);
	}   
	if (DEBUG) 
	    Debug.printtagln (tag, "done.");
	if (DEBUG) 
	    Debug.printtagln (tag, "SELFTEST DONE----------------------");
	if (_selfTest) System.exit (0);
    }

    public void destroy () throws Exception {}

    public int profileSize () {
	return _this_sink.size ();
    }

    // Alarm events
    protected class TellInit implements QueueElementIF {
	public SecureHash guid;
	public TellInit (SecureHash g) { guid = g; }
	public String toString () {
	    return "(TellInit guid=" + guid + ")";
	}
    }

    protected class TellStsTimeout implements QueueElementIF {
	public SecureHash guid;
	public TellStsTimeout (SecureHash g) { guid = g; }
	public String toString () {
	    return "(TellStsTimeout guid=" + guid + ")";
	}
    }

    protected class TellDhTimeout implements QueueElementIF {}

    protected static final void BUG (String msg) {
	Carp.die (msg + "  This is a bug in ostore.mac.HMACDHHandler.  " +
		"Please let Sean know about it ASAP.");
    }

    synchronized protected void dispatch (QueueElementIF item) {
	try {
	    if (SEAN_DEBUG) Debug.printtagln (tag, "dispatching " + item + ".");
	    _classifier.dispatch (item);
	}
	catch (SinkException e) {
	    BUG (tag + ".dispatch: could not dispatch " + item);
	}
    }
}

