/**
 * Party Info
 *
 * @author    Gifford Cheung
 * @version     $Id: PartyInfo.java,v 1.7 2002/09/09 05:04:15 srhea Exp $
 */

package ostore.mac;
import ostore.util.*;
import ostore.security.*;
import java.util.*;

// Security stuff
import java.security.*;
import java.security.spec.*;
import javax.crypto.*;
import javax.crypto.spec.*;
import javax.crypto.interfaces.*;
import cryptix.jce.provider.*;
import cryptix.jce.provider.key.*;
import cryptix.jce.provider.random.*;

/**
 * Party information for DHMessage and MacStage to keep track
 * of what stage of STS is established between this computer
 * and another party <br></br>
 * This is tailored specifically for internal {@link HMACDHHandler} usage
 */
public class PartyInfo {

    // CONSTANT DEFINITIONS
    /**
     * Uninitiated State.
     */
    public static int UNINIT      = 0;
    /**
     * Waiting for DHMessage.RESPOND from other party
     */
    public static int WAITFORRESPONSE = 1;
    /**
     * Waiting for DHMessage.FINALIZE from other party
     */
    public static int WAITFORFINAL    = 2;
    /**
     * A connection has been established and 
     * <tt>secret</tt> ready for use.
     */
    public static int ESTABLISHED     = 3;

    // VARIABLES
    /**
     * My public Diffie-Hellman key
     */
    public DHPublicKey dhpub; // my key
    /**
     * My private Diffie-Hellman key
     */
    public DHPrivateKey dhpri; // my key
    /**
     * Our little secret (generated from the DH handshake, acts as seed
     * for other shared secrets, in this case the HMAC)
     */
    public SecretKey secret;
    /**
     * The other party's  public Diffie-Hellman key
     */
    public DHPublicKey dhpubOther; // their key
    /**
     * Their public signature key
     */
    public QSPublicKey sigpub; // their public signature key
    /**
     * The my current state during the handshake
     */
    public int state;
    /**
     * A counter for number of restarts during a handshake session.
     * An implementing MacHandler will decide what a reasonable ceiling
     * for this is. Should be set to zero upon a sucessful handshake session.
     */
    public int sessionRetries;

    /**
     * Timer token for use with session timeouts.
     */
    public Object timerToken;

    /**
     * Keeps track of DH keys that have been used with this DH key pair
     * as a protection against replay attacks.
     *
     * TODO: if too big need to stop talking to this person. threshold
     * at 100, then we'd have to worry about an attacker changing their
     * guid over and over again to fill up memory space. DDOS attack.
     */
    public Vector usedKeys;
     

    PartyInfo( int State ) {
        state = State;
        usedKeys = new Vector();
    }

    PartyInfo() {
        state = UNINIT;
        usedKeys = new Vector();
    }

    public String toString () {
	String result = "(PartyInfo state=";
	switch (state) {
	    case 0: result += "UNINIT";          break;
	    case 1: result += "WAITFORRESPONSE"; break;
	    case 2: result += "WAITFORFINAL";    break;
	    case 3: result += "ESTABLISHED";     break;
	}
	result += ", retries=" + sessionRetries + ")";
	return result;
    }

    public void print() {
	System.out.println (toString ());
    }

    /*
    public void print() {
        byte[] temp;
        System.out.println("Printing PartyInfo");

        // dhpub
        if (dhpub != null) {
            temp = dhpub.getY().toByteArray();
            System.out.println("(+) My QS Public Diffie Hellman: " +
                                ByteUtils.print_bytes(temp, 0, temp.length) );
        } else {
            System.out.println("(+) My QS Public Diffie Hellman: null");
        }

        // dhpri
        if (dhpri != null) {
            temp = dhpri.getX().toByteArray();
            System.out.println("(+) My QS Private Diffie Hellman: " +
                                ByteUtils.print_bytes(temp, 0, temp.length) );
        } else {
            System.out.println("(+) My QS Private Diffie Hellman: null");
        }

        // secret     
        if (secret != null) {
            System.err.println("(+) Secret Key: " + secret);
            System.err.println("(+) Secret Key: " +
                secret.getEncoded().length + " bytes " +
                "Format: " + secret.getFormat() + " " +
                ByteUtils.print_bytes(secret.getEncoded(), 
                    0, secret.getEncoded().length));
        } else {
            System.out.println("(+) Secret Key: null");
        }

		// dhpubOther
        if (dhpubOther != null) {
            temp = dhpubOther.getY().toByteArray();
            System.out.println("(+) Other's QS Public Diffie Hellman: " +
                                ByteUtils.print_bytes(temp, 0, temp.length) );
        } else {
            System.out.println("(+) Other's QS Public Diffie Hellman: null");
        }

        // sigpub
        if (sigpub != null) {
            System.out.println("(+) Other's public sig: " +
                                new SHA1Hash (sigpub));
        } else {
            System.out.println("(+) Other's public sig: null");
        }

        // state
        System.out.print("(+) State: ");
        switch (state) {
            case 0:  System.out.println("UNINIT"); break;
            case 1:  System.out.println("WAITFORRESPONSE"); break;
            case 2:  System.out.println("WAITFORFINAL"); break;
            case 3:  System.out.println("ESTABLISHED"); break;
        }
        
        // sesstionRetries
        System.out.println("(+) sessionRetries: " + sessionRetries);
   
        // timerToken
        System.out.println("(+) timerToken: " + timerToken);    
    }
    */
        
}
