/**
 * Copyright (c) 2001 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.network;
import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StreamTokenizer;
import java.net.InetAddress;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import ostore.dispatch.Signal;
import ostore.util.AssertionViolatedException;
import ostore.util.ByteUtils;
import ostore.util.Carp;
import ostore.util.CountBuffer;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.util.InputBuffer;
import ostore.util.NodeId;
import ostore.util.Pair;
import ostore.util.PriorityQueue;
import ostore.util.QSClone;
import ostore.util.Queue;
import ostore.util.QuickSerializable;
import ostore.util.QuickSerializable;
import ostore.util.TypeTable;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.NoSuchStageException;
import seda.sandStorm.api.NoSuchStageException;
import seda.sandStorm.api.ProfilableIF;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SingleThreadedEventHandlerIF;
import seda.sandStorm.api.SinkClosedEvent;
import seda.sandStorm.api.SinkClosedException;
import seda.sandStorm.api.SinkException;
import seda.sandStorm.api.SinkFlushedEvent;
import seda.sandStorm.api.SinkFullException;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.core.BufferElement;
import seda.sandStorm.internal.SinkProxy;
import seda.sandStorm.lib.aSocket.ATcpClientSocket;
import seda.sandStorm.lib.aSocket.ATcpConnectFailedEvent;
import seda.sandStorm.lib.aSocket.ATcpConnection;
import seda.sandStorm.lib.aSocket.ATcpInPacket;
import seda.sandStorm.lib.aSocket.ATcpListenFailedEvent;
import seda.sandStorm.lib.aSocket.ATcpListenSuccessEvent;
import seda.sandStorm.lib.aSocket.ATcpServerSocket;
import seda.sandStorm.lib.aSocket.aSocketInputStream;
import soss.network.TransitStub;

/**
 * A stage that supports multiple virtual OceanStore nodes in a single JVM
 * instance.  A single instance of this stage can be used to provide
 * network support for N virtual OceanStore nodes, in place of using N
 * separtate ostore.network.Network stages to acheive the same goal.
 *
 * Lock Dependency Graph:
 * <ul>
 * <item> nodeid_to_connstate precedes conn_to_connstate
 * <item> conn_to_connstate precedes mystery_connections
 * <item> nodeid_to_connstate precedes connstate
 * <item> conn_to_connstate precedes connstate
 * <item> connstate precedes close_me
 * </ul>
 *
 * @author     Sean C. Rhea
 * @version    $Id: MultiplexedNetwork.java,v 1.8 2002/09/08 03:33:33 srhea Exp $
 */
public class MultiplexedNetwork implements EventHandlerIF, ProfilableIF {

    protected String TAG = "MNetwork";
    protected static final boolean DEBUG = false;
    protected static final boolean DEBUG_SERIALIZATION = false;

    /**
     * All the state we need about a connection to another node.
     */
    protected class ConnectionState {
	
	protected long time_established;
	protected ATcpConnection client_conn, server_conn;
	protected ConnectionStream client_is, server_is;
	protected boolean client_closed, server_closed;
	protected Set ids_represented;
	protected NodeId lowest_id_represented;
	protected LinkedList pending_outb_msgs;

	protected synchronized ConnectionStream stream (ATcpConnection c) {
	    if (c == client_conn)
		return client_is;
	    else if (c == server_conn)
		return server_is;
	    else {
		BUG ("doesn't match either stream");
		return null;
	    }
	}

	protected synchronized ATcpConnection preferred_outb_conn () {
	    if (client_conn == null)
		return server_conn;
	    else if (server_conn == null)
		return client_conn;
	    else if (ids_represented == null) 
		return client_conn;
	    else {
		// Choose between them.

		if (lowest_id_represented == null) {
		    Iterator i = ids_represented.iterator ();
		    lowest_id_represented = (NodeId) i.next ();
		    while (i.hasNext ()) {
			NodeId n = (NodeId) i.next ();
			if (n.less_than (lowest_id_represented))
			    lowest_id_represented = n;
		    }
		    if (DEBUG) Debug.printtagln (TAG,
			    "Lowest id this conn is " +
			    lowest_id_represented + ".");
		}

		long close_time = System.currentTimeMillis () + 10*1000;
		if (lowest_nodeid.less_than (lowest_id_represented)) {
		    if (DEBUG) Debug.printtagln (TAG,
			    "We are the server.  Closing the client.");
		    close_client ();
		    return server_conn;
		}
		else {
		    if (DEBUG) Debug.printtagln (TAG,
			    "We are the client.  Closing the server.");
		    close_server ();
		    return client_conn;
		}
	    }
	}

	protected synchronized void close_client () {
	    long close_time = System.currentTimeMillis () + 10*1000;
	    if ((client_conn != null) && (! client_closed)) {
		synchronized (close_me) {
		    close_me.insert (client_conn, close_time, false);
		    client_closed = true;
		}
	    }
	}

	protected synchronized void close_server () {
	    long close_time = System.currentTimeMillis () + 10*1000;
	    if ((server_conn != null) && (! server_closed)) {
		synchronized (close_me) {
		    close_me.insert (server_conn, close_time, false);
		    server_closed = true;
		}
	    }
	}
    }

    /**
     * Maps NodeIds to the Classifiers within this JVM.
     */
    protected Map nodeid_to_classifier;

    /**
     * A list of NodeIds within this JVM.
     */
    protected LinkedList local_nodeids = new LinkedList ();

    /**
     * The Classifier for node_id_0, used to dispatch DelayWrappers to.
     */
    protected Classifier delay_classifier;

    /**
     * The NodeId for node_id_0, used to get Alarms on.
     */
    protected NodeId lowest_nodeid;

    /** 
     * Maps ATcpConnection objects to ConnectionState objects;
     */
    protected Map conn_to_connstate = new HashMap ();

    /** 
     * Maps NodeId objects to the ConnectionState objects in whose
     * ids_represented sets they reside.
     */
    protected Map nodeid_to_connstate = new HashMap ();

    protected SinkIF this_sink;

    /**
     * Used to time out connections without IdentityMessages.
     */
    protected PriorityQueue mystery_connections = new PriorityQueue (10);

    /**
     * Connections which we want to close.
     */
    protected PriorityQueue close_me = new PriorityQueue (10);

    /**
     * For emulation of large networks.
     */
    protected Map node_id_to_vertex = new HashMap ();

    /**
     * For emulation of large networks.
     */
    protected TransitStub ts;

    /**
     * For debugging.
     */
    protected int open_connections;
    protected Object open_connections_lock = new Object ();

    /**
     * Set things up.
     */
    public void init (ConfigDataIF config) throws Exception {

	// Set up some debug stuff.

	TAG = config.getStage().getName();
        if (DEBUG) Debug.turnOn (TAG);

	// Set up the TypeTable, in case no one else does.

	ostore.util.Types.registerAllTypes ();

	// Read in any topology information.

	String topology_graph = config.getString ("topology_graph");
	String topology_mapping = config.getString ("topology_mapping");

	if ((topology_graph == null) ^ (topology_mapping == null)) {
	    Carp.die (TAG + ": must specify topology_graph " +
		    " and topology_mapping together.");
	}
	if (topology_graph != null) {
	    read_topology_info (topology_graph, topology_mapping);
	}
    
	// Set this_sink.

  	if(config.getBoolean("Profile")) {
	    config.getManager().getProfiler().add(
		    "ostore.network.MultiplexedNetwork", this);
	    this_sink = new SinkProxy(config.getStage().getSink(),
		    config.getManager(),
		    config.getStage().getWrapper());
	}
	else
	    this_sink = config.getStage().getSink();

	// Create all of our filters.

	Filter nm_filter = new Filter ();
	if (! nm_filter.requireType (
		    Class.forName ("ostore.network.NetworkMessage")))
	    BUG (TAG + ": could not require type.");
	if (! nm_filter.requireValue ("inbound", new Boolean (false)))
	    BUG (TAG + ": could not require outbound.");

	// Read in all of the node_ids we represent, find their
	// classifiers, and register our filters.

	int node_id_count = config.getInt ("node_id_count");
	nodeid_to_classifier = new HashMap (node_id_count);
	for (int i = 0; i < node_id_count; ++i) {
	    NodeId n = new NodeId (config.getString ("node_id_" + i));
	    Classifier c = Classifier.getClassifier (n);
	    nodeid_to_classifier.put (n, Classifier.getClassifier (n));
	    local_nodeids.addLast (n);
	    c.subscribe (nm_filter, this_sink);
	    if (i == 0) {
		// Choose node_id_0's classifier to recieve our
		// DelayWrappers through and register our Alarms on.

		Filter filter = new Filter ();
		if (! filter.requireType (Class.forName (
				"ostore.network.Network$DelayWrapper")))
		    BUG (TAG + ": could not require type.");
		c.subscribe (filter, this_sink);
		lowest_nodeid = n;
	    }
	    else {
		if (n.less_than (lowest_nodeid)) 
		    lowest_nodeid = n;
	    }

	    int port = n.port ();
	    // Open the server socket
	    if (DEBUG) Debug.printtagln (TAG, 
		    "Openning server socket on port " + port + ".");
	    new ATcpServerSocket (port, this_sink);  
	}
	delay_classifier = Classifier.getClassifier (lowest_nodeid);

	if (DEBUG) Debug.printtagln (TAG,
		"Lowest id this server is " + lowest_nodeid + ".");

	Signal.getSignal (lowest_nodeid).register (
		Signal.ALRM, TAG, this_sink);
	Signal.getSignal (lowest_nodeid).alarm (TAG, 10*1000);
    }

    public void handleEvent (QueueElementIF item) 
    throws EventHandlerException {

	if (item instanceof ostore.dispatch.Signal.AlarmSignal) {
	    handle_alarm ();
	}
	else if (item instanceof ATcpConnection) {
	    handle_atcp_connection ((ATcpConnection) item);
	}
	else if (item instanceof ATcpInPacket) {
	    handle_atcp_inpacket ((ATcpInPacket) item);
	}
	else if (item instanceof DelayWrapper) {
	    handle_outbound_network_message (((DelayWrapper) item).outb, 
		    false /* msg has already been delayed */);
	}
	else if ((item instanceof NetworkMessage) && 
		(((NetworkMessage) item).inbound == false)) {
	    handle_outbound_network_message ((NetworkMessage) item, 
		    true /* allow delay if needed */);
	}
	else if (item instanceof SinkClosedEvent) {
	    handle_sink_closed_event ((SinkClosedEvent) item);
	}
	else if (item instanceof ATcpConnectFailedEvent) {
	    handle_atcp_connect_failed_event ((ATcpConnectFailedEvent) item);
	}
	else if (item instanceof ATcpListenSuccessEvent) {
	    ATcpListenSuccessEvent e = (ATcpListenSuccessEvent) item;
	    ATcpServerSocket s = e.getSocket ();
	    int port = s.getPort ();
	    System.out.println (TAG + " now listening on port " + port + ".");
	}
	else if (item instanceof ATcpListenFailedEvent) {
	    ATcpListenFailedEvent e = (ATcpListenFailedEvent) item;
	    ATcpServerSocket s = e.getSocket ();
	    int port = s.getPort ();
	    System.err.println ("ERROR: Network could not listen on " + 
		    "port " + port + ".  There is probably another " + 
		    "program using the port.  Dying...");
	    System.exit (1);
	}
	else {
	    Carp.die (TAG + ": got unknown event of type " + 
		    item.getClass ().getName () + ": " + item + ".");
	}
    }

    public void handleEvents(QueueElementIF items[]) 
    throws EventHandlerException {
	for(int i=0; i<items.length; i++)
	    handleEvent(items[i]);
    }

    public void destroy() throws Exception {
    }

    protected void read_topology_info (
	    String topology_graph, String topology_mapping) 
	    throws IOException {

	// Open the graph file.
	try {
	    ts = new TransitStub (topology_graph);
	}
	catch (IOException e) {
	    Carp.die (TAG + ": caught exception when creating " +
		    "graph: " + e + ".");
	}

	// Read the IP:Port to Vertex mapping
	StreamTokenizer tok = null;
	try {
	    FileInputStream is = new FileInputStream (topology_mapping);
	    Reader reader = 
		new BufferedReader (new InputStreamReader (is));
	    tok = new StreamTokenizer (reader);
	}
	catch (Exception e) {
	    Carp.die (TAG + ": caught exception " + e + " trying " +
		    "to read topology mapping " + topology_mapping);
	}

	tok.resetSyntax ();
	tok.whitespaceChars ((int) ' ', (int) ' ');
	tok.wordChars ((int) 'a', (int) 'z');
	tok.wordChars ((int) 'A', (int) 'Z');
	tok.wordChars ((int) '0', (int) '9');
	tok.wordChars ((int) ':', (int) ':');
	tok.wordChars ((int) '.', (int) '.');
	tok.wordChars ((int) '-', (int) '-');

	int lineno = 1;
	tok.nextToken ();
	while (tok.ttype == StreamTokenizer.TT_EOL) {
	    tok.nextToken ();
	    lineno++;
	}

	while (tok.ttype != StreamTokenizer.TT_EOF) {

	    // Read the IP:Port
	    if (tok.ttype != StreamTokenizer.TT_WORD) {
		Carp.die (TAG + ": expected IP:Port on line " + 
			lineno + ": " + tok);
	    }
	    NodeId nid = null;
	    try {
		nid = new NodeId (tok.sval);
	    }
	    catch (Exception e) {
		Carp.die (TAG + ": Bad IP:Port string on " +
			"line " + lineno + ": " + e);
	    }

	    // Read the Vertex
	    tok.nextToken ();
	    if (tok.ttype != StreamTokenizer.TT_WORD) {
		Carp.die (TAG + ": expected IP:Port on line " + 
			lineno + ": " + tok);
	    }
	    Integer vertex = null;
	    try {
		vertex = new Integer (Integer.parseInt (tok.sval));
	    }
	    catch (NumberFormatException e) {
		Carp.die (TAG + ": number format error reading " +
			"vertex on line " + lineno + ": " + e);
	    }

	    // Put them in the mapping.
	    if (node_id_to_vertex.containsKey (nid)) {
		Carp.die (TAG + ": duplicate node id " + nid +
			" in topology mapping " + topology_mapping);
	    }
	    node_id_to_vertex.put (nid, vertex);

	    tok.nextToken ();
	    while (tok.ttype == StreamTokenizer.TT_EOL) {
		tok.nextToken ();
		lineno++;
	    }
	}
    }

    protected void handle_outbound_network_message (
	    NetworkMessage msg, boolean delay) {

	if (DEBUG) Debug.printtagln (TAG, "Network got " + msg 
		+ " to send to " + msg.peer + ".");

	// possibly delay the message

	if (delay && (ts != null)) {
	    Integer Dst = (Integer) node_id_to_vertex.get (msg.peer);
	    if (Dst == null) {
		Carp.die (TAG + ": got message " + msg + " to " + msg.peer + 
			", who is not in the topology mapping.");
	    }
	    Integer Src = (Integer) node_id_to_vertex.get (msg.sender);
	    if (Src == null) {
		Carp.die ("We (" + msg.sender + ") are somehow " +
			"not in the topology mapping.");
	    }
	    int dst = Dst.intValue ();
	    int src = Src.intValue ();
	    long type_code = TypeTable.type_code (msg);
	    CountBuffer cb = new CountBuffer ();
	    cb.add (0);
	    cb.add (msg);
	    int size = cb.size ();
	    TransitStub.PathInfo pi = ts.shortest_path (
		    src, dst, size, true, type_code);
	    long delay_ms = Math.round (pi.route_time * 1000.0);
	    //long delay_ms = (Math.round (pi.route_time * 1000.0) / 10) + 1;
	    if (delay_ms != 0) {
		if (DEBUG) Debug.printtagln (TAG, 
			"Delaying msg from " + dst + 
			" to " + src + " by " + delay_ms + " ms.");
		delay_classifier.dispatch_later (
			new DelayWrapper (msg), delay_ms);
		return;
	    }
	}

	// Local messages are easy.
	Classifier local = (Classifier) nodeid_to_classifier.get (msg.peer);
	if (local != null) {
	    
	    NetworkMessage inb = (NetworkMessage) QSClone.qs_clone (msg);
	    /*NetworkMessage inb = null;
	    try {
		inb = (NetworkMessage) msg.clone ();
	    }
	    catch (CloneNotSupportedException e) {
		Carp.out ("Message " + msg + " must support clone.");
		e.printStackTrace (System.err);
		Carp.die ("Dying...");
	    }*/
	    inb.peer = msg.sender;
	    inb.sender = null;
	    inb.inbound = true;

	    try {
		local.dispatch (inb);
	    } catch (SinkFullException e) {
		BUG (TAG + ".dispatch: could not dispatch " + inb);
	    }

	    if (DEBUG) Debug.printtagln (TAG,
		    "Local message " + inb + " dispatched.");

	    return;
	}

	// If this is the first time we have sent a message to a node,
	// we need to open up a connection to it first.

	ConnectionState connstate = null;
	synchronized (nodeid_to_connstate) {
	    connstate = (ConnectionState) nodeid_to_connstate.get (msg.peer);
	    if (connstate == null) {

		// This is the first time we have heard of this peer.
		// Queue the msg and open up a connection to them.

		if (DEBUG) Debug.printtagln (TAG, 
			"Openning new socket to " + msg.peer + ".");

		connstate = new ConnectionState ();
		connstate.pending_outb_msgs = new LinkedList ();
		connstate.pending_outb_msgs.addLast (msg); 

		new ATcpClientSocket (
			msg.peer.address (), msg.peer.port (), this_sink);

		nodeid_to_connstate.put (msg.peer, connstate);

		return;
	    }
	}

	synchronized (connstate) {
	    
	    ATcpConnection conn = connstate.preferred_outb_conn ();
	    if (conn == null) {

		// We've tried to establish a connection, but it's not yet
		// open.

		if (DEBUG) Debug.printtagln (TAG, 
			"Socket exists but is not yet open.");

		if (connstate.pending_outb_msgs == null) {
		    // TODO: connection has been closed beneath us.  We
		    // should re-open it or something.
		}
		else {
		    connstate.pending_outb_msgs.addLast (msg); 
		}
		
		return;
	    }

	    // There is already a connection open to this destination;
	    // enqueue the message to it.

	    if (DEBUG) Debug.printtagln (
		    TAG, "Sending message out on network.");

	    enqueue_msg (conn, msg);
	}
    }

    protected void enqueue_msg (ATcpConnection conn, NetworkMessage msg) {
	try {
	    NetworkOutputBuffer buffer = new NetworkOutputBuffer ();
	    buffer.add ((int) 0);
	    buffer.add (msg);
	    buffer.overwrite (0, buffer.size () - 4);
	    if (DEBUG_SERIALIZATION) 
		System.err.println (TAG + ": " + msg + ", " + buffer);
	    buffer.enqueue (conn);
	}
	catch (SinkException e) {
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS) 
		Carp.out (TAG + ": got SinkException on enqueue of " + 
			"outgoing message.  Dropping it.");
	}
	catch (Exception e) {
	    Carp.out(TAG + ": serialize error on type " +
		    msg.getClass () + ".");
	    e.printStackTrace(System.err);
	    Carp.die(e);
	}
    }

    protected void handle_atcp_connection (ATcpConnection conn) {

	synchronized (open_connections_lock) {
	    ++open_connections;
	}

	NodeId peer = new MyNodeId (conn.getPort (), conn.getAddress ());

	if (DEBUG) Debug.printtagln (TAG, 
		"Network got ATcpConnection event from " + peer + ".");

	ConnectionState connstate = null;
	if (conn.getServerSocket () == null) {
	    // We initiated this connection.  It should be in
	    // nodeid_to_connstate.

	    if (DEBUG) Debug.printtagln (TAG, 
		    "We initiated this connection.");

	    synchronized (nodeid_to_connstate) {
		connstate = (ConnectionState) nodeid_to_connstate.get (peer);
		if (connstate == null)
		    BUG (TAG + ": connstate can't be null.");
	    }

	    synchronized (connstate) {

		// Add the connection to the connstate.

		connstate.client_conn = conn;
		connstate.client_is = new ConnectionStream ();
	    }

	    // Record the connection to connstate mapping.

	    synchronized (conn_to_connstate) {
		conn_to_connstate.put (conn, connstate);
	    }
	}
	else {
	    // They initiated this connection.  We have no idea yet what
	    // ports they're listening on.

	    if (DEBUG) Debug.printtagln (TAG, 
		    "This connection was initiated remotely so"
		    + " we don't know the peers listening ports yet.");

	    // We did not initiate this connection, so the port number is
	    // (possibly) different from the port numbers the sender
	    // listens on.  We'll hang onto this connection in anticipation
	    // of an IdentityMessage.  We also store the current time, so
	    // we can timeout if that message never comes.

	    connstate = new ConnectionState ();
	    long now = System.currentTimeMillis ();
	    connstate.server_conn = conn;
	    connstate.server_is = new ConnectionStream ();

	    synchronized (conn_to_connstate) {
		synchronized (mystery_connections) {
		    conn_to_connstate.put (conn, connstate);
		    mystery_connections.insert (conn, now + 10*1000, false);
		}
	    }
	}

	// Send an Identity message to the other side to let them
	// know what ports we're listening on.

	synchronized (connstate) {
	    IdentityMessage ident = new IdentityMessage (peer, local_nodeids);
	    ident.sender = lowest_nodeid;
	    enqueue_msg (conn, ident);

	    // Send any pending messages.

	    if (connstate.pending_outb_msgs != null) {
		if (DEBUG) Debug.printtagln (TAG, "Sending pending messages.");
		while (! connstate.pending_outb_msgs.isEmpty ()) {
		    NetworkMessage msg = (NetworkMessage) 
			connstate.pending_outb_msgs.removeFirst ();
		    enqueue_msg (conn, msg);
		}
	    }
	}

	// And let the packets flow.

	conn.startReader (this_sink);
    }

    protected void handle_atcp_inpacket (ATcpInPacket packet) {

	// Look up the connection state.
	ATcpConnection conn = packet.getConnection ();
	ConnectionState connstate = null;
	synchronized (conn_to_connstate) {
	    connstate = (ConnectionState) conn_to_connstate.get (conn);
	}

	if (connstate == null) {
	    if (DEBUG) Debug.printtagln (TAG, "got ATcpInPacket for " + 
		    conn + " which is unknown to us.");
	    return;
	}

	if (DEBUG) Debug.printtagln (TAG, "adding packet " + packet + 
		", seq " + packet.getSequenceNumber ());

	synchronized (connstate) {

	    // Add the packet to the stream.

	    ConnectionStream is = connstate.stream (conn);
	    is.add_packet (packet);
	}

	while (true) {
	    
	    InputBuffer ib = null;
	    synchronized (connstate) {

		// And read out any available messages.

		ConnectionStream is = connstate.stream (conn);
		if (! is.message_available ()) 
		    break;

		// Decode the message.

		ib = is.get_next_message_buffer ();
		if (DEBUG_SERIALIZATION) 
		    System.err.println (TAG + ": " + ib);
	    }

	    NetworkMessage msg = null;   
	    try {
		msg = (NetworkMessage) TypeTable.new_object (ib);
	    }
	    catch (Exception e) {
		if (DebugFlags.LOG_BAD_NETWORK_EVENTS) {
		    NodeId peer = new MyNodeId (
			    conn.getPort (), conn.getAddress ());
		    Carp.out (TAG + ": Received message from " + peer + 
			    " that caused exception " + e + 
			    " on decode.  Dropping it.  " + 
			    "(Stack trace follows.)"); 
		    e.printStackTrace (System.err);
		}
		return;
	    }

	    if (msg instanceof IdentityMessage) {
		handle_identity_msg ((IdentityMessage) msg, connstate, conn);
	    }
	    else {

		// Check that the sender is valid.
		if (! msg.sender.address ().equals (conn.getAddress ()))
		    continue; // liar

		if (DEBUG) Debug.printtagln (TAG,
			"Received a message from " + msg.sender + ".");

		Classifier classifier = null;
		synchronized (nodeid_to_classifier) {
		    classifier = (Classifier) 
			nodeid_to_classifier.get (msg.peer);
		}

		if (classifier == null)
		    BUG (TAG + ": no classifier for " + msg.peer);
		
		msg.peer = msg.sender;
		msg.sender = null;
		msg.inbound = true;

		if (DEBUG) Debug.printtagln (TAG, "Message is " + msg + ".");

		try {
		    classifier.dispatch (msg);
		} catch (SinkException e) {
		    BUG (TAG + ".dispatch: could not dispatch " + msg);
		}

		if (DEBUG) Debug.printtagln (TAG,
			"Message correctly decoded and dispatched.");
	    }
	}
    }

    protected void handle_identity_msg (IdentityMessage ident, 
	    ConnectionState connstate, ATcpConnection conn) {

	Iterator i = ident.node_ids.iterator (); 
	NodeId lowest = null;
	while (i.hasNext ()) {
	    NodeId n = (NodeId) i.next ();
	    if ((lowest == null) || n.less_than (lowest))
		lowest = n;
	    if (! n.address ().equals (conn.getAddress ())) {
		if (DebugFlags.LOG_BAD_NETWORK_EVENTS) {
		    Debug.printtagln (TAG, "Node on " + conn + " is a liar: " 
			    + "n=" + n + ", conn=" + conn + 
			    ".  List follows:");
		    int k = 0;
		    Iterator j = ident.node_ids.iterator (); 
		    while (j.hasNext ())
			Debug.printtagln (TAG, (k++) + ": " + j.next ());
		}
		return;
	    }
	}

	synchronized (nodeid_to_connstate) {
	    synchronized (connstate) {
		if (connstate.ids_represented != null) {
		    if (DEBUG) Debug.printtagln (TAG, "got second " +
			    "IdentityMessage for connstate.");
		    return;
		}
	    }

	    ConnectionState loweststate = (ConnectionState)
		nodeid_to_connstate.get (lowest);
	    if (loweststate == null) {
		nodeid_to_connstate.put (lowest, connstate);
		loweststate = connstate;
	    }

	    synchronized (loweststate) {
		loweststate.ids_represented = 
		    new HashSet (ident.node_ids.size ());
		loweststate.ids_represented.addAll (ident.node_ids);
	    }

	    i = ident.node_ids.iterator (); 
	    while (i.hasNext ()) {
		NodeId n = (NodeId) i.next ();
		ConnectionState other = 
		    (ConnectionState) nodeid_to_connstate.get (n);

		if ((other != null) && (other != loweststate)) {
		    if (DEBUG) Debug.printtagln (TAG, 
			    "got second IdentityMessage claiming " + 
			    "to represent " + n);

		    synchronized (conn_to_connstate) {
			synchronized (other) {
			    if (other.client_conn != null)
				conn_to_connstate.remove (other.client_conn);
			    if (other.server_conn != null)
				conn_to_connstate.remove (other.server_conn);
			    other.close_client ();
			    other.close_server ();
			}
		    }
		}

		nodeid_to_connstate.put (n, loweststate);
	    }
	}
    }

    protected class DelayWrapper implements QueueElementIF {
	public NetworkMessage outb;
	public DelayWrapper (NetworkMessage outb) {
	    this.outb = outb;
	}
    }

    protected void handle_sink_closed_event (SinkClosedEvent sce) {
	if (sce.sink instanceof ATcpConnection) {
	    handle_closed_connection ((ATcpConnection) sce.sink);
	}
	else {
	    BUG (TAG + ": got SinkClosedEvent for: " + sce.sink);
	}
    }

    protected void handle_closed_connection (ATcpConnection conn) {
	if (DEBUG) Debug.printtagln (TAG, "handling closed " + conn);

	synchronized (open_connections_lock) {
	    --open_connections;
	}

	ConnectionState connstate = null;
	synchronized (conn_to_connstate) {
	    connstate = (ConnectionState) conn_to_connstate.remove (conn);
	}

	if (connstate != null) {
	    synchronized (connstate) {
		if (connstate.client_conn == conn)
		    connstate.client_conn = null;
		else if (connstate.server_conn == conn)
		    connstate.server_conn = null;
		else 
		    BUG (TAG + ": neither?");
	    }
	}
    }

    protected void handle_atcp_connect_failed_event (
	    ATcpConnectFailedEvent failure) {

	BUG (TAG + ": TODO: connection failed.  " + failure + 
             failure.getSocket().getAddress() + " " +
             failure.getSocket().getPort());
    }

    /**
     * Defined by ProfilableIF
     */
    public int profileSize () {
	return this_sink.size ();
    }

    /**
     * For convenience.
     */
    protected static final void BUG (String msg) {
	Carp.die (msg + "  This is a bug in ostore.network.Network.  " +
		"Please let Sean know about it ASAP.");
    }

    /**
     * Generally, we don't want to close sinks immediately, as there may be
     * data still on the wire; instead, we just put them in this close_me
     * priority queue, and they will be closed sometime after the time
     * equal to their priority has passed.
     */
    protected void handle_alarm () {

	synchronized (open_connections_lock) {
	    Carp.out (TAG + ": " + open_connections +
		    " TCP connections currently open.");
	}

	LinkedList already_closed = null;

	long now = System.currentTimeMillis ();
	synchronized (close_me) {

	    while (true) {
		if (close_me.empty ())
		    break;

		long top = close_me.topPriority (false);

		if (top < now)
		    break;

		ATcpConnection conn_to_close = 
		    (ATcpConnection) close_me.extract (false);

		try {
		    conn_to_close.close (this_sink);
		}
		catch (SinkClosedException e) {
		    if (DEBUG) Debug.printtagln (TAG,
			    "Tried to close already closed "
			    + "client connection.");
		    if (already_closed == null)
			already_closed = new LinkedList ();
		    already_closed.addLast (conn_to_close);
		}
	    }
	}

	if (already_closed != null) {
	    while (! already_closed.isEmpty ()) {
		ATcpConnection conn = 
		    (ATcpConnection) already_closed.removeFirst ();
		handle_closed_connection (conn);
	    }
	}

	synchronized (conn_to_connstate) {
	    synchronized (mystery_connections) {

		while (true) {
		    if (mystery_connections.empty ())
			break;

		    long top = mystery_connections.topPriority (false);

		    if (top > now)
			break;

		    ATcpConnection conn_to_close = 
			(ATcpConnection) mystery_connections.extract (false);

		    if (conn_to_connstate.containsKey (conn_to_close))
			continue;

		    if (already_closed == null)
			already_closed = new LinkedList ();
		    already_closed.addLast (conn_to_close);
		}
	    }
	}

	if (already_closed != null) {
	    while (! already_closed.isEmpty ()) {
		ATcpConnection conn = 
		    (ATcpConnection) already_closed.removeFirst ();
		try {
		    conn.close (this_sink);
		}
		catch (SinkClosedException e) {
		    if (DEBUG) Debug.printtagln (TAG,
			    "Tried to close already closed "
			    + "client connection.");
		    handle_closed_connection (conn);
		}
	    }
	}

	Signal.getSignal (lowest_nodeid).alarm (TAG, 10*1000);
    }
}

