/**
 * FaultAdmin.java
 *
 * @author	Jeremy Stribling
 * @version	$Id: FaultAdmin.java,v 1.1 2003/02/08 00:49:37 strib Exp $
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.network.patchwork;

import java.util.HashMap;
import java.util.Random;
import ostore.network.NetworkMessage;
import ostore.util.NodeId;
import ostore.util.Debug;
import ostore.util.DebugFlags;

/**
 * This is a data structure that decides whether or not a message
 * should be dropped on its way to a stage.
 *
 * @author Jeremy Stribling
 * @version    $Id: FaultAdmin.java,v 1.1 2003/02/08 00:49:37 strib Exp $
 */
public class FaultAdmin {

    private HashMap faultRates;
    private Random rng;

    private boolean _accept_faults;

    private InjectFaultMsg allNodesIFM;

    /**
     * @param accept_faults   only accept faults iff true (i.e. pass a value 
     *                        from a config file here)
     */    
    public FaultAdmin( boolean accept_faults ) {

	_accept_faults = accept_faults;
	if( !_accept_faults ) {
	    return;
	}

	faultRates = new HashMap();
	Random preRng = new Random();
	long rng_starter = preRng.nextLong();
	rng = new Random( rng_starter );
	allNodesIFM = null;
    }

    /**
     * Called when a new <code>InjectFaultMsg</code> is received 
     * by a stage.
     */
    public void updateFaults( InjectFaultMsg msg ) {

	if( !_accept_faults ) {
	    return;
	}

	if( msg.allNodes ) {
	    // set the allNodes value, unless it's 0, then throw it out
	    if( msg.value == 0 ) {
		allNodesIFM = null;
	    } else {
		allNodesIFM = msg;
	    }
	} else {
	    if( msg.nodeToDrop == null ) {
		if( msg.value == 0 ) {
		    faultRates.remove( msg.peer );
		} else {
		    faultRates.put( msg.peer, msg );
		}
	    } else {
		if( msg.value == 0 ) {
		    faultRates.remove( msg.nodeToDrop );
		} else {
		    faultRates.put( msg.nodeToDrop, msg );
		}
	    }
	}
    }

    /**
     * Decide whether or not to drop a message, based on previously
     * received loss rates.
     */
    public boolean dropMessage( NetworkMessage inbound ) {

	if( !_accept_faults ) {
	    return false;
	}

	// can add code here to always accept certain types of messages

	boolean dropIt = false;
	// first, check the allNodes rfm
	if( allNodesIFM != null ) {
	    dropIt = dropMessage( allNodesIFM );
	    if( allNodesIFM.value == 0 ) {
		allNodesIFM = null;
	    }
	}
	// then, try the individual fault
	InjectFaultMsg ifm = (InjectFaultMsg) faultRates.get( inbound.peer );
	if( ifm != null ) {
	    dropIt = dropMessage( ifm );
	    if( ifm.value == 0 ) {
		faultRates.remove( inbound.peer );
	    }
	}
	return dropIt;
    }

    private boolean dropMessage( InjectFaultMsg msg ) {

	if( msg != null ) {
	    if( msg.isPercent && 
		(Math.random()*100 < msg.value) ) {
		// drop is as a percent
		return true;
	    } else if( !msg.isPercent && 
		       (msg.value > 0) ) {
		// drop it as an absolute value
		msg.value--;
		return true;
	    }
	}
	return false;
    }

}
