/**
 * RecalcPatch.java
 *
 * @author	Jeremy Stribling
 * @version	$Id: RecalcPatch.java,v 1.12 2003/05/20 01:58:57 hweather Exp $
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.network.patchwork;

import java.util.TreeSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Vector;
import ostore.network.NetworkNodeStatusMsg;
import ostore.util.InvalidConfigFileException;
import ostore.util.Carp;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import ostore.util.NodeId;
import ostore.util.Clock;
import ostore.util.Pair;
import ostore.util.PriorityQueue;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.UnknownEventException;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SingleThreadedEventHandlerIF;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.api.SinkFullException;   
import seda.sandStorm.api.StagesInitializedSignal;   

/**
 * A Patch that recalculate the <code>PatchworkTable</code> and sends out 
 * deltas.<br><br>
 *
 * Also, it currently is hardcoded to send out messages when it notices that
 * a node is down (loss rate > 95%).<br><br>
 *
 * @author Jeremy Stribling
 * @version    $Id: RecalcPatch.java,v 1.12 2003/05/20 01:58:57 hweather Exp $
 */
public class RecalcPatch extends Patch {

    // nodes currently alive and being monitored
    private TreeSet _alive_nodes;

    // nodes currently dead and being monitored
    private TreeSet _dead_nodes;

    // the userDatas and sinks that want to hear about everyone
    private LinkedList _all_userdata_sinks;

    public RecalcPatch( NodeId self_node_id, Classifier classifier, 
		  PatchworkTable table, FaultAdmin fault_admin, 
			Period period ) {
	super( self_node_id, classifier, table, fault_admin, period );
	class_tag = "RecalcPatch[" + _self_node_id + "]";
	_type = TYPE_RECALC;
	DEBUG = DebugFlags.NETWORK_PATCH_RECALC;
	// put self in Period
	_period.changePeriod( _self_node_id, 0 );

	_alive_nodes = new TreeSet();
	_dead_nodes = new TreeSet();
	_all_userdata_sinks = new LinkedList();

	// set an alarm, to get started
	setNextAlarm();
    }

    public void addNode( NodeId node, int level, Object userData, 
			 SinkIF sink ) {
	final String tag = class_tag + ".addNode";
	if( DEBUG ) {
	    Debug.printtagln( tag, "Adding node " + node + " at level " + 
			      level + " with userdata " + userData + 
			      " and sink " + sink );
	}
	// assume it's a live node if we don't have a dead record for it
	if( !_dead_nodes.contains( node ) ) {
	    _alive_nodes.add( node );
	}
	// catalogue the userData and sink
	LinkedList l = (LinkedList) _userdata_sinks.get( node );
	if( l == null ) {
	    l = new LinkedList();
	    _userdata_sinks.put( node, l );
	}
	Pair p = new Pair( userData, sink );
	if( !l.contains( p ) ) {
	    l.add( p );
	    if( DEBUG ) {
		Debug.printtagln( tag, "REALLY added node " + node ); 
	    }
	}
	return;
    }

    public void addMonitorAll( Object userData, SinkIF sink ) {

	Pair p = new Pair( userData, sink );
	if( !_all_userdata_sinks.contains( p ) ) {
	    _all_userdata_sinks.add( p );
	}
    }

    public void removeNode( NodeId node, Object userData, SinkIF sink ) {
	final String tag = class_tag + ".removeNode";
	if( DEBUG ) {
	    Debug.printtagln( tag, "Removing node " + node + 
			      " with userdata " + userData + 
			      " and sink " + sink );
	}
	LinkedList l = (LinkedList) _userdata_sinks.get( node );
	if( l != null ) {
	    Pair p = new Pair( userData, sink );
	    l.remove( p );
	}
	if( l == null || l.isEmpty() ) {
	    // remove from all sets
	    _alive_nodes.remove( node );
	    _dead_nodes.remove( node );
	    _table.remove( node );
	}
	return;
    }

    protected void addToQueue( NodeId node ) {
	final String tag = class_tag + ".addToQueue";
	if( DEBUG ) {
	    Debug.printtagln( tag, "called, does nothing." );
	}
	// does nothing
	return;
    }
    
    protected void setNextAlarm() {
	AlarmMsg alarm = new AlarmMsg( _self_node_id, _type );
	long reltime = _period.getPeriod( _self_node_id );
	if( reltime < 0 ) {
	    return;
	}
	_classifier.dispatch_later(alarm, reltime);
    }

    private void setNextAlarm( NodeId node, long time ) {
	final String tag = class_tag + ".setNextAlarm(node,time)";
	if( DEBUG ) {
	    Debug.printtagln( tag, "called, does nothing." );
	}
	// does nothing
	return;
    }

    protected void makeObservation( NodeId peer, long value, byte type ) {
	final String tag = class_tag + ".makeObservation";
	if( DEBUG ) {
	    Debug.printtagln( tag, "called, does nothing." );
	}
	// does nothing
	return;
    }

    protected void handleProbeResponse( ProbeMsg response, 
					long timeRecvd ) {

	final String tag = class_tag + ".handleProbeResponse";
	if( DEBUG ) {
	    Debug.printtagln( tag, "called, does nothing." );
	}
	// does nothing
	return;	

    }

    protected void handleAlarm( AlarmMsg alarm ) {
	final String tag = class_tag + ".handleAlarm";
	// recalculate and set the next alarm
	PatchworkTable.Delta d = _table.recalculate();
	if( DEBUG ) {
	    Debug.printtagln( tag, "New table: " + _table );
	}
	// send it out to everyone
	dispatch( new DeltaMsg( d ) );


	// now, let's look into all nodes we know about, and see if their
	// status has changed

	// for now, we hardcode the simple case, where we assume
	// others are only interested if an alive node now has a current
	// loss rate above 95, or a dead node has a current loss rate under 5.
	//
	// NOTE: store in Vectors first, so no concuurrent modifications
	Iterator iter = _alive_nodes.iterator();
	LinkedList aliveToRemove = new LinkedList();
	while( iter.hasNext() ) {
	    NodeId currNode = (NodeId) iter.next();
	    PatchworkStats stats = _table.getStats( currNode, 
						    Patch.TYPE_LOSS );
	    //if( stats != null && stats.getCurrentMean() > 95 ) {
	    if( stats != null && stats.getCurrentMin() != 0 ) {
		aliveToRemove.add( currNode );
	    }
	    
	}
	iter = _dead_nodes.iterator();
	LinkedList deadToRemove = new LinkedList();
	while( iter.hasNext() ) {
	    NodeId currNode = (NodeId) iter.next();
	    PatchworkStats stats = _table.getStats( currNode, 
						    Patch.TYPE_LOSS );
	    if( stats != null && stats.getCurrentMean() < 5 ) {
		deadToRemove.add( currNode );
	    }
	    
	}

	// now actually removed the marked ones
	LinkedList msgs = new LinkedList();
	for( int i = 0; i < aliveToRemove.size(); i++ ) {
	    NodeId currNode = (NodeId) aliveToRemove.get(i);
	    _alive_nodes.remove( currNode );
	    _dead_nodes.add( currNode );
	    LinkedList l = (LinkedList) _userdata_sinks.get( currNode );
	    for( int j = 0; j < l.size(); j++ ) {
		Pair p = (Pair) l.get(j);
		msgs.add( new Pair( new NodeStatusMsg(currNode, 
						     NodeStatusMsg.STATUS_DEAD,
						      p.first), 
				    p.second) );
	    }
	    // also do it for everyone on the all queue
	    sendNodeStatusToAll( currNode, NodeStatusMsg.STATUS_DEAD );
	}
	for( int i = 0; i < deadToRemove.size(); i++ ) {
	    NodeId currNode = (NodeId) deadToRemove.get(i);
	    _dead_nodes.remove( currNode );
	    _alive_nodes.add( currNode );
	    LinkedList l = (LinkedList) _userdata_sinks.get( currNode );
	    for( int j = 0; j < l.size(); j++ ) {
		Pair p = (Pair) l.get(j);
		msgs.add( new Pair( new NodeStatusMsg(currNode, 
						    NodeStatusMsg.STATUS_ALIVE,
						      p.first), 
				    p.second) );
	    }
	    // also do it for everyone on the all queue
	    sendNodeStatusToAll( currNode, NodeStatusMsg.STATUS_ALIVE );
	}
	for( int i = 0; i < msgs.size(); i++ ) {
	    Pair p = (Pair) msgs.get(i);
	    enqueue( (QueueElementIF) p.first, (SinkIF) p.second );
	}

    }

    private void sendNodeStatusToAll( NodeId node, int status ) {
	for( int j = 0; j < _all_userdata_sinks.size(); j++ ) {
	    Pair p = (Pair) _all_userdata_sinks.get(j);
	    enqueue(new NodeStatusMsg(node, status, p.first), 
		    (SinkIF) p.second);
	}
    }

    protected void handleTimeout( ProbeTimeout timeout ) {
	final String tag = class_tag + ".handleTimeout";
	if( DEBUG ) {
	    Debug.printtagln( tag, "called, does nothing." );
	}
	// does nothing
	return;
    }

    protected void handleProbe( ProbeMsg inProbe ) {
	final String tag = class_tag + ".handleProbe";
	if( DEBUG ) {
	    Debug.printtagln( tag, "called, does nothing." );
	}
	// does nothing
	return;
    }

    protected void sendProbe( ProbeMsg probe ) {
	final String tag = class_tag + ".sendProbe";
	if( DEBUG ) {
	    Debug.printtagln( tag, "called, does nothing." );
	}
	// does nothing
	return;
    }

    protected long registerResponse( ProbeMsg response ) {
	final String tag = class_tag + ".registerResponse";
	if( DEBUG ) {
	    Debug.printtagln( tag, "called, does nothing." );
	}
	// does nothing
	return -1;
    }

    protected boolean registerTimeout( ProbeTimeout pt ) {
	final String tag = class_tag + ".registerTimeout";
	if( DEBUG ) {
	    Debug.printtagln( tag, "called, does nothing." );
	}
	// does nothing
	return false;
    }

    protected int getAndUpdateSeqNo( NodeId node ) {
	final String tag = class_tag + ".getAndUpdateSeqNo";
	if( DEBUG ) {
	    Debug.printtagln( tag, "called, does nothing." );
	}
	// does nothing
	return -1;
    }

    protected void handleNetworkNodeStatusMsg( NetworkNodeStatusMsg msg ) {

	final String tag = class_tag + ".handleNetworkNodeStatusMsg";

	if( DEBUG ) {
	    Debug.printtagln( tag, "Got NetworkNodeStatusMsg: " + msg );
	}

	// if in alive nodes, remove and warn
	if( _alive_nodes.contains(msg.dead_node) ) {
	    _alive_nodes.remove( msg.dead_node );
	    _dead_nodes.add( msg.dead_node );
	    LinkedList v = (LinkedList) _userdata_sinks.get( msg.dead_node );
	    for( int i = 0; i < v.size(); i++ ) {
		Pair p = (Pair) v.get(i);
		enqueue(new NodeStatusMsg(msg.dead_node, 
					  NodeStatusMsg.STATUS_DEAD,
					  p.first), 
			(SinkIF) p.second);
	    }
	    sendNodeStatusToAll( msg.dead_node, NodeStatusMsg.STATUS_DEAD );
	}
	
    }
    
}
