/**
 * EventCountStage.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.oil;

import java.lang.reflect.Field;
import ostore.util.QSIO;
import ostore.util.InvalidConfigFileException;
import ostore.util.NodeId;
import ostore.util.Carp;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.UnknownEventException;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.api.SinkFullException;

/**
 *  EventCountStage watches for events of certain types, and counts them.
 *  Pretty simple.  To use, create an EventCountStage stage in the config
 *  file, and in the initargs sections have the line:
 *  <pre>events Event1 Event2 . . . EventN</pre>
 *  where events 1 through N are the events you want to keep track of.
 *
 *  When you want to know the count of a certain even, send out a 
 *  EventCountRequestMsg, and this stage will send you back an
 *  EventCountMsg with the count.
 *
 *  Currently this functionality is pretty simple, but possibly later we
 *  can add support for rates, thresholds, etc. 
 *
 * @author	Jeremy Stribling
 * @version	$Id: EventCountStage.java,v 1.7 2002/07/22 20:54:53 srhea Exp $
 */
public class EventCountStage implements EventHandlerIF
{
    ////////////////////       CONSTANTS	////////////////////
    private boolean DEBUG = DebugFlags.OIL_EVENTCOUNTSTAGE;
    private static final String class_tag = "EventCountStage"; 
    private static final String EVENTS_KEY = "events";

    private static String[] eventStrings;
    private static Class[] eventClasses;
    private static int[] eventCounts;
    private int totalEvents = 0;
    private int totalEventsThreshold = 50;

    private static Class eventCountRequestClass;
    private static final String eventCountRequestName = 
		"ostore.oil.EventCountRequestMsg";
    private static Field eventCountRequestField;
    private static final String eventCountRequestFieldName = 
		"messageClassName";
    private Classifier classifier;

    ////////////////////  STATIC CONSTRUCTORS	////////////////////

    static {

//	Debug.turnOn( class_tag + ".handleEvents" );
	Debug.turnOn( class_tag + ".handleEvent" );
	Debug.turnOn( class_tag + ".init" );
	Debug.turnOn( class_tag + ".destroy" );

	try {
	  eventCountRequestClass = Class.forName( eventCountRequestName );
	  eventCountRequestField = eventCountRequestClass.getDeclaredField( 
			eventCountRequestFieldName );
        } catch( ClassNotFoundException cnfe ) {
	  Carp.die( cnfe );
        } catch( NoSuchFieldException nsfe ) {
	  Carp.die( nsfe );
 	} catch( SecurityException se ) {
	  Carp.die( se );
	}

    }

    ////////////////////      CONSTRUCTORS	////////////////////
    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////

    /* Specified by seda.sandStorm.api.EventHandlerIF */
    public void handleEvents(QueueElementIF elemarr[]) 
	throws EventHandlerException
    {
	final String tag = class_tag+".handleEvents";
	if( DEBUG ) Debug.printtagln( tag, "called: "
				      + elemarr.length + " elements" );
	for( int i=0; i<elemarr.length; i++ ) {
	    handleEvent( elemarr[i] );
	}
	if( DEBUG ) Debug.printtagln( tag, "done" );
    }

    /* Specified by seda.sandStorm.api.EventHandlerIF */
    public void handleEvent(QueueElementIF element) 
	throws EventHandlerException
    {
	final String tag = class_tag+".handleEvent";
	boolean eventSatisfied = false;

	// if it's a request for a count
	if( eventCountRequestClass.isInstance( element ) ) {
	  EventCountRequestMsg ecrm = (EventCountRequestMsg) element;
	  for( int i = 0; i < eventStrings.length; i++ ) {
	    if( eventStrings[i].equals( ecrm.messageClassName ) ) {
	      EventCountMsg ecm = new EventCountMsg( eventStrings[i], 
						     eventCounts[i] );
	      try {
	        classifier.dispatch( ecm );
	      } catch ( SinkFullException sfe ) {
		//TODO: Better way to deal with SinkFull
		try {
		  Thread.currentThread().sleep( 5000 );
		} catch( InterruptedException ie ) {
		  //ignore
		}
	      }
	      eventSatisfied = true;
	      break;
	    }
	  }
	} else {
	  for( int i = 0; i < eventClasses.length; i++ ) {
	    if( eventClasses[i].isInstance( element ) ) {
	      eventCounts[i] = eventCounts[i] + 1;
	      eventSatisfied = true;
	      totalEvents++;
	    }
          }
	}

	if( !eventSatisfied ) {
	    throw new UnknownEventException( "Unknown event " +
 					     "passed filters", element );
        }

	if( totalEvents % totalEventsThreshold == 0 ) {
	  if( DEBUG ) {
	    Debug.printtagln( tag, "----------------------------" );
	    for( int i = 0; i < eventStrings.length; i++ ) {
	       Debug.printtagln( tag, "Caught " + eventCounts[i] +
	  			      " events of type " +
				      eventStrings[i] );
            }
	  }
	}

//	if( DEBUG ) Debug.printtagln( tag, "done" );
    }

    /* Specified by seda.sandStorm.api.EventHandlerIF */
    public void init(ConfigDataIF config) throws Exception
    {
	final String tag = class_tag+".init";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	////////////////////////////////
	// Get Config data

	eventStrings = config.getStringList( EVENTS_KEY );
	eventClasses = new Class[ eventStrings.length ];
	for( int i = 0; i < eventStrings.length; i++ ) {
	    if( DEBUG ) Debug.printtagln( tag, "Finding class: " +
					       eventStrings[i] );
	    try {
              eventClasses[i] = Class.forName( eventStrings[i] );
	    } catch ( ClassNotFoundException cnfe ) {
	      Carp.die( cnfe );
	    }
	}
	eventCounts = new int[ eventStrings.length ];

	////////////////////////////////
	// Initialize Filters
	
	classifier = Classifier.getClassifier(
                new NodeId (config.getString ("node_id")));
	SinkIF my_sink = config.getStage().getSink();

	// Subscribe to all events we are supposed to observe
	Filter filter;
	for( int i = 0; i < eventClasses.length; i++ ) {
	  filter = new Filter();
	  if( !filter.requireType( eventClasses[i] ) ||
	      !classifier.subscribe( filter, my_sink ) ) {
	    throw new InvalidConfigFileException( "Classifier does not " + 
						  "recognize class: " +
						  eventStrings[i] );
	  }
	  // watch for requests about the counts of this message as well
	  filter = new Filter();
	  if( !filter.requireType( eventCountRequestClass ) ||
	      !filter.requireValue( eventCountRequestField, 
				    eventStrings[i] ) ||
	      !classifier.subscribe( filter, my_sink ) ) {
	    throw new InvalidConfigFileException( "Classifier does not " + 
						  "recognize field: " +
						eventCountRequestFieldName);
	  }
	  if( DEBUG ) Debug.printtagln( tag, "watching event: " +
                                             eventStrings[i] );
        }

	if( DEBUG ) Debug.printtagln( tag, "done"  );	
    }
    
    /* Specified by seda.sandStorm.api.EventHandlerIF */
    public void destroy()
    {
	final String tag = class_tag+".destroy";
	if( DEBUG ) Debug.printtagln( tag, "destroyed" );
    }

    ////////////////////     INNER CLASSES	////////////////////

}
