/**
 * HMM.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.oil;

import java.util.Random;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Arrays;
import java.util.Comparator;
import java.text.NumberFormat;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;
import ostore.util.OutputBuffer;
import ostore.util.QSException;
import ostore.util.TypeTable;
import ostore.util.Types;
import ostore.util.ByteUtils;
import ostore.util.QSInt;
import ostore.util.QSVector;
import ostore.util.Carp;
import ostore.util.Debug;
import ostore.util.NodeId;
import ostore.util.StopWatch;

// FIXME!!
// Add "final" declaration and test time improvements.

/**
 * Hidden Markov Model: a statistical tool which models a chain of observed
 * variables as the output of a chain of hidden states.  Given such a chain of
 * observations, it can deduce the relationships among the hidden states and
 * between the hidden states and the output variables.  It can thus also predict
 * future hidden states and output variables.
 *
 * @see Model
 * 
 * @author	Dennis Geels
 * @version	$Id: HMM.java,v 1.31 2002/07/20 19:38:03 srhea Exp $
 */

public class HMM implements SegmentedModel
{
    ////////////////////       CONSTANTS	////////////////////

    private final static boolean DEBUG = ostore.util.DebugFlags.OIL_HMM;
    private final static boolean DEBUG_TIME = true;
    private final static int DEBUG_NONE = 0;
    private final static int DEBUG_SOME = 1;
    private final static int DEBUG_ALL  = 2;
    private final static int DEBUG_LEVEL = DEBUG ? DEBUG_SOME : DEBUG_NONE;

    private final static String class_tag = "HMM";
    private final static String time_tag = class_tag+".time";
    private final static int MAX_NUM_ITERATIONS = 1000;

    
    ///////// Initialization codes /////////

    private final static int INIT_TRANSITIONS_MASK	= 0x00ff;
    /**
     * Use random transition priors.
     */
    public final static int INIT_TRANSITIONS_RANDOM 	= 0x0001;
    /**
     * Use random transition priors, but weight self-transitions heavily.
     */
    public final static int INIT_TRANSITIONS_CLUSTER 	= 0x0002;

    /**
     * Initialize <code>HMM</code> with priors which are useful for a
     * clustering application.
     */
    public final static int INIT_CODE_CLUSTERING = INIT_TRANSITIONS_CLUSTER;

    
    ////////////////////    STATIC VARIABLES 	////////////////////

    /**
     * These variables are hacked in for testing purposed.  Proper use of
     * the StopWatch class may be added later.
     */
    private static NodeId dummy;
    static {
	try {
	    dummy = new NodeId("0.0.0.0:0");
	} catch( Exception e ) {
	}
    }
    private static StopWatch stopwatch = StopWatch.getInstance( dummy );
    
    ////////////////////       VARIABLES	////////////////////

    /* Convergence Threshold
     * The inference algorithm will iterate until the log-likelihood
     * converges to within <code>epsilon * sequence.size()<code>.
     */
    private double epsilon;

    /* IDs
     * These variables store the indices of all states and outputs which
     * this HMM tracks.  We ignore all other states and outputs when performing
     * our computations.  When we do come accross a foreign state or output, we
     * can trigger an event to communicate with other carriers to swap in more
     * information. 
     */
    // FIXME : make these variables private again.
    protected QSVector states;		// The id's of the states stored here.
    private HashMap state_set;		// id->index map.
    private QSVector outputs;		// The id's of the outputs observed here.
    private HashMap output_set;		// id->index map.
    
    /* Sequence
     * We store an array of recent outputs.  We use this sequence to estimate
     * the hidden states.
     */
    private QSVector sequence;	// Observed output values (y_t).

    // FIXME : make these variables private again.
    /* Model Parameters
     */
    /* Maps ID->Double: E[\sum_t(q_t)|y] */
    protected Array marginals;
    /* Maps ID->Array: E[\sum_t(q_t,q_t+1))|y] */
    protected Matrix transitions;
    /* Maps ID->Array: E[\sum_t(q_t,y_t)|y] */
    protected Matrix emissions;
    /* Maps ID->Double: E[q_0|y] */
    private Array priors;
    /* Maps ID->Double: E[q_T-1|y] (See clear()) */

    /*
     * flag used to avoid re-running <code>sync_arrays</code> unnecessarily
     */
    private boolean arrays_in_sync;

    /*
     * Temporary arrays
     * We make them class variables and allocate them once per call to
     * <code>EM</code>, to avoid a large memory allocation on
     * every invocation of <code>E_step</code> or <code>M_step</code>.
     */
    
    /*
     * Sufficient statistics (prior knowledge)
     * These arrays mirror <code>marginals</code>,
     * <code>transitions</code>, <code>emissions</code>, and
     * <code>priors</code>.  They will be initialized by
     * <code>fill_arrays</code>.
     */
    private double[] marginal_ss;	// marginals.toArray
    private double[][] transition_ss;	// transitions.toArray
    private double[][] emission_ss;	// emissions.toArray
    private double[] prior_ss;		// priors.toArray

    /*
     * Expected sufficient statistics
     * These arrays hold the new information calculated from the current
     * probabilities and sequence by <code>E_step</code>. 
     */
    private double[] marginal_ess;	// E[\sum_t(q)]
    private double[][] transition_ess; 	// E[\sum_t(q_t,q_t+1)]
    private double[][] emission_ess;	// E[\sum_t(q_t,y_t)]
    private double[] prior_ess;		// E[q_0]
    
    /* Probabilities
     * These arrays are derived from the sufficient statistics (the sum of
     * both sets), and in turn are used to calculate the expected
     * sufficient statistics.
     */
    private double[][] transition_p;	// MLE p(q_t,q_t+1)
    private double[][] emission_p;	// MLE p(q_t,y_t)
    private double[] prior_p;		// MLE p(q_0)
    
    /*
     * Helper statistics
     * These arrays are used within <code>E_step</code>.
     */
    private double[][] alpha;	// p(q_t|y_0,...,y_t)
    private double[][] beta;	// p(y_t+1,...,y_T|q_t)/p(y_0,...,y_t)
    private double[] normals;	// p(y_t|y_0,...,y_t-1)
    
    ////////////////////      CONSTRUCTORS	////////////////////

    /**
     * Create a new <code>HMM</code>.
     * Initialize the states and sufficient statistic priors using the
     * specified prototype.
     *
     * @param type	A code specifying how to initialize the sufficient
     * 			statistic priors should be generated. (See
     * 			INIT_CODE_* above).
     * @param num_states	The number of hidden states to assume.
     * @param num_outputs	The total number of outputs to expect (this
     * 				variable is not binding; it just produces
     * 				better emission priors).
     * @param epsilon	The convergence threshold.  The inference algorithm
     * 			will iterate until the log-likelihood converges to
     * 			within <code>epsilon<code> times the sequence length.
     */
    public HMM( int type, int num_states, int num_outputs, double epsilon )
    {
	this.epsilon = epsilon;

  	states = new QSVector( num_states );
  	state_set = new HashMap();
	// Just use integers 0,...,M for state names.
	QSInt state;
	for( int i=0; i<num_states; i++ ) {
	    state = new QSInt(i);
	    states.add( state );
	    state_set.put( state, new Integer(i));
	}

	// Don't know the output names yet.
  	outputs = new QSVector();
  	output_set = new HashMap();

  	sequence = new QSVector();

	// Initialize the sufficient statistics
	//  Make sure that the marginals add up correctly.
	marginals = new Array(1);
	transitions = init_transitions( states, 
					type & INIT_TRANSITIONS_MASK );
	emissions = init_emissions( states, num_outputs );
	priors = new Array( .00000001 );	// Uniform and negligible.

	// Remember to initialize *_ss, *_p arrays
	arrays_in_sync = false;
	    
	Debug.printtagln( class_tag, "HMM Ready" );
    }
    
    /**
     * Constructs a <code>HMM</code> from its
     * <code>serialize</code>d form.
     */
    public HMM( InputBuffer buffer ) throws QSException
    {
	if( DEBUG ) Debug.printtagln( class_tag, "deserializing..." );
	
	epsilon = buffer.nextDouble();
	
	// Read most of the data from a stored <code>Delta</code>.
	Delta all = (Delta)buffer.nextObject();
	states = all.states;
	marginals = all.marginals;
	transitions = all.transitions;
	emissions = all.emissions;

	// Keep state_set updated.
	state_set = new HashMap( states.size() );
	for( int i=0; i<states.size(); i++ ) {
	    state_set.put( states.get(i), new Integer(i) );
	}

	// Now deserialize the sequence and priors
	sequence = (QSVector)buffer.nextObject();
	priors = (Array)buffer.nextObject();
	
	if( DEBUG ) {
	    Debug.printtagln( class_tag, "states: " + states.size() );
	    Debug.printtagln( class_tag, "priors: " + priors.size());
	    Debug.printtagln( class_tag, "marginals: " + marginals.size() );
	    Debug.printtagln( class_tag, "transitions: " + transitions.size() );
	    Debug.printtagln( class_tag, "transitions: " + transitions.full_size() );
	    Debug.printtagln( class_tag, "emissions: " + emissions.size() );
	    Debug.printtagln( class_tag, "emissions: " + emissions.full_size());
	    Debug.printtagln( class_tag, "sequence: " + sequence.size());
	}
	
	// Fill the output set with outputs from the sequence.
	outputs = new QSVector();
	output_set = new HashMap();
	rebuild_output_set();

	// Remember to initialize *_ss, *_p arrays
	arrays_in_sync = false;
    }    

    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////

    /**
     * Fill in the sufficient statistics priors for transitions between
     * states.  The transitions must not be uniform, because our
     * gradient-descent algorithm requires some unevenness in order to 
     * differentiate between states.
     * Normalize each row, so that the sum matches the state marginal.
     *
     * @param states	The ID's of the states in the HMM.
     * @param type	A code specifying how to initialize the priors.
     * 			(See INIT_TRANSITIONS_* above).
     * @return a <code>Matrix</code> containing random transition counts.
     */
    private Matrix init_transitions( QSVector states, int type )
    {
	final String tag = class_tag+".init_transitions";
	Matrix new_transitions = new Matrix();
	Array row;
	int num_states = states.size();
	QuickSerializable id1, id2;
	Random rands = new Random();
	rands.setSeed( 0 );	// Repeatability is nice for now.
	double rand, sum;
	double[] raw = new double[num_states];
	
	for( int i=0; i<num_states; i++ ) {
	    sum = 0;
	    for( int j=0; j<num_states; j++ ) {
		rand = rands.nextDouble();
		sum += rand;
		raw[j] = rand;
	    }
	    if( type == INIT_TRANSITIONS_CLUSTER ) {
		// Weight self-transitions heavily. - 50%
		sum -= raw[i];
		raw[i] = sum;
		sum *= 2;
	    }

	    id1 = (QuickSerializable)states.get(i);
	    row = new Array(0);
	    for( int j=0; j<num_states; j++ ) {
		id2 = (QuickSerializable)states.get(j);
		row.put( id2, raw[j]/sum );
	    }
	    new_transitions.put( id1, row );
	}
	if( DEBUG ) Debug.printtagln( tag, "done: " + new_transitions );
	return new_transitions;
    }

    /**
     * Fill in the sufficient statistics priors for emissions from states.
     * The only option currently is a uniform distribution.
     * Normalize each row, so that the sum matches the state marginal.
     *
     * @param states	The ID's of the states in the HMM.
     * @return a <code>Matrix</code> containing uniform emissions.
     */
    private Matrix init_emissions( QSVector states, int num_outputs)
    {
	final String tag = class_tag+".init_emissions";
	Matrix new_emissions = new Matrix();
	QuickSerializable id1;
	double default_prob = (double)1/num_outputs;
	
	for( int i=0; i<states.size(); i++ ) {
	    id1 = (QuickSerializable)states.get(i);
	    new_emissions.put( id1, new Array(default_prob) );
	}
	if( DEBUG ) Debug.printtagln( tag, "done: " + new_emissions );
	return new_emissions;
    }

    /* Specified by ostore.oil.Model interface */
    public synchronized void record( QuickSerializable output )
    {
	final String tag = class_tag+".record";
	if( DEBUG ) Debug.printtagln( tag, "Saw " + output );
	// Add the output to our list.
	sequence.add( output );

	if( !output_set.containsKey( output ) ) {
	    // FIXME!! return a flag?
	    if( DEBUG ) Debug.printtagln( tag, "new output: " + output );
	    rebuild_output_set();
	}

	if( DEBUG ) Debug.printtagln( tag, "Done" );
	return;
    }

    /* Specified by ostore.oil.Model interface */
    public void record( QuickSerializable[] observations )
    {
	for( int i=0; i<observations.length; i++ ) {
	    record( observations[i] );
	}
    }

    /* Specified by ostore.oil.Model interface */
    public synchronized void clear( int num ) 
    {
	final String tag = class_tag+".clear";
	if( DEBUG ) Debug.printtagln( tag, "called: " + num );

	// Clear the sequence, but don't update the output set.  It will
	// (lazily) be rebuilt when a new output is recorded.
	sequence.removeRange( 0, num );

	//FIXME option a: Would a uniform (more flexible) prior be better?
	priors.clear();
	priors.set_default( 1.0/states.size() );

	// FIXME!! option b:
	// Update <code>priors</code> to reflect the <b>end</b> of the
	// sequence we just processed.
//	int T = sequence.size();
//  	for( int i=0; i<states.size(); i++ ) {
//    	    priors.put( states.get(i), (alpha[i][T-1]*beta[i][T-1]) );
//  	}
	

	if( DEBUG ) Debug.printtagln( tag, "done" );
    }

    /* Specified by ostore.oil.Model interface */
    public synchronized void clear() 
    {
	final String tag = class_tag+".clear";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	// See clear( int ).
	sequence.removeAllElements();
	priors.clear();
	priors.set_default( 1.0/states.size() );

	if( DEBUG ) Debug.printtagln( tag, "done" );
    }

    /**
     * Reconstructs <code>outputs</code> and <code>output_set</code> based
     * on the contents of the current <code>sequence</code>.
     * Adds all observations from the sequence to the output set,
     * discarding all other (old) outputs.
     */
    private void rebuild_output_set()
    {
	final String tag = class_tag+".rebuild_output_set";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	// Clear the output set, then re-observe the sequence.
	output_set.clear();
	outputs.removeAllElements();

	for( int i=0; i<sequence.size(); i++ ) {
	    QuickSerializable output = sequence.get(i);
	    if( !output_set.containsKey( output ) ) {
		output_set.put( output, new Integer(outputs.size()) );
		outputs.add( output );
	    }
	}
	
	// Remember to initialize *_ss, *_p arrays
	arrays_in_sync = false;
	
	if( DEBUG ) Debug.printtagln( tag, "done" );
    }

    /* Specified by ostore.oil.Model interface */
    public synchronized QuickSerializable[] outliers()
    {
	final String tag = class_tag+".outliers";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	// This method will most likely be used for reverse-lookup segment
	// paging.  Return the observations that are not strongly bound to
	// any state.
	QSVector outs = new QSVector();
	int M = states.size();
	int N = outputs.size();

	// Make sure that <code>*_ss</code>, <code>*_p</code> arrays are
	// initialized properly.
	sync_arrays();

	//FIXME
	if( DEBUG ) Debug.printtagln( tag, Matrix.toString( emission_ss ) );

	// Scan the emission counts for a column without any
	// significant entries.
	double threshold = 2;	// File seen in cluster twice
    columns:
	for( int j=0; j<N; j++ ) {
	    for( int i=0; i<M; i++ ) {
		if( emission_ss[i][j] > threshold ) {
		    // Good enough.  Try next output.
		    continue columns;
		}
  	    }
	    QuickSerializable outlier = outputs.get(j);
	    if( DEBUG ) Debug.printtagln( tag, "outlier (" + j + "): " +
					  outlier );
	    outs.add( outlier );
  	}
	
	return outs.toArray();
    }
    
    /* specified by ostore.oil.SegmentedModel */
    public synchronized void add_segment( SegmentedModel.Segment s )
    {
	final String tag = class_tag+".add_segment";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	if( !(s instanceof Segment) ) {
	    throw new IllegalArgumentException( "Not a Segment:\t"+s );
	}
	Segment hmms = (Segment)s;

	QuickSerializable state = hmms.state;
	if( state_set.containsKey( state ) ) {
	    if( DEBUG ) Debug.printtagln( tag, "Already have Segment for "
					  + " id \"" + state + "\"" );
	} else {
	    state_set.put( state, new Integer(states.size()) );
	    states.add(state);
	}
	marginals.put( state, hmms.marginal );
	transitions.put( state, hmms.transitions );
	emissions.put( state, hmms.emissions );

	// Remember to initialize *_ss, *_p arrays
	arrays_in_sync = false;
	    
	if( DEBUG ) Debug.printtagln( tag, "Done" );
	return;
    }

    /* Specified by ostore.oil.Model interface */
    public synchronized void add_delta( Model.Delta d )
    {
	final String tag = class_tag+".add_delta";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	if( !(d instanceof Delta) ) {
	    throw new IllegalArgumentException( "Not a Delta:\t" + d );
	}
	Delta hmmd = (Delta)d;

	// Only add information for the states we already have.
	QuickSerializable state;
	for( int i=0; i<hmmd.states.size(); i++ ) {
	    state = hmmd.states.get(i);
	    if( state_set.containsKey( state ) ) {
		marginals.add( state, hmmd.marginals.get( state ) );
		transitions.add( state, hmmd.transitions.get( state ) );
		emissions.add( state, hmmd.emissions.get( state ) );
	    } else {
		if( DEBUG ) Debug.printtagln( tag, "Ignoring state \""
					      +state+ "\"");
	    }
	}

	// Remember to initialize *_ss, *_p arrays
	arrays_in_sync = false;
	    
	if( DEBUG ) Debug.printtagln( tag, "done" );
	return;
    }

    /* Specified by ostore.oil.Model interface */
    public synchronized double loglikelihood()
    {
	final String tag = class_tag+".loglikelihood";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	// Make sure that <code>*_ss</code>, <code>*_p</code> arrays are
	// initialized properly.
	sync_arrays();
	
	// A single run of the alpha recursion can calculate the likelihood.
	double log_likelihood = alpha( sequence.size() );
	
	if( DEBUG ) Debug.printtagln( tag, "done: " + log_likelihood );
	return log_likelihood;
    }
 
    /* Specified by ostore.oil.Model interface */
    /**
     * Updates the internal model parameters using the EM algorithm.
     * Newly calculated Expected Sufficient Statistics are bundled and
     * returned as a <code>Delta</code>.  They are also added to this
     * model's sufficient statistics, so they will affect future
     * calculations.
     *
     * @see Model#recalculate
     *
     * @return a <code>Delta</code> containing the expected
     * transition, emission, and marginal counts for each state.  If no
     * observations have been made, a <code>null</code> delta will be returned.
     */
    public synchronized Model.Delta recalculate()
    {
	final String tag = class_tag+".recalculate";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	Delta d; 
	if( sequence.size() > 0 ) {
	    // Prepare, then run <code>EM</code>.  Then absorb the new expected
	    // sufficient statistics and clear the <code>sequence</code>.
	    EM( MAX_NUM_ITERATIONS );
	    d = construct_delta();
	    add_delta(d);
	    
	    // Now forget that old sequence.
	    clear();
	} else {
	    // Nothing has been observed.
	    d = null;
	}

	if( DEBUG ) Debug.printtagln( tag, "done" );
	return d;
    }
    
    /* specified by ostore.oil.SegmentedModel */
    /**
     * Selects the <code>num</code> most relevant <code>Segment</code>s,
     * discarding the rest.
     * This method chooses the top <code>Segment</code>s found by
     * {@link #predict}.
     *
     * @param num	The number of <code>Segment</code>s to retain.
     */
    public synchronized void choose_segments( int num )
    {
	final String tag = class_tag+".choose_segments";
	if( DEBUG ) Debug.printtagln( tag, "called" );	

	if( num < states.size() ) {
	    throw new IllegalArgumentException( "HMM does not contain " +
						num + " Segments" );
	}
	HashSet to_keep = new HashSet( num );
	// FIXME!! What is a useful horizon?
	// Weigh future equally with past, for now.
	int horizon = sequence.size();
	Prediction pred = (Prediction)predict( horizon );

	// Add the highest-weighted states from the prediction.
	for( int i=0; i<num && i<pred.states.length; i++ ) {
	    QuickSerializable state = pred.states[i];
	    to_keep.add( state );
	}

	QSVector rejects = new QSVector( states.size() - num );
	for( int i=0; i<states.size(); i++ ) {
	    QuickSerializable state = states.get(i);
	    if( ! to_keep.contains( state ) ) {
		rejects.add( state );
	    }
	}
	
	if( to_keep.size() < num ) {
	    if( DEBUG ) {
		Debug.printtagln( tag, "only " + to_keep.size() +
				  " segments predicted - choosing random "
				  + "segments to fill quota" );
	    }
	    Random r = new Random( 0 );	// Repeatability is nice for now.

	    while( to_keep.size() < num ) {
		// We select states with reinsertion.  
		int nominee = r.nextInt( rejects.size() );
		to_keep.add( rejects.remove( nominee ) );
	    }
	}
	// Finally, remove rejected <code>Segment</code>s.
	for( int i=0; i<rejects.size(); i++ ) {
	    if( DEBUG ) Debug.printtagln( tag, "dropping segment: " +
					  rejects.get(i) );
	    remove_segment( rejects.get(i) );
	}
	if( DEBUG ) {
	    for( int i=0; i<states.size(); i++ ) {
		Debug.printtagln( tag, "kept segment: " + states.get(i) );
	    }
	}

	if( DEBUG ) Debug.printtagln( tag, "done" );	
    }

    /**
     * Removes all information concerning a <code>Segment</code> from this
     * <code>Model</code>. 
     *
     * @param state	The ID for the <code>Segment</code> to drop.
     */
    private void remove_segment( QuickSerializable state )
    {
	// FIXME
	// Remember to initialize *_ss, *_p arrays
	arrays_in_sync = false;
    }

    /* Specified by ostore.oil.Model interface */
    /**
     * Estimates the current and future occurrences of each state.
     * This method uses much of the same algorithm as <code>recalculate</code>
     * to estimate marginal counts for each state, but extends the current
     * sequence of observations with a series of observations to new
     * imaginary observation.  Thus the states expected to occurr in the
     * near future (estimated by transition probabilities) are considered
     * in addition to the states believed to have just occurred.
     *
     * @param horizon	The number of observations to look into the future
     * 
     * @see Model#predict
     *
     * @return a <code>Prediction</code> containing the estimated state
     * marginal counts, sorted.
     */
    public synchronized Model.Prediction predict( int horizon )
    {
	final String tag = class_tag+".predict";
	if( DEBUG ) Debug.printtagln( tag, "called: horizon == " + horizon );
	if( DEBUG_TIME ) stopwatch.start( "predict" );

	int M = states.size();
	int T = sequence.size() + horizon;

	// Make sure that <code>*_ss</code>, <code>*_p</code> arrays are
	// initialized properly.
	sync_arrays();
	
	// Step 1: Run alpha- and beta-recursions to calculate estimate states
	double log_likelihood = alpha( T );
	beta( T );

	// Step 2: calculate estimated state marginals
	// save marginal_ess for later.

	// FIXME - is marginal_ess_save always empty here?
	double[] marginal_ess_save = marginal_ess;
	marginal_ess = new double[M];
	for( int t=0; t<T; t++ ) {
	    for( int i=0; i<M; i++ ) {
		marginal_ess[i] += alpha[i][t]*beta[i][t];
	    }
	}
	
	// Step 3: Sort the states by their estimated marginal counts
	// (highest first).
	QuickSerializable[] ids = states.toArray();
	Arrays.sort( ids, new Comparator() {
		public int compare( Object o1, Object o2 ) {
		    int i1 = ((Integer)state_set.get( o1 )).intValue();
		    int i2 = ((Integer)state_set.get( o2 )).intValue();
		    // states with lower marginals should rank higher.
		    // Don't worry about -0 vs. 0, etc.		    
		    return ((marginal_ess[i1] < marginal_ess[i2]) ? 1 :
			    ((marginal_ess[i1] > marginal_ess[i2]) ? -1 : 0 ));
		}
	    } );

	// Step 4: Copy marginal_ess[] into the sorted order.
	double[] marginal_pred = new double[M];
	for( int i=0; i<M; i++ ) {
	    marginal_pred[i] =
		marginal_ess[((Integer)state_set.get(ids[i])).intValue()];
	}
	marginal_ess = marginal_ess_save;
	
	// Step 5: Create Prediction
	// FIXME!!
	// Sparsify prediction?
	// FIXME!! log_likelihood positive?
	Prediction prediction = new Prediction( ids, marginal_pred,
						log_likelihood );

  	if( DEBUG_LEVEL >= DEBUG_SOME ) {
	    Debug.printtagln( tag, "marginal_pred: " + prediction );
	}
	if( DEBUG_TIME ) {
	    Debug.printtagln( time_tag, "predict ran in " +
			      stopwatch.stop( "predict" ) + " ms." );
	}
	if( DEBUG ) Debug.printtagln( tag, "done" );
	return prediction;
    }

    
    /**
     * Estimates the model parameters using the Expectation-Maximization
     * method.  The method loops across <code>E_step</code> and
     * <code>M_step</code> until either the log-likelihood converges or the
     * maximum number of iterations has been completed.
     *
     * @see #M_step
     * @see #E_step
     *
     * @param max_iterations	The maximum number of iterations to
     * complete before returning.
     *
     * @return the log-likelihood of the <code>sequence</code> given the
     * most recent (probably convergent) model parameters.
     */
    private double EM( int max_iterations )
    {
	final String tag = class_tag+".EM";
	// A few commonly used limits:
	int M = states.size();
	int N = outputs.size();
	int T = sequence.size();
	
	if( DEBUG ) Debug.printtagln( tag, "Sequence length: " + T +
				      "\tM: " + M + "\tN: " + N );
	if( DEBUG ) Debug.printtagln( tag, "Free/Total Memory:\t" +
				      Runtime.getRuntime().freeMemory() + "\t" +
				      Runtime.getRuntime().totalMemory() );
	if( DEBUG_TIME ) stopwatch.start( "EM" );

	if( T == 0 ) {
	    return Double.NaN;	    // We haven't seen any outputs to analyze.
	}
	if( DEBUG_LEVEL == DEBUG_ALL ) {
	    // Print out sequence
	    for( int t=0; t<T; t++ ) {
		Object output = sequence.get(t);
		Debug.printtagln( tag, "sequence["+ t + "]: " +
				  output +" (" + output_set.get(output) +")");
	    }
	}

	// Make sure that <code>*_ss</code>, <code>*_p</code> arrays are
	// initialized properly.
	sync_arrays();

	/*
	 * Likelihoods
	 * We store the log-likelihood calculated during the last two
	 * iterations, in order to detect convergence.
	 */
	double log_likelihood = Double.NEGATIVE_INFINITY;
	double last_ll = log_likelihood;

	// Now loop through the Expectation-Maximization method.
	for( int n=0; n<max_iterations; n++ ) {

	    if( DEBUG ) Debug.printtagln( tag, "iteration #" + n );
	    if( DEBUG_TIME ) stopwatch.start( "E_step" );

	    log_likelihood = E_step();

	    if( DEBUG_TIME ) {
		Debug.printtagln( time_tag, "E_step ran in " +
				  stopwatch.split("E_step") + " ms." );
	    }
	    if( DEBUG ) Debug.printtagln( tag, "Log-likelihood:\t" +
					  log_likelihood );

	    if( Math.abs(log_likelihood - last_ll) < (epsilon * T) ) {
		if( DEBUG ) Debug.printtagln( tag, "\t*** CONVERGED ***" );
		break;
	    } 
	    last_ll = log_likelihood;

	    if( DEBUG_TIME ) stopwatch.start( "E_step" );
	    
	    M_step();
	    
	    if( DEBUG_TIME ) {
		Debug.printtagln( time_tag, "M_step ran in " +
				  stopwatch.split( "M_step" ) + " ms." );
	    }
	}

	if( DEBUG_TIME ) {
	    Debug.printtagln( time_tag, "EM ran in " +
			      stopwatch.stop( "EM" ) + " ms." );
	}
	return log_likelihood;
    }

    /**
     * Perform the "expectation" step of the EM algorithm.
     * Use an inference algorithm, here an adaptation of the Baum-Welch
     * "forwards" and "backwards" sweeps.
     *
     * This method reads the probability matrices
     * (<code>marginal_p</code>, etc.)
     * And fills in the expected sufficient statistics
     * (<code>marginal_ess</code>, etc.)
     *
     * @return the log-likelihood of the output sequence given the current
     * model parameters.  This likelihood can be used to test convergence
     * of EM.
     */
    private double E_step()
    {
	final String tag = class_tag+".E_step";
	if( DEBUG ) Debug.printtagln( tag, "called" );
	// TODO: use local shadows for the global temp arrays.

	double log_likelihood;		// log p(y_0,...,y_t)
	int M = states.size();
	int N = outputs.size();
	int T = sequence.size();
	int output_index;

	// Step 1: alpha recursion
	log_likelihood = alpha( T );
	
	// Step 2: beta recursion
	beta( T );

	// Step 3: Calculate expected sufficient statistics.

	// Step 3.0; clear ess matrices;
	for( int i=0; i<M; i++ ) {
	    marginal_ess[i] = 0;
	    prior_ess[i] = 0;
	    for( int j=0; j<M; j++ ) {
		transition_ess[i][j] = 0;
	    }
	    for( int j=0; j<N; j++ ) {
		emission_ess[i][j] = 0;
	    }
	}

	// Step 3.1: priors
	for( int i=0; i<M; i++ ) {
	    if( Double.isNaN(alpha[i][0]*beta[i][0]) ) {
		Carp.out( "Nan! " + i + " " + alpha[i][0] + " " + beta[i][0]);
	    }
	    prior_ess[i] = (alpha[i][0]*beta[i][0]);
	}

	if( DEBUG_TIME ) stopwatch.start( "marginal" );

	// Step 3.2: state marginal: sum across 'gamma'
	for( int t=0; t<T; t++ ) {
	    for( int i=0; i<M; i++ ) {
		marginal_ess[i] += alpha[i][t]*beta[i][t];
	    }
	}
	
	if( DEBUG_TIME ) {
	    Debug.printtagln( time_tag, "marginal recursion ran in " +
			      stopwatch.stop( "marginal" ) + " ms." );
	    stopwatch.start( "transition" );
	}
	
	// Step 3.3: transitions: sum across 'xi'

	// TODO: Sparsify this loop?
	// this loop is O(MMT), the bottleneck of our algorithm.  To speed
	// it up, we would have to approximate the transition counts with a
	// different estimator.
	for( int t=0; t<T-1; t++ ) {
	    output_index =
		((Integer)output_set.get(sequence.get(t+1))).intValue();

	    for( int i=0; i<M; i++ ) {
		for( int j=0; j<M; j++ ) {
		    transition_ess[i][j] +=
			((alpha[i][t] * beta[j][t+1] * transition_p[i][j] *
			  emission_p[j][output_index]) / normals[t+1]);
		}
	    }
	}

	if( DEBUG_TIME ) {
	    Debug.printtagln( time_tag, "transition recursion ran in " +
			      stopwatch.stop( "transition" ) + " ms." );
	    stopwatch.start( "emission" );
	}
	
	// Step 3.4: emissions: sum across 'gamma .* sequence'
	for( int t=0; t<T; t++ ) {
	    output_index = ((Integer)output_set.get(sequence.get(t))).intValue();
	    for( int i=0; i<M; i++ ) {
		emission_ess[i][output_index] += alpha[i][t]*beta[i][t];
	    }
	}

	if( DEBUG_TIME ) {
	    Debug.printtagln( time_tag, "emission recursion ran in " +
			      stopwatch.stop( "emission" ) + " ms." );
	}

	// Done.  All expected sufficient statistics are calculated.
	if( DEBUG ) Debug.printtagln( tag, "done" );
	return( log_likelihood );
    }

    /**
     * Performs the alpha (forward) recursion for the expectation step of
     * the EM algorithm.  This method is used in both <code>E_step</code>
     * and <code>predict</code>.  It understands how to iterate beyond the
     * observed sequence, providing one method of state prediction.
     *
     * @param T	The number of time steps to iterate (may be greater than
     * the sequence length).
     *
     * @return the log-likelihood of the output sequence given the current
     * model parameters.
     *
     * @see #beta
     * @see #E_step
     */
    private double alpha( int T )
    {
	final String tag = class_tag+".alpha";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	double log_likelihood = 0;		// log p(y_0,...,y_t)
	int M = states.size();
	int sequence_length = sequence.size();
	int output_index;

	// Re-allocate the temp arrays if they are either null or too small
	// Add a little extra room because <code>T</code> may grow quickly.
	if((alpha == null) ||
	   (alpha.length < M) || (alpha[0].length < T)) {

	    alpha = new double[M][T+10];	// p(q_t|y_0,...,y_t)
	    beta = new double[M][T+10]; // p(y_t+1,...,y_T|q_t)/p(y_0,...,y_t)
	    normals = new double[T+10];	// p(y_t|y_0,...,y_t-1)
	}
	
	if( DEBUG_TIME ) stopwatch.start( "alpha" );

	// Initialize alpha[0][t].
	normals[0] = 0;
	if( sequence_length > 0 ) {
	    output_index =
		((Integer)output_set.get(sequence.get(0))).intValue();
	    for( int i=0; i<M; i++ ) {
		alpha[i][0] = (prior_p[i] * emission_p[i][output_index]);
		// Now update normalization factors.
		normals[0] += alpha[i][0];
	    }
	} else {
	    for( int i=0; i<M; i++ ) {
		alpha[i][0] = prior_p[i];
		// Now update normalization factors.
		normals[0] += alpha[i][0];
	    }
	}	    
	// Normalize alpha.
	for( int i=0; i<M; i++ ) {
	    alpha[i][0] = alpha[i][0] / normals[0];
	}
	log_likelihood = Math.log(normals[0]);
	
	// Iterate forward to end of sequence
	for( int t=1; t<sequence_length; t++ ) {
	    output_index =
		((Integer)output_set.get(sequence.get(t))).intValue();
	    normals[t] = 0;
	    
	    // Update all states at time t
	    for( int i=0; i<M; i++ ) {
		alpha[i][t] = 0;
		// merge info from all states at time t-1 into time t
		for( int j=0; j<M; j++ ) {
		    alpha[i][t] += (alpha[j][t-1] * transition_p[j][i] *
				    emission_p[i][output_index]);
		}
		// Now update normalization factors.
		normals[t] += alpha[i][t];
	    }
	    // Normalize alpha.
	    for( int i=0; i<M; i++ ) {
		alpha[i][t] = alpha[i][t] / normals[t];
	    }
	    log_likelihood += Math.log(normals[t]);
	}	// end alpha recursion

	// Iterate further, across imaginary outputs.
	// Only transitions matter.
	for( int t=Math.max(sequence_length,1); t<T; t++ ) {
	    normals[t] = 0;
	    
	    // Update all states at time t
	    for( int i=0; i<M; i++ ) {
		alpha[i][t] = 0;
		// merge info from all states at time t-1 into time t
		for( int j=0; j<M; j++ ) {
		    alpha[i][t] += (alpha[j][t-1] * transition_p[j][i]);
		}
		// Now update normalization factors.
		normals[t] += alpha[i][t];
	    }
	    // Normalize alpha.
	    for( int i=0; i<M; i++ ) {
		alpha[i][t] = alpha[i][t] / normals[t];
	    }
	    log_likelihood += Math.log(normals[t]);
	}	// end alpha recursion

	if( DEBUG_LEVEL >= DEBUG_SOME ) {
	    // Print out alpha (transposed, for readability)
	    Debug.printtagln( tag, "max alpha\':" );
	    double [][] alpha_prime = Matrix.transpose(alpha);
	    int display_m = (DEBUG_LEVEL==DEBUG_ALL) ? M : 4;
	    int[][] indices = new int[T][display_m];
	    Matrix.max( alpha_prime, indices );
	    Debug.printtagln( tag, Matrix.toString( alpha_prime, indices ) );
	}
	
	if( DEBUG_TIME ) {
	    Debug.printtagln( time_tag, "alpha recursion ran in " +
			      stopwatch.stop( "alpha" ) + " ms." );
	}
	if( DEBUG ) Debug.printtagln( tag, "done" );
	return log_likelihood;
    }

    /**
     * Performs the beta (backward) recursion for the expectation step of
     * the EM algorithm.  This method is used in both <code>E_step</code>
     * and <code>predict</code>.  It understands how to iterate beyond the
     * observed sequence, providing one method of state prediction.
     *
     * @param T	The number of time steps to iterate (may be greater than
     * the sequence length).
     *
     * @see #alpha
     * @see #E_step
     */
    private void beta( int T )
    {
	final String tag = class_tag+".beta";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	int M = states.size();
	int sequence_length = sequence.size();
	int output_index;

	if( DEBUG_TIME ) stopwatch.start( "beta" );

	// Initialize beta.
	for( int i=0; i<M; i++ ) {
	    beta[i][T-1] = 1;
	}

	// Iterate backwards to end of sequence
	for( int t=T-2; t>=Math.max(0,sequence_length-1); t-- ) {
	    // update all states at time t
	    for( int i=0; i<M; i++ ) {
		beta[i][t] = 0;
		// merge info from all states at time t+1 into time t
		for( int j=0; j<M; j++ ) {
		    beta[i][t] += (beta[j][t+1] * transition_p[i][j]);
		}
	    }
	    // Normalize beta.
	    for( int i=0; i<M; i++ ) {
		beta[i][t] = beta[i][t] / normals[t+1];
	    }
	}
	// Iterate backwards across sequence
	for( int t=sequence_length-2; t>=0; t-- ) {
	    output_index =
		((Integer)output_set.get(sequence.get(t+1))).intValue();

	    // update all states at time t
	    for( int i=0; i<M; i++ ) {
		beta[i][t] = 0;
		// merge info from all states at time t+1 into time t
		for( int j=0; j<M; j++ ) {
		    beta[i][t] += (beta[j][t+1] * transition_p[i][j] *
				   emission_p[j][output_index]); 
		}
	    }
	    // Normalize beta.
	    for( int i=0; i<M; i++ ) {
		beta[i][t] = beta[i][t] / normals[t+1];
	    }
	}	// end beta recursion

	if( DEBUG_LEVEL >= DEBUG_SOME ) {
	    // Print out beta (transposed)
	    Debug.printtagln( tag, "max beta\':" );
	    double [][] beta_prime = Matrix.transpose(beta);
	    int display_m = (DEBUG_LEVEL==DEBUG_ALL) ? M : 4;
	    int[][] indices = new int[T][display_m];
	    Matrix.max( beta_prime, indices );
	    Debug.printtagln( tag, Matrix.toString( beta_prime, indices ) );
	}

	if( DEBUG_TIME ) {
	    Debug.printtagln( time_tag, "beta recursion ran in " +
			      stopwatch.stop( "beta" ) + " ms." );
	}
    }
    
    /**
     * Performs the "maximization" step of the EM algorithm.
     * Apply the Baum-Welch updates to estimate the new parameters.
     * This method reads the sufficient statistic matrices
     * (<code>marginal_ss</code>, <code>marginal_ess</code>, etc.)
     * And fills in the probability matrices
     * (<code>marginal_p</code>, etc.)
     * This method must be called before <code>E_step</code> or
     * </code>alpha</code> if the sufficient statistics have been modified
     * (e.g. by <code>fill_arrays</code> or a previous iteration of
     * <code>E_step</code>).
     */
    private void M_step()
    {
	final String tag = class_tag+".M_step";
	if( DEBUG ) Debug.printtagln( tag, "called" );
	
	int M = states.size();
	int N = outputs.size();
	// Allocate probability arrays if null or too small.
	if((emission_p == null) ||
	   (emission_p.length < M) || (emission_p[0].length < N)) {
	    
	    transition_p = new double[M][M];	// MLE p(q_t,q_t+1)
	    emission_p = new double[M][N];	// MLE p(q_t,y_t)
	    prior_p = new double[M];		// MLE p(q_0)
	}
	
	double prior_norm = 0;
	double row_norm;

	for( int i=0; i<M; i++ ) {
	    row_norm = marginal_ss[i] + marginal_ess[i];
	    
	    // Convert transitions
	    for( int j=0; j<M; j++ ) {
		transition_p[i][j] =
		    ((transition_ess[i][j]+transition_ss[i][j])/row_norm);
	    }
	    
	    // Convert emissions
	    for( int j=0; j<N; j++ ) {
		emission_p[i][j] =
		    ((emission_ess[i][j]+emission_ss[i][j])/row_norm);
	    }

	    // Calculate sum of prior_*ss
	    prior_norm += prior_ess[i]+prior_ss[i];
	}	    
	// Convert priors second, now that we have prior_norm.
	for( int i=0; i<M; i++ ) {
	    prior_p[i] = ((prior_ess[i]+prior_ss[i])/prior_norm);
	}

	if( DEBUG_LEVEL >= DEBUG_SOME ) {
	    int display_m = (DEBUG_LEVEL==DEBUG_ALL) ? M : 4;
	    int display_n = (DEBUG_LEVEL==DEBUG_ALL) ? N : Math.min(N,4);
	    int[] indices = new int[display_m];
	    int[][] m_indices = new int[M][display_m];

	    // Print out marginals
	    Array.max( marginal_ess, indices );
	    Debug.printtagln( tag, "marginal_ess: " +
			      Array.toString( marginal_ess, indices ) );

	    Array.max( marginal_ss, indices );
	    Debug.printtagln( tag, "marginal_ss: " +
			      Array.toString( marginal_ss, indices ) );
	    
	    // Print out probability matrices
	    // priors
	    Array.max( prior_p, indices );
	    Debug.printtagln( tag, "prior_p: " +
			      Array.toString( prior_p, indices ) );
	    // transitions
	    Matrix.max( transition_p, m_indices );
	    Debug.printtagln( tag, "transition_p: " +
			      Matrix.toString( transition_p, m_indices ) );

	    // emissions
	    m_indices = new int[M][display_n];
	    Matrix.max( emission_p, m_indices );
	    Debug.printtagln( tag, "emission_p: " +
			      Matrix.toString( emission_p, m_indices ) );
	}
	
	if( DEBUG ) Debug.printtagln( tag, "done" );
	return;
    }

    /**
     * Synchronizes internal arrays to their sparse counterparts.
     * The <code>*_ss[]</code> and <code>*_p[]</code> arrays are
     * initialized, and the <code>*_ess[]</code> are allocated and zeroed.
     * <p>
     * This method must be called before <code>E_step</code> or
     * </code>alpha</code> if the model parameters or output set have been
     * modified (e.g. by <code>add_delta</code> or
     * <code>rebuild_output_set</code>).
     */
    private void sync_arrays()
    {
	final String tag = class_tag+".sync_arrays";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	if( arrays_in_sync ) {
	    if( DEBUG ) Debug.printtagln( tag, "already synchronized" );
	} else {
	    fill_arrays();
	    M_step();
	    arrays_in_sync = true;
	}
	if( DEBUG ) Debug.printtagln( tag, "done" );
	return;
    }
    
    /**
     * Initializes the sufficient statistic arrays (<code>*_ss[]</code>).
     * Clears the expected sufficient statistic arrays
     * (<code>*_ess[]</code>).
     * Reads <code>marginals</code>, <code>transitions</code>,
     * <code>emissions</code>, and <code>priors</code>.
     * This method must be called before <code>M_step</code> if the model
     * parameters or output set have been modified (e.g. by
     * <code>add_delta</code> or <code>record</code>).
     */
    private void fill_arrays()
    {
	final String tag = class_tag+".fill_arrays";
	if( DEBUG ) Debug.printtagln( tag, "called" );
	int M = states.size();
	int N = outputs.size();

	// Re-allocate the temp arrays if they are either null or too small.
	if((emission_ss == null) ||
	   (emission_ss.length < M) || (emission_ss[0].length < N)) {
	    
	    marginal_ss = new double[M];		// marginals.toArray
	    transition_ss = new double[M][M];	// transitions.toArray
	    emission_ss = new double[M][N];		// emissions.toArray
	    prior_ss = new double[M];		// priors.toArray
	    
	    marginal_ess = new double[M];		// E[\sum_t(q)]
	    transition_ess = new double[M][M]; 	// E[\sum_t(q_t,q_t+1)]
	    emission_ess = new double[M][N];	// E[\sum_t(q_t,y_t)]
	    prior_ess = new double[M];		// E[q_0]
	}
	
	QuickSerializable id1;
	Array row;
	for( int i=0; i<M; i++ ) {
	    id1 = states.get(i);

	    // Convert marginals
	    marginal_ss[i] = marginals.get(id1);
	    marginal_ess[i] = 0;

	    // Convert transitions
	    row = transitions.get(id1);
	    for( int j=0; j<M; j++ ) {
		transition_ss[i][j] = row.get(states.get(j));
		transition_ess[i][j] = 0;
	    }
	    
	    // Convert emissions
	    row = emissions.get(id1);
	    for( int j=0; j<N; j++ ) {
		emission_ss[i][j] = row.get(outputs.get(j));
		emission_ess[i][j] = 0;
	    }
	    
	    // Convert priors
	    prior_ss[i] = priors.get(id1);
	    prior_ess[i] = 0;
	}
	if( DEBUG ) Debug.printtagln( tag, "done" );
	return;
    }
    
    /**
     * Creates a <code>Delta</code> containing information extracted
     * from <code>marginal_ess</code>, <code>transition_ess</code>,
     * and <code>emission_ess</code>.
     *
     * @return a <code>Delta</code> storing the sufficient statistics
     * from <code>*_ess</code>, which were written by <code>EM</code>.
     */
    private Delta construct_delta()
    {
	final String tag = class_tag+".construct_delta";
	if( DEBUG ) Debug.printtagln( tag, "called" );
	Array new_marginals = new Array();
	Matrix new_transitions = new Matrix();
	Matrix new_emissions = new Matrix();

	int M = states.size();
	int N = outputs.size();
	QuickSerializable id1;
	Array row;
	
	for( int i=0; i<M; i++ ) {
	    id1 = states.get(i);

	    // FIXME!! Drop states with low marginal?
	    // Sparsify emissions, and drop then?
	    // Sparsify based on threshold, not count.
	    
	    // Extract marginals
	    new_marginals.put( id1, marginal_ess[i] );

	    // Extract transitions
	    row = new Array();
	    for( int j=0; j<M; j++ ) {
		row.put( states.get(j), transition_ess[i][j] );
	    }
	    new_transitions.put( id1, row );

	    // Extract emissions
	    row = new Array();
	    for( int j=0; j<N; j++ ) {
		row.put( outputs.get(j), emission_ess[i][j] );
	    }
	    new_emissions.put( id1, row );
	}
	// FIXME!!
	// Print out results for debugging
	if( DEBUG_LEVEL >= DEBUG_SOME ) {
	    int[][] m_indices = new int[N][4];

	    // emissions
	    double[][] emission_prime = Matrix.transpose(emission_p);
	    Matrix.max( emission_prime, m_indices );
	    Debug.printtagln( tag, "emission_prime: " +
			      Matrix.toString( emission_prime, m_indices ) );
	}
	// FIXME!!! drop meaningless stats
	// Sparsify arrays?

	if( DEBUG ) Debug.printtagln( tag, "done" );
	return new Delta( states, new_marginals,
			  new_transitions, new_emissions );
    }
    
    /* Specified by ostore.util.QuickSerializable */
    public void serialize( OutputBuffer buffer )
    {
	buffer.add( epsilon );

	// Use the <code>Delta</code> machinery to compress storage.
	Delta all = new Delta( states, marginals,
			       transitions, emissions );
	buffer.add( all );

	// Now store the info related to our specific sequence
	buffer.add( sequence );
	buffer.add( priors );
    }


    ////////////////////////////////////////////////////////////////
    ////////////////////     INNER CLASSES	////////////////////

    /**
     * A <code>HMM.Delta</code> object contains a diff of the sufficient
     * statistics for an <code>HMM</code>.
     *
     * @author	Dennis Geels
     * @version	$Id: HMM.java,v 1.31 2002/07/20 19:38:03 srhea Exp $
     */

    public static class Delta implements Model.Delta
    {
	////////////////////     CONSTANTS	////////////////////
	////////////////////  STATIC VARIABLES 	////////////////////
	////////////////////     VARIABLES	////////////////////
	
	/**
	 * The set of states referenced in the sufficient statistics.
	 * The order is unimportant.
	 */
	public QSVector states;

	/**
	 * Estimated marginal counts for each state.
	 */
	public Array marginals;
	/**
	 * Estimated count of state-state transitions.
	 */
	public Matrix transitions;
	/**
	 * Estimated count of state-output emissions.
	 */
	public Matrix emissions;
	    
	
	////////////////////    CONSTRUCTORS	////////////////////

	/**
	 * Constructs a <code>Delta</code> with the specified
	 * <code>HMM</code> data.  The statistics should only contain
	 * entries for the specified <code>states</code>.
	 *
	 * @param states	The set of states contained in this
	 * 			<code>Delta</code>
	 * @param marginals	A <code>QuickSerializable->double</code>
	 * mapping from states to their marginal sufficient statistic.
	 * @param transitions	A <code>QuickSerializable->Array</code>
	 * mapping from states to their transition sufficient statistics.
	 * @param emissions	A <code>QuickSerializable->Array</code>
	 * mapping from states to their emission sufficient statistics.
	 */
	public Delta( QSVector states, Array marginals,
		      Matrix transitions, Matrix emissions )
	{
	    this.states = states;
	    this.marginals = marginals;
	    this.transitions = transitions;
	    this.emissions = emissions;
	}

	/**
	 * Construct a <code>Delta</code> from its
	 * <code>QuickSerializable</code> form.
	 * See {@link #to_bytes(byte[], int[])}.
	 */
//  	public Delta( byte [] data, int [] offset ) throws QSException
//  	{
//  	    // First deserialize the state ids
//  	    states = new QSVector( data, offset );
	
//  	    // Deserialize the maps for decompression
//  	    // First find the location of the maps relative to here.
//  	    // Adjust for the space taken up by the int itself.
//  	    int map_offset_shift = ByteUtils.bytesToInt( data, offset ) - 4;
//  	    int[] map_offset = {(offset[0] + map_offset_shift)};
	    
//  	    int size = ByteUtils.bytesToInt( data, map_offset );
//  	    HashMap state_map = new HashMap( size );
//  	    QuickSerializable id;
//  	    int type, index;

//  	    for( int i=0; i<size; i++ ) {
//  		type = ByteUtils.bytesToInt( data, map_offset );
//  		id = TypeTable.new_object( type, data, map_offset );
//  		index = ByteUtils.bytesToInt( data, map_offset );
//  		state_map.put( new Integer(index), id );
//  	    }
	
//  	    size = ByteUtils.bytesToInt( data, map_offset );
//  	    HashMap output_map = new HashMap( size );
//  	    for( int i=0; i<size; i++ ) {
//  		type = ByteUtils.bytesToInt( data, map_offset );
//  		id = TypeTable.new_object( type, data, map_offset );
//  		index = ByteUtils.bytesToInt( data, map_offset );
//  		output_map.put( new Integer(index), id );
//  	    }
	
//  	    // Now double back and deserialize the statistics.
//  	    marginals = new Array( data, offset, state_map );
//  	    transitions = new Matrix( data, offset, state_map, state_map );
//  	    emissions = new Matrix( data, offset, state_map, output_map );
	    
//  	    // Finally, return the correct offset
//  	    offset[0] = map_offset[0];
//  	}

	/**
	 * Constructs a <code>Delta</code> from its
	 * <code>serialize</code>d form.
	 */
	public Delta( InputBuffer buffer ) throws QSException
	{
	    throw new NoSuchMethodError();
	    // FIXME out-of-order trick.
	}
	
	////////////////////   STATIC METHODS 	////////////////////
	//////////////////// 	  METHODS	////////////////////
	
	/* Specified by ostore.oil.Model.Delta interface */
	public Model.Delta sum( Model.Delta other )
	{
	    if( !(other instanceof Delta) ) {
		throw new IllegalArgumentException( "Not an Delta:\t" +
						    other );
	    }
	    Delta o = (Delta)other;

	    // First find the union of both sets of states
	    QSVector sum_states = new QSVector();
	    // Could use HashSet here:
	    HashMap state_set = new HashMap(states.size()+o.states.size());
	    QuickSerializable state;
	    for( int i=0; i<states.size(); i++ ) {
		state = states.get( i );
		sum_states.add( state );
		state_set.put( state, state );
	    }
	    for( int i=0; i<o.states.size(); i++ ) {
		state = o.states.get( i );
		if( !state_set.containsKey( state ) ) 
		    sum_states.add( state );
	    }

	    // Now add in the statistics from <code>other</code>.
	    Array sum_marginals = (Array)marginals.clone();
	    sum_marginals.add( o.marginals );
	    Matrix sum_transitions = (Matrix)transitions.clone();
	    sum_transitions.add( o.transitions );
	    Matrix sum_emissions = (Matrix)emissions.clone();
	    sum_emissions.add( o.emissions );
	   
	    return new Delta( sum_states, sum_marginals,
				  sum_transitions, sum_emissions );
	}

	/* Specified by ostore.util.QuickSerializable */
//  	public synchronized void to_bytes( byte [] data, int [] offset )
//  	{
//  	    // First write out the states.  We effectively duplicate this
//  	    // information, because the state_map should contain the same
//  	    // id's.  The waste is small, though, and this way we need not
//  	    // check for a one-to-one relationship.
//  	    states.to_bytes( data, offset );
	    
//  	    // We need to store the compression maps after all the
//  	    // sufficient statistics.
//  	    // Save room here so we can store the maps' offset later.
//  	    int[] map_offset = {offset[0]};
//  	    offset[0] += ByteUtils.SIZE_INT;
	    
//  	    // Use compression for the statistics.
//  	    HashMap state_map = new HashMap( states.size() );
//  	    HashMap output_map = new HashMap();

//  	    marginals.to_bytes( data, offset, state_map );
//  	    transitions.to_bytes( data, offset, state_map, state_map );
//  	    emissions.to_bytes( data, offset, state_map, output_map );

//  	    // Patch in the relative offset of the compression maps
//  	    ByteUtils.intToBytes( (offset[0]-map_offset[0]), data, map_offset );

//  	    // Store the compression maps
//  	    ByteUtils.intToBytes( state_map.size(), data, offset );
//  	    Iterator ids = state_map.keySet().iterator();
//  	    QuickSerializable id;
//  	    int index;
//  	    while( ids.hasNext() ) {
//  		id = (QuickSerializable)ids.next();
//  		// First store the id
//  		ByteUtils.intToBytes( id.type_code(), data, offset );
//  		id.to_bytes( data, offset );
//  		// Then store its mapped index
//  		index = ((Integer)(state_map.get(id))).intValue();
//  		ByteUtils.intToBytes( index, data, offset );
//  	    }

//  	    ByteUtils.intToBytes( output_map.size(), data, offset );
//  	    ids = output_map.keySet().iterator();
//  	    while( ids.hasNext() ) {
//  		id = (QuickSerializable)ids.next();
//  		// First store the id
//  		ByteUtils.intToBytes( id.type_code(), data, offset );
//  		id.to_bytes( data, offset );
//  		// Then store its mapped index
//  		index = ((Integer)(output_map.get(id))).intValue();
//  		ByteUtils.intToBytes( index, data, offset );
//  	    }
	    
//  	    return;
//  	}

	/* Specified by ostore.util.QuickSerializable */
	public void serialize( OutputBuffer buffer )
	{
	    throw new NoSuchMethodError();
	    // FIXME out-of-order trick.
	}

	
    }	// end class Delta
    

    /**
     * A <code>Prediction</code> object contains an array of states
     * and a matching array of weights.  The arrays are sorted, with the
     * most likely (highest-weighted) states first.  The probability of
     * each state is computed using the <code>HMM</code> Maximum Likelihood
     * estimates.
     *
     * @author	Dennis Geels
     * @version	$Id: HMM.java,v 1.31 2002/07/20 19:38:03 srhea Exp $
     *
     * @see HMM#predict
     * @see Model.Prediction
     */
    public static class Prediction implements Model.Prediction
    {
	////////////////////     CONSTANTS	////////////////////
	////////////////////  STATIC VARIABLES 	////////////////////
	////////////////////     VARIABLES	////////////////////

	/**
	 * The states predicted to occur soon.
	 */
	public QuickSerializable[] states;
	
	/**
	 * The estimated weight (confidence) for each state.
	 * Each weight is equal to the number of times the state is
	 * expected to occur in the near future.
	 */
	public double[] weights;

	/**
	 * The logarithm of the likelihood of this <code>Prediction</code>
	 * given the current model parameters (<code>p(D|\theta)</code>).
	 */
	public double log_likelihood;
	
	////////////////////    CONSTRUCTORS	////////////////////

	/**
	 * Creates a new <code>Prediction</code> with the specified
	 * states and their weights.
	 */
	public Prediction( QuickSerializable[] states,
			   double[] weights, double log_likelihood )
	{
	    this.states = states;
	    this.weights = weights;
	    this.log_likelihood = log_likelihood;
	}
	
	/**
	 * Constructs a <code>Prediction</code> from its
	 * <code>serialize</code>d form.
	 */
	public Prediction( InputBuffer buffer ) throws QSException
	{
	    int len = buffer.nextInt();
	    states = new QuickSerializable[len];
	    weights = new double[len];

	    for( int i=0; i<len; i++ ) {
		states[i] = (QuickSerializable)buffer.nextObject();
		weights[i] = buffer.nextDouble();
	    }

	    log_likelihood = buffer.nextDouble();
	}
	

	////////////////////   STATIC METHODS 	////////////////////
	//////////////////// 	  METHODS	////////////////////

	/**
	 * Produces a brief human-readable summary of this
	 * <code>Prediction</code>.
	 *
	 * @return a <code>String</code> containing the top few predicted
	 * states, ordered by weight.
	 */
	public synchronized String toString()
	{
	    NumberFormat nf = NumberFormat.getInstance();
	    StringBuffer sbuf = new StringBuffer( "(ll: " );
	    sbuf.append( nf.format(log_likelihood) );
	    sbuf.append( "){" );

	    int display_m = (DEBUG_LEVEL==DEBUG_ALL) ? states.length : 4;
	    int num_frac_digits = 4;
	    for( int i=0; i<display_m; i++ ) {
		sbuf.append( '<' );
		sbuf.append( states[i].toString() );
		sbuf.append( ':' );
		// Set the formatter for values.
		nf.setMinimumIntegerDigits( 1 );
		nf.setMaximumFractionDigits( num_frac_digits );
		nf.setMinimumFractionDigits( num_frac_digits );
		sbuf.append( nf.format( weights[i] ) );
		sbuf.append( ">" );
		sbuf.append(',');
	    }
	    sbuf.setCharAt( sbuf.length()-1, '}' );

	    return sbuf.toString();
	}

	/* Specified by ostore.util.QuickSerializable */
	public void serialize( OutputBuffer buffer )
	{
	    buffer.add( states.length );
	    for( int i=0; i<states.length; i++ ) {
		buffer.add( states[i] );
		buffer.add( weights[i] );
	    }
	    buffer.add( log_likelihood );
	}
    }

    /**
     * A <code>Segment</code> object contains the internal model state
     * associated with a single state.  It can be used to swap the
     * information to/from an <code>HMM</code>.
     *
     * @author	Dennis Geels
     * @version	$Id: HMM.java,v 1.31 2002/07/20 19:38:03 srhea Exp $
     */
    public static class Segment implements SegmentedModel.Segment
    {
	////////////////////     CONSTANTS	////////////////////
	////////////////////  STATIC VARIABLES 	////////////////////
	////////////////////     VARIABLES	////////////////////

	/**
	 * The ID for the state whose information is stored in
	 * <code>Segment</code>. 
	 */
	public QuickSerializable state;

	/**
	 * Estimated marginal count for this state
	 */
	public double marginal;
	
	/**
	 * Estimated count of transitions out of this state.
	 */
	public Array transitions;
	
	/**
	 * Estimated count of output emissions from this state.
	 */
	public Array emissions;
	
	////////////////////    CONSTRUCTORS	////////////////////

	/**
	 * Create a new <code>Segment</code>.
	 *
	 * @param state	The ID for the state whose information is stored in
	 * 		this <code>Segment</code>.
	 * @param marginal	Estimated marginal count for this state
	 * @param transitions	Estimated count of transitions out of this state.
	 * @param emissions	Estimated count of output emissions from this state.
	 */
	public Segment( QuickSerializable state, double marginal,
			    Array transitions, Array emissions )
	{
	    this.state = state;
	    this.marginal = marginal;
	    this.transitions = transitions;
	    this.emissions = emissions;
	}
	
	/**
	 * Constructs a <code>Segment</code> from its
	 * <code>serialize</code>d form.
	 */
	public Segment( InputBuffer buffer ) throws QSException
	{
	    state = (QuickSerializable)buffer.nextObject();
	    marginal = buffer.nextDouble();
	    transitions = (Array)buffer.nextObject();
	    emissions = (Array)buffer.nextObject();
	}

	////////////////////   STATIC METHODS 	////////////////////
	//////////////////// 	  METHODS	////////////////////

	/* Specified by ostore.oil.SegmentedModel.Segment */
	public synchronized QuickSerializable id()
	{
	    return state;
	}
	
	/* Specified by ostore.util.QuickSerializable */
	public void serialize( OutputBuffer buffer )
	{
	    buffer.add( state );
	    buffer.add( marginal );
	    buffer.add( transitions );
	    buffer.add( emissions );
	}	
    }
}
