/**
 * HMMPredictor.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.oil;

import ostore.dispatch.Classifier;
import ostore.util.QuickSerializable;
import ostore.util.Carp;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.util.InvalidConfigFileException;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.UnknownEventException;
import seda.sandStorm.api.QueueElementIF;

/**
 * A stage which uses a Hidden Markov Model to predict the future of a
 * monitored event stream.
 * This class extends <code>HMMStage</code>, adding support for
 * <code>Predictor</code>-specific config tags and ipmlementing the
 * <code>handleEvent</code> method.
 *
 * @see HMM
 *
 * @author	Dennis Geels
 * @version	$Id: HMMPredictor.java,v 1.9 2002/07/22 20:54:53 srhea Exp $
 */

public class HMMPredictor extends HMMStage
{
    ////////////////////       CONSTANTS	////////////////////

    private static final boolean DEBUG = DebugFlags.OIL_HMMPREDICTOR;

    // A few strings used in the config files:
    private static final String PERIOD_STR	= "period";
    private static final String HORIZON_STR	= "horizon";
    private static final String HISTORY_STR	= "history";
    
    ////////////////////    STATIC VARIABLES 	////////////////////
    ////////////////////       VARIABLES	////////////////////

    /**
     * The number of events to observe before producing and emitting a new
     * <code>Prediction</code>. 
     */
    protected int prediction_period;
    
    /**
     * The number of observations to look into the future
     *
     * @see HMM#predict
     */
    protected int prediction_horizon;
    
    /**
     * The number of past events to consider when making predictions.
     * Effectively the amount of history to retain between predictions.
     */
    protected int history_size;
    
    /**
     * The total number of observations so far.
     */
    protected int num_observed;
    
    ////////////////////  STATIC CONSTRUCTORS	////////////////////

    static {
	class_tag = "HMMPredictor";
    }
    
    ////////////////////      CONSTRUCTORS	////////////////////
    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////


    /* Specified by seda.sandStorm.api.EventHandlerIF */
    /**
     * Processes the monitored event stream.
     *
     * @exception UnknownEventException if <code>elem<code> is not of
     * the expected event types.
     */
    public void handleEvent(QueueElementIF elem) throws UnknownEventException
    {
	final String tag = class_tag+".handleEvent";
	if( DEBUG ) Debug.printtagln( tag, name + " called:\t" + elem );

	if( observed_type.isInstance( elem ) ) {

	    // Extract the desired field from the Event.
	    QuickSerializable observation = null;
	    try {
		observation = (QuickSerializable)accessor.get(elem);
	    } catch( IllegalAccessException iae ) {
		// Should never happen.
		Carp.die( iae );
	    }

	    // Add the observation to the list.
	    model.record( observation );
	    num_observed++;

	    // Now make a prediction based on the list of recent
	    // observations.
	    if( (num_observed >= (prediction_period+history_size)) &&
		((num_observed % prediction_period) == 0) ) {

		if( DEBUG ) {
		    // Predict nothing - run EM on the current sequence
		    // This prediction will show the accuracy of
		    // the most recent prediction, as well as the degree to
		    // which the next prediction depends on the current
		    // state.
		    Model.Prediction p = model.predict( 0 );
		    Debug.printtagln( tag, "Prediction: " + p.toString() );
		}
		    
		Model.Prediction p = model.predict( prediction_horizon );
		if( DEBUG ) Debug.printtagln( tag, "Prediction: " +
					      p.toString() );
		// Don't worry about dropped predictions.  They're only an
		// optimization anyway.
		classifier.dispatch_lossy( p );
		// Now shift the history window over for the next prediction.
		model.clear( prediction_period );
	    }
	    
	} else {
	    throw new UnknownEventException( "Not a " +
					     observed_type.getName(), elem);
	}
	if( DEBUG ) Debug.printtagln( tag, name + " done" );
    }

    /* Specified by seda.sandStorm.api.EventHandlerIF */
    /**
     * Initializes this <code>HMMPredictor</code> using the configuration
     * parameters.
     *
     * @exception InvalidConfigFileException if the
     * <code>ConfigDataIF</code> contains an invalid parameter or does not
     * contain a required field.
     * @exception InvalidTypeException if a parameter specifies a
     * <code>Class</code> or <code>Field</code> which does not implement
     * the proper java interface.
     *
     * @see HMMStage#init
     */
    public void init(ConfigDataIF config) throws Exception
    {
	final String tag = class_tag+".init";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	// Initialize the shared HMMStage fields.
	super.init( config );

	// Initialize generic Predictor parameters

	prediction_period = config.getInt( PERIOD_STR );
	if( prediction_period <= 0 )
	    throw new InvalidConfigFileException(name, PERIOD_STR);
	if( DEBUG ) {
	    Debug.printtagln( tag, "will make predictions every " +
			      prediction_period + " event" + 
			      ((prediction_period==1) ? "." : "s.") );
	}

	prediction_horizon = config.getInt( HORIZON_STR );
	if( prediction_horizon <= 0 )
	    throw new InvalidConfigFileException(name, HORIZON_STR);
	if( DEBUG ) {
	    Debug.printtagln( tag, "will predict " + prediction_horizon +
			      " event" + ((prediction_horizon==1)?"":"s") +
			      " into the future." );
	}

	history_size = config.getInt( HISTORY_STR );
	if( history_size <= 0 )
	    throw new InvalidConfigFileException(name, HISTORY_STR);
	if( DEBUG ) {
	    Debug.printtagln( tag, "will base predictions on most recent "
			      + history_size + " event" +
			      ((prediction_horizon==1)?".":"s.") );
	}

	num_observed = 0;

	if( DEBUG ) Debug.printtagln( tag, "done: " + name );	
    }

    ////////////////////     INNER CLASSES	////////////////////
}
