/**
 * NetSync.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.oil;

import ostore.util.Carp;
import ostore.util.ByteUtils;
import java.net.DatagramSocket;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.io.IOException;

/**
 * A test file used to determine clock skew between machines.
 * <p>
 * This class should generally only be used from within
 * <code>pond/run/calculate-skew</code>.  See that script for usage.
 *
 * @author	Dennis Geels
 * @version	$Id: NetSync.java,v 1.4 2002/09/13 23:30:08 geels Exp $
 */

public class NetSync
{
    ////////////////////       CONSTANTS	////////////////////

    private final static int PORT = 14440;
    private final static int MAX_ITERATIONS = 500;
    private final static int MIN_ITERATIONS = 100;
    private final static double epsilon = 0.1;	// milliseconds
    
    ////////////////////    STATIC VARIABLES 	////////////////////

    /**
     * We only have a singleton.  Do it the easy way for now.
     */
    private static DatagramSocket socket;
    private static byte[] buffer = new byte[ByteUtils.SIZE_LONG];
    private static int[] offset = new int[1];
    private static DatagramPacket packet =
	new DatagramPacket( buffer, ByteUtils.SIZE_LONG );
	
    ////////////////////       VARIABLES	////////////////////
    ////////////////////  STATIC CONSTRUCTORS	////////////////////
    ////////////////////      CONSTRUCTORS	////////////////////
    ////////////////////     STATIC METHODS 	////////////////////

    public static void main( String args[] ) throws IOException {
	socket = new DatagramSocket( PORT );
	boolean am_master;
	
	if( args.length > 0 ) {
	    am_master = true;
	    Carp.out( "Attaching to " + args[0] );
	    InetAddress addr = InetAddress.getByName( args[0] );
	    socket.connect( addr, PORT );
	} else {
	    am_master = false;
	    Carp.out( "Listening..." );
	}
	if( !am_master ) {
	    // slave
	    while( true ) {
		boolean done = pong_ping();
		if( done ) {
		    break;
		}
	    }
	} else {
	    // master.
	    // Keep average of these differences until it converges.

	    double[] skews = new double[MAX_ITERATIONS];
	    double[] avgs = new double[MAX_ITERATIONS];
	    double[] stddevs = new double[MAX_ITERATIONS];
	    double[] recent_avgs = new double[MIN_ITERATIONS];
	    double skew, skew_sum = 0;
	    int stopped_at = MAX_ITERATIONS-1;
	    double avg_avg = 0;

	    // Start at a negative index, and ignore negative iterations.
	    // This allows us to easily ignore the warm-up rounds.
	    final int start_index = -2;
	    for( int i=start_index; i< MAX_ITERATIONS; i++ ) {

		skew = ping_pong( (long)i );
		if( i>= 0 ) {
		    skews[i] = skew;
		    skew_sum += skew;
		    avgs[i] = mean( skews, i );

		    // Stop if the past several averages are nearly identical.
		    if( i >= MIN_ITERATIONS-1 ) {
			System.arraycopy( avgs, i-MIN_ITERATIONS+1,
					  recent_avgs, 0, MIN_ITERATIONS );
			avg_avg = mean( recent_avgs, MIN_ITERATIONS-1 );
			stddevs[i] = std_dev( recent_avgs, avg_avg,
					      MIN_ITERATIONS-1 );
			if( stddevs[i] < epsilon ) {
			    // Close enough.
			    // Send one last message to shut slave down.
			    ping_pong( (long)MAX_ITERATIONS-1 );
			    stopped_at = i;
			    break;
			}
		    }
		}
	    }
	    Carp.out( "This machine is " + avg_avg + " ms slow" );

	    // Print out averages, for debugging.
	    // Skip first 10. (first 1 is always off).
	    for( int i=0; i<=stopped_at; i++ ) {
		Carp.out( "i=" + i + " skew: " + skews[i] +
			  "\tavg: " + avgs[i] + "\tstddev: " + stddevs[i] );
	    }
	}
    }
    
    //////////////////// 	METHODS		////////////////////
    
    /**
     * Send a datagram to the slave, recording send and receive times.
     * Guess that the remote client is half the RTT away, and calculate the
     * difference between that time and his timestamp.
     *
     * @param counter	The data to send in the datagram.
     * @return The difference between the timestamp in the response and the
     * time we expect it to have been sent.
     */
    private static double ping_pong( long counter ) throws IOException
    {
	long send_time, remote_time, receive_time;
	double skew;
	// include sequence number
	offset[0] = 0;
	ByteUtils.longToBytes( counter, buffer, offset );
	// send
	send_time = System.currentTimeMillis();
	socket.send( packet );
	// receive response
	offset[0] = 0;
	socket.receive( packet );
	receive_time = System.currentTimeMillis();		
	remote_time = ByteUtils.bytesToLong( buffer, offset );
	skew = (remote_time - (send_time +
			       0.5*(receive_time-send_time)));
	Carp.out( "ping: " + counter + "\t skew: " + skew );
	return skew;
    }
    
    /**
     * Receive a message from the master and send a timestamped response.
     *
     * @return <code>true</code> iff the received message contains a
     * sequence number equal to <code>MAX_ITERATIONS-1</code>.
     */
    private static boolean pong_ping() throws IOException
    {
	// receive
	offset[0] = 0;
	socket.receive( packet );
	long sequence_num = ByteUtils.bytesToLong( buffer, offset );
	// timestamp
	offset[0] = 0;
	long local_time = System.currentTimeMillis();
	ByteUtils.longToBytes( local_time, buffer, offset );
	// send response
	socket.send( packet );
	return( sequence_num == (MAX_ITERATIONS-1) );
    }

    /**
     * Calculate the mean of a set of numbers.
     *
     * @param nums	The numbers to average.
     * @param last	The index of the last valid entry.
     */
    private static double mean( double[] nums, int last )
    {
	double sum = 0;
	for( int i=0; i<=last; i++ ) {
	    sum += nums[i];
	}
	return (sum/(last+1));
    }
    
    /**
     * Calculate the standard deviation of a set of numbers.
     *
     * @param nums	The numbers.
     * @param mean	The mean of those numbers.
     * @param last	The index of the last valid entry.
     */
    private static double std_dev( double[] nums, double mean, int last )
    {
	double sum_squares = 0;
	for( int i=0; i<=last; i++ ) {
	    sum_squares += ((nums[i] - mean) * (nums[i] - mean));
	}
	return Math.sqrt(sum_squares/last);
    }
    
    ////////////////////     INNER CLASSES	////////////////////
}
