/**
 * SafeBigInt.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.security;
import java.math.BigInteger;
import java.util.Random;
import ostore.util.ByteUtils;
import ostore.util.Carp;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;
import ostore.util.QuickSerializable;
import ostore.util.OutputBuffer;

/**
 * BigInteger is broken in two respects for use with our cryptographic
 * code; this class makes it hard to run into such problems.
 *
 * @author   Sean C. Rhea
 * @version  $Id: SafeBigInt.java,v 1.10 2002/07/20 22:19:35 srhea Exp $
 */
public class SafeBigInt implements QuickSerializable {

    /**
     * Don't construct negative numbers.
     */
    public SafeBigInt (byte [] val) {
	boolean is_zero = true;
	for (int i = 0; i < val.length; ++i) {
	    if (val [i] != 0) {
		is_zero = false;
		break;
	    }
	}
	_bi = new BigInteger ((is_zero ? 0 : 1), val);
    }

    public SafeBigInt (int signum, byte [] magnitude) {
	_bi = new BigInteger (signum, magnitude);
    }

    public SafeBigInt (int bitLength, int certainty, Random rnd) {
	_bi = new BigInteger (bitLength, certainty, rnd);
    }

    public SafeBigInt (int numBits, Random rnd) {
	_bi = new BigInteger (numBits, rnd);
    }

    public SafeBigInt (String val) {
	_bi = new BigInteger (val);
    }

    public SafeBigInt (String val, int radix) {
	_bi = new BigInteger (val, radix);
    }

    public SafeBigInt (BigInteger bi) {
	_bi = new BigInteger (bi.toByteArray ());
    }

    public SafeBigInt (InputBuffer buffer) {
	this (from_bytes_helper (buffer));
    }

    private static byte [] from_bytes_helper (InputBuffer buffer) {
	int length = buffer.nextInt ();
	byte [] bytes = new byte [length];
	buffer.nextBytes (bytes, 0, length);
	return bytes;
    }

    private static byte [] from_bytes_helper (byte [] data, int [] offset) {
	int length = ByteUtils.bytesToInt (data, offset);
	byte [] bytes = new byte [length];
	offset [0] += ByteUtils.memcpy (bytes, 0, data, offset [0], length);
	return bytes;
    }



    public void serialize (OutputBuffer buffer) {
	byte [] bytes = toByteArray ();
	buffer.add (bytes.length);
	buffer.add (bytes);
    }

    public SafeBigInt add (SafeBigInt val) {
	return new SafeBigInt (_bi.add (val._bi));
    }

    public SafeBigInt multiply (SafeBigInt val) {
	return new SafeBigInt (_bi.multiply (val._bi));
    }

    public SafeBigInt modInverse (SafeBigInt m) {
	return new SafeBigInt (_bi.modInverse (m._bi));
    }

    public boolean isProbablePrime (int certainty) {
	return _bi.isProbablePrime (certainty);
    }

    public int compareTo (SafeBigInt val) {
	return _bi.compareTo (val._bi);
    }

    public int bitLength () {
	return _bi.bitLength ();
    }

    public boolean equals (Object other) {
	SafeBigInt rhs = (SafeBigInt) other;
	return _bi.equals (rhs._bi);
    }

    public BigInteger toBigInteger () {
	return _bi;
    }

    public SafeBigInt mod (SafeBigInt m) {
	return new SafeBigInt (_bi.mod (m._bi));
    }

    public SafeBigInt gcd (SafeBigInt val) {
	return new SafeBigInt (_bi.gcd (val._bi));
    }

    public SafeBigInt divide (SafeBigInt val) {
	return new SafeBigInt (_bi.divide (val._bi));
    }

    public SafeBigInt shiftLeft (int n) {
	return new SafeBigInt (_bi.shiftLeft (n));
    }

    public SafeBigInt modPow (SafeBigInt exponent, SafeBigInt m) {
	return new SafeBigInt (_bi.modPow (exponent._bi, m._bi));
    }

    public int intValue () {
	return _bi.intValue ();
    }

    public String toString () {
	return _bi.toString ();
    }

    public String toString (int radix) {
	return _bi.toString (radix);
    }

    /**
     * Strips leading zeroes off <code>super.toByteArray ()</code>.
     */
    public byte [] toByteArray () {
	byte [] result = _bi.toByteArray ();
	int index = 0;
	while (result [index] == 0)
	    ++index;
	if (index > 0) {
	    byte [] tmp = new byte [result.length - index];
	    for (int i = 0; i < tmp.length; ++i)
	    	tmp [i] = result [i + index];
	    result = tmp;
	}
	return result;
    }

    public static SafeBigInt valueOf (long val) {
	return new SafeBigInt (BigInteger.valueOf (val));
    }

    private BigInteger _bi;
}

