/**
 * SignedQSTree.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.security;
import ostore.util.*;
import ostore.dataobj.*;
import java.security.*;
import java.util.*;

/**
 * SignedQSTree
 *
 * @author   Sean C. Rhea
 * @version  $Id: SignedQSTreePath.java,v 1.10 2002/07/20 19:38:03 srhea Exp $
 */
public class SignedQSTreePath extends SignedQS {

    public SignedQSTreePath (SignedQSTree source, int elt_num) {
	//System.out.println ("elt_num = " + elt_num);
	_user_data = source._user_data;
	_signature = source._signature;
	_qs_signature = source._qs_signature;
	_qs_signature_type = source._qs_signature_type;

	QSArray elts = (QSArray) source.user_data ();
	_elt = (QuickSerializable) elts.get (elt_num);
	//System.out.println ("elt = " + _elt);
	_elt_num = elt_num;

	SecureHash [] tree = SignedQSTree.compute_tree (elts);
	//System.out.println ("tree.length = " + tree.length);
	int leaves = (tree.length + 1) / 2;
	//System.out.println ("leaves = " + leaves);
	int walker = tree.length - leaves + elt_num;
	while (walker > 2) {
	    int sibling = (walker % 2 == 0) ? (walker - 1) : (walker + 1);
	    _sibling_hashes.addLast (tree [sibling]);
	    //System.out.println ("sibling = " + sibling);
	    //System.out.println ("tree[sibling] = " + tree[sibling]);
	    int parent = (walker - 1) / 2;
	    walker = parent;
	}
    }

    public SignedQSTreePath (InputBuffer buffer) throws QSException {
	super (buffer);
	_elt = (QuickSerializable) buffer.nextObject ();
	_elt_num = buffer.nextInt ();
	int path_len = buffer.nextInt ();
	while (path_len-- > 0)
	    _sibling_hashes.addLast ((SecureHash) buffer.nextObject ());
    }

    public QuickSerializable user_data () { 
	return _elt;
    }



    public void serialize (OutputBuffer buffer) {
	super.serialize (buffer);
	buffer.add (_elt);
	buffer.add (_elt_num);
	buffer.add (_sibling_hashes.size ());
	for (Iterator i = _sibling_hashes.iterator (); i.hasNext (); ) 
	    buffer.add ((SecureHash) i.next ());
    }

    public boolean verify (PublicKey pubkey, Signature engine) 
    throws InvalidKeyException, SignatureException, TypeTable.NoSuchTypeCode {
	
	SecureHash hash = new SHA1Hash (_elt);
	//System.out.println ("hash = " + hash);
	int tree_height = _sibling_hashes.size () + 1;
	int leaves = 1;  while (tree_height-- > 0) leaves *= 2;
	int tree_len = 2 * leaves - 1;
	//System.out.println ("tree_len = " + tree_len);
	int walker = tree_len - leaves + _elt_num;
	Iterator i = _sibling_hashes.iterator ();
	while (walker > 2) {
	    SecureHash sibling_hash = (SecureHash) i.next ();
	    //System.out.println ("sibling_hash = " + sibling_hash);
	    SecureHashBuffer b = new SHA1HashBuffer ();
	    if (walker % 2 == 1) { // we're on the left, sibling on the right
		// Don't want type code.
		hash.serialize (b);  sibling_hash.serialize (b);
	    }
	    else { // we're on the right, sibling on the left
		// Don't want type code.
		sibling_hash.serialize (b);  hash.serialize (b);  
	    }
	    hash = b.toHash ();
	    //System.out.println ("hash = " + hash);
	    int parent = (walker - 1) / 2;
	    walker = parent;
	}
	SignedQSTree.Root r = (SignedQSTree.Root) _user_data;
	SecureHash expected = (walker % 2 == 1) ? r.left : r.right;
	//System.out.println ("expected = " + expected);
	if (! expected.equals (hash)) return false;
	return super.verify (pubkey, engine);
    }

    public String toString () {
	return "(SignedQSTreePath " + user_data () + ")";
    }

    private QuickSerializable _elt;
    private int _elt_num;
    private LinkedList _sibling_hashes = new LinkedList ();
}

