/**
 * SigShare.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.security.thresh;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.math.BigInteger;
import java.security.MessageDigest;
import ostore.security.SafeBigInt;

import ostore.util.AssertionViolatedException;
import ostore.util.ByteUtils;
import ostore.util.Debug;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;
import ostore.util.OutputBuffer;
import ostore.util.QSException;
import ostore.util.QuickSerializable;
import ostore.util.TypeTable;

/**
 * Signature Shares Class.
 * Associates a signature share with an id and wraps a static verifier.
 *
 * <p>Reference: "Practical Threshold Signatures",
 * Victor Shoup (sho@zurich.ibm.com), IBM Research Paper RZ3121, 4/30/99
 *
 * @author  <a href=mailto:sweis@theory.lcs.mit.edu>Steve Weis</a> 
 * @version $Id: SigShare.java,v 1.19 2002/08/02 19:53:08 geels Exp $
 */
public class SigShare implements QuickSerializable {
   
// Constants and variables
//............................................................................
   
    private int id;
    private SafeBigInt sig;
    private Verifier sigVerifier;

    private static final String class_tag = "SigShare";


// Constructors
//............................................................................

    public SigShare(int id, SafeBigInt sig, Verifier sigVerifier) {
	this.id = id;
	this.sig = sig;
	this.sigVerifier = sigVerifier;
    }
   
    public boolean equals (Object other) {
	SigShare rhs = (SigShare) other;
	return (this.id == rhs.id) && (this.sig.equals (rhs.sig)) &&
	    (this.sigVerifier.equals (rhs.sigVerifier));
    }

    public SigShare (InputBuffer buffer) throws QSException {
	id = buffer.nextInt ();
	sig = (SafeBigInt) buffer.nextObject ();
	sigVerifier = (Verifier) buffer.nextObject ();
    }


    public void serialize (OutputBuffer buffer) {
	int i = id;
	buffer.add (i);
	buffer.add (sig);
	buffer.add (sigVerifier);
    }


// Public Methods
//............................................................................
  
    /**
     * Return this share's id. Needed for Lagrange interpolation
     *
     * @return the id of this key share
     */
    public int getId() { return id; }

    /**
     * Return a SafeBigInt representation of this signature
     *
     * @return a SafeBigInt representation of this signature
     */
    public SafeBigInt getSig() { return sig; }

    /**
     * Return this signature's verifier
     *
     * @return A verifier for this signaute
     */
    public Verifier getSigVerifier() { return sigVerifier; }

    /**
     * Return a byte array representation of this signature
     *
     * @return a byte array representation of this signature
     */
    public byte[] getBytes() { return sig.toByteArray(); }

    public String toString() {
	return "Sig["+id+"]: " + sig.toString();
    }

    private static final SafeBigInt TWO = SafeBigInt.valueOf(2l);
    private static final SafeBigInt FOUR = SafeBigInt.valueOf(4l);

    public boolean verify (byte [] data, ThresholdPublicKey public_key) {
	SafeBigInt n = public_key.n;

	SafeBigInt x = (new SafeBigInt(data)).mod(n);	   
	int l = public_key.l;
	SafeBigInt delta = SigShare.factorial(l);
	SafeBigInt xtilde = x.modPow(FOUR.multiply(delta),n);

	try {
	    MessageDigest md = MessageDigest.getInstance("SHA");
	    md.reset();
	    Verifier ver = getSigVerifier();
	    SafeBigInt v = ver.getGroupVerifier();
	    SafeBigInt vi = ver.getShareVerifier();
	    md.update(v.toByteArray());
	    md.update(xtilde.toByteArray());
	    md.update(vi.toByteArray());
	    SafeBigInt xi = getSig();
	    md.update(xi.modPow(TWO,n).toByteArray());
	    SafeBigInt vz = v.modPow(ver.getZ(),n);
	    SafeBigInt vinegc = vi.modPow(ver.getC(),n).modInverse(n);
	    md.update(vz.multiply(vinegc).mod(n).toByteArray());

	    SafeBigInt xtildez = xtilde.modPow(ver.getZ(),n);

	    // TODO: CHECK PAPER
	    SafeBigInt xineg2c =
		xi.modPow(ver.getC(),n).modInverse(n);
	    // According to Shoup, pg. 8 this should be
	    // xi.modPow(TWO,n).modPow(ver.getC(),n).modInverse(n);

	    // Something to do with working in Q_n since every
	    // element is a square

	    md.update(xineg2c.multiply(xtildez).mod(n).toByteArray());

	    byte [] dig = md.digest ();
	    SafeBigInt result = new SafeBigInt(dig).mod(n);

	    return result.equals(ver.getC());
	}
	catch (Exception ex) {
	    debug("Encountered error while attempting to check verifiers");
	    ex.printStackTrace();
	}
	    
	return false;
    }

// Static methods
//............................................................................

    public static boolean verify(byte[] data, byte [] combined_sig,
	ThresholdPublicKey public_key) throws ThresholdSigException
    {
	int l = public_key.l;
	SafeBigInt n = public_key.n;
	SafeBigInt e = public_key.e;

	if (NativeIF.available) {
	    return NativeIF.verify_c_helper (data, combined_sig, l,
		    n.toBigInteger (), e.toBigInteger ());
	}
	else {
	    SafeBigInt w = new SafeBigInt (combined_sig);
	    SafeBigInt x = (new SafeBigInt(data)).mod(n);	   
	    SafeBigInt delta = SigShare.factorial(l);
	    SafeBigInt eprime = delta.multiply(delta).shiftLeft(2);
	    SafeBigInt xeprime = x.modPow(eprime,n);
	    SafeBigInt we = w.modPow(e,n);

	    return (xeprime.compareTo(we) == 0);
	}
    }

    public static boolean verify(byte[] data, SigShare[] sigs, 
	ThresholdPublicKey public_key) throws ThresholdSigException
    {
	byte [] combined_sig = package_up (sigs, public_key);
	return verify (data, combined_sig, public_key);
    }

    public static byte [] package_up (SigShare[] sigs, 
	ThresholdPublicKey public_key) throws ThresholdSigException
    {
	final String tag = "SigShare.package_up";

	int k = public_key.k;
	int l = public_key.l;
	SafeBigInt n = public_key.n;
	SafeBigInt e = public_key.e;

	// TOD: move this elsewhere.  Sanity Check - make sure there are at
	// least k unique sigs out of l possible

	boolean[] haveSig = new boolean[l];
	for (int i=0; i<k; i++) {
	    //	    debug("Checking sig " + sigs[i].getId());
	    if (sigs[i] == null)
		throw new ThresholdSigException("Null signature");
	    if (haveSig[sigs[i].getId()-1])
		throw new ThresholdSigException("Duplicate signature: " +
						sigs[i].getId());
	    haveSig[sigs[i].getId()-1] = true;
	}

	SafeBigInt delta = SigShare.factorial(l);
	SafeBigInt from_c = null, from_java = null;
	if (NativeIF.available) {
	    BigInteger [] xsigs = new BigInteger [k];
	    int [] xlambdas = new int [k];
	    for (int i=0; i<k; i++) {
		xsigs [i] = sigs [i].getSig ().toBigInteger ();
		xlambdas [i] = 
		    SigShare.lambda (sigs [i].getId (), sigs, delta);
		if (DEBUG) Debug.printtagln (tag, "xlambda [" + i + "] = " +
			xlambdas [i]);
		
	    }
	    from_c = new SafeBigInt (NativeIF.package_up_c_helper (l, 
			n.toBigInteger (), e.toBigInteger (), 
			xsigs, xlambdas, DEBUG));
	}
	
	if (! NativeIF.available || DEBUG) {
	    from_java = SafeBigInt.valueOf (1l);
	    for (int i=0; i<k; i++) {
		SafeBigInt lambda = new SafeBigInt (BigInteger.valueOf (
			SigShare.lambda (sigs[i].getId (), sigs, delta)));
		SafeBigInt sig = sigs[i].getSig ();
		SafeBigInt smpl = sig.modPow (lambda, n);
		from_java = from_java.multiply (smpl);
		if (DEBUG) {
		    Debug.printtagln (tag, "lambda [" + i + "] = " +
			    lambda.toString (16));
		    Debug.printtagln (tag, "sig [" + i + "] = " +
			    sig.toString (16));
		    Debug.printtagln (tag, "smpl [" + i + "] = " +
			    smpl.toString (16));
		    Debug.printtagln (tag, "w = " + from_java.toString (16));
		}
	    }
	    from_java = from_java.mod (n);
	}

	if (DEBUG && NativeIF.available) {
	    if (! from_java.equals (from_c)) {
		throw new AssertionViolatedException (tag + 
			": NOT EQUAL:\nfrom_c = " + from_c.toString (16) + 
			",\nfrom_java = " + from_java.toString (16));
	    }
	}

	return (NativeIF.available ? 
		from_c.toByteArray () :
		from_java.toByteArray ());
    }

    /**
     * Returns the factorial of the given integer as a SafeBigInt
     *
     * @return l
     */
    private static SafeBigInt factorial(int l) {
	SafeBigInt x = SafeBigInt.valueOf(1l);
	for (int i=1; i<=l; i++)
	    x = x.multiply(SafeBigInt.valueOf((long)i));

	return x;
    }
   

    /**
     * Compute lagarange interpolation points
     * Reference: Shoup, pg 7.
     *
     * @param p - a polynomial to evaluate these points on
     * @param ik - a point in S
     * @param S - a set of k points in {0...l}
     * @param delta - the factorial of the group size
     *
     * @return the Lagarange interpolation of these points at 
     */
    private static int lambda(int ik, SigShare[] S,
	    SafeBigInt delta) {
	// lambda(id,l) = PI {id!=j, 0<j<=l} (i-j')/(id-j')
	int value = delta.intValue ();

	for (int i=0; i<S.length; i++)
	    if (S[i].getId() != ik)
		value *= S[i].getId();

	for (int i=0; i<S.length; i++)
	    if (S[i].getId() != ik)
		value /= (S[i].getId()-ik);
	   
	return value;
    }



// Debugging
//............................................................................
   
    private final static boolean DEBUG = false;

    static {
	if (DEBUG) Debug.turnOn ("SigShare");
    }
   
    private static void debug(String s)
    {
	System.err.println("SigShare: " + s);
    }
}

