/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.bloom;
import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.HashMap;
import java.util.Random;
import java.util.Vector;
import java.util.LinkedList;
import java.util.Iterator;
import java.io.PrintStream;
import java.io.FileOutputStream;
import java.io.BufferedOutputStream;
import java.io.FileNotFoundException;
import java.io.StreamTokenizer;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.Reader;
import ostore.util.Pair;
import ostore.util.PriorityQueue;
import soss.network.TransitStub;

/**
 * Creates scf's based on a graph, a number of participating nodes, and a
 * number of files per node.
 * 
 * Example invocation:
 * 
 * rm /tmp/bar/*; ./run-java ostore.bloom.CreateSossTest /tmp/bar 1 ../regression/graphs/ts495-1.hier 100 100 1000 true 4 3 999
 * @author Ben Y. Zhao
 */
public class CreateSossTest {

  public static int get_node(Set available_nodes, Random rand) {
    int choice = rand.nextInt() % available_nodes.size();
    Iterator i = available_nodes.iterator();
    Integer result = (Integer) i.next();
    for (int j = 0; j < choice; ++j)
      result = (Integer) i.next();
    i.remove();
    return result.intValue();
    
  }
  
  public static void bfs (Integer src, Map neighbor_sets, Set visited,
			  int hops_left) {
    
    if (hops_left > 0) {
      Set neighbors = (Set) neighbor_sets.get(src);
      for (Iterator i = neighbors.iterator(); i.hasNext(); ) {
	Integer dst = (Integer) i.next();
	visited.add(dst);
      }
      for (Iterator i = neighbors.iterator(); i.hasNext(); ) {
	Integer dst = (Integer) i.next();
	bfs(dst, neighbor_sets, visited, hops_left - 1);
      }
    }
  }
  
  public static double avg_reachability (Map neighbor_sets, int depth) {
    if (depth == 0) return 0.0;
    
    int total = 0;
    int count = 0;
    
    for (Iterator i = neighbor_sets.keySet().iterator();
	 i.hasNext(); ) {
      
      Integer src = (Integer) i.next();
      Set neighbors = (Set) neighbor_sets.get(src);
      Set visited = new HashSet();
      visited.add(src);
      bfs(src, neighbor_sets, visited, depth);
      total += visited.size() - 1;
      count += 1;
    }
    
    return ((double) total) / count;
  }
  
  
  public static void main (String [] args) {
    
    try {
      
      int argnum = 0;
      String /* 1 */ outputdir           = args[argnum++];
      long   /* 2 */ seed                = Long.parseLong(args[argnum++]);
      String /* 3 */ graph               = args[argnum++];
      int    /* 4 */ files_per_node      = Integer.parseInt(args[argnum++]);
      int    /* 5 */ participating_nodes = Integer.parseInt(args[argnum++]);
      int    /* 6 */ total_requests      = Integer.parseInt(args[argnum++]);
      int    /* 7 */ interdomain         = Integer.parseInt(args[argnum++]);
      long   /* 14 */ overlay_seed       = Long.parseLong(args[argnum++]);
      
      int reads_per_node = total_requests / participating_nodes;
      boolean interdomain_aware = (interdomain != -1);
      
      TransitStub ts = new TransitStub (graph);
      int total_nodes = ts.vertices();
      
      // Build the overlay network
      Random overlay_rand = new Random(overlay_seed);
      
      System.out.println("Creating test.scf.");
      PrintStream scf = null;
      {
	FileOutputStream fos = 
	  new FileOutputStream(outputdir + "/test.scf");
	BufferedOutputStream bos = new BufferedOutputStream(fos);
	scf = new PrintStream(bos);
      }
      
      scf.println("Seed " + seed);
      scf.println("Graph " + graph);
      scf.println("DesiredSuccesses " + participating_nodes);
      
      Set available_nodes = new HashSet();
      for (int i = 0; i < total_nodes; ++i) {
	available_nodes.add(new Integer(i));
      }
      
      int fedport = total_nodes - 1;
      available_nodes.remove(new Integer(fedport));
      
      scf.println("Algorithm ostore.tapestry.impl.FederationBuilder");
      scf.println("Arguments 0");
      scf.println("ParticipatingNodes 1");
      scf.println("Node 2 192.168.0.1:" + fedport + " " + 
		   participating_nodes);
      
      scf.println("Algorithm ostore.tapestry.impl.TapestryBuilder");
      scf.println("Arguments 3 192.168.0.1:" + fedport + " static " + seed);
      scf.println("ParticipatingNodes " + participating_nodes);
      
      // Create the neighbor graph.
      Map neighbor_sets = new HashMap();
      for (int i = 0; i < participating_nodes; ++i) {
	neighbor_sets.put(new Integer(get_node(available_nodes, 
					       overlay_rand)), 
			  new HashSet());
      }
      
      if (! interdomain_aware) {
	PriorityQueue pq = 
	  new PriorityQueue(neighbor_sets.size() * neighbor_sets.size());
	
	for (Iterator i = neighbor_sets.keySet().iterator(); i.hasNext(); ) {
	  Integer src = (Integer) i.next();
	  
	  for (Iterator j = neighbor_sets.keySet().iterator(); j.hasNext();) {
	    Integer dst = (Integer) j.next();
	    
	    if (src.intValue() > dst.intValue()) {
	      soss.network.TransitStub.PathInfo p = 
		ts.shortest_path(src.intValue(), dst.intValue(), 0, false, 0);
	      double route_time = p.route_time;
	      long rt_us = Math.round(route_time * 1000000);
	      pq.insert(new Pair(p, new Pair(src, dst)), 
			rt_us, false);
	    }
	  }
	}
	
	while (! pq.empty()) {
	  Pair q = (Pair) pq.extract(false);
	  Pair p = (Pair) q.second;
	  soss.network.TransitStub.PathInfo pi = 
	    (soss.network.TransitStub.PathInfo) q.first;
	  
	  Integer src = (Integer) p.first;
	  Set sneighbors = (Set) neighbor_sets.get(src);
	  
	  /*	  if (sneighbors.size() < bloom_degree) {
	    Integer dst = (Integer) p.second;
	    Set dneighbors = (Set) neighbor_sets.get(dst);
	    if (dneighbors.size() < bloom_degree) {
	    System.out.println("Adding neighbors " + src + " " + dst);
	    sneighbors.add(dst);
	    dneighbors.add(src);
	    }
	    } */
	}
      } else {
	PriorityQueue inter_pq = 
	  new PriorityQueue(neighbor_sets.size() * neighbor_sets.size());
	PriorityQueue intra_pq = 
	  new PriorityQueue(neighbor_sets.size() * neighbor_sets.size());
	
	for (Iterator i = neighbor_sets.keySet().iterator(); i.hasNext(); ) {
	  Integer src = (Integer) i.next();
	  
	  for (Iterator j = neighbor_sets.keySet().iterator();
	       j.hasNext(); ) {
	    Integer dst = (Integer) j.next();
	    
	    if (src.intValue() > dst.intValue()) {
	      soss.network.TransitStub.PathInfo p = 
		ts.shortest_path(src.intValue(), dst.intValue(), 0, false, 0);
	      double route_time = p.route_time;
	      long rt_us = Math.round(route_time * 1000000);
	      if (p.interdomain_edges == null)
		intra_pq.insert(new Pair(p, new Pair(src, dst)), 
				rt_us, false);
	      else if (interdomain > 0)
		inter_pq.insert(new Pair(p, new Pair(src, dst)), 
				rt_us, false);
	    }
	  }
	}

	while (! intra_pq.empty()) {
	  Pair q = (Pair) intra_pq.extract(false);
	  Pair p = (Pair) q.second;
	  soss.network.TransitStub.PathInfo pi = 
	    (soss.network.TransitStub.PathInfo) q.first;
	  
	  Integer src = (Integer) p.first;
	  //	  Set sneighbors = (Set) neighbor_sets.get(src);
	  
	  /*	  if (sneighbors.size() < bloom_degree) {
	    Integer dst = (Integer) p.second;
	    Set dneighbors = (Set) neighbor_sets.get(dst);
	    if (dneighbors.size() < bloom_degree) {
	      System.out.println("Adding neighbors " + src 
				 + " " + dst);
	      sneighbors.add(dst);
	      dneighbors.add(src);
	    }
	    } */
	}
	
	Map crossed = new HashMap();
	while (! inter_pq.empty()) {
	  
	  Pair q = (Pair) inter_pq.extract(false);
	  Pair p = (Pair) q.second;
	  soss.network.TransitStub.PathInfo pi = 
	    (soss.network.TransitStub.PathInfo) q.first;
	  
	  boolean too_many_crossed = false;
	  if (pi.interdomain_edges != null) {
	    for (Iterator i = pi.interdomain_edges.iterator();
		 i.hasNext(); ) {
	      Integer edge =(Integer) i.next();
	      Integer crossings =(Integer) crossed.get(edge);
	      if ((crossings != null) &&
		  (crossings.intValue() >= interdomain)) {
		too_many_crossed = true;
		break;
	      }
	    }
	  }
	  
	  if (! too_many_crossed) {
	    Integer src = (Integer) p.first;
	    Set sneighbors = (Set) neighbor_sets.get(src);
	    
	    /*	    if (sneighbors.size() < bloom_degree) {
	      Integer dst = (Integer) p.second;
	      Set dneighbors = (Set) neighbor_sets.get(dst);
	      if (dneighbors.size() < bloom_degree) {
		sneighbors.add(dst);
		dneighbors.add(src);
		System.out.println("Adding neighbors " + src 
				   + " " + dst);
		if (pi.interdomain_edges != null) {
		  for (Iterator i = pi.interdomain_edges.iterator();
		       i.hasNext(); ) {
		    Integer edge = (Integer) i.next();
		    Integer crossings = (Integer) crossed.get(edge);
		    System.out.println ("IDE " + edge + 
					" being crossed to connect " + src +
					" and " + dst);
		    if (crossings == null)
		      crossed.put(edge, new Integer(1));
		    else
		      crossed.put(edge, 
				  new Integer(crossings.intValue() + 1));
		  }
		}
	      }
	      }*/
	  }
	}
	
	int total_crossings = 0;
	for (Iterator i = crossed.keySet().iterator(); i.hasNext(); ) {
	  Integer edge = (Integer) i.next();
	  Integer crossings = (Integer) crossed.get(edge);
	  total_crossings += crossings.intValue();
	}
	System.out.println ("Avg Interdomain Crossings/Edge = " +
			    (((double) total_crossings) / crossed.size()));
	
      }
      
      /*      {
	double total = 0;
	for (int i = 1; i <= bloom_depth; ++i) {
	  double reach = avg_reachability(neighbor_sets, i);
	  System.out.println("Avg reachability at depth " + i + 
			     " = " +(reach - total));
	  total = reach;
	}
	System.out.println("Total avg reachability = " + total);
	}*/
      
      // Free up some memory.
      ts = null;
      System.gc();
      
      // Print the node lines for each node.
      
      for (Iterator i = neighbor_sets.keySet().iterator(); i.hasNext(); ) {
	
	Integer src = (Integer) i.next();
	//scf.println ("printing " + src);
	Set neighbors = (Set) neighbor_sets.get(src);
	scf.print ("Node " + (3 + neighbors.size()) + " 192.168.0.1:" +
		   src.intValue() + " " + outputdir + "/file-list." +
		   src.intValue() + " " + outputdir);
	for (Iterator j = neighbors.iterator(); j.hasNext(); ) {
	  Integer dst = (Integer) j.next();
	  scf.print(" 192.168.0.1:" + dst.intValue());
	}
	scf.println("");
      }
      
      scf.close();
      
      // Finally, assign each of the objects to one machine's disk.
      
      Random rand = new Random(seed); // make independent of graph construction
      System.out.println("Assigning objects to nodes.");
      
      // Print them out.
      
      int [] all_files = new int [neighbor_sets.size() * files_per_node];
      
      int client_num = 0;
      for (Iterator i = neighbor_sets.keySet().iterator();
	   i.hasNext(); ) {
	Integer client = (Integer) i.next();
	FileOutputStream fos = 
	  new FileOutputStream (outputdir +"/file-list." + client.intValue());
	BufferedOutputStream bos = new BufferedOutputStream(fos);
	PrintStream os = new PrintStream(bos);
	
	for (int j = 0; j < files_per_node; ++j) {
	  int file = rand.nextInt() & 0x7fffffff;
	  int size = 10240;  // all files are 10k
	  os.println("file-" + file + "\t" + size);
	  all_files [client_num * files_per_node + j] = file;
	}
	os.close();
	++client_num;
      }

      // Create read lists.
      
      rand = new Random(seed);  // make independent of graph construction
      client_num = 0;
      for (Iterator i = neighbor_sets.keySet().iterator();
	   i.hasNext(); ) {
	Integer client = (Integer) i.next();
	FileOutputStream fos = 
	  new FileOutputStream (outputdir + "/read-file." +client.intValue());
	BufferedOutputStream bos = new BufferedOutputStream(fos);
	PrintStream os = new PrintStream(bos);
	
	for (int j = 0; j < reads_per_node; ++j) {
	  int file = rand.nextInt() & 0x7fffffff;
	  file %= all_files.length;
	  os.println(all_files[file]);
	}
	os.close();
	++client_num;
      }
      
    } 
    catch (RuntimeException e) {
      System.err.println("Exception caught:");
      e.printStackTrace();
      System.exit(1);
    }
    catch (Exception e) {
      System.err.println("Exception caught:");
      e.printStackTrace();
      System.exit(1);
    } 
  }
}

