/*
 *  @author     Ben Y. Zhao
 *  @version    $Id: DynamicNotifyMsg.java,v 1.18 2003/03/05 19:41:36 strib Exp $
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.tapestry.impl;
import ostore.network.NetworkMessage;
import ostore.util.NodeId;
import ostore.util.NonceMsg;
import ostore.util.SecureHash;
import ostore.util.SHA1Hash;
import ostore.util.TypeTable;
import ostore.util.ByteUtils;
import ostore.util.QSException;
import ostore.util.OutputBuffer;
import ostore.util.InputBuffer;
import ostore.util.QuickSerializable;


/**
 * DynamicNotifyMsg.java
 *
 * A message that travels between nodes for the express purpose of informing a
 * node about the existence of another.  This message class actually embeds in
 * two types of functionality. One is the specialized notify message sent
 * from an integrated node X to the newly inserting node N, informing N of X
 * so that N can use X to fill a hole. Of course X might also be a node being
 * inserted in parallel with N.  The other is a more generic notify message
 * between any two tapestry nodes, for the purposes of filling a hole (should
 * not happen if insertion works well) and optimizing nodes' routing tables.
 *
 */
public class DynamicNotifyMsg 
extends NonceMsg implements QuickSerializable {

  public SecureHash src;
  //the GUID of the source of the message, the notifier
  public SecureHash dest;
  //the GUID of the destination of the message, the node being notified
  public int type;
  //a type specifying whether this is an optimization or a holefiller or both
  public int minlevel;
  //the minimum level above which the notifier can be added. This comes in
  //handy during insertion. If not specified, the default value is 0.
  public GuidAddr [] list;

  public static final int OPTIMIZE_TYPE = 0;
  public static final int FILLHOLE_TYPE = 1;
  public static final int DMCAST = 2;
  public static final int GENERAL_TYPE = 3;
  public static final int RESPONSE_TYPE = 4;

  protected boolean internal = false;

  public DynamicNotifyMsg (NodeId peer, SecureHash src, SecureHash dest,
			   int type) {
    super (peer);
    this.src = src;
    this.dest = dest;
    this.type = type;
    this.minlevel = 0;
  }

  public DynamicNotifyMsg (NodeId peer, SecureHash src, SecureHash dest,
			   int type, int minlevel) {
    super (peer);
    this.src = src;
    this.dest = dest;
    this.type = type;
    this.minlevel = minlevel;
  }

  public DynamicNotifyMsg (InputBuffer buffer) throws QSException {
    super(buffer);
    src = (SecureHash) buffer.nextObject();
    dest = (SecureHash) buffer.nextObject();
    type = buffer.nextInt();
    minlevel = buffer.nextInt();
    int size = buffer.nextInt();
    NodeId nid;
    SecureHash guid;
    if (size != 0) {
      list = new GuidAddr [size];
      for (int i=0; i< size; i++) {
	nid = (NodeId) buffer.nextObject();
	guid = (SecureHash) buffer.nextObject();
	if (guid.isNull())
	  guid = null;
	list[i] = new GuidAddr(guid, nid);
      }
    } else list = null;
  }

  public void serialize (OutputBuffer buffer) {
    super.serialize (buffer);
    buffer.add(src);
    buffer.add(dest);
    buffer.add(type);
    buffer.add(minlevel);
    if (list == null || list.length ==0)
      buffer.add(0);
    else {
      buffer.add(list.length);
      for (int i=0; i<list.length; i++) {
	//assuming no null entries here
	buffer.add(list[i].getAddr());
	if (list[i].getGuid() == null)
	  buffer.add(SHA1Hash.NULL_HASH);
	else buffer.add(list[i].getGuid());
      }
    }
  }

  public Object clone () throws CloneNotSupportedException {
    DynamicNotifyMsg result = (DynamicNotifyMsg) super.clone();
    result.src = src;
    result.dest = dest;
    result.type = type;
    result.minlevel = minlevel;
    result.list = list;
    return result;
  }

  public String toString() {
    if (type == OPTIMIZE_TYPE) {
      return "(DynamicNotifyMsg for optimization) minlevel="+minlevel+
	  ", peer="+peer+", src="+src+", dest="+dest;
    } else if (type == FILLHOLE_TYPE) {
      String s= "(DynamicNotifyMsg for filling holes) minlevel="+minlevel+
	  ", peer="+peer+", src="+src+", dest="+dest+"\n";
      for (int i=0; i<list.length; i++) {
	s+=" "+list[i];
      }
      return s;
    } else if (type == RESPONSE_TYPE ) {
      return "(DynamicNotifyMsg in response mode, peer="+peer+
	", src="+src+", dest="+dest;
    } else {
      return "(General DynamicNotifyMsg) minlevel="+minlevel+
	  ", peer="+peer+", src="+src+", dest="+dest;
    }
  }

}
