/**
 * DynamicObjMapMsg.java
 *
 * This message is used to trafer object location information between nodes
 * in the dynamic insertion/deletion processes.  Each message encapsulates
 * the publishInfo for all replicas of a particular object GUID.  This
 * message is used in two instances. 1) During insertion, as part of the
 * Dynamic Directed Multicast ({@link DynamicDMCastMsg}) that contacts
 * need-to-know nodes, which potentially transfer objects rooted at them to
 * the new node.  2) During voluntary deletes, when the leaving node packages
 * objects for which it's acting as root, and sends them to find new
 * surrogates in a multicast message.  The flag we set in the type determines
 * the context of its usage.
 * 
 * @author Ben Y. Zhao
 * @version $Id: DynamicObjMapMsg.java,v 1.22 2003/06/06 00:08:58 strib Exp $ */

package ostore.tapestry.impl;
import ostore.network.NetworkMessage;
import ostore.util.NodeId;
import ostore.util.NonceMsg;
import ostore.util.SecureHash;
import ostore.util.TypeTable;
import ostore.util.ByteUtils;
import ostore.util.QSException;
import ostore.util.QuickSerializable;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Iterator;
import java.util.Set;
import ostore.util.OutputBuffer;
import ostore.util.InputBuffer;
import ostore.util.QuickSerializable;

public class DynamicObjMapMsg 
extends NonceMsg implements QuickSerializable {

  public NodeId src;
  public SecureHash objguid; //linkedsets of objects to node(s)
  public LinkedList pinfo;
  public int objcount;    //how many of these msgs are coming
  public int seq;         //which msg is this, starts at 1, not 0
  // if objcount == seq, means its the last message
  // if objcount == 0, means there are no objects to be sent over
  public int type;
  public static final int INSERTION = 1;
  public static final int DELETION = 2;
  public static final int DELETIONACK = 3;
  public static final int DELETION_NACK = 4;

  /**
   * if type == DELETIONACK or DELETION_NACK, then this is the list
   * of nodes we've already looked at, and are all leaving
   */
  public HashSet ignoreNodes;
  /**
   * if type == DELETION_NACK, this is the next suggested surrogate node
   */
  public NodeId nextSurrogate;

  public DynamicObjMapMsg (NodeId peer, NodeId src, SecureHash objguid, 
			   LinkedList pinfo, int objcount, int seq) {
    super (peer);
    this.src = src;
    this.objguid = objguid;
    this.pinfo = pinfo;   //linkedlist of publish info objects
    this.objcount = objcount;
    this.seq = seq;
    this.type = INSERTION;
  }

  public DynamicObjMapMsg (NodeId peer, NodeId src, SecureHash objguid, 
			   LinkedList pinfo, int objcount, int seq, int type) {
    super (peer);
    this.src = src;
    this.objguid = objguid;
    this.pinfo = pinfo;   //linkedlist of publish info objects
    this.objcount = objcount;
    this.seq = seq;
    this.type = type;
  }

  public void addIgnoreNodes(HashSet toIgnore) {
    ignoreNodes = toIgnore;
  }

  public void addNextSurrogate (NodeId next) {
    nextSurrogate = next;
  }

  public DynamicObjMapMsg (InputBuffer buffer) throws QSException {
    super(buffer);
    src = (NodeId) buffer.nextObject();
    objguid = (SecureHash) buffer.nextObject();

    int numofpinfos = buffer.nextInt();
    PublishInfo pi;

    pinfo = new LinkedList();
    for (int i=0; i < numofpinfos; i++) {
      pi = (PublishInfo) buffer.nextObject();
      pinfo.add(pi);
    }

    objcount = buffer.nextInt();
    seq = buffer.nextInt();
    type = buffer.nextInt();
    int size = buffer.nextInt();
    if (size == 0) ignoreNodes = null;
    else {
      ignoreNodes = new HashSet();
      for (int i=0; i<size; i++) {
	ignoreNodes.add(buffer.nextObject());
      }
    }
    if (type == DELETION_NACK)
      nextSurrogate = (NodeId) buffer.nextObject();
  }

  public void serialize (OutputBuffer buffer) {
    super.serialize (buffer);
    buffer.add(src);
    buffer.add(objguid);
    
    buffer.add(pinfo.size());

    PublishInfo pi;
    synchronized( pinfo ) {
	for (Iterator j=pinfo.iterator(); j.hasNext(); ) {
	    pi = (PublishInfo) j.next();
	    buffer.add(pi);
	}
    }

    buffer.add(objcount);
    buffer.add(seq);
    buffer.add(type);
    if (ignoreNodes == null || ignoreNodes.isEmpty())
      buffer.add(0);
    else {
      buffer.add(ignoreNodes.size());
      for (Iterator it = ignoreNodes.iterator(); it.hasNext(); ) {
	buffer.add((NodeId) it.next());
      }
    }
    if (type == DELETION_NACK)
      buffer.add(nextSurrogate);
    //note: nextsurrogate had better not be null
  }

  public Object clone () throws CloneNotSupportedException {
    DynamicObjMapMsg result = (DynamicObjMapMsg) super.clone();
    result.src = src;
    result.objguid = objguid;
    result.pinfo.addAll(pinfo);
    result.objcount = objcount;
    result.seq = seq;
    result.type = type;
    //    result.mac = mac;
    return result;
  }

  public String toString() {
    if (type == INSERTION)
      return "(DynamicObjMapMsg.INSERTION from/to:"+peer+" for guid="+
	objguid+", #" +seq+" of "+objcount+"\n" + pinfo+ ")";
    else if (type == DELETION)
      return "(DynamicObjMapMsg.DELETION from/to:"+peer+" for guid="+
	objguid+", #" +seq+" of "+objcount+"\n" + pinfo+ ")";
    else
      return "(DynamicObjMapMsg.DELETIONACK from/to:"+peer+" for guid="+
	objguid+", #" +seq+" of "+objcount+"\n" + pinfo+ ")";
  }

}
