/**
 * MulticastState.java
 *
 * This class encapsulates state kept for acknowledged multicast messages
 * which are waiting to be acknowledged.  We keep one of these objects per
 * node.  It needs to keep state to tell which multicast is which, and when
 * all the acks from a particular multicast has come back, and also where the
 * parent node for that multicast is, so an ack can be sent up the tree. It
 * keeps several data structures, including a HashMap from neighbor nodeId to
 * waiting acknowledgements, a HashMap from nodeId of new node to parent in
 * acknowledgement tree.
 *
 * @author Ben Y. Zhao
 * @version $Id: MulticastState.java,v 1.8 2003/04/26 00:27:45 strib Exp $ */

package ostore.tapestry.impl;
import ostore.util.NodeId;
import ostore.util.SecureHash;
import java.util.HashSet;
import java.util.HashMap;
import java.util.TreeSet;

public class MulticastState {

  private SecureHash self_guid;
  private HashMap neighborToMcast; //neighbor nodeid->hashset(Guid of new node)
  private HashMap nodeMap;         //Guid of new node --> NodeState
  public HashMap indexToLocks;     //level+":"+index->HashSet of locked Guids
  public final boolean DEBUG = ostore.util.DebugFlags.TAPESTRY_DYNAMIC_TCLIENT;

  public MulticastState(SecureHash self_guid) {
    this.self_guid = self_guid;
    this.neighborToMcast = new HashMap();
    this.nodeMap = new HashMap();
    this.indexToLocks = new HashMap();
  }

  public HashMap getNodeMap() {
    return nodeMap;
  }

  public SecureHash getGuid() {
    return self_guid;
  }

  public HashSet getMCasts(NodeId neighbor) {
    return (HashSet) neighborToMcast.get(neighbor);
  }

  public int getAlpha(SecureHash newnodeguid) {
    NodeState ns = (NodeState) nodeMap.get(newnodeguid);
    if (ns != null) return ns.alpha;
    return -1;
  }

  public NodeId getMCParent(SecureHash newnodeguid) {
    NodeState ns = (NodeState) nodeMap.get(newnodeguid);
    if (ns != null) return ns.parent;
    return null;
  }

  public NodeId getMCAddr(SecureHash newnodeguid) {
    NodeState ns = (NodeState) nodeMap.get(newnodeguid);
    if (ns != null) return ns.addr;
    return null;
  }

  public NodeState getNodeState(SecureHash newnodeguid) {
    return (NodeState) nodeMap.get(newnodeguid);
  }

  public NodeState addNodeState(SecureHash newguid, int alpha,
			   NodeId parent, NodeId addr, TreeSet locks) {
    NodeState ns = new NodeState(newguid, alpha, parent, addr, locks);
    nodeMap.put(newguid, ns);
    return ns;
  }

  public void addLocks(SecureHash newnodeguid, TreeSet locks) {
    NodeState ns = (NodeState) nodeMap.get(newnodeguid);
    if (ns == null) {
      ns = new NodeState(newnodeguid);
      nodeMap.put(newnodeguid, ns);
    }
    ns.locks = locks;
  }

  public void addMCAddr(SecureHash newnodeguid, NodeId addr) {
    NodeState ns = (NodeState) nodeMap.get(newnodeguid);
    if (ns == null) {
      ns = new NodeState(newnodeguid);
      nodeMap.put(newnodeguid, ns);
    }
    ns.mcaddrs.add(addr);
  }

  public void addMCParent(SecureHash newnodeguid, NodeId parent) {
    NodeState ns = (NodeState) nodeMap.get(newnodeguid);
    if (ns == null) {
      ns = new NodeState(newnodeguid);
      nodeMap.put(newnodeguid, ns);
    }
    ns.parent = parent;
  }

  public void addAlpha(SecureHash newnodeguid, int alpha) {
    NodeState ns = (NodeState) nodeMap.get(newnodeguid);
    if (ns == null) {
      ns = new NodeState(newnodeguid);
      nodeMap.put(newnodeguid, ns);
    }
    ns.alpha = alpha;
  }

  public int calcAndAddAlpha(SecureHash newnodeguid,
			     SecureHash currnodeguid) {
    int alpha = calcAlpha(newnodeguid, currnodeguid);
    NodeState ns = (NodeState) nodeMap.get(newnodeguid);
    if (ns == null) {
      ns = new NodeState(newnodeguid);
      nodeMap.put(newnodeguid, ns);
    }
    if (DEBUG)
      System.err.println("Adding alpha="+alpha+" for guid="+newnodeguid);
    ns.alpha = alpha;
    return alpha;
  }

  public int calcAlpha(SecureHash newnodeguid,
			  SecureHash currnodeguid) {
    int [] newnode = RoutingTable.digits(newnodeguid);
    int [] currnode = RoutingTable.digits(currnodeguid);
    int i;
    for (i=0; i< newnode.length; i++) {
      if (newnode[i] != currnode[i]) {
	break;
      }
    }
    return i;
  }

  public void addMCast(SecureHash newnodeguid, NodeId neighbor) {
    HashSet hs;
    if (neighborToMcast.containsKey(newnodeguid)) {
      hs = (HashSet) neighborToMcast.get(newnodeguid);
      hs.add(neighbor);
    } else {
      hs = new HashSet();
      hs.add(neighbor);
      neighborToMcast.put(newnodeguid, hs);
    }
  }

  /**
   * Method to remove a mcast reference as part of handling acks
   * @return whether the acks for this particular guid are all done
   */
  public boolean ackMCast(SecureHash newnodeguid, SecureHash fromGuid,
			  NodeId fromAddr)
    throws MissingObjectException {
    HashSet hs;
    boolean result = false;
    
    if (neighborToMcast.containsKey(fromAddr)) {
      hs = (HashSet) neighborToMcast.get(fromAddr);
      if (!hs.remove(newnodeguid))
	throw new MissingObjectException("Missing multicaststate data");
    }
    
    NodeState ns = (NodeState) nodeMap.get(newnodeguid);
    if (ns == null)
      throw new MissingObjectException("Missing multicaststate data");
    else if (ns.mcaddrs.remove(fromAddr)){
      //we removed it from the mc neighbor list, see if we're done w/ locks
      result = (ns.mcaddrs.isEmpty() && (ns.locks==null||ns.locks.isEmpty())
		&& ns.notified);
      if (DEBUG)
	System.err.println("Removed lock for guid:"+newnodeguid+
			   " after ack from node:"+fromAddr);
    } else {
      //it wasn't in the mc neighbor list, see if it's in the lock list
      if (ns.locks !=null && ns.locks.remove(fromGuid))
	result = (ns.mcaddrs.isEmpty() && ns.locks.isEmpty() && ns.notified);
      else {
	if (DEBUG)
	  System.err.println("Missing mcstate: newguid="+newnodeguid
			     +", from neighbor:"+fromAddr+
			     ", not in neighborToMcast && no matching locks.");
	throw new MissingObjectException("Missing multicaststate data");
      }
    }
    //    if (result) {
      //we're returning true, meaning this guy is done
      //TODO: should do anything else here?
    //    }
    return result;
  }

  public boolean setNotify(SecureHash src) {
    NodeState ns = (NodeState) nodeMap.get(src);
    if (ns == null) {
      if (DEBUG)
	System.err.println("Unexpected lookup for nodestate for guid:"+src);
      return false;
    }
    if (ns.notified && DEBUG)
      System.err.println("MulticastState: extra piggybacked PingMsg received."
			 +" .. Ignoring.");
    ns.notified = true;
    if (ns.mcaddrs.isEmpty() && (ns.locks==null || ns.locks.isEmpty())) {
      //TODO: should do anything else here?
      return true;
    }
    return false;
  }

  /**
   * Function to clean out the state for a particular node insertion.
   * This should be called when a node is sending up a dynamimDMAck
   * for a particular node.
   */
  public void cleanSlate(NodeId newAddr, SecureHash newGuid) {
    neighborToMcast.remove(newAddr);
    nodeMap.remove(newGuid);
  }
}
