/**
 * PublishInfo.java 
 *
 * The class that encapsulates metadata about a particular replica of an
 * object. The primary info of note is the GUID of its current location, but
 * PublishInfo also keeps track of the distance to this replica, the lasthop
 * taken to get here, and the level of the next hop on the publication path.
 * 
 * @author Ben Y. Zhao
 * @version $Id: PublishInfo.java,v 1.25 2002/11/27 02:02:12 hweather Exp $ */

package ostore.tapestry.impl;
import ostore.util.SecureHash;
import ostore.util.AssertionViolatedException;
import ostore.util.QSException;
import ostore.tapestry.api.TapestryTag;
import ostore.util.NodeId;
import ostore.util.OutputBuffer;
import ostore.util.InputBuffer;
import ostore.util.QuickSerializable;
import ostore.util.TypeTable;

public class PublishInfo 
  implements Comparable, QuickSerializable {

  /** The GUID of this replica's location
   */
  public SecureHash src;
  /** The distance in ms to this replica's location
   */
  public long distance;
  public TapestryTag tag;
  /** The NodeId of the last hop traversed on the publication path.
   */
  public NodeId lasthop;
  /** The routing level that we exit this node at, on the publication path.
   ** Using an index starting with 0.  If this node is the root, the value
   ** is set to -1.
   */

  public PublishInfo (SecureHash src, TapestryTag tag, NodeId lasthop) {
    this.src = src;
    this.distance = Long.MAX_VALUE;
    this.tag = tag;
    this.lasthop = lasthop;
  }
  
  public PublishInfo (SecureHash src, long distance, TapestryTag tag,
		      NodeId lasthop) {
    this.src = src;
    this.distance = distance;
    this.tag = tag;
    this.lasthop = lasthop;
  }
  
  public PublishInfo (InputBuffer buffer) throws QSException {
    src = (SecureHash) buffer.nextObject();
    distance = buffer.nextLong();
    if (buffer.nextBoolean())
      tag = (TapestryTag) buffer.nextObject();
    else tag = null;
    lasthop = (NodeId) buffer.nextObject();
  }

  public void serialize (OutputBuffer buffer) {
    //    super.serialize (buffer);
    buffer.add(src);
    buffer.add(distance);
    if (tag == null) buffer.add(false);
    else {
      buffer.add(true);
      buffer.add (tag);
    }
    buffer.add(lasthop);
  }

  public int compareTo (Object o) {
    if (true)
	throw new AssertionViolatedException
       ("don't use this function; it does not work properly.");
    PublishInfo other = (PublishInfo) o;
    long diff = distance - other.distance;
    if (diff != 0)
	return (int) diff;
    if (equals (other))
	return 0;
    return hashCode () - other.hashCode ();
  }

  /**
   * Don't look at distance.
   */
  public boolean equals (Object other) {
    if (! (other instanceof PublishInfo))
    	return false;
    PublishInfo rhs = (PublishInfo) other;
    if (! src.equals (rhs.src))
	return false;

    if (tag == null)
	return rhs.tag == null;
    else if (rhs.tag == null)
	return false;
    else if (rhs.tag == tag)
	return true;
    else {
        long this_tag = TypeTable.type_code(tag);
	long rhs_tag = TypeTable.type_code(rhs.tag);
	return (this_tag == rhs_tag);
	//return tag.equals (rhs.tag);
    }
  }

  /**
   * Don't look at distance.
   */
   public int hashCode () {
   	int result = src.hashCode ();
	if (tag != null) {
	    result ^= tag.hashCode ();
	}
	return result;
   }

    /**
     * Returns a human-readable representation of this
     * <code>PublishInfo</code>. */
    public String toString()
    {
      String str = new String();

      str += "<PublishInfo";
      str += " tag=" + tag;
      str += " distance=" + distance;
      str += " src=" + RoutingTable.printGuid(src);
      str += " lasthop=" + lasthop;
      str += ">";
      
      return str;
    }
}

