/**
 * PatchworkProbeMsg.java
 *
 * @author Jeremy Stribling
 * @version    $Id: PatchworkProbeMsg.java,v 1.6 2002/08/05 07:23:04 strib Exp $
 */

package ostore.tapestry.patchwork;
import ostore.network.NetworkMessage;
import ostore.util.ByteUtils;
import ostore.util.NodeId;
import ostore.util.InputBuffer;
import ostore.util.OutputBuffer;
import ostore.util.QuickSerializable;
import ostore.util.QSException;
import ostore.tapestry.api.TapestryRouteMsg;

/**
 * This is a probing message sent out by the <code>PatchworkStage</code>
 * in order to measure statistics of other nodes.
 *
 * @author Jeremy Stribling
 * @version    $Id: PatchworkProbeMsg.java,v 1.6 2002/08/05 07:23:04 strib Exp $
 */
public class PatchworkProbeMsg extends NetworkMessage
    implements QuickSerializable {

    private int[] filler;

    public static final int LATENCY = 0;
    public static final int BANDWIDTH_1 = 1;
    public static final int BANDWIDTH_2 = 2;
    public static final int LOSS = 3;

    public static final int MINIMUM_SIZE = 21;

    public int type;
    public int fillerSize;
    public int seqNo;
    public int numLost;

    public boolean response;

    /**
     * Constucts a probing message of a certain type, with a given
     * approximate size.  The size needs to be at least 
     * <code>MINIMUM_SIZE</code>, and if difference between it and 
     * <code>MINIMUM_SIZE</code> is not divisible by 4 it will be 
     * rounded down to a number that is.
     *
     * @param peer  the node to probe
     * @param totalSize   how big the message should be
     * @param type  the type of the message
     * @param response  whether this is a response
     */
    public PatchworkProbeMsg (NodeId peer, int totalSize, int type, 
			      boolean response) {

	super (peer, false /* outbound */);

	this.response = response;

	// make sure the sender knows it needs to be at least some bytes
	if( totalSize < MINIMUM_SIZE ) {
	  throw new RuntimeException( "Size of PatchworkProbeMsg must " +
				       "be at least " + MINIMUM_SIZE + 
				       "bytes" );
	}

	if( (type != LATENCY) && (type != BANDWIDTH_1) &&
	    (type != BANDWIDTH_2) && (type != LOSS) ) {
	  throw new RuntimeException( "Invalid type for PatchworkProbeMsg" );
	} else {
	  this.type = type;
	}

	fillerSize = totalSize - MINIMUM_SIZE;
	int remainder = fillerSize % 4;	

	// make sure the fillerSize is divisible by 4
	fillerSize -= remainder;

	// make filler to fill the rest of the array size
	filler = new int[ fillerSize/4 ];

    }

    public PatchworkProbeMsg (InputBuffer buffer) throws QSException {
	super( buffer );
	type = buffer.nextInt();
	seqNo = buffer.nextInt();
	response = buffer.nextBoolean();
	if( response || type != LOSS ) {
	    numLost = buffer.nextInt();
	    if( type == BANDWIDTH_1  || type == BANDWIDTH_2 ) {
		fillerSize = buffer.nextInt();
		filler = new int[fillerSize/4];
	    }
	}
    }

    public void serialize (OutputBuffer buffer) {
	super.serialize( buffer );
	buffer.add( type );
	buffer.add( seqNo );
	buffer.add( response );
	// serialize these things for all types except
	// loss non-responses.  trying to cut the probe size down.
	if( response || type != LOSS ) {
	    buffer.add( numLost );
	    if( type == BANDWIDTH_1 || type == BANDWIDTH_2 ) {
		buffer.add( fillerSize );
		for (int i = 0; i < (fillerSize/4); ++i)
		    buffer.add (filler [i]);
	    }
	}
    }

    public Object clone () throws CloneNotSupportedException {
	PatchworkProbeMsg newone = (PatchworkProbeMsg) super.clone ();
	newone.type = type;
	newone.seqNo = seqNo;
	newone.response = response;
	newone.numLost = numLost;
	newone.fillerSize = fillerSize;
	newone.filler = new int[fillerSize/4];
	return filler;
    }

    public String toString () {
	return "(PatchworkProbeMsg, type " + type + ", fillerSize " + 
		fillerSize + ", response: " + response + ", seqNo: "
	        + seqNo + ", numLost: " + numLost + " "
		+ super.toString () + ")";
    }
}

