/**
 * PatchworkStage.java
 *
 * @author	Jeremy Stribling
 * @version	$Id: PatchworkStage.java,v 1.26 2003/03/17 04:28:04 strib Exp $
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.tapestry.patchwork;

import java.util.HashMap;
import java.util.TreeSet;
import java.util.Iterator;
import java.util.LinkedList;
import ostore.util.InvalidConfigFileException;
import ostore.util.Carp;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import ostore.network.NetworkNodeStatusMsg;
import ostore.tapestry.api.TapestryNodeDetachMsg;
import ostore.tapestry.api.TapestryIntegrateMsg;
import ostore.tapestry.impl.DynamicPtrMsg;
import ostore.tapestry.impl.RoutingTable;
import ostore.tapestry.impl.RouteEntry;
import ostore.util.NodeId;
import ostore.util.Clock;
import ostore.util.SecureHash;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.UnknownEventException;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.api.SinkFullException;   
import seda.sandStorm.api.StagesInitializedSignal;   
import seda.sandStorm.core.ssTimer;
import ostore.oil.Model;

/**
 * <p>
 * The <code>PatchworkStage</code> code actively collects network
 * statistics for the purpose of optimizing Tapestry routing.  It keeps
 * track of these statistics in a <code>PatchworkModel</code> object that
 * implements the <code>ostore.oil.Model</code> interface.  Other models
 * that clone this model can be constructed by sending a 
 * <code>PatchworkDeltaRequestMsg</code> to the stage.  The
 * <code>PatchworkDeltaResponseMsg</code> sent back will contain a
 * <code>PatchworkModel.Delta</code> that can be added to the local
 * clone of the model through the model's <code>add_delta</code>
 * function.  The stage will also automatically compute a delta
 * periodically and send out a response message, based on the
 * <code>recalc_period</code> tag in the config file.</p><p>
 *
 * The configuration file for this stage may contain the following
 * tags:</p><p>
 *
 * <code>latency_period</code>: the period of the latency measurements
 * of the stage in in milliseconds, i.e. how often the stage actively collects
 * latency statistics of nodes in the routing table.  If this tag is 
 * not present, it defaults to a value of 5000 milliseconds.  Specify a 
 * 0 to never collect latency measurements.</p><p>
 *
 * <code>bandwidth_period</code>: the period of the bandwidth measurements 
 * of the stage in in milliseconds, i.e. how often the stage actively collects
 * bandiwdth statistics.  If this tag is not present, it defaults to a 
 * value of 10000 milliseconds.  Specify a 0 to never collect
 * bandwidth measurements.</p><p>
 *
 * <code>loss_period</code>: the period of the loss measurements
 * of the stage in in milliseconds, i.e. how often the stage actively collects
 * loss statistics.  If this tag is not present, it defaults to a 
 * value of 1000 milliseconds.  Specify a 0 to never collect
 * loss measurements.</p><p>
 *
 * <code><i>type</i>_op</code> and <code><i>type</i>_factor</code>:
 * These options control how the probing period varies between levels
 * of the routing table.  Level 0 of the routing table (neighbors that
 * share only the first digit) are probed with a period
 * <code><i>type</i>_period</code>, as specified by the above options.
 * The probing period of the nodes in level <i>n</i> of the routing table
 * would be <code><i>type</i>_period</code> affected by
 * <code>( <i>type</i>_op <i>type<i>_factor )</code> <i>n</i> times.
 * <code><i>type</i>_op</code> can be +, -, *, /, or ^.  For example,
 * to specify an exponential dropoff function for the loss 
 * measurements, you would set <code><i>loss_op</code> to * and
 * <code>loss_factor</code> to 2.</p><p>
 *
 * <code>recalc_period</code>: the period of recalculation of the 
 * <code>PatchworkModel</code> of the stage in in milliseconds.  If this 
 * tag is not present, it defaults to a value of 20000 milliseconds.  
 * Specify a 0 to never automatically recalculate the table.</p><p>
 *
 * <code>respond_period</code>: how often responses get sent back to
 * probing peers.  Beacons are collected, and acknowledgements 
 * are grouped before sent back, to save bandwidth.  Specified in
 * milliseconds. If this tag is not present, it defaults to a value 
 * of 16000 milliseconds.  Specify a 0 to never respond to any beacons.</p><p>
 *
 * <code>timeout</code>: the timeout of statistic collection packets in
 * milliseconds.  If no acknowledgement is received within this amount
 * of time, the packet is considered lost.  If this tag is not present,
 * it defaults to a value of 120000 milliseconds.</p><p>
 *
 * <code>optimize_on</code>: set this to <code>true</code> to enable
 * this stage to optimize the routing table.  Currently, this only
 * involves checking for 100% packet loss and switching to a backup
 * node as appropriate.</p><p>
 *
 * <code>accept_faults</code>: set this to <code>true</code> to allow
 * this stage to be directed to drop probing packets.  Then 
 * <code>PatchworkFaultMsg</code>s can be sent to it specifying
 * how many or what percentage of packets to drop.</p><p>
 *
 * To change the periods of the stage during runtime, send out a 
 * <code>PatchworkChangePeriodMsg</code> message, specifying
 * what type of period to change and to what to change it.  Change it
 * to 0 to turn that particular period off.</p><p>
 *
 * Currently, the statistics collected by this stage per node are:</p><p>
 *
 * <b>Latency</b>: the total amount of round-trip delay of sending one 
 * packet to the node, in milliseconds.  Measured by sending one packet at t1,
 * receiving an acknowledgement at t2, and letting latency = t2 - t1.</p><p>
 *
 * <b>Bandwidth</b>: the bottleneck bandwidth between this node and 
 * another node, in bytes/milliseconds (or roughly KBytes/sec).  Measured
 * using Sender-Based Packet Pair where two packets of exactly the same
 * size are sent at the same time and acknowledgement is received at times 
 * s1 and s2, then bandwidth = (size of packet)/(s2 - s1).  Current size 
 * of packet is 1461 bytes and clock is of millisecond accuracy, so largest
 * measurable bandwidth is 1461 bytes/ms.</p><p>
 *
 * <b>Loss</b>: the percentage of packets lost (as determined by 
 * acknowledgements from the other node) since measurements were 
 * last taken.</p>
 *
 * @author 	Jeremy Stribling
 * @version	$Id: PatchworkStage.java,v 1.26 2003/03/17 04:28:04 strib Exp $
 */

public class PatchworkStage implements EventHandlerIF
{
    ////////////////////       CONSTANTS	////////////////////
    private boolean DEBUG = DebugFlags.TAPESTRY_PATCHWORK_PATCHWORKSTAGE;
    private static final String class_tag = "PatchworkStage";

    private static final String TIMEOUT_KEY = "timeout";
    private static final int TIMEOUT_DEFAULT = 120000;
    private static final String TIMEOUT_P_KEY = "timeout_packets";
    private static final int TIMEOUT_P_DEFAULT = 25;
    private static final int M_COUNT_SCALAR = 4;

    private static final int OP_ADD = 0;
    private static final int OP_MULT = 1;
    private static final int OP_SUB = 2;    
    private static final int OP_DIV = 3;    
    private static final int OP_EXP = 4;

    private static final String LATENCY_PERIOD_KEY = "latency_period";
    private static final String LATENCY_OP_KEY = "latency_op";
    private static final String LATENCY_FACTOR_KEY = "latency_factor";
    private static final int LATENCY_PERIOD_DEFAULT = 5000;
    private static final int LATENCY_OP_DEFAULT = OP_ADD;
    private static final int LATENCY_FACTOR_DEFAULT = LATENCY_PERIOD_DEFAULT;
    private static final String BANDWIDTH_PERIOD_KEY = "bandwidth_period";
    private static final String BANDWIDTH_OP_KEY = "bandwidth_op";
    private static final String BANDWIDTH_FACTOR_KEY = "bandwidth_factor";
    private static final int BANDWIDTH_PERIOD_DEFAULT = 10000;
    private static final int BANDWIDTH_OP_DEFAULT = OP_ADD;
    private static final int BANDWIDTH_FACTOR_DEFAULT 
	= BANDWIDTH_PERIOD_DEFAULT;
    private static final String LOSS_PERIOD_KEY = "loss_period";
    private static final String LOSS_OP_KEY = "loss_op";
    private static final String LOSS_FACTOR_KEY = "loss_factor";
    private static final int LOSS_PERIOD_DEFAULT = 1000;
    private static final int LOSS_OP_DEFAULT = OP_ADD;
    private static final int LOSS_FACTOR_DEFAULT = LOSS_PERIOD_DEFAULT;
    private static final String RECALC_PERIOD_KEY = "recalc_period";
    private static final int RECALC_PERIOD_DEFAULT = 10000;
    private static final String RESPOND_PERIOD_KEY = "respond_period";
    private static final int RESPOND_PERIOD_DEFAULT = 5000;
    private static final String SECOND_CHANCES_T_KEY = "second_chances_thresh";
    private static final int SECOND_CHANCES_T_DEFAULT = 2;
    private static final String SECOND_CHANCES_R_KEY = "second_chances_resp";
    private static final int SECOND_CHANCES_R_DEFAULT = 2;
    private static final String OPTIMIZE_ON_KEY = "optimize_on";
    private static final String ACCEPT_FAULTS_KEY = "accept_faults";

    private static final int BANDWIDTH_MTU = 1461;

    // over what span of time to batch requests (in ms)
    private static final int BATCHING_PERIOD = 10;

    private static final double LOSS_THRESHOLD = 90.0;

    private Classifier classifier;

    private int latency_period;
    private int latency_op;
    private double latency_factor;
    private int bandwidth_period;
    private int bandwidth_op;
    private double bandwidth_factor;
    private int loss_period;
    private int loss_op;
    private double loss_factor;
    private int recalc_period;
    private int respond_period;
    private int second_chances_thresh;
    private int second_chances_resp;

    // TreeSets for finding next smallest alarm time
    private TreeSet latencyNextTimes;
    private TreeSet bandwidthNextTimes;
    private TreeSet lossNextTimes;
    private TreeSet respondNextTimes;

    private boolean optimize_on;

    private boolean accept_faults;

    // How long to wait for packets before moving on (in ms)
    private int timeout;
    // how many probe packets to send without getting a response
    private int timeout_packets;

    // from routing table
    private int digits_per_guid;
    private int digit_values;

    // max value of m_count = M_COUNT_SCALAR*digits_per_guid
    private int m_count_max;

    private SinkIF my_sink;
    private RoutingTable routing_table;

    // Contain the relevant timestamp in Long format
    // indexed by NodeId.
    private HashMap lastLatencyProbes;
    private HashMap lastBandwidthProbes;
    private HashMap lastBandwidthResps;

    // number of individual nodes being monitored
    private int numLatencyNodes = 0;
    private int numBandwidthNodes = 0;
    private int numLossNodes = 0;

    // contains the BeaconMemory for each node
    private HashMap lastLossProbes;

    // Contains a count of probes received per peer, 
    // in BeaconCollector format
    private HashMap beaconsReceived;

    // Patchwork observation model
    private PatchworkModel model;

    // which nodes are hibernating
    private LinkedList hiberNodes;

    private boolean allNodes_isPercent;
    private int allNodes_value;

    // self node id
    NodeId node_id;

    private static final String FAULT_TAG = class_tag + ".faultStuff";

    ////////////////////  STATIC CONSTRUCTORS	////////////////////

    static {

	Debug.turnOn( class_tag + ".init" );
	//Debug.turnOn( class_tag + ".dispatch" );
	//Debug.turnOn( class_tag + ".handleEvent" );
	Debug.turnOn( class_tag + ".handleAlarmMsg" );
	Debug.turnOn( class_tag + ".handleDeltaRequest" );
	Debug.turnOn( class_tag + ".handleRoutingTableAvailable" );
	//Debug.turnOn( class_tag + ".handleProbe" );
	Debug.turnOn( class_tag + ".handleFaultMsg" );
	Debug.turnOn( class_tag + ".setAlarm" );
	Debug.turnOn( class_tag + ".handleOptimize" );
	//Debug.turnOn( class_tag + ".respondToProbeMsg" );
	Debug.turnOn( FAULT_TAG );

    }

    ////////////////////      CONSTRUCTORS	////////////////////
    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////

    /**
     * Specified by seda.sandStorm.api.EventHandlerIF 
     */
    public void handleEvents(QueueElementIF elemarr[]) 
	throws EventHandlerException
    {

	// make sure all events are handled at the same time
	long timeReceived = Clock.current_date(node_id).getTime();
	
	final String tag = class_tag+".handleEvents";
	if( DEBUG ) Debug.printtagln( tag, "called: "
				      + elemarr.length + " elements" );
	for( int i=0; i<elemarr.length; i++ ) {
	    handleEvent( elemarr[i], timeReceived );
	}
	if( DEBUG ) Debug.printtagln( tag, "done" );
    }

    /** 
     * Specified by seda.sandStorm.api.EventHandlerIF 
     */
    public void handleEvent(QueueElementIF element) 
	throws EventHandlerException
    {

	handleEvent( element, Clock.current_date(node_id).getTime() );

    }

    /**
     * Handle events given the exact time they were received by this stage
     */
    private void handleEvent( QueueElementIF element, long timeReceived ) {
      try
      {
	final String tag = class_tag+".handleEvent." + node_id;
	if( DEBUG ) Debug.printtagln( tag, "called" );

        if (DEBUG) Debug.printtagln (tag, "got " + element.getClass () +
                " event at " + timeReceived + "ms: " + element + ".");

	if( element instanceof StagesInitializedSignal ) {

	} else if( element instanceof TapestryIntegrateMsg ) {
	    
	    routing_table = null;

	} else if( element instanceof PatchworkAlarmMsg ) {

	  handleAlarmMsg( (PatchworkAlarmMsg) element, timeReceived );

        } else if( element instanceof RoutingTable.Available ) {
	
	  handleRoutingTableAvailable( (RoutingTable.Available) element );
	
	} else if( element instanceof PatchworkProbeMsg ) {
	
	  handleProbe( (PatchworkProbeMsg) element, timeReceived );

	} else if( element instanceof PatchworkDeltaRequestMsg ) {

	  handleDeltaRequest();

	} else if( element instanceof PatchworkChangePeriodMsg ) {

	  handleChangePeriod( (PatchworkChangePeriodMsg) element );

	} else if( element instanceof PatchworkFaultMsg ) {

	  handleFaultMsg( (PatchworkFaultMsg) element );

	} else if( element instanceof NetworkNodeStatusMsg ) {

	  handleNetworkNodeStatusMsg( (NetworkNodeStatusMsg) element );

	} else {

	  Carp.error( "Dropping unexpected message: " + element );

	}

	if( DEBUG ) Debug.printtagln( tag, "done" );
      }
      catch(Exception e)
      {
        Carp.dieGracefully(e, node_id); 
      }
    }	

    /**
     * It's time to take some measurements!
     */
    private void handleAlarmMsg( PatchworkAlarmMsg msg, long timeReceived ) {

	final String tag = class_tag+".handleAlarmMsg."+node_id;
	if (DEBUG) Debug.printtagln( tag, " called with type " + msg.type );

	int m_count;
	long m_period;
	HashMap measureMap;

	// find the right period
	if( msg.type == PatchworkAlarmMsg.RECALC_MODEL ) {
	  m_period = recalc_period;
	} else {
	  m_period = respond_period;
	}

	// no routing table yet
	if( routing_table == null ) {
	  if( DEBUG ) Debug.printtagln( tag, "No routing table yet" );
	  if( msg.type == PatchworkAlarmMsg.LATENCY || 
	      msg.type == PatchworkAlarmMsg.BANDWIDTH ||
	      msg.type == PatchworkAlarmMsg.LOSS ) {
	      m_period = Clock.current_date(node_id).getTime() - timeReceived;
	  }
	  setAlarm( msg.type, m_period );
	  return;
	}

	// if it's a recalc message, then recalculate
	if( msg.type == PatchworkAlarmMsg.RECALC_MODEL ) {

	  handleDeltaRequest();
	
	  // if we want to optimize, now's the time
	  if( optimize_on ) {
	    handleOptimize();
	  }

	  setAlarm( msg.type, recalc_period );
	  return;
	}

	// if it's a respond message, respond to all peers
	if( msg.type == PatchworkAlarmMsg.RESPOND ) {
	    
	    BeaconCollector bc;
	    for( int j = 0; j < msg.nodes.size(); j++ ) {
		bc = (BeaconCollector) 
		    beaconsReceived.get( getGuid((NodeId) msg.nodes.get(j)) );
		if( bc != null ) {

		    // let them know how many have arrived
		    PatchworkProbeMsg ppm= new PatchworkProbeMsg(bc.peer(),
					        PatchworkProbeMsg.MINIMUM_SIZE,
						        PatchworkProbeMsg.LOSS,
								   true );
		    ppm.seqNo = bc.lastSeqNo();
		    ppm.numLost = bc.numLost();
		    bc.reset();
		    dispatch( ppm );

		    // dispatch 2 extra time for redundancy
		    //dispatch( ppm );
		    //dispatch( ppm );
		    if( DEBUG ) Debug.printtagln( FAULT_TAG, 
						  "[" + node_id + "]: " +
						  "Sent beacon-response" 
						  + " to " + bc.peer() + 
						  " for " + ppm.numLost + 
						  " beacons (seqno = "
						  + ppm.seqNo + ")");
		} 
	    }
	    setAlarm( msg.type, 
		      Clock.current_date(node_id).getTime() - timeReceived );
	    return;
	}

	// find the measurement map
	if( msg.type == PatchworkAlarmMsg.BANDWIDTH ) {
	  measureMap = lastBandwidthProbes;
	} else if( msg.type == PatchworkAlarmMsg.LATENCY ) {
	  measureMap = lastLatencyProbes;
	} else {
	  measureMap = lastLossProbes;
	}

	// measure every peer in routing table level i if:
        //    m_count is divisible by (i+1)
	//    that peer has no outstanding pings	
	/*
        for( int i = 0; i < digits_per_guid; i++ ) {

	  if( m_count % (i+1) == 0 ) {
	*/
	if( DEBUG ) Debug.printtagln( tag, "scanning table level " + 
					       msg.level );

	NodeId currpeer;

	// go through all the nodes in the message
	// (msg.nodes won't be null, it's new'd in the constructor of alarm)

	for( int j = 0; j < msg.nodes.size(); j++ ) {

	    currpeer = (NodeId) msg.nodes.get(j);
	    
	    // if this is for a loss, just send it out
	    if( msg.type == PatchworkAlarmMsg.LOSS) {
		
		BeaconMemory bm = 
		    (BeaconMemory) measureMap.get( getGuid(currpeer) );
		if( bm == null ) {
		    int level  = -1;
		    int digit = -1;
		    boolean found = false;
		    boolean foundLevel = false;
TIME_FIRSTFOR:	    for( int k = 0; k < digits_per_guid; k++ ) {
			  
                        RouteEntry entry;
			for( int l = 0; l < digit_values; l++ ) {
			    entry = routing_table.getEntry( k, l );
			    if( entry != null ) { 
				for( int m = 0; m < entry.size(); m++ ) {
				    if( !entry.getAtIndex(m)
					.equals(currpeer)) {
					continue;
				    }
				    level = k;
				    digit = l;
				    found = true;
				    foundLevel = true;
				    // break the loop is this is a first entry
				    if( m == 0 ) {
					break TIME_FIRSTFOR;
				    }
				}
			    }
			}
			// if we found it as a backup, but have already
			// gone through a level without finding it at all,
			// bail.
			if( found && !foundLevel ) {
			    break TIME_FIRSTFOR;
			}
			foundLevel = false;
                    }
		    // if it's not in our routing table, ignore it
		    if( level != -1 ) {
			bm = new BeaconMemory( level, digit );
			if( DEBUG ) Debug.printtagln( tag, 
						      node_id + 
						      " new bm for " + 
						      currpeer );
			measureMap.put( getGuid( currpeer ), bm );
		    } else {
			continue;
		    }
		}

		// first, make sure we haven't timed out
		// also make sure it's not a backup
		RouteEntry re = routing_table.getEntry(bm.level,bm.digit);
		if( optimize_on && bm.getNumNoResp() > timeout_packets &&
		    !hiberNodes.contains(getGuid(currpeer)) && 
		    re != null && re.getFirst() != null /*&&
							  re.getFirst().equals(currpeer)*/) {
		    
		    PatchworkNodeStatusMsg pnsm =  
			new PatchworkNodeStatusMsg( currpeer, 
						    getGuid(currpeer),
					    PatchworkNodeStatusMsg.HIBERNATING,
						    bm.level);
		    removeNode( pnsm );
		    dispatch( pnsm );
		    hiberNodes.add( getGuid(currpeer) );

		}

		PatchworkProbeMsg ppm = new PatchworkProbeMsg( currpeer, 
						PatchworkProbeMsg.MINIMUM_SIZE,
						        PatchworkProbeMsg.LOSS,
							       false );
		ppm.seqNo = bm.nextSeqNo();
		dispatch( ppm );
		
		if( DEBUG ) Debug.printtagln( tag, "Sent beacon to "
					      + currpeer );

	    } else {
		
		Long lastMeasureTime = (Long) measureMap.get( currpeer );
		
		// if the response got in, record one no-loss,
		// if not and time expired then record a loss
		if( lastMeasureTime == null ) {
		    
		    //model.record(new PatchworkModel.Observation( currpeer, 0,
		    //				        PatchworkModel.LOSS ));
		    //if( DEBUG ) Debug.printtagln( tag, "Recorded a no-loss"
		    //				  + " for " + currpeer ); 

		} else if( (timeReceived - lastMeasureTime.longValue()) 
			   >= timeout ) {
		    
		  //model.record( new PatchworkModel.Observation( currpeer,100,
		    //				PatchworkModel.LOSS ));
		    measureMap.put( currpeer, null );
		    lastMeasureTime = null;
		    //if( DEBUG ) Debug.printtagln( tag, "Recorded a no-loss"
		    //			  + " for " + currpeer );

		}

		// more stuff to check if bandwidth
		if( msg.type == PatchworkAlarmMsg.BANDWIDTH && 
		    lastMeasureTime == null ) {
		    
		    Long firstResp = (Long) lastBandwidthResps.get(currpeer);
		    
		    // if second one received, record no-loss
		    if( firstResp == null ) {
			
		    //model.record( new PatchworkModel.Observation( currpeer,0,
			//				 PatchworkModel.LOSS));
		      //if( DEBUG ) Debug.printtagln( tag, "Recorded a no-loss"
				//		    + " for " + currpeer ); 
		      
		    } else if( (timeReceived - firstResp.longValue()) 
			       >= timeout ) {
			
		      //model.record( new PatchworkModel.Observation( currpeer,
			//			   100, PatchworkModel.LOSS ));
			lastBandwidthResps.put( currpeer, null );
			//if(DEBUG) Debug.printtagln( tag, "Recorded a no-loss"
				//		      + " for " + currpeer );
			
		    }
		    
		}
		
		// send out messages if:
		//   non-null peer
		//   last measurement came in or it's for a loss
		//   if bandwidth, both responses came in
		if( ( (lastMeasureTime == null) || 
		      (msg.type == PatchworkAlarmMsg.LOSS) ) &&
		    ( (msg.type != PatchworkAlarmMsg.BANDWIDTH) ||
		      (lastBandwidthResps.get(currpeer) == null) ) ) {
		    
		    if( DEBUG ) Debug.printtagln( tag,
						  "pinging nodeID " + 
						  currpeer + " ( guid " 
						  + routing_table.
						  printGuid(routing_table.
							    node_id_to_guid
							    ( currpeer )) + 
						  ") at digit " +
						  msg.level + ", value " + j + 
						  ".");
		    
		    // probe the peer
		    PatchworkProbeMsg ppm;
		    PatchworkProbeMsg ppm2 = null;
		    if( msg.type == PatchworkAlarmMsg.LATENCY ) {
			ppm = new PatchworkProbeMsg( currpeer, 
						PatchworkProbeMsg.MINIMUM_SIZE,
						     PatchworkProbeMsg.LATENCY,
						     false );
		    } else {
			// bandwidth
			ppm = new PatchworkProbeMsg( currpeer, 
					        PatchworkProbeMsg.MINIMUM_SIZE,
						 PatchworkProbeMsg.BANDWIDTH_1,
						     false );
			ppm2 = new PatchworkProbeMsg( currpeer, 
						      BANDWIDTH_MTU,
						 PatchworkProbeMsg.BANDWIDTH_2,
						      false );
		    }
		      
		    long time = Clock.current_date(node_id).getTime();
		    measureMap.put( currpeer, 
				    new Long( time ) );
		    dispatch( ppm );
		    
		    if( ppm2 != null ) {
			dispatch( ppm2 );
		    }
		    
		}
		  
	    }
	    
	}
       
	if( msg.type == PatchworkAlarmMsg.LATENCY || 
	    msg.type == PatchworkAlarmMsg.BANDWIDTH ||
	    msg.type == PatchworkAlarmMsg.LOSS ) {
	    m_period = Clock.current_date(node_id).getTime() - timeReceived;
	}
	setAlarm( msg.type, m_period );

    }

    /**
     * Save a reference to the routing table
     */
    private void handleRoutingTableAvailable( RoutingTable.Available msg ) {

	final String tag = class_tag+".handleRoutingTableAvailable";

	if( DEBUG ) { 
	  Debug.printtagln( tag, "called with: \n" + msg.table );
	  if( routing_table != null ) {
              Carp.die( "got two versions of the routing table.");
	  }
	}

	routing_table = msg.table;

	beaconsReceived = new HashMap();
	hiberNodes = new LinkedList();

	// create TreeSets for everyone
	NextTimeComparator ntc = new NextTimeComparator();
	latencyNextTimes = new TreeSet( ntc );
	bandwidthNextTimes = new TreeSet( ntc );
	lossNextTimes = new TreeSet( ntc );
	respondNextTimes = new TreeSet( ntc );

	lastLatencyProbes = new HashMap();
	lastBandwidthProbes = new HashMap();
	lastLossProbes = new HashMap();
	lastBandwidthResps = new HashMap();

	digits_per_guid = routing_table.digits_per_guid();
	digit_values = routing_table.digit_values();
	m_count_max = M_COUNT_SCALAR*digits_per_guid;

	model = new PatchworkModel();
	dispatch( new PatchworkAvailable( model ) );

	allNodes_value = 0;
	allNodes_isPercent = true;

	// populate tree set
	latencyNextTimes = updateNextTimes( latencyNextTimes, latency_period, 
					    latency_op, latency_factor, 0 );
	bandwidthNextTimes = updateNextTimes( bandwidthNextTimes, 
					      bandwidth_period, bandwidth_op,
					      bandwidth_factor, 0 );
	lossNextTimes = updateNextTimes( lossNextTimes, loss_period, loss_op, 
					 loss_factor, 0 );
	
	setAlarm( PatchworkAlarmMsg.LATENCY, 0 );
	setAlarm( PatchworkAlarmMsg.BANDWIDTH, 0 );
	setAlarm( PatchworkAlarmMsg.LOSS, 0 );

	// set first alarms only if they're to be probed
	if( recalc_period != 0 ) {
	    setAlarm( PatchworkAlarmMsg.RECALC_MODEL, recalc_period );
	}
	if( respond_period != 0 ) {
	    setAlarm( PatchworkAlarmMsg.RESPOND, respond_period );
	}

    }	

    /**
     * Handle a resp
     */
    private void handleProbe( PatchworkProbeMsg msg, long timeReceived ) {

	final String tag = class_tag+".handleProbe." + node_id;
	if( DEBUG ) Debug.printtagln( tag, "called with " + msg );
	HashMap measureMap;

	// no routing table yet
	if( routing_table == null ) {
	  if( DEBUG ) Debug.printtagln( tag, "No routing table yet" );
	  return;
	}

	// respond if needed . . .
	if( !msg.response ) {
	    respondToProbeMsg( msg );
	    return;
	}
	
	// otherwise take some measurements

	BeaconMemory bm = null;
	Long measureFrom = null;

	// look at the right map to figure out measurement
	if( msg.type == PatchworkProbeMsg.LATENCY ) {
	  measureMap = lastLatencyProbes;
	  measureFrom = (Long) measureMap.get( msg.peer );
	} else if( msg.type == PatchworkProbeMsg.BANDWIDTH_1 ) {
	  measureMap = lastBandwidthProbes;
	  measureFrom = (Long) measureMap.get( msg.peer );
	} else if( msg.type == PatchworkProbeMsg.BANDWIDTH_2 ) {
	  measureMap = lastBandwidthResps;
	  measureFrom = (Long) measureMap.get( msg.peer );
	} else {
	  measureMap = lastLossProbes;
	  bm = (BeaconMemory) measureMap.get( getGuid( msg.peer) );
	}

	// if resp was unexpected, then drop
	if( (measureFrom == null) && (bm == null) ) {
	  if( DEBUG ) Debug.printtagln( tag, "Dropping unexpected resp from "
					     + msg.peer );
	  // if it's bandwidth #2, stop looking for bandwidth #1
	  if( msg.type == PatchworkProbeMsg.BANDWIDTH_2 ) {
	    lastBandwidthProbes.put( msg.peer, null );
	  }
	  return;
	}

	// make the measurement
	if( msg.type == PatchworkProbeMsg.LATENCY ) {

	  // calculate latency

	  long latency = timeReceived - measureFrom.longValue();
	  if( DEBUG ) Debug.printtagln( tag, "Latency to " + msg.peer + 
					     " is " + latency );
	  model.record( new PatchworkModel.Observation( getGuid(msg.peer), 
						    (double) latency, 
					            PatchworkModel.LATENCY ) );

	  // ack this resp
	  measureMap.put( msg.peer, null );

	} else if( msg.type == PatchworkProbeMsg.BANDWIDTH_1 ) {

	  // get ready for next message	
	  lastBandwidthResps.put( msg.peer, new Long( timeReceived ) );
	  if( DEBUG ) Debug.printtagln( tag, "Received bandwidth #1 for " +
					     msg.peer ); 

	  // ack this resp
	  measureMap.put( msg.peer, null );

	} else if( msg.type == PatchworkProbeMsg.BANDWIDTH_2 ) {

	  // calculate bandwidth

	  long lpTime = measureFrom.longValue();

	  // if times are equal
	  if( lpTime == timeReceived ) {
	    if( DEBUG ) Debug.printtagln( tag, "ignoring time compression, " + 
					       "setting diff to 1" );
	    //measureMap.put( msg.peer, null );
	    //return;
	    timeReceived++;
	  } 

	  // otherwise calculate the bandwidth
	  double bandwidth = (BANDWIDTH_MTU/(timeReceived - lpTime ));

	  if( DEBUG ) Debug.printtagln( tag, "Bandwidth to " + msg.peer + 
					     " is " + bandwidth );
	  model.record( new PatchworkModel.Observation( getGuid(msg.peer), 
							bandwidth, 
					         PatchworkModel.BANDWIDTH ) );

	  // ack this resp
	  measureMap.put( msg.peer, null );

	} else {

	  // figure out the number lost and record
	  int numSuccess;
	  int numLost;
	  if( msg.numLost == -1 ) {
	      // then the responder hasn't seen any beacons 
	      // since its last update!
	      numSuccess = 0;
	      numLost = bm.allLost();
	      if( optimize_on && numLost > 0 ) {
		  // make sure it's not already hibernating or a backup
		  // if re is null, then it's been detached from Tapestry
		  RouteEntry re = routing_table.getEntry(bm.level, bm.digit);
		  if (!hiberNodes.contains(getGuid(msg.peer)) && re != null &&
		      re.getFirst() != null /*&& 
					      re.getFirst().equals(msg.peer)*/) {
		      int chances = bm.updateChancesResp();
		      // if we've gotten more that the ok number 
		      // of empty responses, change the node status
		      if( chances > second_chances_resp ) {
			  
			  if( DEBUG )
			      Debug.printtagln( tag, "hibernating "+msg.peer );
			  PatchworkNodeStatusMsg pnsm =  
			      new PatchworkNodeStatusMsg( msg.peer, 
							  getGuid(msg.peer),
					    PatchworkNodeStatusMsg.HIBERNATING,
							  bm.level);
			  removeNode( pnsm );
			  dispatch( pnsm );
			  hiberNodes.add( getGuid(msg.peer) );
		      }
		  }
	      }
	  } else {
	      numLost = bm.updateNumLost( msg.numLost );
	      numSuccess = bm.update( msg.seqNo, numLost );
	  }
	  
	  if( numSuccess == -1 ) {
	      return;
	  } else if( numSuccess > 0 ) {
	      bm.clearChancesResp();
	      hiberNodes.remove( getGuid(msg.peer) );
	  }
	  for( int i = 0; i < numSuccess; i++ ) {
	      // record a 0 for each one received
	      model.record( new PatchworkModel.Observation( getGuid(msg.peer), 
							    0, 
						       PatchworkModel.LOSS ) );
	  }
	  // record 100 for each one lost
	  for( int i = 0; i < numLost; i++ ) {
	      model.record( new PatchworkModel.Observation( getGuid(msg.peer), 
							    100,
						       PatchworkModel.LOSS ) );
	  }
	  if( DEBUG ) Debug.printtagln( FAULT_TAG, "[" + node_id + "]: " + 
					     "Recorded " + msg.numLost + 
					     " losses and " + numSuccess +
					     " non-losses for " + msg.peer );

	}		

    }


    private void respondToProbeMsg (PatchworkProbeMsg msg) {
	final String tag = class_tag + ".respondToProbeMsg." + node_id;
	if (DEBUG) Debug.printtagln (tag, "responding to  PatchworkProbeMsg "
				     + "from " + msg.peer );

	// if it's bandwidth-related, make a new small one and send it
	if( (msg.type == PatchworkProbeMsg.BANDWIDTH_1) ||
	    (msg.type == PatchworkProbeMsg.BANDWIDTH_2) ) {

	  msg = new PatchworkProbeMsg( msg.peer, 
				       PatchworkProbeMsg.MINIMUM_SIZE,
				       msg.type, true );
	  dispatch( msg );

	} else if( msg.type == PatchworkProbeMsg.LOSS ) {
	  // increase the count for this peer (maybe)
	  BeaconCollector bc = 
	      (BeaconCollector) beaconsReceived.get( getGuid(msg.peer) );
	  if( bc == null ) {
	      bc = new BeaconCollector( msg.peer );
	      beaconsReceived.put( getGuid( msg.peer ), bc );	      	      
	  }
	  if( bc.newBC ) {
	      bc.newBC = false;
	      bc.setSeqNo( msg.seqNo );
	      // make a new NextTimes entry for this node
	      // the seq no of this packets should tell you
	      // the level
	      int level = (int) Math.floor(msg.seqNo/1000);
	      NextTime nt = new NextTime( msg.peer, level, respond_period, 
					  loss_op, loss_factor );
	      respondNextTimes.add( nt );
	  }

	  // check the global rejection values:
	  if( accept_faults ) {
	      if( allNodes_isPercent && 
		  (Math.random()*100 < allNodes_value) ) {
		  if( DEBUG ) Debug.printtagln( FAULT_TAG, 
						"(all)rejecting msg/percent =" 
						+ allNodes_value );
		  return;
	      } else if( !allNodes_isPercent && 
			 (allNodes_value > 0) ) {
		  allNodes_value--;
		  if( DEBUG ) Debug.printtagln( FAULT_TAG, 
						"(all)rejecting msg/# left =" 
						+ allNodes_value );
		  return;
	      }
	  }

	  // if we're accepting faults, you might need to drop the message
	  if( accept_faults ) {

	      if( bc.fault_loss_isPercent && 
		  (Math.random()*100 < bc.fault_loss_value) ) {
		  if( DEBUG ) Debug.printtagln( FAULT_TAG, 
						"rejecting msg/percent =" 
						+ bc.fault_loss_value );
		  return;
	      } else if( !bc.fault_loss_isPercent && 
			 (bc.fault_loss_value > 0) ) {
		  bc.fault_loss_value--;
		  if( DEBUG ) Debug.printtagln( FAULT_TAG, 
						"rejecting msg/# left =" 
						+ bc.fault_loss_value );
		  return;
	      }
    
	  }

	  bc.update( msg.seqNo );
	  // . . . but don't repond yet	  

	} else {

	  // Latency. 
	  // flip the response and send back out
	  msg.response = true;
	  msg.inbound = false;
	  dispatch( msg );

	}

    }

    /** 
     * Send back a delta to the requestor 
     */
    private void handleDeltaRequest() {

	final String tag = class_tag+".handleDeltaRequest";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	PatchworkModel.Delta d = (PatchworkModel.Delta) model.recalculate();

	if( DEBUG ) {
	  Debug.printtagln( tag, "Recalculating model" );
	  Debug.printtagln( tag, "Current model:\n" + model );
	  Debug.printtagln( tag, "Current delta:\n" + d );
	}

	dispatch( new PatchworkDeltaResponseMsg( d ) );

    }

    private int transOp( int msgOp ) {

	switch( msgOp ) {
	case PatchworkChangePeriodMsg.OP_ADD:
	    return OP_ADD;
	case PatchworkChangePeriodMsg.OP_MULT:
	    return OP_MULT;
	case PatchworkChangePeriodMsg.OP_SUB:
	    return OP_SUB;
	case PatchworkChangePeriodMsg.OP_DIV:
	    return OP_DIV;
	case PatchworkChangePeriodMsg.OP_EXP:
	    return OP_EXP;
	default:
	    return -1;
	}
	
    }

    /**
     * Updates or inserts the correct NextTime for a single node 
     */
    private TreeSet updateNextTimes( TreeSet nextTimes, NodeId node, 
				     int newPeriod ) {
	Iterator ntIter = nextTimes.iterator();
	boolean changed = false;
	while( !changed && ntIter.hasNext() ) {
	    NextTime nextNT = (NextTime) ntIter.next();		
	    if( nextNT.getNode() != null 
		&& nextNT.getNode().equals( node ) ) {
		// if we found it, remove it and start it over
		changed = true;
		ntIter.remove();
		nextNT.changePeriod( newPeriod );
		nextNT.nextCall();
		// only add it back in if it's a nonzero period
		if( newPeriod > 0 ) {
		    nextTimes.add( nextNT );
		}
	    }
	}
	if( !changed && newPeriod > 0 ) {
	    // it's a new one with a nonzero period, so add it in
	    NextTime newNT = new NextTime( node, newPeriod );
	    nextTimes.add( newNT );
	}
	return nextTimes;
	
    }

    /**
     * Updates all level-based NextTimes in a TreeSet
     */
    private TreeSet updateNextTimes( TreeSet nextTimes, int newPeriod, 
				     int newOp, double newFactor, 
				     int oldPeriod) {

	// if the old period was zero, make new nodes for all levels
	if( oldPeriod == 0 && newPeriod != 0 ) {
	    NextTime nt;
	    for( int i = 0; i < digits_per_guid; i++ ) {
		nt = new NextTime( i, newPeriod, 
				   newOp, newFactor ); 
		nextTimes.add( nt );
	    }
	    return nextTimes;
	} else {
	    Iterator ntIter = nextTimes.iterator();
	    TreeSet tempTS = new TreeSet( new NextTimeComparator() );
	    while( ntIter.hasNext() ) {
		NextTime nextNT = (NextTime) ntIter.next();
		if( nextNT.getLevel() != -1 ) {
		    // if it's a zero period, don't add it back in
		    if( newPeriod == 0 ) {
			continue;
		    }
		    nextNT.changePeriod( newPeriod, newOp, 
					 newFactor );
		    nextNT.nextCall();
		}
		tempTS.add( nextNT );
	    }
	    return tempTS;
	}
    }

    /** 
     * Change one of the calculation periods
     */
    private void handleChangePeriod( PatchworkChangePeriodMsg msg ) {

	final String tag = class_tag+".handleChangePeriod";
	if( DEBUG ) Debug.printtagln( tag, "called with msg " + msg );

	switch( msg.type ) {
	case PatchworkChangePeriodMsg.LATENCY:
	    if( msg.node != null ) {
		// if a node, just update that nodes time
		latencyNextTimes = updateNextTimes( latencyNextTimes, 
						    msg.node, msg.newPeriod );
	    } else {
		// if for whole table, update all periods
		int oldPeriod = latency_period;
		latency_period = msg.newPeriod;
		if( msg.newOp != -1 ) {
		    latency_op = transOp(msg.newOp);
		}
		if( msg.newFactor != -1 ) {
		    latency_factor = msg.newFactor;
		}
		latencyNextTimes = updateNextTimes( latencyNextTimes,
						    latency_period, latency_op,
						    latency_factor, oldPeriod);
	    }
	    break;
	case PatchworkChangePeriodMsg.BANDWIDTH:
	    if( msg.node != null ) {
		// if a node, just update that nodes time
		bandwidthNextTimes = updateNextTimes( bandwidthNextTimes, 
						      msg.node, msg.newPeriod);
	    } else {
		// if for whole table, update all periods
		int oldPeriod = bandwidth_period;
		bandwidth_period = msg.newPeriod;
		if( msg.newOp != -1 ) {
		    bandwidth_op = transOp(msg.newOp);
		}
		if( msg.newFactor != -1 ) {
		    bandwidth_factor = msg.newFactor;
		}
		bandwidthNextTimes = updateNextTimes( bandwidthNextTimes,
						      bandwidth_period, 
						      bandwidth_op,
						      bandwidth_factor, 
						      oldPeriod );
	    }
	    break;
	case PatchworkChangePeriodMsg.LOSS:
	    if( msg.node != null ) {
		// if a node, just update that nodes time
		lossNextTimes = updateNextTimes( lossNextTimes, 
						 msg.node, msg.newPeriod );
	    } else {
		// if for whole table, update all periods
		int oldPeriod = loss_period;
		loss_period = msg.newPeriod;
		if( msg.newOp != -1 ) {
		    loss_op = transOp(msg.newOp);
		}
		if( msg.newFactor != -1 ) {
		    loss_factor = msg.newFactor;
		}
		lossNextTimes = updateNextTimes( lossNextTimes,
						 loss_period, loss_op,
						 loss_factor, oldPeriod);
	    }
	    break;
	// TODO: Note that in the next two cases there is a slight
	// problem.  If someone changes the period to 0 while an
	// alarm is still in the system, and then changes it to be nonzero
	// before that alarm is dispatched, we'll then have two alarms
	// in the system, which is bad.
	case PatchworkChangePeriodMsg.RECALC_MODEL:
	    // disabled for now, cause ChannelHandler interferes
	    //if( recalc_period == 0 ) {
	    //setAlarm( msg.type, msg.newPeriod );
	    //}
	    //recalc_period = msg.newPeriod;
	    break;
	case PatchworkChangePeriodMsg.RESPOND:
	    if( respond_period == 0 ) {
		setAlarm( msg.type, msg.newPeriod );
	    }
	    respond_period = msg.newPeriod;
	    break;
	}

    }

    /** 
     * Handle a fault message
     */
    private void handleFaultMsg( PatchworkFaultMsg msg ) {

	final String tag = class_tag+".handleFaultMsg";
	if( DEBUG ) Debug.printtagln( tag, "called with msg " + msg );

	if( msg.allNodes ) {

	    allNodes_isPercent = msg.isPercent;
	    allNodes_value = msg.value;
	    return;
	}

	// no routing table yet
	if( routing_table == null ) {
	  if( DEBUG ) Debug.printtagln( tag, "No routing table yet" );
	  return;
	}

	// make it specific to the one who sent the msg (or the specified node)
	NodeId nodeToDrop;
	if( msg.nodeToDrop == null ) {
	    nodeToDrop = msg.peer;
	} else {
	    nodeToDrop = msg.nodeToDrop;
	}
	BeaconCollector bc = 
	    (BeaconCollector) beaconsReceived.get( getGuid(nodeToDrop) );
	if( bc == null ) {
	    bc = new BeaconCollector( nodeToDrop );
	    beaconsReceived.put( getGuid(nodeToDrop), bc );
	}
	bc.fault_loss_value = msg.value;
	bc.fault_loss_isPercent = msg.isPercent;

    }

    /** 
     * Handle a NetworkNodeStatus message by removing all nodes with a 
     * certain InetAddress from the routing table
     */
    private void handleNetworkNodeStatusMsg( NetworkNodeStatusMsg msg ) {

	final String tag = class_tag+".handleNetworkNodeStatusMsg";
	if( DEBUG ) Debug.printtagln( tag, "called with msg " + msg );

	// if it's us, that's a big problem.
	if( msg.dead_node.equals( node_id ) ) {
	    Carp.die("Got a NetworkNodeStatus message about myself. Unhappy!");
	}

	// go through the routing table, and find nodes that match this
	// InetAddress
	for( int i = 0; i < digits_per_guid; i++ ) {
	    for( int j = 0; j < digit_values; j++ ) {

		RouteEntry re = routing_table.getEntry( i, j );
		if( re != null ) {
		    for( int k = 0; k < re.size(); k++ ) {
			NodeId nid = re.getAtIndex(k);
			if( msg.dead_node.equals( nid ) ) {

			    // a match! remove this entry.
			    if( DEBUG ) Debug.printtagln( tag, "Removing node "
							  + nid + " because " +
							  "the network said " +
							  "its physical " + 
							  "connection was " + 
							  "closed." );
			    PatchworkNodeStatusMsg pnsm = 
				new PatchworkNodeStatusMsg( nid, getGuid(nid),
						   PatchworkNodeStatusMsg.DEAD,
							    i );
			    removeNode( pnsm );
			    dispatch( pnsm );
			}
		    }
		}
	    }
	}
    }

    /**
     * Tries to optimize the routing table based on the model.
     * Right now, just looks at losses.
     */
     private void handleOptimize() {

	final String tag = class_tag+".handleOptimize";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	// no routing table yet
	if( routing_table == null ) {
	  if( DEBUG ) Debug.printtagln( tag, "No routing table yet" );
	  return;
	}

	// optimize every peer in routing table level i if:
        //    m_count is divisible by (i+1)
        for( int i = 0; i < digits_per_guid; i++ ) {

	    if( DEBUG ) Debug.printtagln( tag, "optimizing table level " + 
					  i );
	    
	    NodeId currpeer;
	    RouteEntry entry;
	    double loss;
	    for( int j = 0; j < digit_values; j++ ) {
		
		entry = routing_table.getEntry( i, j );
		if( entry == null ) {
		    continue;
		}
		
		for( int m = 0; m < entry.size(); m++ ) {

		    currpeer = entry.getAtIndex(m);
		    loss = model.lookupDouble( getGuid(currpeer), 
					       PatchworkModel.LOSS );
		    if( Double.isNaN( loss ) ) {
			// no calculations have been made yet, so skip
			continue;
		    }
		
		    BeaconMemory bm = 
			(BeaconMemory) lastLossProbes.get(getGuid( currpeer ));

		    // maybe there was a reintegration?
		    if( bm == null ) {
			continue;
		    }
		
		    // if loss is 100%, try to switch it with a backup
		    if( loss >= LOSS_THRESHOLD ) {
		    
			if( !hiberNodes.contains(getGuid(currpeer)) ) {
			
			    // update the chances
			    int chances = bm.updateChancesThresh();
			    int totalChances = second_chances_thresh;
			    // figure out how many chances it gets
			    switch( loss_op ) {
			    case OP_ADD:
				for( int k = 0; k < i; k++ ) {
				    totalChances +=(int) second_chances_thresh*
					(loss_factor/loss_period);
				}    
				break;
			    case OP_MULT:
				for( int k = 0; k < i; k++ ) {
				    totalChances *= loss_factor;
				}    
				break;
			    case OP_SUB:
				for( int k = 0; k < i; k++ ) {
				    totalChances -=(int) second_chances_thresh*
					(loss_factor/loss_period);
				}
				break;
			    case OP_DIV:
				for( int k = 0; k < i; k++ ) {
				    totalChances /=(int) loss_factor;
				}    
				break;
			    case OP_EXP:
				for( int k = 0; k < i; k++ ) {
				    totalChances = (int) 
					Math.pow(totalChances, loss_factor);
				}
			    }
			
			    if( totalChances < 0 ) {
				totalChances = 0;
			    }
			
			    // if we haven't exceeded the chances, no worries
			    if( chances < totalChances ) {
				continue;
			    }
			
			    if( DEBUG ) Debug.printtagln( tag, currpeer + 
							  " has run out "+
							  "of all its " + 
							  totalChances +
							  " chances." );
			    /* Commented out.  Let another stage worry?
			       int entrySize = entry.size();
			       double otherLoss;
			       for( int k = 1; k < entrySize; k++ ) {
			       otherLoss = 
			       model.lookupDouble(getGuid(entry.getAtIndex(k)),
			       PatchworkModel.LOSS);
			       if( !Double.isNaN( otherLoss ) && 
			       (otherLoss < LOSS_THRESHOLD) ) {
				// this entry is better!  
				//put the old one one at the end.
				entry.switchRoutes( 0, k );
				// move the bad entry to the end
				for( int l=k+1; l < entrySize; l++ ) {
				entry.switchRoutes( l-1, l );
				} 
				if( DEBUG ) Debug.printtagln( tag, 
				"in the entry (" 
				+ i + ", " + j + 
				"), switching "+ 
				"in index "
				+ k + 
				", loss rates "+
				" were " + loss +
				" vs "
				+ otherLoss );
				break;
				}
				}
			    */
			    // let everyone know
			    PatchworkNodeStatusMsg pnsm =  
				new PatchworkNodeStatusMsg( currpeer, 
							    getGuid(currpeer),
					    PatchworkNodeStatusMsg.HIBERNATING,
						  i /* routing table level */);
			    removeNode( pnsm );
			    dispatch( pnsm );
			    hiberNodes.add( getGuid(currpeer) );
			
			}
		
		    } else {
			bm.clearChancesThresh();
			// clear it if the other is under its threshold
			if( bm.getChancesResp() <= second_chances_resp ) {
			    hiberNodes.remove( getGuid(currpeer) );
			}
		    }
		}
		
	    }
	    
	}

     }

    /**
     * Sets this stage's alarm clock for <code>millis</code> milliseconds
     * from now.
     */	
    private void setAlarm( int type, long millis ) {

	final String tag = class_tag+".setAlarm";
	if( DEBUG ) Debug.printtagln( tag, "called with type " + type + 
				      " and millis " + millis );

	// no routing table yet
	if( routing_table == null ) {
	  if( DEBUG ) Debug.printtagln( tag, "No routing table yet" );
	  return;
	}

	NextTime lowest;
	TreeSet nextTimes;
	int d_period;
	// find the lowest NextTime
	switch( type ) {
	case PatchworkAlarmMsg.LATENCY:
	    d_period = LATENCY_PERIOD_DEFAULT;
	    nextTimes = latencyNextTimes;
	    break;
	case PatchworkAlarmMsg.BANDWIDTH:
	    d_period = BANDWIDTH_PERIOD_DEFAULT;
	    nextTimes = bandwidthNextTimes;
	    break;
	case PatchworkAlarmMsg.LOSS:
	    d_period = LOSS_PERIOD_DEFAULT;
	    nextTimes = lossNextTimes;
	    break;
	case PatchworkAlarmMsg.RESPOND:
	    d_period = RESPOND_PERIOD_DEFAULT;
	    nextTimes = respondNextTimes;
	    break;
	case PatchworkAlarmMsg.RECALC_MODEL:
	    if( recalc_period == 0 ) {
		return;
	    }
	    nextTimes = null;
	    d_period = 0;
	    break;
	default:
	    nextTimes = null;
	    d_period = 0;
	}

	if( nextTimes != null ) {

	    // if there is nothing to set an alarm for, set one anyway
	    if( nextTimes.size() == 0 ) {
		classifier.
		    dispatch_later( new PatchworkAlarmMsg( type ), 
				    d_period );
		return;
	    }

	    lowest = (NextTime) nextTimes.first();
	    // use the millis field as the amount of time passed
	    // while performing the last event
	    // update all others
	    lowest.updateTime( millis );
	    long newAlarm = lowest.getNextTime();
	    if( newAlarm < 0 ) {
		newAlarm = 0;
	    }
	    long upAmount = newAlarm + millis;
	    Iterator ntIter = nextTimes.iterator();
	    // subtract the upAmount from each element of the set
	    // because you're subtracting the same amount from
	    // each element, the ordering will not be affected
	    while( ntIter.hasNext() ) {
		NextTime nextNT = (NextTime) ntIter.next();
		nextNT.updateTime( upAmount );
	    }
	    if( DEBUG ) Debug.printtagln( tag, "dispatching type " + type + 
					  " for level " + lowest.getLevel() + 
					  " for " + newAlarm + " ms " + 
					  "(inserted " + lowest + ")");

	    // if it's a level, add all nodes in level
	    PatchworkAlarmMsg pam = new PatchworkAlarmMsg( type );

	    // make an alarm for all nodes within the batching period
	    do {

		// reinsert this nextTime
		nextTimes.remove( lowest );
		lowest.nextCall();
		nextTimes.add( lowest );

		if( lowest.getLevel() != -1 ) {
		    // we don't set alarms until after we get the routing table
		    // so it can't be null
		    NodeId currpeer;
		    RouteEntry entry;
		    int startIndex = 0;
		    for( int j = 0; j < digit_values; j++ ) {
			
			entry = routing_table.getEntry( lowest.getLevel(), j );
			if( entry == null || entry.equals( node_id ) || 
			    entry.size() == 0) {
			    continue;
			}
			
			// if we're optimizing, then we should check the
			// the depths of the routing table
			if( optimize_on ) {
			    currpeer = entry.getAtIndex(startIndex);
			    startIndex++;
			    if( startIndex == entry.size() ) {
				// start at beginning of next digit
				startIndex = 0;
			    } else {
				// stay on the same digit
				j--;
			    }
			} else {
			    currpeer = entry.getFirst();
			}
			if( currpeer != null && !currpeer.equals( node_id) ) {
			    pam.addNode( currpeer );
			}
		    }
		    
		} else {
		    
		    // if it's not a level, add the node
		    pam.addNode( lowest.getNode() );
		}

		// find the next lowest node
		lowest = (NextTime) nextTimes.first();

	    } while( lowest.getNextTime() < BATCHING_PERIOD );

	    if( newAlarm > 0 ) {
		classifier.dispatch_later( pam, newAlarm );
	    } else {
		dispatch( pam );
	    }
	} else {
	    if( DEBUG ) Debug.printtagln( tag, "dispatching type " + type + 
					  " for " + millis + " ms " );
	   
	    classifier.dispatch_later( new PatchworkAlarmMsg( type ), millis );
	}

    }

    /**
     * An internal method used for dispatching events
     */
    private void dispatch( QueueElementIF event ) {
                                                
	final String tag = class_tag+".dispatch";
	if( DEBUG ) Debug.printtagln( tag, "called with event " + event );

      try {
        classifier.dispatch( event );
      } catch ( SinkFullException sfe ) {
        //TODO: Better way to deal with SinkFull
        try {
          Thread.currentThread().sleep( 5000 );
        } catch( InterruptedException ie ) {
          //ignore
        }
      }                                         
      
    }

    private int stringToOp( String opString ) 
	throws InvalidConfigFileException{

	if( opString == null ) {
	    throw new InvalidConfigFileException( "null passed to stringToOp");
	} else if( opString.equals( "add" ) || opString.equals( "+" ) ) {
	    return OP_ADD;
	} else if( opString.equals( "mult" ) || opString.equals( "*" ) ) {
	    return OP_MULT;
	} else if( opString.equals( "sub" ) || opString.equals( "-" ) ) {
	    return OP_SUB;
	} else if( opString.equals( "div" ) || opString.equals( "/" ) ) {
	    return OP_DIV;
	} else if( opString.equals( "exp" ) || opString.equals( "^" ) ) {
	    return OP_EXP;
	} else {
	  throw new InvalidConfigFileException( "An \"op\" tag cannot be " + 
						opString );
	}
    }
    
    private SecureHash getGuid( NodeId node ) {

	// no routing table yet
	if( routing_table == null ) {
	    if( DEBUG ) Debug.printtagln( class_tag + ".getGuid", 
					  "No routing table yet" );
	    return null;
	}

	return routing_table.node_id_to_guid( node );

    }


    private void removeNode( PatchworkNodeStatusMsg msg ) {

	final String tag = class_tag+".removeNode";

	int digit = RoutingTable.digits(msg.node)[msg.hops];
	RouteEntry re = routing_table.getEntry( msg.hops, digit );

	if( re == null || re.removeEntry( msg.node_id ) != null ) {

	    if( DEBUG ) Debug.printtagln( tag, "Could not remove node " +
					  msg.node_id + " from the RouteEntry "
					  + re + " at (" + msg.hops + ", " +
					  digit + ")" );

	} else {

	    if( DEBUG ) Debug.printtagln( tag, "Removed node " +
					  msg.node_id + " from the RouteEntry "
					  + re + " at (" + msg.hops + ", " +
					  digit + ")" );
	    // send out a Ptr message letting the node know it's been removed
	    dispatch( new DynamicPtrMsg( msg.node_id, msg.node, 
					 msg.hops, true ) );

	}

    }

    /** 
     * Specified by seda.sandStorm.api.EventHandlerIF 
     */
    public void init(ConfigDataIF config) throws Exception
    {
	final String tag = class_tag+".init";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	///////////////////
	// get config info

	latency_period = config.getInt( LATENCY_PERIOD_KEY );
	if( latency_period == -1 ) {
	  latency_period = LATENCY_PERIOD_DEFAULT;
	} else if( latency_period < 0 ) {
	  throw new InvalidConfigFileException( "The " + LATENCY_PERIOD_KEY +
						"tag must be >= 0");
 	}

	String lop = config.getString( LATENCY_OP_KEY );
	if( lop == null ) {
	  latency_op = LATENCY_OP_DEFAULT;
	} else {
	  latency_op = stringToOp( lop );
 	}

	latency_factor = config.getDouble( LATENCY_FACTOR_KEY );
	if( latency_factor == -1 ) {
	  latency_factor = LATENCY_FACTOR_DEFAULT;
	} else if( latency_factor < 0 ) {
	  throw new InvalidConfigFileException( "The " + LATENCY_FACTOR_KEY +
						"tag must be >= 0");
 	}

	if( DEBUG ) Debug.printtagln( tag, "latency_period is " + 
					   latency_period + ", op is " + 
				           latency_op + ", factor is " +
				           latency_factor );

	bandwidth_period = config.getInt( BANDWIDTH_PERIOD_KEY );
	if( bandwidth_period == -1 ) {
	  bandwidth_period = BANDWIDTH_PERIOD_DEFAULT;
	} else if( bandwidth_period < 0 ) {
	  throw new InvalidConfigFileException( "The " + BANDWIDTH_PERIOD_KEY +
						"tag must be >= 0");
 	}

	String bop = config.getString( BANDWIDTH_OP_KEY );
	if( bop == null ) {
	  bandwidth_op = BANDWIDTH_OP_DEFAULT;
	} else {
	  bandwidth_op = stringToOp( bop );
 	}

	bandwidth_factor = config.getDouble( BANDWIDTH_FACTOR_KEY );
	if( bandwidth_factor == -1 ) {
	  bandwidth_factor = BANDWIDTH_FACTOR_DEFAULT;
	} else if( bandwidth_factor < 0 ) {
	  throw new InvalidConfigFileException( "The " + BANDWIDTH_FACTOR_KEY +
						"tag must be >= 0");
 	}

	if( DEBUG ) Debug.printtagln( tag, "bandwidth_period is " + 
					   bandwidth_period  + ", op is " + 
				           bandwidth_op + ", factor is " +
				           bandwidth_factor );

	loss_period = config.getInt( LOSS_PERIOD_KEY );
	if( loss_period == -1 ) {
	  loss_period = LOSS_PERIOD_DEFAULT;
	} else if( loss_period < 0 ) {
	  throw new InvalidConfigFileException( "The " + LOSS_PERIOD_KEY +
						"tag must be >= 0");
 	}

	String lossop = config.getString( LOSS_OP_KEY );
	if( lossop == null ) {
	  loss_op = LOSS_OP_DEFAULT;
	} else {
	  loss_op = stringToOp( lossop );
 	}

	loss_factor = config.getDouble( LOSS_FACTOR_KEY );
	if( loss_factor == -1 ) {
	  loss_factor = LOSS_FACTOR_DEFAULT;
	} else if( loss_factor < 0 ) {
	  throw new InvalidConfigFileException( "The " + LOSS_FACTOR_KEY +
						"tag must be >= 0");
 	}

	if( DEBUG ) Debug.printtagln( tag, "loss_period is " + loss_period
				           + ", op is " + 
				           loss_op + ", factor is " +
				           loss_factor );

	recalc_period = config.getInt( RECALC_PERIOD_KEY );
	if( recalc_period == -1 ) {
	  recalc_period = RECALC_PERIOD_DEFAULT;
	} else if( recalc_period < 0 ) {
	  throw new InvalidConfigFileException( "The " + RECALC_PERIOD_KEY +
						"tag must be >= 0");
 	}

	if( DEBUG ) Debug.printtagln( tag, "recalc_period is " + 
					   recalc_period );

	respond_period = config.getInt( RESPOND_PERIOD_KEY );
	if( respond_period == -1 ) {
	  respond_period = RESPOND_PERIOD_DEFAULT;
	} else if( respond_period < 0 ) {
	  throw new InvalidConfigFileException( "The " + RESPOND_PERIOD_KEY +
						"tag must be >= 0");
 	}

	if( DEBUG ) Debug.printtagln( tag, "respond_period is " + 
					   respond_period );

	second_chances_thresh = config.getInt( SECOND_CHANCES_T_KEY );
	if( second_chances_thresh == -1 ) {
	  second_chances_thresh = SECOND_CHANCES_T_DEFAULT;
	} else if( second_chances_thresh < 0 ) {
	  throw new InvalidConfigFileException( "The " + SECOND_CHANCES_T_KEY +
						"tag must be >= 0");
 	}

	if( DEBUG ) Debug.printtagln( tag, "second_chances_thresh is " + 
					   second_chances_thresh );

	second_chances_resp = config.getInt( SECOND_CHANCES_R_KEY );
	if( second_chances_resp == -1 ) {
	  second_chances_resp = SECOND_CHANCES_R_DEFAULT;
	} else if( second_chances_resp < 0 ) {
	  throw new InvalidConfigFileException( "The " + SECOND_CHANCES_R_KEY +
						"tag must be >= 0");
 	}

	if( DEBUG ) Debug.printtagln( tag, "second_chances_resp is " + 
					   second_chances_resp );

	timeout = config.getInt( TIMEOUT_KEY );
	if( timeout == -1 ) {
	  timeout = TIMEOUT_DEFAULT;
	} else if( timeout <= 0 ) {
	  throw new InvalidConfigFileException( "The " + TIMEOUT_KEY +
						"tag must be greater than 0");
 	}

	if( DEBUG ) Debug.printtagln( tag, "Timeout is " + timeout );

	timeout_packets = config.getInt( TIMEOUT_P_KEY );
	if( timeout_packets == -1 ) {
	  timeout_packets = TIMEOUT_P_DEFAULT;
	} else if( timeout_packets <= 0 ) {
	  throw new InvalidConfigFileException( "The " + TIMEOUT_P_KEY +
						"tag must be greater than 0");
 	}

	if( DEBUG ) Debug.printtagln( tag, "Timeout_packets is " + 
				      timeout_packets );

	optimize_on = config.getBoolean( OPTIMIZE_ON_KEY );
	if( DEBUG ) Debug.printtagln( tag, "Optimize_on is " + optimize_on );

	accept_faults = config.getBoolean( ACCEPT_FAULTS_KEY );
	if( DEBUG ) Debug.printtagln( tag, "accept_faults is " +accept_faults);
	

	//////////////////
	// initialize filters

	node_id = new NodeId( config.getString( "node_id" ) );
	classifier = Classifier.getClassifier( node_id );
	my_sink = config.getStage().getSink();
	Filter filter;

	// Subscribe to all my favorite events

	String[] eventNames = {
	    "seda.sandStorm.api.StagesInitializedSignal",	
	    "ostore.tapestry.impl.RoutingTable$Available",
	    "ostore.tapestry.patchwork.PatchworkDeltaRequestMsg",
	    "ostore.tapestry.patchwork.PatchworkChangePeriodMsg",
	    "ostore.tapestry.patchwork.PatchworkAlarmMsg",
	    "ostore.tapestry.api.TapestryIntegrateMsg",
	    "ostore.network.NetworkNodeStatusMsg"
	};

	String[] inb_eventNames = {
	    "ostore.tapestry.patchwork.PatchworkProbeMsg",
	};

	if( accept_faults ) {

	    // accept fault messages
	    try {
		filter = new Filter();
		String faultName="ostore.tapestry.patchwork.PatchworkFaultMsg";
		Class faultClass = Class.forName( faultName );
		if( DEBUG ) Debug.printtagln( tag, "Requiring class: " +
					      faultName );
		if( !filter.requireType( faultClass ) ||
		    !filter.requireValue ( "inbound", new Boolean(true) ) || 
		    !classifier.subscribe( filter, my_sink ) ) {
		    
		    Carp.die( "Could not require class " + faultName );
		} 

          } catch ( ClassNotFoundException cnfe ) {
            Carp.die( cnfe );
          }

	}
	

	for( int i = 0; i < eventNames.length; i++ ) {
	  filter = new Filter();
	  Class eventClass;
          if( DEBUG ) Debug.printtagln( tag, "Finding class: " +
                                               eventNames[i] );
          try {
            eventClass = Class.forName( eventNames[i] );
	    if( DEBUG ) Debug.printtagln( tag, "Requiring class: " +
	  				       eventNames[i] );
	    if( !filter.requireType( eventClass ) || 
	        !classifier.subscribe( filter, my_sink ) ) {

	      Carp.die( "Could not require class " + eventNames[i] ); 

	    } 

          } catch ( ClassNotFoundException cnfe ) {
            Carp.die( cnfe );
          }

        }

	// require inbounders
	for( int i = 0; i < inb_eventNames.length; i++ ) {
	  filter = new Filter();
	  Class eventClass;
          if( DEBUG ) Debug.printtagln( tag, "Finding class: " +
                                               inb_eventNames[i] );
          try {
            eventClass = Class.forName( inb_eventNames[i] );
	    if( DEBUG ) Debug.printtagln( tag, "Requiring class: " +
	  				       inb_eventNames[i] );
	    if( !filter.requireType( eventClass ) ||
		!filter.requireValue ( "inbound", new Boolean(true) ) || 
	        !classifier.subscribe( filter, my_sink ) ) {

	      Carp.die( "Could not require class " + inb_eventNames[i] ); 

	    } 

          } catch ( ClassNotFoundException cnfe ) {
            Carp.die( cnfe );
          }

        }

    }

    /** 
     * Specified by seda.sandStorm.api.EventHandlerIF 
     */
    public void destroy()
    {
        final String tag = class_tag+".destroy";
        if( DEBUG ) Debug.printtagln( tag, "destroyed" );
    }


    ////////////////////     INNER CLASSES	////////////////////

    /**
     * Keeps track of how many beacons you've sent to a peer.
     */
    private class BeaconMemory {

	int level;
	int digit;
	private int nextSeqNo;
	private int lastAcked;
	private int dupsSeen;
	private int allLost;
	private int chances_thresh;
	private int chances_resp;

	private static final int MAX_DUPLICATES = 2;

	// difference limit for guessing where overlaps occurred
	private static final double OVERLAP_LIMIT = Integer.MAX_VALUE*.5;

	public BeaconMemory( int level, int digit ) {
	    this.level = level;
	    this.digit = digit;
	    nextSeqNo = 1000*level + 1;
	    lastAcked = nextSeqNo-1;
	    dupsSeen = 0;
	    allLost = 0;
	    chances_thresh = 0;
	    chances_resp = 0;
	}

	/** 
	 * Calculates the number of beacons lost before the
	 * given time, knowing that numSuccess of them got
	 * through since the last update.  This is not an
	 * exact calculation, since it depends on the 
	 * loss_period (which could possibly change) and since
	 * the losses aren't guaranteed to be sent out on
	 * a strict period.  But pretty close.
	 *
	 * @param numSuccess number of successful beacons received 
	 *                   by peer
	 * @param beforeTime  the timestamp on the last beacon received
	 *                    by peer
	 * @return the number of beacons lost on their way to the peer
	 */
	public int update( int ackedSeqNo, int numLost ) {

	    if( DEBUG ) Debug.printtagln( FAULT_TAG, "BM.update called with "
					  + ackedSeqNo + " last acked and " 
					  + numLost + " num lost." );

	    // If it's equal to the last one acked, this is
	    // a duplicate, so ignore
	    if( ackedSeqNo == lastAcked ) {
		return -1;
	    }

	    // If it's greater than the last seqNo sent out,
	    // we have a problem
	    if( ackedSeqNo >= nextSeqNo && 
		(ackedSeqNo - nextSeqNo) < OVERLAP_LIMIT ) {
		if( DEBUG ) Debug.printtagln( FAULT_TAG, "Received an ACK "
					      + "for a packet that hasn't " 
					      + "yet been sent!" );
		return -1;
	    }

	    // Figure out how were sent since ackedSeqNo
	    if( ackedSeqNo < lastAcked ) {
		// wrapped around
		lastAcked =- Integer.MAX_VALUE;
	    }
	    int numTotal = ackedSeqNo - lastAcked;
	    lastAcked = ackedSeqNo;
	    allLost = 0;
	    return numTotal - numLost;

	}

	public int nextSeqNo() {
	    int next = nextSeqNo;
	    if( nextSeqNo == Integer.MAX_VALUE ) {
		nextSeqNo = 0;
	    } else {
		nextSeqNo++;
	    }
	    return next;
	}

	// responder got no beacons.  assume all were lost since the
	// last one acknowledged, and return that many.
	public int allLost() {
	    int tempLast = lastAcked;
	    lastAcked = nextSeqNo;
	    if( lastAcked < tempLast ) {
		// wrapped around
		tempLast =- Integer.MAX_VALUE;
	    }
	    int numLost = lastAcked - tempLast;
	    allLost += numLost;
	    return numLost;    
	}

	// deal with lost packets we already knew about
	public int updateNumLost( int numLost ) {
	    return numLost - allLost;
	}

	// another recalc period above the loss threshold
	public int updateChancesThresh() {
	    chances_thresh++;
	    return chances_thresh;
	}

	public void clearChancesThresh() {
	    chances_thresh = 0;
	}

	public int getChancesThresh() {
	    return chances_thresh;
	}

	// another empty or timed out response beacon
	public int updateChancesResp() {
	    chances_resp++;
	    return chances_resp;
	}

	public void clearChancesResp() {
	    chances_resp = 0;
	}

	public int getChancesResp() {
	    return chances_resp;
	}

	// get the number of probes that have not yet been responded to
	public int getNumNoResp() {
	    int tempLast = lastAcked;
	    if( nextSeqNo < tempLast ) {
		// wrapped around
		tempLast =- Integer.MAX_VALUE;
	    }
	    return nextSeqNo - tempLast;
	}

    }

    /**
     * Keeps track of how many beacons have been received from a
     * particular source
     */
    private class BeaconCollector {

	// -1 if it hasn't seen any packets
	private int numLost;
	private int nextSeqNo;
	private NodeId peer;
	// the first one I saw this cycle
	private int firstSeqNo;

	// how to deal with fault injections:
	protected int fault_loss_value = 0;
	protected boolean fault_loss_isPercent = false;
	protected boolean newBC = true;

	// difference limit for guessing where overlaps occurred
	private static final double OVERLAP_LIMIT = Integer.MAX_VALUE*.5;

	public BeaconCollector( NodeId peer ) {
	    this.peer = peer;
	    numLost = -1;
	    nextSeqNo = 1;
	    firstSeqNo = 0;
	}

	public void update( int seqNo ) {

	    if( numLost < 0 ) { 
		// we've seen one!
		numLost = 0;
	    }

	    if( seqNo >= nextSeqNo ) {

		// if it's bigger . . .
		if( seqNo != nextSeqNo ) {
		    // if it's bigger by more the 50% of the biggest int,
		    // assume it was lost a long time ago and ignore
		    int diff = seqNo - nextSeqNo;
		    if( diff < OVERLAP_LIMIT ) {
			// find difference between the two and lose that much
			numLost += diff;
		    }
		}
	        // set the new expected sequence number
		if( seqNo == Integer.MAX_VALUE ) {
		    nextSeqNo = 0;
		} else {
		    nextSeqNo = seqNo+1;
		}

	    } else {
		
		// if it's smaller, but more than the firstSeqNo, 
		// and not a duplicate, undo a lost because it's out
		// of order
		// TODO: Does not deal with duplicates
		//       (assumes they're lost packets)
		if( (seqNo > firstSeqNo) || (nextSeqNo < firstSeqNo) ) {
		    numLost--;
		}

	    }

	}

	public NodeId peer() {
	    return peer;
	}

	public int numLost() {
	    return numLost;
	}

	public int lastSeqNo() {
	    if( nextSeqNo == 0 && firstSeqNo > 0 ) {
		return Integer.MAX_VALUE;
	    } else {
		return nextSeqNo-1;
	    }
	}

	public void reset() {
	    numLost = -1;
	    if( nextSeqNo == 0 ) {
		firstSeqNo = Integer.MAX_VALUE;
	    } else {
		firstSeqNo = nextSeqNo - 1;
	    }	    
		

	}

	public void setSeqNo( int firstNo ) {
	    firstSeqNo = firstNo;
	    nextSeqNo = firstNo + 1;
	}

    }

    private class NextTime implements Comparable {

	private long msToNextTime;
	private int level;
	private int period;
	private int op;
	private double factor;
	private long cachedTime;
	private NodeId node;

	private NextTime( int period ) {

	    this.period = period;
	    cachedTime = -1;
	}

	public NextTime( int level, int period, int op, double factor ) {

	    this.period = period;
	    cachedTime = -1;
	    this.level = level;
	    this.op = op;
	    this.factor = factor;
	    node = null;
	    nextCall();

	}

	public NextTime( NodeId node, int period ) {

	    this.period = period;
	    cachedTime = -1;
	    level = -1;
	    op = -1;
	    factor = -1;
	    this.node = node;
	    nextCall();

	}

	public NextTime( NodeId node, int level, int period, 
			 int op, double factor ) {

	    this.period = period;
	    cachedTime = -1;
	    this.level = level;
	    this.op = op;
	    this.factor = factor;
	    this.node = node;
	    nextCall();

	}

	public void nextCall() {

	    if( cachedTime == -1 ) {
		cachedTime = calcNextCall();
	    }
	    msToNextTime = cachedTime;

	}

	private long calcNextCall() {

	    if( level == -1 ) {
		return period;
	    }

	    long returnTime = period;

	    switch( op ) {
	    case OP_ADD:
		for( int i = 0; i < level; i++ ) {
		    returnTime += factor;
		}    
		break;
	    case OP_MULT:
		for( int i = 0; i < level; i++ ) {
		    returnTime *= factor;
		}    
		break;
	    case OP_SUB:
		for( int i = 0; i < level; i++ ) {
		    returnTime -= factor;
		}    
		break;
	    case OP_DIV:
		for( int i = 0; i < level; i++ ) {
		    returnTime /= factor;
		}    
		break;
	    case OP_EXP:
		for( int i = 0; i < level; i++ ) {
		    returnTime = (long) Math.pow(returnTime, factor);
		}
	    }
	    
	    if( returnTime < 0 ) {
		returnTime = 0;
	    }

	    return returnTime;

	}

	public void changePeriod( int period ) {
	    this.period = period;
	    cachedTime = -1;
	}

	public void changePeriod( int period, int op, double factor ) {
	    changePeriod( period );
	    this.op = op;
	    this.factor = factor;
	}

	public void updateTime( long timePassed ) {
	    msToNextTime -= timePassed;
	}
	
	// return the level only if it's not a Node as well
	public int getLevel() {
	    if( node == null ) {
		return level;
	    } else {
		return -1;
	    }
	}

	public NodeId getNode() {
	    return node;
	}

	public long getNextTime() {
	    return msToNextTime;
	}

	public int compareTo( Object other ) {

	    NextTime otherNT = (NextTime) other;
	    
	    int returnVal;
	    if( otherNT.getNextTime() < msToNextTime ) {	
		returnVal = -1;
	    } else if( otherNT.getNextTime() == msToNextTime ) {
		// perform a tiebreak
		if( otherNT.getLevel() == -1 ) {
		    if( getLevel() == -1 ) {
			// both nodes, so compare by node id hashcode
			returnVal = otherNT.getNode().hashCode() 
			    - node.hashCode();
			// double tiebreak
			if( returnVal == 0 && 
			    !node.equals( otherNT.getNode() ) ) {
			    if( node.less_than( otherNT.getNode() ) ) {
				returnVal = 1;
			    } else {
				returnVal = -1;
			    }
			}
		    } else {
			// node id is always greater than a level
			returnVal = -1;
		    }
		} else {
		    if( getLevel() == -1 ) {
			// node id is always greater than a level
			returnVal = 1;
		    } else {
			// both are levels
			returnVal = otherNT.getLevel() - getLevel();
		    }
		}
	    } else {
		returnVal = 1;
	    }
	    return returnVal;

	}
	
	public boolean equals( Object other ) {

	    NextTime otherNT = (NextTime) other;
	    
	    return (otherNT.getNextTime() == msToNextTime) &&
		(otherNT.getLevel() == getLevel() ) && 
		((otherNT.getNode() == null && node == null) ||
		 (otherNT.getNode() != null && node != null && 
		  otherNT.getNode().equals(node)));

	}

	public String toString() {
	    return "{node " + node + ", level: " + level + ", next time: " + 
		msToNextTime + "}";
	}


    }

    private class NextTimeComparator implements java.util.Comparator {

	public int compare( Object o1, Object o2 ) {
	    // they're only equal if they have the same level
	    NextTime nt1 = (NextTime) o1;
	    NextTime nt2 = (NextTime) o2;
	    return nt2.compareTo( nt1 );
	}

	// implement this as a reference check, 
	// so TreeSet.remove will remove the correct object
	public boolean equals( Object o1, Object o2 ) {
	    return o1.equals( o2 );
	}

    }

}
