/**
 * AckStage.java
 *
 * @author	Jeremy Stribling
 * @version	$Id: AckStage.java,v 1.5 2003/03/27 20:59:42 strib Exp $
 *
 * Copyright (c) 2003 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.util;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import ostore.dispatch.Resender;
import ostore.dispatch.Signal;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkException;
import seda.sandStorm.api.SinkIF;

/**
 * The AckStage is a general tool for adding robustness to NetworkMessages.
 * If a stage wants reliable delivery of messages, they need only extend 
 * this stage, have their reliable messages extend {@link NonceMsg}, and
 * use {@link #sendNonceMsg} to dispatch the messages.  The stage is smart 
 * in that it will piggyback acks onto messages already heading to a node.
 * <br><br>
 * This can also be used to ack end-to-end messages such as TapestryRouteMsgs,
 * by using the methods {@link #forceRegisterNonce} and {@link #forceSendAck}.
 * For normal NetworkMessages that extend NonceMsg, everything is taken
 * care of automatically.
 * <br><br>
 * If a reliable message times out, the AckStage will call back to the 
 * child stage through the {@link #timeoutCallback} method.  This must be 
 * implemented by each child stage separately.
 * <br><br>
 * Note that stages extending this should <b>NOT</b> implement handleEvents
 * and handleEvent.  Instead, they should implement the method 
 * {@link #handle_event}.  Along with all the normal events to expect
 * to be passed into this method, the stage should also expect 
 * {@link NonceAckMsg} and {@link ostore.dispatch.Signal.AlarmSignal}, but 
 * probably will not want to act on them; they are taken care of by the 
 * parent's method, {@link #handle_event}, which should be called at
 * the top of the child's handle_event as: 
 * <code>super.handle_event(event)</code>.
 *
 * @author	Jeremy Stribling
 * @version	$Id: AckStage.java,v 1.5 2003/03/27 20:59:42 strib Exp $
 */
public abstract class AckStage extends StandardStage 
    implements EventHandlerIF {

    private static Object GLOBAL_LOCK = new Object();
    private static short stage_id_counter = 0;
    
    protected short my_stage_id;
    protected Resender resender;

    // nonce counter
    protected int next_nonce;

    // list of nonce messages waiting to be sent
    protected LinkedList waiting_msgs;
    // list of messages that need acks
    protected LinkedList waiting_acks;

    // map of waiting nonce msgs, keyed by nonce
    private HashMap waiting_nonces;

    /**
     * Sets the stage id, the initial nonce, and register for events that
     * the stage cares about, such as {@link NonceAckMsg} and 
     * {@link ostore.dispatch.Signal.AlarmSignal}.
     */
    public void init(ConfigDataIF config) throws Exception {

	super.init( config );

	waiting_msgs = new LinkedList();
	waiting_acks = new LinkedList();

	// create first nonce by taking the lower 32 bits of the time,
	// and concatenating on 32 zeros.  Next nonces can then quickly
	// be created by adding one.  This is guaranteed to be unique
	// across restarts and such.
	long time = Clock.current_date(my_node_id).getTime();
	// get rid of the first 48 bits
	time = time & 0x000000000000FFFF;
	next_nonce = (int) (time << (ByteUtils.SIZE_INT*8/2));

	resender = new Resender( my_node_id, tag );
	waiting_nonces = new HashMap();

	// pick a new stage id
	synchronized( GLOBAL_LOCK ) {
	    my_stage_id = stage_id_counter;
	    stage_id_counter++;
	}

	// now register yourself for plain acks that are only registered
	// for this stage
	Filter filter = new Filter ();
	String ackName = "ostore.util.NonceAckMsg";
	if (! filter.requireType (Class.forName (ackName)))
		BUG (tag + ": could not require type " + ackName);
	if (! filter.requireValue ("stage_id", new Short (my_stage_id)))
	    BUG (tag + ": could not require stage_id = " + my_stage_id + 
		 " for " + ackName);
	if (! filter.requireValue ("inbound", new Boolean (true)))
	    BUG (tag + ": could not require inbound = true for " 
		 + ackName);
	classifier.subscribe (filter, my_sink);

	// listen for the resender events
	Signal.getSignal(my_node_id).register(Signal.ALRM, tag, my_sink);

    }

    public void destroy() {
	resender.clear(true);
    }

    public void handleEvents(QueueElementIF element_array[]) 
    throws EventHandlerException {
	preprocessAcks();
	for (int i = 0; i < element_array.length; ++i)
	    handle_event( element_array[i] );
	postprocessAcks();
    }

    public void handleEvent(QueueElementIF element) 
	throws EventHandlerException {
	preprocessAcks();
	handle_event( element );
	postprocessAcks();
    }

    /**
     * Handles a single event with no extra processing.  Needs to be
     * overloaded by each child class, but should be called as
     * <code>super.handle_event(element)</code> at the top of the
     * overloaded method.
     */
    protected void handle_event( QueueElementIF element )
	throws EventHandlerException {

	final String mtag = tag + ".handle_event";

	if( element instanceof Signal.AlarmSignal ) {
	    handleAlarmSignal();
	} else if( element instanceof NonceAckMsg ) {
	    NonceAckMsg nam = (NonceAckMsg) element;
	    receivedAck( nam );
	    if( nam instanceof NonceMsg ) {
		// ack this nonce message
		waiting_acks.add( nam );
	    }
	}

	// now return to the child class's handle_event

    }

    /**
     * Received a resender timeout event 
     */
    protected void handleAlarmSignal() {

	final String mtag = tag + ".handleAlarmSignal";

	resender.process_queues();
	QueueElementIF[] timedOut = resender.clear(false);
	if( DEBUG ) {
	    Debug.printtagln( mtag, "called with queue size of " + 
			      timedOut.length );
	}
	for( int i = 0; i < timedOut.length; i++ ) {
	    NonceMsg nm = (NonceMsg) timedOut[i];
	    NonceState ns = 
		(NonceState) waiting_nonces.remove(new Integer(nm.getNonce()));
	    if( ns == null ) {
		if( DEBUG ) {
		    Debug.printtagln( mtag, "Got a timeout for nonce " + 
				      nm.getNonce() + " but had no such state."
				      + " Ignoring." );
		}
	    } else {
		timeoutCallback( nm, ns.userData );
	    }
	}
    }
    
    /**
     * Application level timeout.  Must be overloaded by the child class.
     *
     * @param msg  the msg that was never acked
     * @param userData  the data passed into {@link #sendNonceMsg}
     */
    protected abstract void timeoutCallback( NonceMsg msg, Object userData );


    protected void preprocessAcks() {
	// do nothing for now, since the simulator might call handleEvent
	// recursively
    }

    protected void postprocessAcks() {

	final String mtag = tag + ".postprocessAcks";

	// now aggregate acks and send off
	for( int i = 0; i < waiting_acks.size(); i++ ) {
	    NonceMsg toBeAcked = (NonceMsg) waiting_acks.get(i);
	    // see if there's a nonce message going to this peer already
	    boolean found = false;
	    for( int j = 0; j < waiting_msgs.size(); j++ ) {
		NonceAckMsg toBeSent = 
		    ((NonceState) waiting_msgs.get(j)).msg;
		// don't need to check if stage_ids are equal
		// because we assume here that if we're sending a NonceMsg
		// to the other node, it will be caught by the same stage
		// regardless of stage id (since it is not a plain
		// NonceAckMsg)
		if( toBeAcked.peer.equals( toBeSent.peer ) ) {
		    toBeSent.addNonceAck( toBeAcked.getNonce() );
		    if( DEBUG ) {
			Debug.printtagln( mtag, "Piggybacking ack for " +
					  toBeAcked + " onto " + 
					  toBeSent );
		    }
		    found = true;
		    break;
		}
	    }
	    if( !found ) {
		// if there are no messages going there, make a new one
		NonceAckMsg newAck = new NonceAckMsg( toBeAcked.peer );
		newAck.addNonceAck( toBeAcked.getNonce() );
		// make sure this is set to the NonceMsg's stage id,
		// and not my_stage_id, since the same stage on
		// different nodes aren't guaranteed to have the
		// same stage id.
		newAck.setStageId( toBeAcked.getStageId() );
		waiting_msgs.add( new NonceState( newAck, -1, null ) );
	    }
	}

	// now send all the waiting msgs
	for( int i = 0; i < waiting_msgs.size(); i++ ) {
	    NonceState ns = (NonceState) waiting_msgs.get(i);
	    if( ns.timeout < 0 ) { 
		dispatch( ns.msg );
	    } else {
		if( DEBUG ) {
		    Debug.printtagln( mtag, "Sending event: " + ns.msg );
		}
		ns.token = resender.send( ns.msg, ns.timeout );
		waiting_nonces.put( new Integer(((NonceMsg)ns.msg).getNonce()),
				    ns );
	    }
	}

	// make them new LinkedLists here, since the simulator could
	// call handleEvent recursively I think
	waiting_msgs = new LinkedList();
	waiting_acks = new LinkedList();

    }

    /**
     * Returns the currently un-acked nonce that was associated with the
     * given userData.  This association was created through a call to either
     * {@link #sendNonceMsg} or {@link #forceRegisterNonce}.
     * <br><br>
     * NOTE: This is a very slow method (involves a linear search 
     * over all currently un-acked nonces), so it should not be used
     * if speed is of the essence.  Also, if more than one nonce is 
     * associated with the given userData, it will only return the
     * first one found.
     */
    protected int findAssociatedNonce( Object userData ) 
	throws NonceNotFoundException {

	Iterator iter = waiting_nonces.values().iterator();
	while( iter.hasNext() ) {
	    NonceState ns = (NonceState) iter.next();
	    if( userData == ns.userData ) {
		return ((NonceMsg)ns.msg).getNonce();
	    } else if( userData == null || ns.userData == null ) {
		// one (but not both) is null, so it can't be
		continue;
	    } else if( userData.equals( ns.userData ) ) {
		return ((NonceMsg)ns.msg).getNonce();
	    }
	}
	// we couldn't find it
	throw new NonceNotFoundException( "Couldn't find nonce for userData: " 
					  + userData );
    }

    /**
     * Send a reliable NetworkMessage.
     *
     * @param msg the message to send
     * @param timeout  how long to wait for a response
     * @param userData  an arbitrary object that will be passed back 
     *                  on a timeout
     */
    protected void sendNonceMsg( NonceMsg msg, long timeout, 
				 Object userData ) {

	final String mtag = tag + ".sendNonceMsg";
	msg.setNonce( nextNonce() );
	if( DEBUG ) {
	    Debug.printtagln( mtag, "Setting ack on " + msg + " to " +
			      msg.getNonce() );
	}
	msg.setStageId( my_stage_id );
	waiting_msgs.add( new NonceState( msg, timeout, userData ) );
    }

    /**
     * Send a NonceAckMsg (for which you don't want an ack).
     * Allows acks for other messages to piggy back.  This is best used
     * for messages that are already acting as application-level acks
     * for NonceMsgs, but don't themselves require a nonce.
     */
    protected void sendNonceAckMsg( NonceAckMsg msg ) {
	waiting_msgs.add( new NonceState( msg, -1, null ) );
    }

    /**
     * Keep a record of a non-NonceMsg message.  This can be used for
     * end-to-end acks on things like RouteMsgs.  This does not actaully
     * send any messages.
     *
     * @param timeout  how long to wait for a response
     * @param userData  an arbitrary object that will be passed back 
     *                  on a timeout
     * @return the nonce assigned to this message
     */
    protected int forceRegisterNonce( long timeout, Object userData ) {
       	NonceMsg nm = new GenericNonceMsg( null );
	nm.setNonce( nextNonce() );
	nm.setStageId( my_stage_id );
	NonceState ns = new NonceState( nm, timeout, userData );
	ns.token = resender.remember( nm, ns.timeout );
	waiting_nonces.put( new Integer(nm.getNonce()), ns );
	return nm.getNonce();
    }

    /**
     * Send an ack for a non-NonceMsg message.  This can be used for
     * end-to-end acks on things like RouteMsgs, if the nonce is
     * included in the original message
     *
     * @param peer to whom to send an ack
     * @param nonce  the nonce to ack
     */
    protected void forceSendAck( NodeId peer, int nonce, short stage_id ) {
	NonceMsg nm = new GenericNonceMsg( peer );
	nm.setNonce( nonce );
	nm.setStageId( stage_id );
	waiting_acks.add( nm );
    }

    /**
     * Got an ack.
     */
    protected void receivedAck( NonceAckMsg msg ) {

	final String mtag = tag + ".receivedAck";

	// for each nonce included in the ack, let the resender know
	// that it has been received
	int[] nonces = msg.getNonceAcks();
	for( int i = 0; i < nonces.length; i++ ) {
	    Integer nonce = new Integer( nonces[i] );
	    NonceState ns = (NonceState) waiting_nonces.remove( nonce );
	    if( DEBUG ) {
		Debug.printtagln( mtag, "Got ack for " + nonce );
	    }
	    if( ns == null ) {
		if( DEBUG ) {
		    Debug.printtagln( mtag, "Received an ack for nonce " +
				      nonces[i] + " but have no record of " +
				      "that nonce.  Could have already timed "+
				      "out.  Ignoring." );
		}
		continue;
	    }
	    // cancel the alarm
	    resender.cancel( ns.token );
	}
    }

    /**
     * Assign a new nonce.
     */
    private synchronized int nextNonce() {
	// I believe it's safe to just keep on incrementin', even if you
	// increment past the original 16 zeros that were placed at the
	// end of the current time in the original nonce.  Think about it:
	// by the time it increments through all 2^16 values, at least two
	// milliseconds will have passed in real time, so it won't be possible
	// to restart and come back up with the same nonce
	int tmp = next_nonce;
	if( next_nonce == Integer.MAX_VALUE ) {
	    next_nonce = 0;
	} else {
	    next_nonce++;
	}
	return tmp;
    }

    private class NonceState {
	public NonceAckMsg msg;
	public long timeout;
	public Object userData;
	public Object token;

	public NonceState( NonceAckMsg msg, long timeout, 
				 Object userData ) {
	    this.msg = msg;
	    this.timeout = timeout;
	    this.userData = userData;

	}
    }

    protected class GenericNonceMsg extends NonceMsg {
	
	public GenericNonceMsg( NodeId peer ) {
	    super( peer );
	}

    }
    
    protected class NonceNotFoundException extends Exception {
	public NonceNotFoundException( String s ) {
	    super(s);
	}
    }

}
