/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.util;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;

/**
 * A <code>QSHashMap</code> is a wrapper class of {@link java.util.HashMap}
 * that implements {@link ostore.util.QuickSerializable}.  All elements of the
 * <code>QSHashMap</code> must also be {@link ostore.util.QuickSerializable}.
 *
 * Most of the functionality from the {@link java.util.HashMap} has not been
 * ported.  If you require the use of a missing method, implement it.
 *
 *  @author  Jeremy Stribling
 *  @version $Id: QSHashMap.java,v 1.1 2003/01/24 21:02:10 strib Exp $
 */
public class QSHashMap implements QuickSerializable
{
    private HashMap _map;

    /**
     * Constructs a <code>QSHashMap</code>. */
    public QSHashMap() 
    {
	_map = new HashMap();
    }
    
    /**
     * Constructs a <CODE>QSHashMap</CODE> from its
     * {@link ostore.util.QuickSerializable#serialize serialized} form.
     *
     * @param buffer {@link ostore.util.QuickSerializable#serialize serialized}
     *                form of object.  */
    public QSHashMap(InputBuffer buffer) throws QSException
    {
	int size = buffer.nextInt();
	_map = new HashMap();
	for(int i = 0; i < size; i++) {
	    _map.put(getQS( buffer ), getQS( buffer ));
	}
    }

    /**
     * Helper method to get a possibly null object from the buffer
     */
    private QuickSerializable getQS( InputBuffer buffer ) throws QSException {
	if( buffer.nextBoolean() ) {
	    return buffer.nextObject();
	} else {
	    return null;
	}
    }
    
    /**
     * As specified by {link ostore.util.QuickSerializable}.
     */
    public void serialize(OutputBuffer buffer)
    {
	buffer.add(_map.size());
	Iterator i = _map.keySet().iterator();
	while(i.hasNext()) {
	    QuickSerializable key = (QuickSerializable) i.next();
	    addQS( buffer, (QuickSerializable) key );
	    addQS( buffer, (QuickSerializable) _map.get( key ) );
	}
    }
    
    /**
     * Helper method to add a possibly null object to the buffer
     */
    private void addQS( OutputBuffer buffer, QuickSerializable qs ) {
	if( qs != null ) {
	    buffer.add( true );
	    buffer.add( qs );
	} else {
	    buffer.add( false );
	}
    }

    /**
     * Return a hashcode, specified and generated as in 
     * {@link java.util.HashMap}.
     *
     * @return a hashcode which incorporates the hashcodes of all the elements
     * in the <code>QSTreeSet</code> */
    public int hashCode()
    {
	return _map.hashCode();
    }

    protected HashMap getHashMap() {
	return _map;
    }

    /**
     * Compare this <code>QSHashMap</code> to another <code>Object</code>.
     *
     * @return <code>true</code> iff the other <code>Object</code> is also a
     * <code>QSHashMap</code> and contains <code>equal</code> elements. */ 
    public boolean equals(Object o)
    {
	if(o == null || !(o instanceof QSHashMap))
	    return false;
	else
	    return _map.equals(((QSHashMap)o).getHashMap());
    }
    
    ////////////////////////////////////////////////////
    // Methods which mirror java.util.HashMap

    /**
     * Returns a set view of the keys contained in this map.
     *
     * @return a set view of the keys contained in this map */
    public Set keySet()
    {
	return _map.keySet();
    }

    /**
     * Returns a collection view of the values contained in this map.
     *
     * @return a collection view of the values contained in this map */
    public Collection values()
    {
	return _map.values();
    }

    /**
     * Returns the number of mappings in this map.
     *
     * @return the number of mappings in this map */
    public int size() 
    {
	return _map.size();
    }

    /**
     * Returns <tt>true</tt> if this map contains no mappings.
     *
     * @return <tt>true</tt> if this map contains no mappings */
    public boolean isEmpty()
    {
	return _map.isEmpty();
    }
    
    /**
     * Returns <tt>true</tt> if this map contains the specified key.
     *
     * @param o the key to be checked for containment in this map.
     * @return <tt>true</tt> if this map contains the specified key.
     * 
     * @throws ClassCastException if the specified object cannot be compared
     * 		  with the elements currently in the map. */
    public boolean containsKey(QuickSerializable o)
    {
	return _map.containsKey(o);
    }

    /**
     * Returns <tt>true</tt> if this map contains the specified value.
     *
     * @param o the value to be checked for containment in this map.
     * @return <tt>true</tt> if this map contains the specified value.
     * 
     * @throws ClassCastException if the specified object cannot be compared
     * 		  with the elements currently in the map. */
    public boolean containsValue(QuickSerializable o)
    {
	return _map.containsValue(o);
    }

    /**
     * Puts this mapping in the map
     *
     * @throws ClassCastException if the specified object cannot be compared
     * 		  with the elements currently in the set. */
    public QuickSerializable put(QuickSerializable key, 
				 QuickSerializable value)
    {
	return (QuickSerializable) _map.put(key, value);
    }

    /**
     * Returns the object mapped to the given key in this map if it is present.
     *
     * @param o key to lookup
     * @return the object the key is mapped to
     * 
     * @throws ClassCastException if the specified object cannot be compared
     * 		  with the elements currently in the set. */
    public QuickSerializable get(QuickSerializable o)
    {
	return (QuickSerializable) _map.get(o);
    }
    
    /**
     * Removes the given key mapping from this map if it is present.
     *
     * @param o key of the mapping to be removed from this map, if present.
     * @return the object the key was mapped to
     * 
     * @throws ClassCastException if the specified object cannot be compared
     * 		  with the elements currently in the set. */
    public QuickSerializable remove(QuickSerializable o)
    {
	return (QuickSerializable) _map.remove(o);
    }

    /**
     * Removes all mappings from this set. */
    public void clear()
    {
	_map.clear();
    }

}
