/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.util;

import java.util.Set;
import java.util.Map;
import java.util.TreeSet;
import java.util.Iterator;
import java.util.SortedSet;
import java.util.SortedMap;
import java.util.Comparator;
import java.util.Collection;

/**
 * A <code>QSTreeSet</code> is a wrapper class of {@link java.util.TreeSet}
 * that implements {@link ostore.util.QuickSerializable}.  All elements of the
 * <code>QSTreeSet</code> must also be {@link ostore.util.QuickSerializable}.
 *
 * Most of the functionality from the {@link java.util.TreeSet} has not been
 * ported.  If you require the use of a missing method, implement it.
 *
 *  @author  Hakim Weatherspoon
 *  @version $Id: QSTreeSet.java,v 1.4 2003/03/18 01:01:19 jeffpang Exp $
 */
public class QSTreeSet implements QuickSerializable
{
  ////////////////////       CONSTANTS	////////////////////
  ////////////////////    STATIC VARIABLES 	////////////////////
  ////////////////////       VARIABLES	////////////////////
  private TreeSet _treeSet;
  private QSComparator _comparator;

  ////////////////////      CONSTRUCTORS	////////////////////

  /**
   * Constructs a <code>QSTreeSet</code>. */
  public QSTreeSet() 
  {
    _treeSet = new TreeSet();
  }

  public QSTreeSet(QSComparator c)
  {
    _comparator = c;
    _treeSet = new TreeSet(c);
  }

  /**
   * Constructs a <CODE>QSTreeSet</CODE> from its
   * {@link ostore.util.QuickSerializable#serialize serialized} form.
   *
   * @param buffer {@link ostore.util.QuickSerializable#serialize serialized}
   *                form of object.  */
  public QSTreeSet(InputBuffer buffer) throws QSException
  {
    // comparator?
    boolean hasComparator = buffer.nextBoolean();
    if (hasComparator) {
      _comparator = (QSComparator)buffer.nextObject();
      _treeSet = new TreeSet(_comparator);
    } else
      _treeSet = new TreeSet();
    int size = buffer.nextInt();
    for(int i = 0; i < size; i++)
      _treeSet.add(buffer.nextObject());
  }

  /**
   * Constructs a new set containing the same elements as the given sorted
   * set, sorted according to the same ordering.
   *
   * @param s sorted set whose elements will comprise the new set.
   *
   * @throws QSException if any of the elements are not instanceof 
   *                     {@link ostore.util.QuickSerializable}. */
  public QSTreeSet(SortedSet s) throws QSException
  {
    QuickSerializable element = null;
    Iterator i = s.iterator();
    while(i.hasNext())
    { 
      element = (QuickSerializable)i.next();

      if(!(element instanceof QuickSerializable))
	  throw new QSException("QSTreeSet.<init>: element=" + element + 
				" instanceof" + 
				(element==null?null:element.getClass().getName() +
				 ", NOT instanceof QuickSerializable."));
    }
    _treeSet = new TreeSet(s);
  }

  /**
   * Constructs a new set containing the elements in the specified
   * collection, sorted according to the elements' <i>natural order</i>.
   * All keys inserted into the set must implement the <tt>Comparable</tt>
   * interface.  Furthermore, all such keys must be <i>mutually
   * comparable</i>: <tt>k1.compareTo(k2)</tt> must not throw a
   * <tt>ClassCastException</tt> for any elements <tt>k1</tt> and
   * <tt>k2</tt> in the set.
   *
   * @param c The elements that will comprise the new set.
   *
   * @throws ClassCastException if the keys in the given collection are not
   * comparable, or are not mutually comparable.
   * @throws QSException if any of the elements are not instanceof 
   *                     {@link ostore.util.QuickSerializable}. */
  public QSTreeSet(Collection c) throws QSException
  {
    _treeSet = new TreeSet();
    addAll(c);
  }

  ////////////////////     STATIC METHODS 	////////////////////
  //////////////////// 	METHODS		////////////////////

  public void serialize(OutputBuffer buffer)
  {
    // comparator?
    if (_comparator == null)
      buffer.add(false);
    else {
      buffer.add(true);
      buffer.add(_comparator);
    }
    buffer.add(_treeSet.size());
    Iterator i = _treeSet.iterator();
    while(i.hasNext())
      buffer.add((QuickSerializable)i.next());
  }
    
  //  end methods for ostore.util.QuickSerializable interface

  public TreeSet getTreeSet()
  {
    return _treeSet;
  }

 /**
  * Return a hashcode, specified and generated as in {@link java.util.TreeSet}.
  *
  * @return a hashcode which incorporates the hashcodes of all the elements
  * in the <code>QSTreeSet</code> */
  public int hashCode()
  {
    return _treeSet.hashCode();
  }

  /**
   * Compare this <code>QSTreeSet</code> to another <code>Object</code>.
   *
   * @return <code>true</code> iff the other <code>Object</code> is also a
   * <code>QSTreeSet</code> and contains <code>equal</code> elements. */ 
  public boolean equals(Object o)
  {
    if(o == null || !(o instanceof QSTreeSet))
      return false;
    else
      return _treeSet.equals(o);
  }

  /**
   * Create an exact copy of this <CODE>QSTreeSet</CODE>
   * @return new <CODE>QSTreeSet</CODE> identical to this one.
   *         <I>this is a shallow-copy</I> */
  public Object clone() throws CloneNotSupportedException 
  {
    QSTreeSet clone = null;
    try {
      clone = new QSTreeSet((SortedSet)_treeSet);
    } catch (QSException qse) {
    }

    return clone;
  }

  /**
   * Produce a human-readable version of this <code>QSVector</code>,
   * which includes all the elements.
   *
   * @return a <code>String</code> representation of this
   * 		<code>QSTreeSet</code>. */
  public String toString()
  {
    StringBuffer sbuf = new StringBuffer( "{" );
    Iterator i = _treeSet.iterator();
    while(i.hasNext())
    {
      sbuf.append( ' ' );
      sbuf.append( '(' );
      sbuf.append( (QuickSerializable)i.next() );
      sbuf.append( ')' );
    }
    sbuf.append( '}' );
    return sbuf.toString();
  }

  ////////////////////////////////////////////////////
  // Methods which mirror java.util.TreeSet

  /**
   * Returns an iterator over the elements in this set.  The elements
   * are returned in ascending order.
   *
   * @return an iterator over the elements in this set. */
  public Iterator iterator()
  {
    return _treeSet.iterator();
  }

  /**
   * Returns the number of elements in this set (its cardinality).
   *
   * @return the number of elements in this set (its cardinality). */
  public int size() 
  {
    return _treeSet.size();
  }

  /**
   * Returns <tt>true</tt> if this set contains no elements.
   *
   * @return <tt>true</tt> if this set contains no elements. */
  public boolean isEmpty()
  {
    return _treeSet.isEmpty();
  }

  /**
   * Returns <tt>true</tt> if this set contains the specified element.
   *
   * @param o the object to be checked for containment in this set.
   * @return <tt>true</tt> if this set contains the specified element.
   * 
   * @throws ClassCastException if the specified object cannot be compared
   * 		  with the elements currently in the set. */
  public boolean contains(QuickSerializable o)
  {
    return _treeSet.contains(o);
  }

  /**
   * Adds the specified element to this set if it is not already present.
   *
   * @param o element to be added to this set.
   * @return <tt>true</tt> if the set did not already contain the specified
   *         element.
   * 
   * @throws ClassCastException if the specified object cannot be compared
   * 		  with the elements currently in the set. */
  public boolean add(QuickSerializable o)
  {
    return _treeSet.add(o);
  }

  /**
   * Removes the given element from this set if it is present.
   *
   * @param o object to be removed from this set, if present.
   * @return <tt>true</tt> if the set contained the specified element.
   * 
   * @throws ClassCastException if the specified object cannot be compared
   * 		  with the elements currently in the set. */
  public boolean remove(QuickSerializable o)
  {
    return _treeSet.remove(o);
  }

  /**
   * Removes all of the elements from this set. */
  public void clear()
  {
    _treeSet.clear();
  }

  /**
   * Adds all of the elements in the specified collection to this set.
   *
   * @param c elements to be added
   * @return <tt>true</tt> if this set changed as a result of the call.
   *
   * @throws ClassCastException if the elements provided cannot be compared
   *		  with the elements currently in the set. 
   * @throws QSException if any of the elements are not instanceof 
   *                     {@link ostore.util.QuickSerializable}. */
  public boolean addAll(Collection c) throws QSException
  {
    QuickSerializable element = null;
    Iterator i = c.iterator();
    while(i.hasNext())
    {
      element = (QuickSerializable)i.next();

      if(!(element instanceof QuickSerializable))
	  throw new QSException("QSTreeSet.addAll: element=" + element + 
				" instanceof" + 
				(element==null?null:element.getClass().getName() +
				 ", NOT instanceof QuickSerializable."));
    }
    return _treeSet.addAll(c);
  }

  public boolean addAll(QSTreeSet c) throws QSException
  {
    return addAll(c.getTreeSet());
  }

  /**
   * Returns a view of the portion of this set whose elements range from
   * <tt>fromElement</tt>, inclusive, to <tt>toElement</tt>, exclusive.  (If
   * <tt>fromElement</tt> and <tt>toElement</tt> are equal, the returned
   * sorted set is empty.)  The returned sorted set is backed by this set,
   * so changes in the returned sorted set are reflected in this set, and
   * vice-versa.  The returned sorted set supports all optional Set
   * operations.<p>
   *
   * The sorted set returned by this method will throw an
   * <tt>IllegalArgumentException</tt> if the user attempts to insert an
   * element outside the specified range.<p>
   *
   * Note: this method always returns a <i>half-open range</i> (which
   * includes its low endpoint but not its high endpoint).  If you need a
   * <i>closed range</i> (which includes both endpoints), and the element
   * type allows for calculation of the successor a given value, merely
   * request the subrange from <tt>lowEndpoint</tt> to
   * <tt>successor(highEndpoint)</tt>.  For example, suppose that <tt>s</tt>
   * is a sorted set of strings.  The following idiom obtains a view
   * containing all of the strings in <tt>s</tt> from <tt>low</tt> to
   * <tt>high</tt>, inclusive: <pre>
   *     SortedSet sub = s.subSet(low, high+"\0");
   * </pre>
   * 
   * A similar technique can be used to generate an <i>open range</i> (which
   * contains neither endpoint).  The following idiom obtains a view
   * containing all of the strings in <tt>s</tt> from <tt>low</tt> to
   * <tt>high</tt>, exclusive: <pre>
   *     SortedSet sub = s.subSet(low+"\0", high);
   * </pre>
   *
   * @param fromElement low endpoint (inclusive) of the subSet.
   * @param toElement high endpoint (exclusive) of the subSet.
   * @return a view of the portion of this set whose elements range from
   * 	       <tt>fromElement</tt>, inclusive, to <tt>toElement</tt>,
   * 	       exclusive.
   * @throws ClassCastException if <tt>fromElement</tt> and
   *         <tt>toElement</tt> cannot be compared to one another using
   *         this set's comparator (or, if the set has no comparator,
   *         using natural ordering).
   * @throws IllegalArgumentException if <tt>fromElement</tt> is greater than
   *         <tt>toElement</tt>.
   * @throws NullPointerException if <tt>fromElement</tt> or
   *	       <tt>toElement</tt> is <tt>null</tt> and this set uses natural
   *	       order, or its comparator does not tolerate <tt>null</tt>
   *         elements.
   */
  public SortedSet subSet(QuickSerializable fromElement, QuickSerializable toElement)
  {
    return _treeSet.subSet(fromElement, toElement);
  }

  /**
   * Returns a view of the portion of this set whose elements are strictly
   * less than <tt>toElement</tt>.  The returned sorted set is backed by
   * this set, so changes in the returned sorted set are reflected in this
   * set, and vice-versa.  The returned sorted set supports all optional set
   * operations.<p>
   *
   * The sorted set returned by this method will throw an
   * <tt>IllegalArgumentException</tt> if the user attempts to insert an
   * element greater than or equal to <tt>toElement</tt>.<p>
   *
   * Note: this method always returns a view that does not contain its
   * (high) endpoint.  If you need a view that does contain this endpoint,
   * and the element type allows for calculation of the successor a given
   * value, merely request a headSet bounded by
   * <tt>successor(highEndpoint)</tt>.  For example, suppose that <tt>s</tt>
   * is a sorted set of strings.  The following idiom obtains a view
   * containing all of the strings in <tt>s</tt> that are less than or equal
   * to <tt>high</tt>: <pre> SortedSet head = s.headSet(high+"\0");</pre>
   *
   * @param toElement high endpoint (exclusive) of the headSet.
   * @return a view of the portion of this set whose elements are strictly
   * 	       less than toElement.
   * @throws ClassCastException if <tt>toElement</tt> is not compatible
   *         with this set's comparator (or, if the set has no comparator,
   *         if <tt>toElement</tt> does not implement <tt>Comparable</tt>).
   * @throws IllegalArgumentException if this set is itself a subSet,
   *         headSet, or tailSet, and <tt>toElement</tt> is not within the
   *         specified range of the subSet, headSet, or tailSet.
   * @throws NullPointerException if <tt>toElement</tt> is <tt>null</tt> and
   *	       this set uses natural ordering, or its comparator does
   *         not tolerate <tt>null</tt> elements. */
  public SortedSet headSet(QuickSerializable toElement) 
  {
    return _treeSet.headSet(toElement);
  }

  /**
   * Returns a view of the portion of this set whose elements are
   * greater than or equal to <tt>fromElement</tt>.  The returned sorted set
   * is backed by this set, so changes in the returned sorted set are
   * reflected in this set, and vice-versa.  The returned sorted set
   * supports all optional set operations.<p>
   *
   * The sorted set returned by this method will throw an
   * <tt>IllegalArgumentException</tt> if the user attempts to insert an
   * element less than <tt>fromElement</tt>.
   *
   * Note: this method always returns a view that contains its (low)
   * endpoint.  If you need a view that does not contain this endpoint, and
   * the element type allows for calculation of the successor a given value,
   * merely request a tailSet bounded by <tt>successor(lowEndpoint)</tt>.
   * For example, suppose that <tt>s</tt> is a sorted set of strings.  The
   * following idiom obtains a view containing all of the strings in
   * <tt>s</tt> that are strictly greater than <tt>low</tt>: <pre>
   *     SortedSet tail = s.tailSet(low+"\0");
   * </pre>
   *
   * @param fromElement low endpoint (inclusive) of the tailSet.
   * @return a view of the portion of this set whose elements are
   * 	       greater than or equal to <tt>fromElement</tt>.
   * @throws ClassCastException if <tt>fromElement</tt> is not compatible
   *         with this set's comparator (or, if the set has no comparator,
   *         if <tt>fromElement</tt> does not implement <tt>Comparable</tt>).
   * @throws IllegalArgumentException if this set is itself a subSet,
   *         headSet, or tailSet, and <tt>fromElement</tt> is not within the
   *         specified range of the subSet, headSet, or tailSet.
   * @throws NullPointerException if <tt>fromElement</tt> is <tt>null</tt>
   *	       and this set uses natural ordering, or its comparator does
   *         not tolerate <tt>null</tt> elements. */
  public SortedSet tailSet(QuickSerializable fromElement)
  {
    return _treeSet.tailSet(fromElement);
  }


  ////////////////////////////////////////////////////
  // Methods which mirror java.util.AbstractCollection

  /**
   * Returns <tt>true</tt> if this collection contains all of the elements
   * in the specified collection. <p>
   *
   * This implementation iterates over the specified collection, checking
   * each element returned by the iterator in turn to see if it's
   * contained in this collection.  If all elements are so contained
   * <tt>true</tt> is returned, otherwise <tt>false</tt>.
   *
   * @param c collection to be checked for containment in this collection.
   * @return <tt>true</tt> if this collection contains all of the elements
   * 	       in the specified collection.
   * 
   * @see #contains(QuickSerializable)  */
  public boolean containsAll(Collection c) throws QSException
  {
    Iterator e = c.iterator();
    while (e.hasNext())
    {
      Object element = e.next();
      if(!(element instanceof QuickSerializable))
	throw new QSException("QSTreeSet.containsAll: element=" + element + 
			      " instanceof" + 
			      (element==null?null:element.getClass().getName() +
			       ", NOT instanceof QuickSerializable."));
      if(!contains((QuickSerializable)element))
	return false;
    }
    
    return true;
  }

  public boolean containsAll(QSTreeSet c) throws QSException
  {
    return containsAll(c.getTreeSet());
  }

  ////////////////////////////////////////////////////
  // Methods which mirror java.util.AbstractSet

  /**
   * Removes from this set all of its elements that are contained in
   * the specified collection (optional operation).<p>
   *
   * This implementation determines which is the smaller of this set
   * and the specified collection, by invoking the <tt>size</tt>
   * method on each.  If this set has fewer elements, then the
   * implementation iterates over this set, checking each element
   * returned by the iterator in turn to see if it is contained in
   * the specified collection.  If it is so contained, it is removed
   * from this set with the iterator's <tt>remove</tt> method.  If
   * the specified collection has fewer elements, then the
   * implementation iterates over the specified collection, removing
   * from this set each element returned by the iterator, using this
   * set's <tt>remove</tt> method.<p>
   *
   * Note that this implementation will throw an
   * <tt>UnsupportedOperationException</tt> if the iterator returned by the
   * <tt>iterator</tt> method does not implement the <tt>remove</tt> method.
   *
   * @param c elements to be removed from this set.
   * @return <tt>true</tt> if this set changed as a result of the call.
   *
   * @throws    UnsupportedOperationException removeAll is not supported
   *            by this set.
   * @throws QSException if any of the elements are not instanceof 
   *                     {@link ostore.util.QuickSerializable}. 
   * @see #remove(QuickSerializable)
   * @see #contains(QuickSerializable) */
  public boolean removeAll(Collection c) throws QSException
  {
    QuickSerializable element = null;
    Iterator i = c.iterator();
    while(i.hasNext())
    {
      element = (QuickSerializable)i.next();

      if(!(element instanceof QuickSerializable))
	  throw new QSException("QSTreeSet.addAll: element=" + element + 
				" instanceof" + 
				(element==null?null:element.getClass().getName() +
				 ", NOT instanceof QuickSerializable."));
    }
    return _treeSet.removeAll(c);
  }

  public boolean removeAll(QSTreeSet c) throws QSException
  {
    return removeAll(c.getTreeSet());
  }


  /**
   * Returns the comparator used to order this sorted set, or <tt>null</tt>
   * if this tree set uses its elements natural ordering.
   *
   * @return the comparator used to order this sorted set, or <tt>null</tt>
   * if this tree set uses its elements natural ordering. */
  public Comparator comparator()
  {
    return _treeSet.comparator();
  }

  /**
   * Returns the first (lowest) element currently in this sorted set.
   *
   * @return the first (lowest) element currently in this sorted set.
   * @throws    NoSuchElementException sorted set is empty. */
  public Object first()
  {
    return _treeSet.first();
  }

  /**
   * Returns the last (highest) element currently in this sorted set.
   *
   * @return the last (highest) element currently in this sorted set.
   * @throws    NoSuchElementException sorted set is empty.
   */
  public QuickSerializable last()
  {
    return (QuickSerializable)_treeSet.last();
  }

  /**
   * Returns an array containing all of the elements in this collection.  If
   * the collection makes any guarantees as to what order its elements are
   * returned by its iterator, this method must return the elements in the
   * same order.  The returned array will be "safe" in that no references to
   * it are maintained by the collection.  (In other words, this method must
   * allocate a new array even if the collection is backed by an Array).
   * The caller is thus free to modify the returned array.<p>
   *
   * This implementation allocates the array to be returned, and iterates
   * over the elements in the collection, storing each object reference in
   * the next consecutive element of the array, starting with element 0.
   *
   * @return an array containing all of the elements in this collection. */
  public QuickSerializable[] toArray()
  {
    QuickSerializable[] result = new QuickSerializable[_treeSet.size()];
    Iterator e = _treeSet.iterator();
    for(int i=0; e.hasNext(); i++)
      result[i] = (QuickSerializable)e.next();
    return result;
  }

  /**
   * Returns an array with a runtime type is that of the specified array and
   * that contains all of the elements in this collection.  If the
   * collection fits in the specified array, it is returned therein.
   * Otherwise, a new array is allocated with the runtime type of the
   * specified array and the size of this collection.<p>
   *
   * If the collection fits in the specified array with room to spare (i.e.,
   * the array has more elements than the collection), the element in the
   * array immediately following the end of the collection is set to
   * <tt>null</tt>.  This is useful in determining the length of the
   * collection <i>only</i> if the caller knows that the collection does
   * not contain any <tt>null</tt> elements.)<p>
   *
   * If this collection makes any guarantees as to what order its elements
   * are returned by its iterator, this method must return the elements in
   * the same order. <p>
   *
   * This implementation checks if the array is large enough to contain the
   * collection; if not, it allocates a new array of the correct size and
   * type (using reflection).  Then, it iterates over the collection,
   * storing each object reference in the next consecutive element of the
   * array, starting with element 0.  If the array is larger than the
   * collection, a <tt>null</tt> is stored in the first location after the
   * end of the collection.
   *
   * @param  a the array into which the elements of the collection are to
   * 	       be stored, if it is big enough; otherwise, a new array of the
   * 	       same runtime type is allocated for this purpose.
   * @return an array containing the elements of the collection.
   * 
   * @throws NullPointerException if the specified array is <tt>null</tt>.
   * 
   * @throws ArrayStoreException if the runtime type of the specified array
   *         is not a supertype of the runtime type of every element in this
   *         collection. */
  public QuickSerializable[] toArray(QuickSerializable a[])
  {
    int size = _treeSet.size();
    if (a.length < size)
	a = (QuickSerializable[])java.lang.reflect.Array.newInstance(
                                  a.getClass().getComponentType(), size);

    Iterator it=_treeSet.iterator();
    for (int i=0; i<size; i++)
	a[i] = (QuickSerializable)it.next();
    
    if (a.length > size)
	a[size] = null;
    
    return a;
  }
}

