/**
 * ostore_util_sha1_c.c
 *
 * @author   Hakim Weatherspoon
 * @version  $Id: ostore_util_sha1_c.c,v 1.3 2002/09/07 02:19:28 strib Exp $
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 **/

/**
 * <CODE>ostore_util_sha1_c</CODE> implements the
 * <CODE>SHA1NativeIF</CODE> and <CODE>ostore_util_sha1_c.h</CODE> 
 * interface.  Specifically, <CODE>ostore_util_sha1_c</CODE> implements
 * <CODE>SHA-1</CODE>; the message digest scheme described in FIPS 180-1.
 *
 * The result of <CODE>ostore_util_sha1_c</CODE> is used by the 
 * {@link ostore.util.SHA1Hash} and {@link ostore.util.SHA1HashBuffer} 
 * class and by all entities which are SHA1 hashes.
 * (In the prototype, this means Guids and VHashes).
 *
 * @author   Hakim Weatherspoon
 * @version  $Id: ostore_util_sha1_c.c,v 1.3 2002/09/07 02:19:28 strib Exp $
 **/
#include <jni.h>
#include <stdio.h>
#include <sys/time.h>
#include "ostore_util_SHA1NativeIF.h"
#include "ostore_util_sha1_c.h"

#define INIT1_CONST 0x67452301UL
#define INIT2_CONST 0xefcdab89UL
#define INIT3_CONST 0x98badcfeUL
#define INIT4_CONST 0x10325476UL
#define INIT5_CONST 0xc3d2e1f0UL

#define RND1_CONST 0x5a827999UL
#define RND2_CONST 0x6ed9eba1UL
#define RND3_CONST 0x8f1bbcdcUL
#define RND4_CONST 0xca62c1d6UL

#define COUNTMAX 64
#define COUNTMASK COUNTMAX-1

#define STATE_MEMBER_NAME "mdState"
#define INT_ARRAY_SYMBOL "[I"

void zeromem(void *dst, unsigned long len)
{
 unsigned char *mem = (unsigned char *)dst;
 while (len--)
    *mem++ = 0;
}

#define F0(x,y,z)  ( (x&y) | ((~x)&z) )
#define F1(x,y,z)  (x ^ y ^ z)
#define F2(x,y,z)  ((x & y) | (z & (x | y)))
#define F3(x,y,z)  (x ^ y ^ z)

void sha1_init(struct sha1_state* md)
{
  md->AA = INIT1_CONST;
  md->BB = INIT2_CONST;
  md->CC = INIT3_CONST;
  md->DD = INIT4_CONST;
  md->EE = INIT5_CONST;
  md->count = 0;
  zeromem(md->W, 320);
}

/**
 * Compute a the hash for the current block.
 *
 * This is in the same vein as Peter Gutmann's algorithm listed in
 * the back of Applied Cryptography, Compact implementation of
 * "old" NIST Secure Hash Algorithm.
 *
 */
void computeBlock(struct sha1_state* md) 
{
  unsigned int temp=0, a, b, c, d, e, i;
  unsigned int* W = md->W;
  int t = -1;

  // The first 16 ints have the byte stream, compute the rest of
  // the buffer
  for (t = 16; t <= 79; t++) {
    temp = W[t-3] ^ W[t-8] ^ W[t-14] ^ W[t-16];
    /*W[t] = ((temp << 1) | (temp >>(32 - 1)));*/
    W[t] = ROL(temp,1);
  }
  
  a = md->AA;
  b = md->BB;
  c = md->CC;
  d = md->DD;
  e = md->EE;
  
  // Round 1
  for (i = 0; i < 20; i++) {
    /*temp = ((a<<5) | (a>>(32-5))) +
      ((b&c)|((~b)&d))+ e + W[i] + RND1_CONST;*/
    temp = ROL(a, 5) + F0(b,c,d) + e + W[i] + RND1_CONST;
    e = d;
    d = c;
    /*c = ((b<<30) | (b>>(32-30)));*/
    c = ROL(b, 30); 
    b = a;
    a = temp;
  }
  
  // Round 2
  for (i = 20; i < 40; i++) {
    /*temp = ((a<<5) | (a>>(32-5))) +
      (b ^ c ^ d) + e + W[i] + RND2_CONST;*/
    temp = ROL(a, 5) + F1(b,c,d) + e + W[i] + RND2_CONST;
    e = d;
    d = c;
    /*c = ((b<<30) | (b>>(32-30)));*/
    c = ROL(b, 30); 
    b = a;
    a = temp;
  }

  // Round 3
  for (i = 40; i < 60; i++) {
    /*temp = ((a<<5) | (a>>(32-5))) +
      ((b&c)|(b&d)|(c&d)) + e + W[i] + RND3_CONST;*/
    temp = ROL(a, 5) + F2(b,c,d) + e + W[i] + RND3_CONST;
    e = d;
    d = c;
    /*c = ((b<<30) | (b>>(32-30)));*/
    c = ROL(b, 30);
    b = a;
    a = temp;
  }

  // Round 4
  for (i = 60; i < 80; i++) {
    /*temp = ((a<<5) | (a>>(32-5))) +
      (b ^ c ^ d) + e + W[i] + RND4_CONST;*/
    temp = ROL(a, 5) + F3(b,c,d) + e + W[i] + RND4_CONST;
    e = d;
    d = c;
    /*c = ((b<<30) | (b>>(32-30)));*/
    c = ROL(b, 30);
    b = a;
    a = temp;
  }

  md->AA += a;
  md->BB += b;
  md->CC += c;
  md->DD += d;
  md->EE += e;
}

/**
 * Update a byte.
 *
 * @param b	the byte
*/
void update1(int b, struct sha1_state* md)  
{
  int word;
  int offset;
  unsigned int* W = md->W;

  /* compute word offset and bit offset within word the low bits
     of count are inverted to make put the bytes in the write
     order */
  word = ((int)md->count & COUNTMASK) >> 2;
  offset = (~(int)md->count & 3) << 3;
  
  W[word] = (W[word] & ~(0xff << offset)) | ((b & 0xff) << offset);
  
  /* If this is the last byte of a block, compute the partial hash */
  if (((int)md->count & COUNTMASK) == COUNTMASK) {
    computeBlock(md);
  }
  md->count++;
}

void update2(const unsigned char* b, int off, int len, struct sha1_state* md)
{
  unsigned int* W = md->W;
  int word;
  b = b+off;
  
  // Use single writes until integer aligned
  while ((len > 0) &&
	 ((int)md->count & 3) != 0) {
    update1((int)*b++, md);
    len--;
  }


  /* Assemble groups of 4 bytes to be inserted in integer array */
  for (;len >= 4; len -= 4) {
      
    word = ((int)md->count & COUNTMASK) >> 2;

    LOAD32H(W[word],b);
    b+=4;
    md->count += 4;

    if (((int)md->count & COUNTMASK) == 0) {
      computeBlock(md);
    }
  }
	
  /* Use single writes for last few bytes */
  for (; len > 0; len--) {
    update1((int)*b++, md);
  }
}

int digest(unsigned char* hashvalue, int offset, int length, struct sha1_state* md)
{
  unsigned int a, b, c, d, e;

  /* The number of bits before padding occurs */
  ulong64 bits = md->count << 3;	

  update1(0x80, md);

  /* Pad with zeros until length is a multiple of 448 (the last two
     32 ints are used a holder for bits (see above). */
  while ((int)(md->count & COUNTMASK) != 56) {
    update1(0, md);
  }

  md->W[14] = (int)(bits >> 32);
  md->W[15] = (int)(bits & 0xffffffff);

  md->count += 8;
  computeBlock(md);

  a = md->AA;
  b = md->BB;
  c = md->CC;
  d = md->DD;
  e = md->EE;

  // Copy out the result
  hashvalue[offset + 0] = (unsigned char)(a >> 24);
  hashvalue[offset + 1] = (unsigned char)(a >> 16);
  hashvalue[offset + 2] = (unsigned char)(a >> 8);
  hashvalue[offset + 3] = (unsigned char)(a >> 0);
    
  hashvalue[offset + 4] = (unsigned char)(b >> 24);
  hashvalue[offset + 5] = (unsigned char)(b >> 16);
  hashvalue[offset + 6] = (unsigned char)(b >> 8);
  hashvalue[offset + 7] = (unsigned char)(b >> 0);
    
  hashvalue[offset + 8] = (unsigned char)(c >> 24);
  hashvalue[offset + 9] = (unsigned char)(c >> 16);
  hashvalue[offset + 10] = (unsigned char)(c >> 8);
  hashvalue[offset + 11] = (unsigned char)(c >> 0);
    
  hashvalue[offset + 12] = (unsigned char)(d >> 24);
  hashvalue[offset + 13] = (unsigned char)(d >> 16);
  hashvalue[offset + 14] = (unsigned char)(d >> 8);
  hashvalue[offset + 15] = (unsigned char)(d >> 0);
    
  hashvalue[offset + 16] = (unsigned char)(e >> 24);
  hashvalue[offset + 17] = (unsigned char)(e >> 16);
  hashvalue[offset + 18] = (unsigned char)(e >> 8);
  hashvalue[offset + 19] = (unsigned char)(e >> 0);

  sha1_init(md);		// remove the evidence

  return SHA1_LENGTH;
}


/**
 * Produce a <code>String</code> representation for the specified array of
 * <code>byte</code>s.  Print each <code>byte</code> as two hexadecimal
 * digits.
 *
 * @param data	The array to print
 * @param offset 	the start offset in <code>data</code>
 * @param length	the number of <code>byte</code>s to print
 */
void print_bytes (jbyte* data, int offset, int length)
{
  unsigned char buf[2*length+1]; 
  int low_mask = 0x0f;
  int high_mask = 0xf0;
  int buf_pos = 0;
  int i;
  jbyte b;
  buf[2*length] = '\0';
  
  for (i=offset; i < (offset+length); ++i) {
    b = data[i];
    buf[buf_pos++] = digits[(high_mask & b) >> 4];
    buf[buf_pos++] = digits[(low_mask & b)];
  }
  printf("%s", buf);
}

JNIEXPORT void JNICALL Java_ostore_util_SHA1NativeIF_engineUpdate__B(
	     JNIEnv *env         /* in  */,
	     jobject not_used    /* in  */,
	     jbyte input         /* in  */)
{
  struct sha1_state* md;
  jintArray md_state;
  jint* md_state2;
  jclass cls = (*env)->GetObjectClass(env, not_used);
  jfieldID fid = 
        (*env)->GetFieldID(env, cls, STATE_MEMBER_NAME, INT_ARRAY_SYMBOL);
  if(fid == 0)
  {
    printf("\ndamnn!!\n");
  }

  md_state = (*env)->GetObjectField(env, not_used, fid);
  md_state2 = (*env)->GetIntArrayElements(env, md_state, 0);
  md = (struct sha1_state*)md_state2;

  update1((unsigned char)input, md);

  (*env)->ReleaseIntArrayElements(env, md_state, md_state2, 0);
}

JNIEXPORT void JNICALL Java_ostore_util_SHA1NativeIF_engineUpdate___3BII(
	     JNIEnv *env         /* in  */,
	     jobject not_used    /* in  */,
	     jbyteArray input    /* in  */,
	     jint off            /* in  */,
	     jint len            /* in  */)
{
  const jbyte *input_local = (*env)->GetByteArrayElements(env, input, 0);

  struct sha1_state* md;
  jintArray md_state;
  jint* md_state2;
  jclass cls = (*env)->GetObjectClass(env, not_used);
  jfieldID fid = 
        (*env)->GetFieldID(env, cls, STATE_MEMBER_NAME, INT_ARRAY_SYMBOL);
  if(fid == 0)
  {
    printf("\ndamnn!!\n");
  }

  md_state = (*env)->GetObjectField(env, not_used, fid);
  md_state2 = (*env)->GetIntArrayElements(env, md_state, 0);
  md = (struct sha1_state*)md_state2;

  update2((const unsigned char*)(input_local/*+off*/), off, len, md);
 
  (*env)->ReleaseIntArrayElements(env, md_state, md_state2, 0);
  (*env)->ReleaseByteArrayElements(env, input, (jbyte*)input_local, JNI_ABORT);
  
}

JNIEXPORT jint JNICALL Java_ostore_util_SHA1NativeIF_engineDigest___3BII(
	     JNIEnv *env         /* in  */,
	     jobject not_used    /* in  */,
	     jbyteArray buf      /* out */,
	     jint off            /* in  */,
	     jint len            /* in  */)
{
  jbyte* output_local = (*env)->GetByteArrayElements(env, buf, 0);
  int rv = -1;

  struct sha1_state* md;
  jintArray md_state;
  jint* md_state2;
  jclass cls = (*env)->GetObjectClass(env, not_used);
  jfieldID fid = 
        (*env)->GetFieldID(env, cls, STATE_MEMBER_NAME, INT_ARRAY_SYMBOL);
  if(fid == 0)
  {
    printf("\ndamnn!!\n");
  }

  md_state = (*env)->GetObjectField(env, not_used, fid);
  md_state2 = (*env)->GetIntArrayElements(env, md_state, 0);
  md = (struct sha1_state*)md_state2;

  rv = digest((unsigned char*)output_local, (int)off, (int)len, md);

  (*env)->ReleaseIntArrayElements(env, md_state, md_state2, 0);
  (*env)->ReleaseByteArrayElements(env, buf, output_local, 0/*copyback*/);

  return (jint)rv;
}

JNIEXPORT jint JNICALL Java_ostore_util_SHA1NativeIF_engineUpdateAndDigest(
	     JNIEnv *env         /* in  */,
	     jobject not_used    /* in  */,
	     jbyteArray buf      /* out */,
	     jint out_off        /* in  */,
             jint out_len        /* in  */,
	     jbyteArray input    /* in  */,
	     jint in_off         /* in  */,
	     jint in_len         /* in  */)
{
  const jbyte *input_local = (*env)->GetByteArrayElements(env, input, 0);
  jbyte* output_local = (*env)->GetByteArrayElements(env, buf, 0);

  int rv = -1;

  struct sha1_state md;
  sha1_init(&md);
  update2((const unsigned char*)input_local, (int)in_off, (int)in_len, &md);
  rv = digest((unsigned char*)output_local, (int)out_off, (int)out_len, &md);

  /*
  Java_ostore_util_SHA1NativeIF_engineUpdate___3BII(env, not_used, input, in_off, in_len, md);
  Java_ostore_util_SHA1NativeIF_engineDigest___3BII(env, not_used, buf, out_off, out_len, md);
  */

  (*env)->ReleaseByteArrayElements(env, input, (jbyte*)input_local, JNI_ABORT);
  (*env)->ReleaseByteArrayElements(env, buf, output_local, 0/*copyback*/);
  
  return (jint)rv;
}

JNIEXPORT void JNICALL Java_ostore_util_SHA1NativeIF_engineReset(
	     JNIEnv *env         /* in  */,
	     jobject not_used    /* in  */)
{
  /*reset();*/
}

JNIEXPORT void JNICALL Java_ostore_util_SHA1NativeIF_engineInit(
	     JNIEnv *env         /* in  */,
	     jobject not_used    /* in  */)
{
  struct sha1_state* md;
  

  jintArray md_state;
  jint* md_state2;
  jclass cls = (*env)->GetObjectClass(env, not_used);
  jfieldID fid = 
        (*env)->GetFieldID(env, cls, STATE_MEMBER_NAME, INT_ARRAY_SYMBOL);
  int length = -1;

  if(fid == 0)
  {
    printf("\nostore_util_sha1.engineInit: ERROR: damnn!!\n");
    fflush(NULL);
    exit(1);
  }

  length = sizeof(struct sha1_state);
  if(length % 4 != 0)
  {
    printf("\nostore_util_sha1.engineInit: ERROR: length==%d % 4 != 0\n", 
	   length, length%4);
    fflush(NULL);
    exit(1);
  }

  length = length/4;
  md_state = (*env)->NewIntArray(env, length);

  md_state2 = (*env)->GetIntArrayElements(env, md_state, 0);
  md = (struct sha1_state*)md_state2;
  sha1_init(md);

  (*env)->ReleaseIntArrayElements(env, md_state, md_state2, 0);

  (*env)->SetObjectField(env, not_used, fid, md_state);

}

JNIEXPORT jlong JNICALL Java_ostore_util_SHA1NativeIF_testTput(
	     JNIEnv *env         /* in  */,
	     jobject not_used    /* in  */,
	     jint arrSz          /* in  */,
	     jint iterations     /* in  */) 
{

  // first create the data set
  int length = 4096;
  int num_bytes = 20;
  char** data = (char **) malloc( length*sizeof(char*) );
  struct sha1_state* md;
  jintArray md_state;
  jint* md_state2;
  int i,j;
  struct timeval start_time, end_time;
  char *output = (char *) malloc( 20*sizeof(char) );
  jclass cls = (*env)->GetObjectClass(env, not_used);
  jfieldID fid = 
        (*env)->GetFieldID(env, cls, STATE_MEMBER_NAME, INT_ARRAY_SYMBOL);
  if(fid == 0)
  {
    printf("\ndamnn!!\n");
  }

  md_state = (*env)->GetObjectField(env, not_used, fid);
  md_state2 = (*env)->GetIntArrayElements(env, md_state, 0);
  md = (struct sha1_state*)md_state2;

  // seed the randomizer
  srand( time(0) );

  // populate data set
  for( i = 0; i < length; i++ ) {
    data[i] = (char *) malloc( arrSz*sizeof(char) );
    for( j = 0; j < arrSz; j++ ) {
      data[i][j] = (char) (rand()%256);
    }
  }

  gettimeofday(&start_time,0);

  // perform the iterations
  for( i = 0; i < iterations; i++ ) {

    update2(data[i%length],0,arrSz,md);
    digest( output, 0, num_bytes, md );
    // no reset?

  }

  gettimeofday(&end_time,0);

  // free all the memory
  for( i = 0; i < length; i++ ) {
    free(data[i]);
  }
  free(data);

  return (jlong)((float)((end_time.tv_sec-start_time.tv_sec)*1000.0)+
	  (float)(end_time.tv_usec-start_time.tv_usec)/1000.0);

}
	     

