#!/usr/bin/perl -w
#
# Copyright (c) 2002 Regents of the University of California.
# All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#  3. Neither the name of the University nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
#  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
#  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
#  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#  SUCH DAMAGE.

#
# author: Dennis Geels
# $Id: make-release.pl,v 1.6 2003/02/24 01:06:49 geels Exp $
#
# Creates a release package.

use strict;
use File::Path 'rmtree';
use Getopt::Long;
use Cwd ('cwd', 'chdir');
$| = 1;	# command buffering only

####################
# Variables:
my $base_dir = &cwd();
my $seda_jar = "seda.jar";
my $seda_util = "libMDWUtil.so";
my $seda_nbio = "libNBIO.so";
my $seda_dir = "seda";
my @seda_libs = ($seda_jar, $seda_util, $seda_nbio);
my $pond_dir = "pond";
my $ostore_dir = "$pond_dir/ostore";
my $soss_dir = "$pond_dir/soss";
my $ostorelib_dir = "$pond_dir/lib";
my $ostore_jar = "ostore.jar";
my $package_dir = "oceanstore";
my $lib_dir = "$package_dir/lib";
my $run_dir = "$package_dir/run";
my $run_files = "RUNFILES.txt";
my $docs_dir = "$package_dir/docs";
# following libraries assumed to be in CVS pond/lib/
my @cryptix_libs = ("cryptix-jce-api.jar", "cryptix-jce-provider.jar");
my @db_libs = ("db-4.0.14.jar", "libdb_java-4.0.so", "libdb-4.0.so");
my @trash_list;

####################
# Phase 1: Parse options:

sub usage {
    print<<END_USAGE;
    
USAGE:	make-release.pl [-options]

OPTIONS:
	
	--help
		Print this message
	
	--with-java-tarball <tarball>
	-jt	    
		Include the specified Java JDK/JRE tarball
	
	--with-java-dir <directory>
	-jd	    
		Include the specified Java JDK/JRE directory

	--date <CVS date specification>
		Download CVS repositories using the following 	    
		    
	--pond-srcdir <directory>
		Build the OceanStore jar in the specifying existing
		directory, instead of downloading a fresh repository

	--seda-srcdir <directory>
		Build the SEDA jar in the specifying existing
		directory, instead of downloading a fresh repository

	--seda-libdir <directory>
		Copy the SEDA jar and JNI libraries from the specified
		directory, instead of building them

EXAMPLES:
	make-release.pl --with-java-tarball \
	    /usr/local/IBMJava2-13/IBMJava2-SDK-13.tgz

	make-release.pl -d "11:59 PST"

	make-release.pl --seda-srcdir=~/seda --pond-srcdir=..

END_USAGE
}

my %options;
&GetOptions( \%options, "help|?", "date=s", "pond-srcdir=s",
	     "seda-srcdir=s", "seda-libdir=s",
	     "with-java-tarball|jt=s", "with-java-dir|jd=s" )
    or &usage and exit( -1 );

if( $options{"help"} ) {
    &usage and exit( 0 );
}

#FIXME: jd, pond_dir options, USAGE, copy libs properly.

my $cvs_date = $options{"date"};
if( !$cvs_date ) {
    $cvs_date = "now";
}

####################
# Phase 2: build SEDA
&print_phase( "Building SEDA Libraries" );

my $exit_code;
if( $options{"seda-libdir"} ) {
    &print_step( "copying libs from " . $options{"seda-libdir"} );
    foreach (@seda_libs) {
	$exit_code = system( "cp", $options{"seda-libdir"}."/$_", "." );
	&check( $exit_code,
		"Could not copy ". $options{"seda-libdir"}."/$_" );
	push @trash_list, $_;
    }
    goto PHASE_2;	# WARNING! Goto is tricky (especially in perl)
}

########## Fetch Source
if( $options{"seda-srcdir"} ) {
    ########## Use existing directory    
    $seda_dir = $options{"seda-srcdir"};
    if( $seda_dir !~ m%^/% ) {	# relative pathname
	$seda_dir = "$base_dir/$seda_dir";
    } 
    &print_step( "using source directory: $seda_dir" );

} else {
    ########## Download new copy into ./seda
    ########## First clear out any current ./seda directory.
    if( -d $seda_dir ) {
	die( "$seda_dir is in the way.  Move it.\n" );
    }
    &print_step( "downloading" );
    $exit_code =
	system( "cvs", "-Q", "-d",
		':pserver:anonymous@cvs.seda.sourceforge.net:/cvsroot/seda',
		"export", "-D", $cvs_date, "-d", $seda_dir, "seda");
    &check( $exit_code, "Could not download SEDA code" );
    # Convert to absolute pathname:
    $seda_dir = "$base_dir/$seda_dir";
    push @trash_list, $seda_dir;
}

########## Compile
&print_step( "compiling" );

$ENV{"CLASSPATH"} .=  ":$seda_dir/src:.";
if( ! -d "$seda_dir/lib" ) {
    mkdir( "$seda_dir/lib", 0777 ) or die "Could not create $seda_dir/lib";
}
&chdir( "$seda_dir/src/seda" ) or die "Could not change directories\n";

$exit_code = system( "make >& $base_dir/make.seda.out" );
&check( $exit_code, "Could not compile SEDA libraries (see make.seda.out)" );
push @trash_list, "make.seda.out";

########## JAR
&print_step( "building jar" );

&chdir( ".." ) or die "Could not change directories\n";
$exit_code = system( 'jar cf ' . "$base_dir/$seda_jar"
		     . ' `find . -name "*class"`' );
&check( $exit_code, "Could not build SEDA jar" );
push @trash_list, $seda_jar;
&chdir( $base_dir ) or die "Could not change directories\n";

########## JNI libs
&print_step( "copying JNI libs from $seda_dir/lib" );
foreach ($seda_util, $seda_nbio) {
    $exit_code = system( "cp", "$seda_dir/lib/$_", "." );
    &check( $exit_code, "Could not copy $seda_dir/lib/$_" );
    push @trash_list, $_;
}

####################
# Phase 2: build OceanStore
  PHASE_2:
&print_phase( "Building OceanStore Libraries" );

if( $options{"pond-srcdir"} ) {
    ########## Use existing directory
    $pond_dir = $options{"pond-srcdir"};
    if( $pond_dir !~ m%^/% ) {	# relative pathname
	$pond_dir = "$base_dir/$pond_dir";
    } 
    &print_step( "using source directory: $pond_dir" );
    # reassign subdirectories also.
    $ostore_dir = "$pond_dir/ostore";
    $soss_dir = "$pond_dir/soss";
    $ostorelib_dir = "$pond_dir/lib";

} else {
    ########## Download new copy into ./pond
    ########## First clear out any current ./pond directory.
    if( -d $pond_dir ) {
	die( "$pond_dir is in the way.  Move it.\n" );	
    }
    
    ##########    
    &print_step( "downloading" );
    my $module;
    foreach $module ($ostore_dir, $soss_dir, $ostorelib_dir) {
	$exit_code =
	    system( "cvs", "-Q", "export", "-D", "$cvs_date", $module );
	&check( $exit_code, "Could not download OceanStore code" );
    }
    push @trash_list, $pond_dir;
    
    # Make SEDA librairies available to OceanStore compile:
    $exit_code = system( "cp", @seda_libs, $ostorelib_dir );
    &check( $exit_code, "Could not move SEDA libs" );
}

##########
&print_step( "compiling" );

&chdir( $ostore_dir ) or die "Could not change directories\n";
$exit_code = system( "make >& $base_dir/make.ostore.out" );
&check( $exit_code,
	"Could not compile OceanStore libraries (see make.ostore.out)" );
push @trash_list, "make.ostore.out";

##########
&print_step( "building jar" );

&chdir( ".." ) or die "Could not change directories\n";
$exit_code = system( 'jar cf ' . "$base_dir/$ostore_jar" .
		     ' `find . -name "*class"`' );
&check( $exit_code, "Could not build OceanStore jar" );
push @trash_list, $ostore_jar;
&chdir( $base_dir ) or die "Could not change directories\n";

####################
# Phase 3: copy over rest of package
&print_phase( "Assembling Package" );

########## Create oceanstore/ directory
# First clear out any current ./oceanstore directory.
if( -d $package_dir ) {
    die( "$package_dir is in the way.  Move it.\n" );
}
mkdir( $package_dir, 0777 ) or die "Could not create $package_dir";
push @trash_list, $package_dir;

##########
&print_step( "Copying libraries" );
mkdir( $lib_dir, 0777 ) or die "Could not create $lib_dir";
my @libs = (@seda_libs, $ostore_jar,
	    map( "$ostorelib_dir/$_", (@cryptix_libs, @db_libs) ),
	    glob( "$ostorelib_dir/libostore*" ) );
# Include Tapas?
my $tapas_jar = "$ostorelib_dir/tapas.jar";
push( @libs, $tapas_jar ) if( -f $tapas_jar );

$exit_code = system( "cp", @libs, $lib_dir );
&check( $exit_code, "Copy failed" );

##########
&print_step( "Copying RUNFILES" );

mkdir( $run_dir, 0777 ) or die "Could not create $run_dir";
open( RUNFILES, "<$run_files" ) or die "Could not open $run_files";
my @lines = <RUNFILES>;
close( RUNFILES );
foreach (@lines) {
    s/#.*//;	# remove comments
    next unless /\w/;	# skip empty lines
    chomp;
    my $src = "../run/$_";
    my $dest = "$run_dir/$_";
    if( -d $src ) {
	mkdir( $dest, 0777 ) or die "Could not create $dest";
    } else {
	$exit_code = system( "cp", $src, $dest );
	&check( $exit_code, "Could not copy $src" );
    }
}

##########
&print_step( "Copying docs" );
mkdir( $docs_dir, 0777 ) or die "Could not create $docs_dir";
$exit_code = system( "cp", "docs/README.licenses",
		     glob("docs/LICENSE.*"), $docs_dir );
&check( $exit_code, "Copy failed" );

##########
&print_step( "Copying README" );
$exit_code = system( "cp", "README", $package_dir );
&check( $exit_code, "Copy failed" );

########## Include Java?
if( $options{"with-java-tarball"} ) {
    my $tarball = $options{"with-java-tarball"};
    if( $tarball !~ m%^/% ) {	# relative pathname
	$tarball = "$base_dir/$tarball";
    } 
    &print_step( "Copying JRE" );
    &chdir( $package_dir ) or die "Could not change directories\n";    
    $exit_code = system( "tar", "xzf", $tarball );
    &check( $exit_code, "Could not untar JRE" );
    &chdir( $base_dir ) or die "Could not change directories\n";
} elsif( $options{"with-java-dir"} ) {
    &print_step( "Copying JRE" );
    $exit_code = system( "cp", "-R", $options{"with-java-dir"}, $package_dir );
    &check( $exit_code, "Could not copy JRE" );
};

##########
&print_step( "Compressing Package" );
$exit_code = system( "tar", "czf", "$package_dir.tgz", $package_dir );
&check( $exit_code, "tar failed" );

####################
# Phase 4: Cleanup
&print_phase( "Cleaning Up" );

foreach (@trash_list) {
    if( -d ) {
#	print "Removing directory: $_\n";
	&rmtree( $_ ) or die "Could not remove $_";
    } elsif( -f ) {
#	print "Removing file: $_\n";
	unlink( $_ ) or die "Could not remove $_";
    }
}
&print_phase( "DONE" );

########################################
########################################

# Throw a big banner out onto STDOUT.    
sub print_phase {
    print "\n\n";
    print "*"x60;
    print "\n* @_ \n";
    print "*"x60;
    print "\n\n";
}

# Throw a small banner out onto STDOUT.    
sub print_step {
    print "\t @_ \n";
}

# Print a slightly useful error message
sub check {
    my ($exit_code, $error_str) = @_;
    if( $exit_code != 0 ) {	# system exit codes are inverted
	warn "\t $error_str \n";
	warn "\t process exited with value " . ($exit_code >> 8) .
	    ", signal " . ($exit_code & 255) . "\n";
	exit( -1 );
    }
}
    
BEGIN {
    # Find the 
    my $java_home;
    if( -d "/opt/IBMJava2-13" ) {
	$java_home = "/opt/IBMJava2-13";
    } elsif( -d "/usr/local/IBMJava2-13" ) {
	$java_home = "/usr/local/IBMJava2-13";
    } elsif( -d "/usr/local/jdk" ) {
	$java_home = "/usr/local/jdk";
    } else {
	die "Could not find a valid JDK";
    }
    $ENV{"PATH"} = "$java_home/bin:" . $ENV{PATH};
    if( defined $ENV{"CLASSPATH"} ) {
	$ENV{"CLASSPATH"} .= ":$java_home/jre/lib/rt.jar";
    } else {
	$ENV{"CLASSPATH"} = "$java_home/jre/lib/rt.jar";
    }

    # BUG: SEDA breaks jikes.
#      # Try to use jikes even for SEDA.
#      my $jikes = `which jikes`;
#      if( defined $jikes ) {
#  	chomp $jikes;
#  	if( -x $jikes ) {
#  	    $ENV{"JAVAC"} = $jikes;
#  	    print "Using $jikes\n";
#  	}
#      }
}
