#!/usr/bin/perl -w
#
# Copyright (c) 2001 Regents of the University of California.
# All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#  3. Neither the name of the University nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
#  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
#  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
#  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#  SUCH DAMAGE.

#
# $Id: cfgcat.pl,v 1.10 2003/02/10 19:39:28 geels Exp $
#
# Concatinates several config files desiged to run on the same machine into
# one.  Allows for running multiple "virtual" OceanStore servers on a
# single physical machine.
#
# The header from the first file becomes the header for the unified
# configuration file.
# Depending on the value of $use_multiplex_network or the --multiplexed option, 
# This script removes all ".*Network.*" stages and appends a single
# "YANetwork" stage that will serve all virtual nodes.

use strict;
use Getopt::Long;
$Getopt::Long::order = $REQUIRE_ORDER;	# Stop processing at filenames.

# Replace all Network stages with a single YANetwork stage
my $use_multiplex_network = 1;	# on by default
# Assume all virtual nodes share a single YANetwork stage
my $single_network_stage_hack = "false";
# Retry failed connections forever
my $infinite_connect_retry = "false";

my $connection_cache_size = 256;
# listening socket allows 511 incoming connections, so the connection
# cache size is not a hard limit.  It may overfill temporarily.

sub usage {
    print <<"EOUsage";
	USAGE: cfgcat [options] <config 1> ... <config N>

	OPTIONS:
		--help			- prints this message.
		--(no)multiplexed	- (do not) use YANetwork stage.
		--single_net		- assume single YANetwork stage per
					  machine.  Implies --multiplexed.
		--retry			- retry failed connections forever.

EOUsage
}

# First parse options
my %options;
&GetOptions( \%options, "help|?", "multiplexed!", "single_net", "retry" )
    or &usage and exit( 1 );

if( $options{"help"} ) {
    &usage and exit( 0 );
}
if( $options{"single_net"} ) {
    $single_network_stage_hack = "true";
    # implies --multiplexed.
    $options{"multiplexed"} = 1;
}
if( defined $options{"multiplexed"} ) {
    $use_multiplex_network = $options{"multiplexed"};
}
if( $options{"retry"} ) {
    $infinite_connect_retry = "true";
}

my @filenames = @ARGV;
my $filename;
my @nodeids;
my $first_file = 1;	# "true" for first file only.


# Loop across the configuration files, extracting and renaming the stages.
foreach $filename (@filenames) {
    my $thisnodeid;

    open (IN, "<$filename");

    # Scan the header for the "node_id" field
    while (<IN>) {
	# find the node id
	if (m/^[^\#]*node_id\s+(\S+)$/) {
	    $thisnodeid = $1;
	    push @nodeids, $thisnodeid;
	}

	# print the header of the first file only
	if( $first_file ) {
	    print; 
	}

	# The header ends when the stages begin.
	if (m/^[^\#]*<stages>/) {
	    last;
	}
    }
    $first_file = 0;	# set to false.

    # Now loop across the stage entries in the body of the file.
    STAGE_LOOP: while (1) {
	my $stage;
	my $done_with_body = 0;	# false

	if( ! defined( $thisnodeid ) ) {
	    die "ERROR: no global \"node_id\" field in $filename\n";
	}
	
        # phase 1:  find start of each stage
	while (<IN>) {
	    if (m/^([^\#]*)<([^\/]+)>(.*)$/) {
		# The start of a stage entry; append the NodeId to the name
		$stage = $2;

		if( $use_multiplex_network && $stage =~ m/Network/ ) {
		    # Skip to the end of this stage.
		    while( <IN> ) {
			if (m/^([^\#]*)<\/$stage>(.*)$/) {
			    next STAGE_LOOP;
			}			    
		    }
		} else {
		    print "$1<$stage-$thisnodeid>$3\n";
		    last;	# goto phase 2.
		}

	    } elsif (m/^[^\#]*<\/stages>/) {
		# No more stages in this file.
		last STAGE_LOOP;
	    } else {
		# Just noise between stages.  Might as well save it.
		print;
	    }
	}
	if (! defined( $stage ) ) {
	    # EOF.
	    last STAGE_LOOP;
	}

	# phase 2: add local "node_id" definition.
	while (<IN>) {
	    if (m/^([^\#]*)<initargs>/) {
		# insert NodeId
		print; 
		print "\t\tnode_id $thisnodeid\n";
	    } elsif (m/^([^\#]*)<\/$stage>(.*)$/) {
		# append the NodeId to the stage name
		print "$1</$stage-$thisnodeid>$2\n";
		next STAGE_LOOP;  # done with phase 2.
	    } else {
		# Other stage content.  Just copy.
		print;
	    }
	}
	# Shouldn't reach end of loop (EOF) in phase 2.
	die "ERROR: unending stage definition: $stage\n" ;
    }
    close (IN);
}

# Finally, add YANetwork stage and configuration file footer.
if ($use_multiplex_network)
{
  my ($i, @nodeid_defs);
  for ($i = 0; $i < @nodeids; ++$i) {
    push @nodeid_defs, "\t\tnode_id_$i $nodeids[$i]\n";
  }

  print <<HERE;

	<YANetwork>
    	    class ostore.network.YANetwork
    	    queueThreshold 100000
    	    <initargs>
		max_conn $connection_cache_size
		min_free_frac 0.1
		single_network_hack $single_network_stage_hack
		infinite_connect_retry $infinite_connect_retry
		node_id_count $i
@nodeid_defs
    	    </initargs>
    	</YANetwork>
HERE
}

#  Close out the config file.
print <<HERE;
    </stages>
</sandstorm>
HERE
