#!/usr/bin/perl -w
#
# Copyright (c) 2001 Regents of the University of California.
# All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#  3. Neither the name of the University nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
#  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
#  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
#  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#  SUCH DAMAGE.

#
# author: Dennis Geels
# $Id: fetch-planetlab-list.pl,v 1.13 2003/03/13 21:04:31 strib Exp $
#
# Fetches and parses the list of PlanetLab nodes.
# Then prints out a command that will export the list into a sh environment

use strict;
use Getopt::Long;

my $randomized;
my $use_blacklist = 1;
my $app_file = "x-pl.app";
my $use_app_file = 1;
&GetOptions( "randomized:i" => \$randomized,
	     "blacklist!" => \$use_blacklist,
	     "checkappfile!" => \$use_app_file ) or &usage and exit( -1 );

sub usage {
    print <<'EOUsage';
	USAGE: fetch-planetlab-list.pl [options]
	OPTIONS:
		--randomized [seed]	Return the hosts in a random order.
		--noblacklist		Do not blacklist hosts found in infospect.
		--nocheckappfile	Do not check $app_file.
EOUsage
}

# Downloads the file named by the specified URL.  Returns the name.
sub fetch {
    my $url = shift;
    my $local_file;
    if( $url =~ m%/([^/]+)$% ) {
	$local_file = $1;
    } else {
	die( "could not parse filename out of $url\n" );
    }
    my @wget_cmd = ("wget", "-nv");
    
    my $exit_code = system( @wget_cmd, $url );
    if( $exit_code != 0 ) {
	die( "ERROR: @wget_cmd failed" );
    }
    return $local_file;
}

##############################
# First get total list.

#my $url = "www.planet-lab.org/xml/sites.xml";
#my $url = "www.planet-lab.org/db/scripts/data/nodes.txt";
#my $url = "www.planet-lab.org/db/nodes/production_hosts.php";
my $url = "www.planet-lab.org/db/nodes/production_ips.php";
my $local_file = &fetch( $url );

open( NODES, "<$local_file" ) or die( "ERROR: could not open $local_file" );
my @nodes = <NODES>;
close( NODES ) or die( "ERROR: since when does close fail?" );
unlink $local_file or die( "ERROR: could not remove $local_file" );

##############################
# Now get the blacklist from Infospect

# Maps a name to a badness counter.  Badness > 0 means bad.
my %bad_nodes;
if( $use_blacklist ) {
    $url = "infospect.planet-lab.org/index.php";
    $local_file = &fetch( $url );
    
    open( ISSUES, "<$local_file" ) or die( "ERROR: could not open $local_file" );
    my @bad_lines = <ISSUES>;
    close( ISSUES ) or die( "ERROR: since when does close fail?" );
    unlink $local_file or die( "ERROR: could not remove $local_file" );

    my $node;
    foreach (@bad_lines) {
	if( /(\d+\.\d+\.\d+\.\d+)/ ) {
	    $node = $1;
#	print "blacklisted: $1\n";
	    $bad_nodes{$node}++;	# increment badness count
	} elsif( defined $node ) {
	    # Check subsequent report lines.
	    if( /web server fails/ ) {
		# Decrement the badness count; we don't care about the web server.
		$bad_nodes{$node}--;
	    } elsif( /respond to finger/ ) {
		# Decrement the badness count; we don't care about fingerd.
		$bad_nodes{$node}--;
	    } elsif( /node/ ) {
		# Some other problem. Increment count a lot.
		$bad_nodes{$node} += 10;
	    } elsif( /<h/ ) {
		undef $node;	# end of node list.
	    }
	}
    }
}

##############################
# Add a personal blacklist.

if( $ENV{"OSTORE_BLACKLIST"} ) {
    my @bad_nodes = split /\s+/, $ENV{"OSTORE_BLACKLIST"};
    foreach my $bad_node (@bad_nodes) {
	$bad_nodes{$bad_node} = 1;
    }
}
#print join( ' ', keys %bad_nodes) . "\n";

##############################
# Blacklist nodes that are missing from the local all-pairs-pings file
my %app_nodes;

if( $use_app_file ) {
    open( APP, "<$app_file" ) or die;
    my $ignore = <APP>;
    $ignore = <APP>;
    my $host_line = <APP>;
    my @hosts = split /\s+/, $host_line;
    foreach (@hosts) {
#	print "app node: $_\n";
	$app_nodes{$_} = 1;
    }
    close( APP ) or die;
}

my @good_nodes;
for my $index ( 0 .. $#nodes ) {
    chomp $nodes[$index];
    my $node = $nodes[$index];
    next unless $node =~ /\w/;	# skip empty lines    
#    print "node: \"$node\"\n";
    if( (defined $bad_nodes{$node}) and ($bad_nodes{$node} > 0) ) {
	print "skipping bad node: $node\n";
    } elsif( $use_app_file and (! $app_nodes{$node}) ) {
	print "not in $app_file: $node\n";
    } else {
	push @good_nodes, $node;
    }
}

if( defined $randomized ) {
    my @randomized_nodes = ();
    my $seed = ($randomized or 0);
    srand( $seed );
    for( 1 .. @good_nodes ) {
	my $index = int rand @good_nodes;
	push @randomized_nodes, splice( @good_nodes, $index, 1 );
    }
    print( "export OSTORE_SERVERS=\"@randomized_nodes\"\n" );
} else {
    print( "export OSTORE_SERVERS=\"@good_nodes\"\n" );
}
