#!/usr/bin/perl -w

use strict;
use Getopt::Std;
use File::Copy;
use Tools;
use vars qw($opt_h $opt_a $opt_i $opt_c $opt_d $opt_u $opt_w $opt_l $opt_x $opt_k $opt_g);

sub usage
{
    select(STDERR);
    print <<'EOUsage';

run-tput-test.pl [options]
  Options:
    -h           - Display this help message.
    -a           - Enable the archive.  By default, the inner ring does
                   not initiate the inner ring.
    -i           - Inline the archive.  This option has no effect if the
                   archive is disabled.  By default, when the archive is
                   enabled, it runs asynchronously with the updates.
    -c MACHINES  - Use MACHINES machines to apply load.  If the number
                   of machines requested is greater than the requested
                   update load, the extra machines will remain idle
                   through the test.
    -d UPDT_LOAD - Generate UPDT_LOAD amount of load during the test.
                   Load is measured by the maximum number of outstanding
                   updates.  If specified as a range, the test will sweep
                   across the range described by UPDT_LOAD.  The format
                   for notating the range is described below.
    -u UPDT_SIZE - Sweep the size of the update submitted by test stages
                   across the range describe by UPDT_SIZE.  The format
                   for notating the range is described below.  Note, the
                   size of the data object is set to twice the length of
                   the update size.
    -w WARM_SIZE - Warm the caches and JIT by updating WARM_SIZE
                   megabytes of data before starting the test.  By
                   default, the test will warm the caches by updating
                   one megabyte of data.
    -l LENGTH    - Run each test for LENGTH seconds.
    -x ROC       - Relocate the test in the X-cluster to start using
                   machine xROC.millennium.berkeley.edu.
    -k KEY_SIZE  - Use keys of of length KEY_SIZE bits.  Valid values
                   are 512 and 1024.  The default value is 512.
    -g           - Include a histogram of the individual update
                   latencies in the result log.

  Ranges can be specified in a very flexible manner.  Most generally,
  each range has a 'start' value, an 'end' value, a 'skip operator',
  and a 'skip value'.  The skip operator is one of the four basic
  arithmetic operators (+, -, *, and /).  During execution, when the
  range operator is '+' or '*', the range leads to execution as
    for (x = $start; x <= $end; x = x $skip_op $skip_value) { ; }
  When the operator is '-' or '/', the range leads to execution as
    for (x = $start; x >= $end; x = x $skip_op $skip_value) { ; }

  A range is notated as '$start-$end/$skip_op$skip_value'.  Several
  shortened forms are also accepted.  These forms include '$start' which
  executes only one point, '$start-$end' which uses the '+' operator
  with a skip value of 1, and '$start-$end/$skip_value' which defaults
  to use the '+' operator.  Examples of valid ranges include

    o  28
    o  28-56
    o  2-10/2
    o  128-1024/*2
    o  2048-128//2
EOUsage

    exit(1);
}

#  Constants.
my $TPUT_RESULT_LOG;
my $MASTER_MACHINE_OFFSET = 5;
my $TESTER_MACHINE_OFFSET = 6;

#  Test parameters.
my $SKIP_ARCHIVE = 'true';
my $INLINE_ARCHIVE = 'false';
my $TEST_MACHINES = 1;
my $TEST_LENGTH = 10;
my $INIT_SIZE = 1048576;
my $CLUSTER_MACHINE = -1;
my $KEY_SIZE = 512;
my $HISTOGRAM = 'false';

my $INNER_CFG = "./inner-startrek.cfg";

my @UPDATE_LOAD_SPEC;
my $UPDATE_LOAD_START = 1;
my $UPDATE_LOAD_END   = 1;
my $UPDATE_LOAD_OP    = '+';
my $UPDATE_LOAD_SKIP  = 1;

my @UPDATE_SIZE_SPEC;
my $UPDATE_SIZE_START = 1024;
my $UPDATE_SIZE_END   = 1024;
my $UPDATE_SIZE_OP    = '+';
my $UPDATE_SIZE_SKIP  = 1;

#  Get the user-defined parameters.
my $opts_ok = getopts('haic:l:x:k:d:u:w:g');
if ($opt_h) { &usage; }
if ($opt_a) { $SKIP_ARCHIVE = 'false'; }
if ($opt_i) { $INLINE_ARCHIVE = 'true'; $SKIP_ARCHIVE = 'false'; }
if ($opt_c) { $TEST_MACHINES = $opt_c; }
if ($opt_l) { $TEST_LENGTH = $opt_l; }
if ($opt_x) { $CLUSTER_MACHINE = $opt_x; }
if ($opt_k) { $KEY_SIZE = $opt_k; }
if ($opt_w) { $INIT_SIZE *= $opt_w; }
if ($opt_g) { $HISTOGRAM = 'true'; }
if ($opt_d)
{
    @UPDATE_LOAD_SPEC = &parse_range_spec($opt_d);
    my @spec = @UPDATE_LOAD_SPEC;
    $UPDATE_LOAD_START = shift(@spec);
    $UPDATE_LOAD_END   = shift(@spec);
    $UPDATE_LOAD_OP    = shift(@spec);
    $UPDATE_LOAD_SKIP  = shift(@spec);
}
if ($opt_u)
{
    @UPDATE_SIZE_SPEC = &parse_range_spec($opt_u);
    my @spec = @UPDATE_SIZE_SPEC;
    $UPDATE_SIZE_START = shift(@spec);
    $UPDATE_SIZE_END   = shift(@spec);
    $UPDATE_SIZE_OP    = shift(@spec);
    $UPDATE_SIZE_SKIP  = shift(@spec);
}

#  The command line is valid if getopts returned without error and
#  there are no arguments left over.  $#ARGV returns the number of
#  remaining arguments minus one.
if ((! $opts_ok) || ($#ARGV != -1)) { &usage; }

#  Check that the set of input parameters is valid.
if ($KEY_SIZE != 512 && $KEY_SIZE != 1024)
{
    my $old_fd = select(STDERR);
    print<<EOErrMsg;
ERROR:  Invalid key size.
EOErrMsg
    select($old_fd);
    &usage();
    exit(1);
}
if ($CLUSTER_MACHINE == -1)
{
    my $old_fd = select(STDERR);
    print<<EOErrMsg;
ERROR:  Undefined cluster machine.
EOErrMsg
    select($old_fd);
    &usage();
    exit(1);
}

#  Compute the result log name.
$TPUT_RESULT_LOG = "/tmp/tput_result_log.noar.$$.log";
if (($SKIP_ARCHIVE eq 'false') && ($INLINE_ARCHIVE eq 'true'))
{
    $TPUT_RESULT_LOG = "/tmp/tput_result_log.inar.$$.log";
}
elsif ($SKIP_ARCHIVE eq 'false')
{
    $TPUT_RESULT_LOG = "/tmp/tput_result_log.dear.$$.log";
}

#  Set the environment variable for the keys.
my $key_cmd;
if ($KEY_SIZE == 512)
{
    $ENV{KEYSHARE_PREFIX} = "ir";
    $key_cmd = "KEYSHARE_PREFIX=ir";
}
elsif ($KEY_SIZE == 1024)
{
    $ENV{KEYSHARE_PREFIX} = "ir-1024-4-2";
    $key_cmd = "KEYSHARE_PREFIX=ir-1024-4-2";
}
else
{
    print(STDERR "Script error.");
}

#  Create the inner ring config files with the right archive settings.
copy($INNER_CFG, "/tmp/inner.cfg");
my %inner_vals =
  ("skip_archive"   => $SKIP_ARCHIVE,
   "inline_archive" => $INLINE_ARCHIVE,
  );
&override("/tmp/inner.cfg", %inner_vals);

#  Run experiments to collect data.
my $update_size = $UPDATE_SIZE_START;
my $update_size_comparator = '<=';
if ($UPDATE_SIZE_OP eq '-' || $UPDATE_SIZE_OP eq '/')
{
    $update_size_comparator = '>=';
}
while (eval("$update_size $update_size_comparator $UPDATE_SIZE_END"))
{
    my $dobj_size = $update_size * 2;
    if ($dobj_size > $update_size + 512 * 1024) {
	$dobj_size = $update_size + 512 * 1024;
    }

    my $update_load = $UPDATE_LOAD_START;
    my $update_load_comparator = '<=';
    if ($UPDATE_LOAD_OP eq '-' || $UPDATE_LOAD_OP eq '/')
    {
        $update_load_comparator = '>=';
    }
    while (eval("$update_load $update_load_comparator $UPDATE_LOAD_END"))
    {
        #  Create configuration files.
        my $master_cfg  =
          &write_master_config_file($TEST_MACHINES, $update_load, $dobj_size,
                                    $update_size, $TEST_LENGTH, $INIT_SIZE);
        my @tester_cfgs =
          &write_tester_config_files($TEST_MACHINES);
        my $exp_cfg     =
          &write_experiment_file($TEST_MACHINES);
        my $tput_fork_log = "/tmp/tput_fork_log.$$.log";

        #  Run the experiment.
        print("Running tput test: $key_cmd ./run-experiment $exp_cfg\n" .
              "  TestMachines = $TEST_MACHINES\n" .
              "  UpdateLoad   = $update_load\n" .
              "  DobjSize     = $dobj_size\n" .
              "  UpdateSize   = $update_size\n" .
              "  TestLength   = $TEST_LENGTH\n" .
              "  InitLength   = $INIT_SIZE\n" .
              "  KeySize      = $KEY_SIZE\n");

        my $pid = &fork_process($tput_fork_log,
                                "./run-experiment", $exp_cfg);
        if(! &returned_zero($pid))
        {
            die("Experiment returned non-zero.\n");
        }

        #  Read the result log.
        my $result_log_name = &get_result_log_name($pid);
        my $total_ops       = &analyze_result_log($result_log_name);

        #  Compute the results.
        my $ops_tput  = $total_ops / $TEST_LENGTH;
        my $data_tput = ($total_ops * $update_size) / $TEST_LENGTH;
        print("  OpsTput      = $ops_tput\n" .
              "  DataTput     = " . ($data_tput/1024/1024) ." MB/s\n");

        #  Write to log.
        &write_result_to_log($SKIP_ARCHIVE, $INLINE_ARCHIVE,
                             $TEST_MACHINES, $update_load, $TEST_LENGTH,
                             $KEY_SIZE, $dobj_size, $update_size,
                             $total_ops, $total_ops, $ops_tput, $data_tput);

        if ($HISTOGRAM eq 'true')
        {
            my $histogram = &read_latency_histogram($result_log_name);
            write_to_log("\n" . $histogram . "\n");
        }

        #  Remove the files we created.
        unlink($master_cfg, @tester_cfgs, $exp_cfg,
               $tput_fork_log, $result_log_name);

        $update_load = eval("$update_load $UPDATE_LOAD_OP $UPDATE_LOAD_SKIP");
    }

    $update_size = eval("$update_size $UPDATE_SIZE_OP $UPDATE_SIZE_SKIP");
}

#  Remove the inner ring configuration file.
unlink("/tmp/inner.cfg");

exit(0);


sub parse_range_spec
{
    my $spec = shift(@_);

    my $start = 0;
    my $end   = 0;
    my $op    = '+';
    my $skip  = 1;

    #  (\d+)                  The start of the range...
    #  (-(\d+))?              ...optionally a '-' and the end of the range...
    #  (\/([-\+\*\/])?(\d+))? ...optionally a '/', an optional
    #                         operation (-, +, *, or /) and the skip value.
    if ($spec =~ /(\d+)(-(\d+))?(\/([-\+\*\/])?(\d+))?/ )
    {
        $start = $1;
        $end   = $start;
        if (defined($3)) { $end = $3; }
        if (defined($5)) { $op = $5; }
        if (defined($6)) { $skip = $6; }
    }

    #print("start=$start \n");
    #print("end=$end   \n");
    #print("op=$op    \n");
    #print("skip=$skip  \n");

    if (($op eq '+') || ($op eq '*'))
    {
        if ($start > $end)
        {
          die("Invalid range specification.\n" .
              "  start=$start, end=$end, operation=$op\n");
        }
    }
    else
    {
        if ($end > $start)
        {
          die("Invalid range specification.\n" .
              "  start=$start, end=$end, operation=$op\n");
        }
    }

    my @range;
    push(@range, $start);
    push(@range, $end);
    push(@range, $op);
    push(@range, $skip);

    return @range;
}

sub write_result_to_log
{
    my $skip          = shift(@_);
    my $inline        = shift(@_);
    my $test_machines = shift(@_);
    my $update_load   = shift(@_);
    my $init_size     = shift(@_);
    my $test_length   = shift(@_);
    my $key_size      = shift(@_);
    my $dobj_size     = shift(@_);
    my $update_size   = shift(@_);
    my $total_ops     = shift(@_);
    my $ops_tput      = shift(@_);
    my $data_tput     = shift(@_);

    my $log_msg =
      "SkipArchive=${skip} " .
      "InlineArchive=${inline} " .
      "TestMachines=${test_machines} " .
      "UpdateLoad=${update_load} " .
      "TestLength=${test_length}s " .
      "InitSize=${init_size}MB " .
      "KeySize=${key_size} " .
      "DobjSize=${dobj_size} " .
      "UpdateSize=${update_size} " .
      "TotalOps=${total_ops} " .
      "UpdateTput=${ops_tput}ops/s " .
      "DataTput=${data_tput}bytes/s\n";

    write_to_log($log_msg);

    return;
}

sub write_to_log
{
    my $log_msg = shift(@_);

    open(RESULT_LOG, ">> $TPUT_RESULT_LOG") ||
      die("Could not open result log ($TPUT_RESULT_LOG).");
    print(RESULT_LOG "$log_msg");
    close(RESULT_LOG);

    return;
}

sub get_result_log_name
{
    my $pid = shift(@_);

    my @internal_logs = glob("/tmp/*-${pid}*-Master-internal.log");
    (scalar(@internal_logs) == 1) ||
      die("Cannot determine result log name.\n");


    my $log = shift(@internal_logs);
    return $log;
}

sub analyze_result_log
{
    my $log = shift(@_);

    open(RESULT_LOG, "$log") || die("Could not open result log.\n");
    my $total_ops = 0;
    while(<RESULT_LOG>)
    {
        if (/NumOps=(.*)/)
        {
            $total_ops += $1;
        }
    }

    close(RESULT_LOG);

    return $total_ops;
}

sub read_latency_histogram
{
    my $log = shift(@_);

    open(RESULT_LOG, "$log") || die("Could not open result log.\n");
    my $histogram = '';
    while(<RESULT_LOG>)
    {
        if (/TotalLatStats/ || /^\s+\d+\s+\d+$/)
        {
            $histogram .= $_;
        }
    }

    close(RESULT_LOG);

    return $histogram;
}

sub write_experiment_file
{
    my $test_machines = shift(@_);

    my $exp_cfg = "/tmp/tput.$$.exp";
    open(EXP, ">$exp_cfg") ||
      die("Could not create experiment configuration file ($exp_cfg).\n");
    my $old_fd = select(EXP);
    my $fed = $CLUSTER_MACHINE - 1;
    my $ir0 = $CLUSTER_MACHINE + 1 - 1;
    my $ir1 = $CLUSTER_MACHINE + 2 - 1;
    my $ir2 = $CLUSTER_MACHINE + 3 - 1;
    my $ir3 = $CLUSTER_MACHINE + 4 - 1;
    my $master = $CLUSTER_MACHINE + 5 - 1;
    print <<EOExp;
hosts      ROC
topology   NONE
Federation \${run_dir}/federation.cfg  fed     daemon $fed
RP         \${run_dir}/rp-startrek.cfg          static  daemon $fed
Ring0      /tmp/inner.cfg              static  daemon $ir0
Ring1      /tmp/inner.cfg              static  daemon $ir1
Ring2      /tmp/inner.cfg              static  daemon $ir2
Ring3      /tmp/inner.cfg              static  daemon $ir3
Master     /tmp/tput-master.$$.cfg     static  mortal $master
EOExp

    for (my $i = 0; $i < $test_machines; $i++)
    {
        my $tester_cfg = "/tmp/tput-tester-$i.$$.cfg";
        my $machine_num = $CLUSTER_MACHINE + $TESTER_MACHINE_OFFSET + $i - 1;
        print("Tester$i $tester_cfg static  daemon $machine_num\n");
    }
    close(EXP);
    select($old_fd);

    return $exp_cfg;
}

sub write_master_config_file
{
    my $test_stages = shift(@_);
    my $update_load = shift(@_);
    my $dobj_size   = shift(@_);
    my $update_size = shift(@_);
    my $test_length = shift(@_);
    my $init_size   = shift(@_);

    my $master_cfg = "/tmp/tput-master.$$.cfg";
    open(MASTER, ">$master_cfg") ||
      die("Could not create master configuration file ($master_cfg).\n");
    my $old_fd = select(MASTER);
    &write_master_header();
    &write_master_stage($test_stages, $update_load, $dobj_size,
                        $update_size, $test_length, $init_size);
    &write_master_footer();
    close(MASTER);
    select($old_fd);

    return $master_cfg;
}

sub write_tester_config_files
{
    my $test_machines = shift(@_);

    my $master_machine = $CLUSTER_MACHINE + $MASTER_MACHINE_OFFSET;
    #  The port number is an OceanStore magic number.  For now, we
    #  always start at port 3630.
    my $master = "x$master_machine.millennium.berkeley.edu:3630";

    my @tester_config_files;
    for (my $i = 0; $i < $test_machines; $i++)
    {
        my $tester_cfg = "/tmp/tput-tester-$i.$$.cfg";
        push(@tester_config_files, $tester_cfg);
        open(TESTER, ">$tester_cfg") ||
          die("Could not create tester configuration file ($tester_cfg).\n");

        my $old_fd = select(TESTER);

        &write_tester_header();
        &write_tester_stage($master);
        &write_tester_footer();

        close(TESTER);
        select($old_fd);
      }

    return @tester_config_files;
}

sub write_master_header
{
    print <<'EOHeader';
<sandstorm>

<global>
  defaultThreadManager TPSTM
  <TPSTM>
    governor false
    governorDelay 2000
    governorMaxThreads 10
  </TPSTM>

  <TPPTM>
    numCpus 4
    maxThreads 4
  </TPPTM>

  <profile>
    enable false
    delay 1000
    filename ${SSProfileFile}
    graphfilename ${SSGraphFile}
    graph false
    sockets false
  </profile>

  <initargs>
    node_id ${NodeID}
    profile false
  </initargs>
</global>

<stages>
  #  We need to have Network and Tapestry stages to generate the
  #  proper 'Ready' messages that the run-scripts look for.
  <Network>
    class ostore.network.Network
    queueThreshold 1000
    <initargs>
    </initargs>
  </Network>

  <Router>
    class ostore.tapestry.startrek.StartrekRouter
    queueThreshold 1000

    <initargs>
      pkey ${NodePublicKey}
      skey ${NodePrivateKey}
      dynamic_route ${TapestryMode}
    </initargs>
  </Router>

  <TClient>
    class ostore.tapestry.impl.StaticTClient
    queueThreshold 1000
    <initargs>
      federation ${FederationID}
      dynamic_route ${TapestryMode}
    </initargs>
  </TClient>

EOHeader
}

sub write_master_stage
{
    my $num_testers = shift(@_);
    my $update_load = shift(@_);
    my $dobj_size   = shift(@_);
    my $update_size = shift(@_);
    my $test_length = shift(@_);
    my $init_size   = shift(@_);

    print <<EOStage;
  <TputMicroBenchmarkMaster>
    class ostore.apps.test.TputMicroBenchmarkMaster
    queueThreshold 1000
    <initargs>
        NumTesters $num_testers
        UpdateLoad $update_load
        DobjSize   $dobj_size
        UpdateSize $update_size
        TestLength $test_length
        SettleTime 3000
        InitSize   $init_size
        ResultLog  \${InternalLog}
    </initargs>
  </TputMicroBenchmarkMaster>
EOStage
}

sub write_master_footer
{
    print <<'EOFooter';
</stages>

</sandstorm>
EOFooter
}

sub write_tester_header
{
    print <<'EOHeader';
<sandstorm>
  <global>
  defaultThreadManager TPSTM
    <TPSTM>
    governor false
    governorDelay 2000
    governorMaxThreads 10
    </TPSTM>

    <TPPTM>
    numCpus 4
    maxThreads 4
    </TPPTM>

    <profile>
      enable false
      delay 1000
      filename ${SSProfileFile}
      graphfilename ${SSGraphFile}
      graph false
      sockets false
    </profile>

    <initargs>
        node_id ${NodeID}
        Profile false
    </initargs>
  </global>

<stages>
  <Network>
    class ostore.network.Network
    queueThreshold 1000
    <initargs>
    </initargs>
  </Network>

  <Router>
    class ostore.tapestry.startrek.StartrekRouter
    queueThreshold 1000

    <initargs>
      pkey ${NodePublicKey}
      skey ${NodePrivateKey}
      dynamic_route ${TapestryMode}
    </initargs>
  </Router>

  <TClient>
    class ostore.tapestry.impl.StaticTClient
    queueThreshold 1000
    <initargs>
      federation ${FederationID}
      dynamic_route ${TapestryMode}
    </initargs>
  </TClient>

  <ClientStage>
    class ostore.client.ClientStage
    queueThreshold 1000
    <initargs>
      UseDynamicTapestry false
      ProvideApiServer   false
      LogEvents          false
      LogFilename        /tmp/log
      LogSampleRate      1
    </initargs>
  </ClientStage>

  <ReplicaStage>
    class ostore.replica.ReplicaStage
    queueThreshold 1000
    <initargs>
    </initargs>
  </ReplicaStage>

  <ReadStage>
    class ostore.read.ReadStage
    queueThreshold 1000
    <initargs>
    </initargs>
  </ReadStage>

  <DTree>
    class ostore.dtree.DTreeNode
    queueThreshold 1000
    <initargs>
    </initargs>
  </DTree>

  <RequestorStage>
    class ostore.archive.RequestorStage
    queueThreshold 1000
    <initargs>
      RequestCardinality        20
      CooperativeCaching        false
    </initargs>
  </RequestorStage>

  <Cache>
    class ostore.cache.MemoryCache
    queueThreshold 1000
    <initargs>
      CacheSize 314572800 # 300 MB
      cachedir ${CacheDir}
      Publish false
    </initargs>
  </Cache>

EOHeader
}

sub write_tester_stage
{
    my $master = shift(@_);

    print <<EOStage;
  <TputMicroBenchmark>
    class ostore.apps.test.TputMicroBenchmark
    queueThreshold 1000
    <initargs>
      Master             $master
      rp_pkeyfile        \${RPPublicKey}
      principal_pkeyfile \${PrincipalPublicKey}
      principal_skeyfile \${PrincipalPrivateKey}
    </initargs>
  </TputMicroBenchmark>
EOStage
}

sub write_tester_footer
{
    print <<EOFooter;
</stages>
</sandstorm>
EOFooter
}
