#!/usr/bin/perl -w

use strict;
use Tools;
use Getopt::Std;
#use File::Copy;
#use POSIX;
use vars qw($opt_h $opt_a $opt_s $opt_v $opt_f $opt_c $opt_d $opt_u
	    $opt_w $opt_l $opt_x $opt_k $opt_g $opt_n $opt_z $opt_t);

sub usage
{
    select(STDERR);
    print <<'EOUsage';

run-tput-test.pl [options]
  Options:
    -h           - Display this help message.
    -t           - indicate use of tapas, not tapestry
    -a           - Enable the archive.  By default, the inner ring does
                   not initiate the inner ring.
    -s           - Use Serializer node(s) instead of a set of Inner Ring.
		   Inherently the archive is on: disables -a option.
    -f           - Disseminate Fragments.  false by default.
    -n SERS      - Use SERS Serializers.  Distribute the clients evenly
                   over the Serializers.  Implies -s.  Default value is 1.
    -v           - Verbose output.  That is, save the internal log of 
                   the different archiver stages.
    -c MACHINES  - Use MACHINES machines to apply load.  If the number
                   of machines requested is greater than the requested
                   update load, the extra machines will remain idle
                   through the test.
    -d UPDT_LOAD - Generate UPDT_LOAD amount of load during the test.
                   Load is measured by the maximum number of outstanding
                   updates.  If specified as a range, the test will sweep
                   across the range described by UPDT_LOAD.  The format
                   for notating the range is described below.
    -u UPDT_SIZE - Sweep the size of the update submitted by test stages
                   across the range describe by UPDT_SIZE.  The format
                   for notating the range is described below.  Note, the
                   size of the data object is set to twice the length of
                   the update size.
    -w WARM_SIZE - Warm the caches and JIT by updating WARM_SIZE
                   megabytes of data before starting the test.  By
                   default, the test will warm the caches by updating
                   one megabyte of data.
    -l LENGTH    - Run each test for LENGTH seconds.
    -x ROC	 - Relocate the test in the X-cluster to start using
                   machine xROC.millennium.berkeley.edu.
		   NOTE: if no -x option is specified, the test will use
		   the machines listed in the OSTORE_SERVERS environment
		   variable.
    -k KEY_SIZE  - Use keys of of length KEY_SIZE bits.  Valid values
                   are 512 and 1024.  The default value is 512.
    -g           - Include a histogram of the individual update
                   latencies in the result log.
    -z		 - Include the Visdemo stage for each node.

  Ranges can be specified in a very flexible manner.  Most generally,
  each range has a 'start' value, an 'end' value, a 'skip operator',
  and a 'skip value'.  The skip operator is one of the four basic
  arithmetic operators (+, -, *, and /).  During execution, when the
  range operator is '+' or '*', the range leads to execution as
    for (x = $start; x <= $end; x = x $skip_op $skip_value) { ; }
  When the operator is '-' or '/', the range leads to execution as
    for (x = $start; x >= $end; x = x $skip_op $skip_value) { ; }

  A range is notated as '$start-$end/$skip_op$skip_value'.  Several
  shortened forms are also accepted.  These forms include '$start' which
  executes only one point, '$start-$end' which uses the '+' operator
  with a skip value of 1, and '$start-$end/$skip_value' which defaults
  to use the '+' operator.  Examples of valid ranges include

    o  28
    o  28-56
    o  2-10/2
    o  128-1024/*2
    o  2048-128//2
EOUsage

    exit(1);
}

# Set a nice signal handler.
$SIG{INT} = \&slow_death;
$SIG{QUIT} = \&slow_death;
$SIG{HUP} = \&slow_death;
$SIG{PIPE} = \&slow_death;
$SIG{TERM} = \&slow_death;

#  Constants.
my $TPUT_RESULT_LOG;
my $PING_DELAY = 10000;

#  Bookkeeping
my $run_experiment_pid;
my $key_cmd;

#  Test parameters.
my $SKIP_ARCHIVE = 'true';
my $TEST_MACHINES = 1;
my $NUM_SERIALIZERS = 1;
my $TEST_LENGTH = 10;
my $INIT_SIZE = 1048576;
my @SERVERS = split /\s+/, $ENV{"OSTORE_SERVERS"} if( $ENV{"OSTORE_SERVERS"} );
my $KEY_SIZE = 512;
my $HISTOGRAM = 'false';
my $USE_SERIALIZERS = 'false';
my $VERBOSE = 'false';

my @UPDATE_LOAD_SPEC;
my $UPDATE_LOAD_START = 1;
my $UPDATE_LOAD_END   = 1;
my $UPDATE_LOAD_OP    = '+';
my $UPDATE_LOAD_SKIP  = 1;

my @UPDATE_SIZE_SPEC;
my $UPDATE_SIZE_START = 1024;
my $UPDATE_SIZE_END   = 1024;
my $UPDATE_SIZE_OP    = '+';
my $UPDATE_SIZE_SKIP  = 1;

#archive parameters
my $DISSEMINATE = 'false';
my $NUM_STORAGE_NODES = 32;
my $NUM_STORAGE_MACHINES = 16;
my $NUM_FRAGS = 32;		# Must align with REQUEST_CARDINALITY
my $REQUEST_CARDINALITY = 24;	# Must align with NUM_FRAGS

# Visdemo stage
my $VISDEMO = 'false';

# Tapas stage
my $use_tapas = 0;

#  Get the user-defined parameters.
my $opts_ok = getopts('hasvfgtc:d:u:w:l:x:k:n:z');
if ($opt_z) { $VISDEMO = 'true'; }
if ($opt_h) { &usage; }
if ($opt_a) { $SKIP_ARCHIVE = 'false'; }
if ($opt_s) { $USE_SERIALIZERS = 'true'; }
if ($opt_v) { $VERBOSE = 'true'; }
if ($opt_f) { $DISSEMINATE = 'true'; }
if ($opt_c) { $TEST_MACHINES = $opt_c; }
if ($opt_l) { $TEST_LENGTH = $opt_l; }
if ($opt_n) { $NUM_SERIALIZERS = $opt_n; $USE_SERIALIZERS = 'true' }
if ($opt_x) { @SERVERS = map "x$_.millennium.berkeley.edu", ($opt_x .. 42); }
if ($opt_k) { $KEY_SIZE = $opt_k; }
if ($opt_w) { $INIT_SIZE *= $opt_w; }
if ($opt_g) { $HISTOGRAM = 'true'; }
if ($opt_t) { $use_tapas = 1; }
if ($opt_d)
{
    @UPDATE_LOAD_SPEC = &parse_range_spec($opt_d);
    my @spec = @UPDATE_LOAD_SPEC;
    $UPDATE_LOAD_START = shift(@spec);
    $UPDATE_LOAD_END   = shift(@spec);
    $UPDATE_LOAD_OP    = shift(@spec);
    $UPDATE_LOAD_SKIP  = shift(@spec);
}
if ($opt_u)
{
    @UPDATE_SIZE_SPEC = &parse_range_spec($opt_u);
    my @spec = @UPDATE_SIZE_SPEC;
    $UPDATE_SIZE_START = shift(@spec);
    $UPDATE_SIZE_END   = shift(@spec);
    $UPDATE_SIZE_OP    = shift(@spec);
    $UPDATE_SIZE_SKIP  = shift(@spec);
}

#  The command line is valid if getopts returned without error and
#  there are no arguments left over.  $#ARGV returns the number of
#  remaining arguments minus one.
if ((! $opts_ok) || ($#ARGV != -1)) { &usage; }

#  Check that the set of input parameters is valid.
if ($KEY_SIZE != 512 && $KEY_SIZE != 1024)
{
    print STDERR "ERROR:  Invalid key size.\n";
    &usage();
    exit(1);
}
if( ! @SERVERS ) {
    print STDERR "ERROR:  Must define OSTORE_SERVERS or specify -x option.\n";
    &usage();
    exit(1);
}
if( ($USE_SERIALIZERS eq 'true') and $opt_a ) {
    print STDERR "WARNING: -a flag ignored when serializers enabled.\n";
}


#  Compute the result log name.
$TPUT_RESULT_LOG = &get_tput_result_log_name();

#  Set the environment variable for the keys.
if ($KEY_SIZE == 512) {
    $ENV{KEYSHARE_PREFIX} = "ir";
    $key_cmd = "KEYSHARE_PREFIX=ir";
} else {
    $ENV{KEYSHARE_PREFIX} = "ir-1024-4-2";
    $key_cmd = "KEYSHARE_PREFIX=ir-1024-4-2";
}


##############################
#  Run experiments to collect data.
my $update_size = $UPDATE_SIZE_START;
my $update_size_comparator = '<=';
if ($UPDATE_SIZE_OP eq '-' || $UPDATE_SIZE_OP eq '/')
{
    $update_size_comparator = '>=';
}
while (eval("$update_size $update_size_comparator $UPDATE_SIZE_END"))
{
    my $dobj_size = $update_size * 2;
    if ($dobj_size > $update_size + 512 * 1024) {
	$dobj_size = $update_size + 512 * 1024;
    }

    my $update_load = $UPDATE_LOAD_START;
    my $update_load_comparator = '<=';
    if ($UPDATE_LOAD_OP eq '-' || $UPDATE_LOAD_OP eq '/')
    {
        $update_load_comparator = '>=';
    }
    while (eval("$update_load $update_load_comparator $UPDATE_LOAD_END"))
    {
        #  Modify configuration files.
	my @master_config_directives = 
	    ("set Master.TputMicroBenchmarkMaster.NumTesters $TEST_MACHINES\n",
	     "set Master.TputMicroBenchmarkMaster.UpdateLoad $update_load\n",
	     "set Master.TputMicroBenchmarkMaster.DobjSize $dobj_size\n",
	     "set Master.TputMicroBenchmarkMaster.UpdateSize $update_size\n",
	     "set Master.TputMicroBenchmarkMaster.TestLength $TEST_LENGTH\n",
	     "set Master.TputMicroBenchmarkMaster.InitSize $INIT_SIZE\n");
	
	my $exp_cfg = &write_experiment_file($TEST_MACHINES,
					     @master_config_directives);

        my $tput_fork_log = "$Tools::tmp_dir/tput_fork_log.$$.log";

	my @args = ();
	push @args, ($tput_fork_log, "./run-experiment");
	if($use_tapas) {
	  push @args, "--tapas"; 
	}
	push @args, $exp_cfg;

        $run_experiment_pid = &fork_process(@args);

        #  Run the experiment.
        print("Running tput test: $key_cmd ./run-experiment $exp_cfg\n" .
	      "pid=$run_experiment_pid\tlog=$tput_fork_log\n" .
              "  TestMachines = $TEST_MACHINES\n" .
              "  UpdateLoad   = $update_load\n" .
              "  DobjSize     = $dobj_size\n" .
              "  UpdateSize   = $update_size\n" .
              "  TestLength   = $TEST_LENGTH\n" .
              "  InitLength   = $INIT_SIZE\n" .
              "  KeySize      = $KEY_SIZE\n");

        if(! &returned_zero($run_experiment_pid))
        {
            die("Experiment returned non-zero.\n");
        }

        #  Read the result log.
        my $result_log_name = &get_result_log_name($run_experiment_pid);
        my $total_ops       = &analyze_result_log($result_log_name);

        #  Compute the results.
        my $ops_tput  = $total_ops / $TEST_LENGTH;
        my $data_tput = ($total_ops * $update_size) / $TEST_LENGTH;
        print("  OpsTput      = $ops_tput\n" .
              "  DataTput     = " . ($data_tput/1024/1024) ." MB/s\n");

        #  Write to log.
        &write_result_to_log($SKIP_ARCHIVE,
                             $TEST_MACHINES, $update_load, $TEST_LENGTH,
                             $KEY_SIZE, $dobj_size, $update_size,
                             $total_ops, $total_ops, $ops_tput, $data_tput);

        if ($HISTOGRAM eq 'true')
        {
            my $histogram = &read_latency_histogram($result_log_name);
            write_to_log("\n" . $histogram . "\n");
        }

        #  Remove the files we created.
        unlink($exp_cfg, $tput_fork_log, $result_log_name);

        $update_load = eval("$update_load $UPDATE_LOAD_OP $UPDATE_LOAD_SKIP");
    }

    $update_size = eval("$update_size $UPDATE_SIZE_OP $UPDATE_SIZE_SKIP");
}
# Done!
exit(0);


sub parse_range_spec
{
    my $spec = shift(@_);

    my $start = 0;
    my $end   = 0;
    my $op    = '+';
    my $skip  = 1;

    #  (\d+)                  The start of the range...
    #  (-(\d+))?              ...optionally a '-' and the end of the range...
    #  (\/([-\+\*\/])?(\d+))? ...optionally a '/', an optional
    #                         operation (-, +, *, or /) and the skip value.
    if ($spec =~ /(\d+)(-(\d+))?(\/([-\+\*\/])?(\d+))?/ )
    {
        $start = $1;
        $end   = $start;
        if (defined($3)) { $end = $3; }
        if (defined($5)) { $op = $5; }
        if (defined($6)) { $skip = $6; }
    }

    #print("start=$start \n");
    #print("end=$end   \n");
    #print("op=$op    \n");
    #print("skip=$skip  \n");

    if (($op eq '+') || ($op eq '*'))
    {
        if ($start > $end)
        {
          die("Invalid range specification.\n" .
              "  start=$start, end=$end, operation=$op\n");
        }
    }
    else
    {
        if ($end > $start)
        {
          die("Invalid range specification.\n" .
              "  start=$start, end=$end, operation=$op\n");
        }
    }

    my @range;
    push(@range, $start);
    push(@range, $end);
    push(@range, $op);
    push(@range, $skip);

    return @range;
}

#  Compute the result log name.
sub get_tput_result_log_name
{
    my $type = "noar";
    $type = "inar" if ($SKIP_ARCHIVE eq 'false');
    $type = "only" if ($USE_SERIALIZERS eq 'true');

    my $log = "$Tools::tmp_dir/tput_result_log.$type.$$.log";    
    return $log;
}

sub write_result_to_log
{
    my $skip          = shift(@_);
    my $test_machines = shift(@_);
    my $update_load   = shift(@_);
    my $init_size     = shift(@_);
    my $test_length   = shift(@_);
    my $key_size      = shift(@_);
    my $dobj_size     = shift(@_);
    my $update_size   = shift(@_);
    my $total_ops     = shift(@_);
    my $ops_tput      = shift(@_);
    my $data_tput     = shift(@_);

    my $log_msg =
      "SkipArchive=${skip} " .
      "TestMachines=${test_machines} " .
      "UpdateLoad=${update_load} " .
      "TestLength=${test_length}s " .
      "InitSize=${init_size}MB " .
      "KeySize=${key_size} " .
      "DobjSize=${dobj_size} " .
      "UpdateSize=${update_size} " .
      "TotalOps=${total_ops} " .
      "UpdateTput=${ops_tput}ops/s " .
      "DataTput=${data_tput}bytes/s\n";

    write_to_log($log_msg);

    return;
}

sub write_to_log
{
    my $log_msg = shift(@_);

    open(RESULT_LOG, ">> $TPUT_RESULT_LOG") ||
      die("Could not open result log ($TPUT_RESULT_LOG).");
    print(RESULT_LOG "$log_msg");
    close(RESULT_LOG);

    return;
}

sub get_result_log_name
{
    my $pid = shift(@_);

    my @internal_logs =
	glob( "$Tools::tmp_dir/*-${pid}*-Master-internal.log");

    die("Cannot determine result log name.\n") if (@internal_logs != 1);

    my $log = shift(@internal_logs);
    return $log;
}

sub analyze_result_log
{
    my $log = shift(@_);

    open(RESULT_LOG, "$log") || die("Could not open result log.\n");
    my $total_ops = 0;
    while(<RESULT_LOG>)
    {
        if (/NumOps=(.*)/)
        {
            $total_ops += $1;
        }
    }

    close(RESULT_LOG);

    return $total_ops;
}

sub read_latency_histogram
{
    my $log = shift(@_);

    open(RESULT_LOG, "$log") || die("Could not open result log.\n");
    my $histogram = '';
    while(<RESULT_LOG>)
    {
        if (/TotalLatStats/ || /^\s+\d+\s+\d+$/)
        {
            $histogram .= $_;
        }
    }

    close(RESULT_LOG);

    return $histogram;
}

sub write_experiment_file
{
    my ($test_machines, @master_config_directives) = @_;

    my $exp_cfg = "$Tools::tmp_dir/tput.$$.exp";
    open(EXP, ">$exp_cfg") ||
      die("Could not create experiment configuration file ($exp_cfg).\n");
    my $old_fd = select(EXP);
    my $fed = 0;
    my $master;

    # Experiment header
    print "hosts	@SERVERS\n";
    print "Federation	\${run_dir}/federation.cfg fed daemon $fed\n";

    # Next, either RP and Inner Ring or Archivers (serializers)
    if( $USE_SERIALIZERS eq 'false' ) {
	my($ir0, $ir1, $ir2, $ir3) = (1 .. 4);
	$master = 5;
	print "RP \${run_dir}/rp.cfg static daemon $fed\n";
	print "Ring0 \${run_dir}/inner.cfg static daemon $ir0\n";
	print "Ring1 \${run_dir}/inner.cfg static daemon $ir1\n";
	print "Ring2 \${run_dir}/inner.cfg static daemon $ir2\n";
	print "Ring3 \${run_dir}/inner.cfg static daemon $ir3\n";
	print "set Ring*.Inner.skip_archive $SKIP_ARCHIVE\n";	
	print "set Ring*.Inner.num_frags $NUM_FRAGS\n";	
    } else {
	my $serializer_offset = $fed + 1;
	for (my $i = 0; $i < $NUM_SERIALIZERS; $i++) {
	    my $machine_num = $serializer_offset + $i;
	    print("Archiver$i \${run_dir}/archiver.cfg " .
		  "static daemon $machine_num\n");
	    print "set Archiver$i.ArchiverStage.archiver_number $i\n";	    
	}
	print "set Archiver*.ALL.LogEvents $VERBOSE\n";
	print "set Archiver*.ALL.Verify false\n";

	$master = $serializer_offset + $NUM_SERIALIZERS;
    }

    # Master with per-iteration directives
    print "Master \${cfg_dir}/tput-master.cfg static mortal $master\n";
    print @master_config_directives;

    # Testers
    my $tester_offset = $master + 1;
    for (my $i = 0; $i < $test_machines; $i++)
    {
        my $machine_num = $tester_offset + $i;
	if( $USE_SERIALIZERS eq 'false' ) {
	    # Use normal ClientStage.
	    print("Tester$i \${cfg_dir}/tput-tester.cfg " .
		  "static  daemon $machine_num\n");
	} else {
	    # Use FakeClient.
	    print("Tester$i \${cfg_dir}/tput-fake-tester.cfg " .
		  "static  daemon $machine_num\n");
	    # Partition the test machines equally among the serializers.
	    my $archiver_number = int(($i/$TEST_MACHINES)*$NUM_SERIALIZERS);
	    print "set Tester$i.FakeClient.archiver_number $archiver_number\n";
	    print "set Tester$i.FakeClient.rp_pkeyfile " .
			"\${Archiver${archiver_number}PublicKey}\n";
	    print "set Tester$i.FakeClient.num_frags $NUM_FRAGS\n";
	    print "set Tester$i.TputMicroBenchmark.rp_pkeyfile " .
			"\${Archiver${archiver_number}PublicKey}\n";
	}
    }
    print "set Tester*.RequestorStage.AlarmTimeout 180000\n";
    print "set Tester*.RequestorStage.RequestCardinality " .
	"$REQUEST_CARDINALITY\n";
#    print "set Tester*.Cache.CacheSize 314572800\n";	# 300 MB
    print "set Tester*.BlockBufferCache.cache_size 524288000\n"; # 500 MB 314572800\n";	# 300 MB
    

    # Dissemination set creator, fragment storage nodes
    if($DISSEMINATE eq 'true') {
	print "Creator \${run_dir}/creator.cfg static daemon $fed\n";
	print "set Creator.SetCreatorStage.SetSize $NUM_FRAGS\n";

	for( my $i = 0; $i < $NUM_STORAGE_NODES; $i++) {
	    my $machine_num =
		($tester_offset + $test_machines + ($i % $NUM_STORAGE_MACHINES));
	    print("StorageServer$i \${run_dir}/storage.cfg "
		  . "static daemon $machine_num\n");
	}
	print "set StorageServer*.ArchivalStorageStage.AlarmTimeout 180000\n";

	# Additional archive stages for other nodes
	print "add Tester* RecoalesceBlk.stg\n";
	
	if( $USE_SERIALIZERS eq 'false' ) {
	    print "add Ring* Disseminator.stg\n";
	    print "set Ring*.DisseminatorStage.AlarmTimeout 180000\n";
	} else {
	    print "add Archiver* Disseminator.stg\n";
	}
    }
    else
    {
      print "set Tester*.RequestorStage.UseDtreeRegardless true\n";
    }

    # Other overrides:
    print "remove * StorageManager\n";

    print "set *.TClient.PingDelay $PING_DELAY\n";
    print "set *.MemoryCache.Publish false\n";
    print "set *.RequestorStage.SendCachePin true\n";
    print "set *.RecoalesceBlkStage.SendCachePin true\n";
    print "set *.CacheManager.Publish false\n";
    print "set *.BlockBufferCache.UseDisk false\n";
    print "set *.DisseminatableBufferCache.UseDisk false\n";

    if( $VISDEMO eq 'true' ) {
	print "add * Visdemo.stg";
    }

    close(EXP);
    select($old_fd);

    return $exp_cfg;
}

# try to kill run-experiment nicely.
sub slow_death {
    print(  "*** Caught SIG@_ ***\n" );
    if( defined $run_experiment_pid ) {
	print "  Killing run-experiment...\n";
	kill 2, $run_experiment_pid;
	# Now wait for it to die.
	&returned_zero( $run_experiment_pid );
    }
    &exit_all( 1 );
}
