/**
 * DisplayManager.java
 *
 * Copyright (c) 2002 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
package Sonar; 
 
import java.awt.*; 
import java.awt.event.*; 
import java.util.*; 
import javax.swing.*; 
import java.lang.Math; 
import java.io.*; 
import java.util.zip.*; 
import java.net.InetAddress;

import Sonar.event.*; 
import Sonar.Dialog.*; 
import Sonar.util.*; 
import Sonar.PacketAnalyzers.*; 

import ostore.util.NodeId;

/**
 * This class is listening to the main panel on the mainFrame and generates
 * events whenver a node or edge are clicked. It can also be used for scrolling
 * or zooming. DisplayManager could havebeen integrated with the MainFrameClass,
 * but Kamin wanted to leave that class to VisualCafe.
 *
 * ***Note: This version of DisplayManager is heavily modified for oceanstore,
 * and like the rest of Sonar, will not work with TinyOS packets.
 *
 * @author	Kamin Whitehouse <kamin@cs.berkeley.edu>
 * @author	Jason Lee <jlee81@cs.berkeley.edu>
 */

public class DisplayManager extends PaintTools implements java.awt.event.MouseListener, java.awt.event.MouseMotionListener, NodePainter, EdgePainter, NodeEventListener, EdgeEventListener, NodeClickedEventListener, EdgeClickedEventListener, NodeDialogContributor, GroupDialogContributor, EdgeDialogContributor //, java.io.Serializable 
{ 
    //these vectors hold all people that want to paint on the edge, nodes, or
    //over the entire screen 
    
    //a list of all objects that want to paint nodes 
    protected Vector nodePainters;
    //a list of all objects that want to paint edges 
    protected Vector edgePainters;
    //a list of all objects that want to paint the screen 
    protected Vector screenPainters;
   
    //these vectors contain all people who want to add to the node/edge
    //properties dialog 
    
    //a list of all objects that want to add to the node properties Dialog 
    protected Vector nodeDialogContributors;
    protected Vector groupDialogContributors;
    //a list of all objects that want to add to the edge properties dialog 
    protected Vector edgeDialogContributors;
  
    //these vectors contain all the registered listeners for node and edge
    //click events 
    protected Vector nodeClickedEventListeners; 
    protected Vector edgeClickedEventListeners; 
 
    //these hashtables hold nodeInfo and edgeInfo objects, which hold display
    //information about specific nodes and edges (e.g. color, image) 
    protected static Hashtable proprietaryNodeInfo; 
    protected static TwoKeyHashtable proprietaryEdgeInfo;
    //For Edges with type, specific time to live
    protected static TwoKeyHashtable extraEdgeInfo; 
 
    protected MenuManager menuManager; 
 
    //these variables describe how the mouse interacts with the network 
    protected boolean selectMode; 
    protected boolean handMode; 
    protected boolean zoomMode; 
     	 
    protected int pressedXCoord=-1; 
    protected int pressedYCoord=-1; 
    protected static long refreshRate;//refresh rate 

    protected double zoomFactor;//the factor by which the screen is enlarged 
 
    //the thread that runs in the background and refreshes the screen 
    //periodically 
    //protected static Thread refreshScreenThread;

    /**** Oceanstore-specific ***/
    //protected static int maxEdgeWidth = 10;
    
    public boolean showTapestryArrows = false;
    public boolean showTapestryEdges = false; 
    public boolean showNodeIds = false;

    public static boolean rolePlaying = true; //flag to choose to identify nodes
    public static final int NEVER_SET = -1; //Defaults to imac
    public static final int NON_RESPONSIVE = 0;
    public static final int CLIENT = 1;
    public static final int INNER_RING = 2;
    public static final int FILE_SERVER = 3;
    public static final int BOTH_IR_FS = 4;
    public static final int RESP_PARTY = 5;
    public static final int SET_CREATOR = 6;
    public static final int REPLICA = 7;

    protected boolean showGuid = false;
    protected boolean showPort = true;


    /**** Oceanstore-specific ***/
    
    public MainFrame mainFrame;

    public long edgeEventLongevity = 2 * 1000;
    public long nodeEventLongevity = 10 * 1000;
    // if modified alert the other listeners by an event

    //public GridLocationAnalyzer locationAnalyzer;

    //------------------------------------------------------------------ 
    //CONSTRUCTOR 
    DisplayManager(MainFrame pMainFrame) 
    { 
	mainFrame = pMainFrame;
	nodePainters = new Vector(); 
	edgePainters = new Vector(); 
	screenPainters = new Vector(); 
	nodeDialogContributors = new Vector(); 
	groupDialogContributors = new Vector();
	edgeDialogContributors = new Vector(); 
	nodeClickedEventListeners = new Vector(); 
	edgeClickedEventListeners = new Vector();	 
	proprietaryNodeInfo = new Hashtable(); 
	proprietaryEdgeInfo = new TwoKeyHashtable();
	extraEdgeInfo = new TwoKeyHashtable();

	locationAnalyzer = new GridLocationAnalyzer(this);

	//register to receive all mouse clicks on the display panel 
	//pMainFrame.getGraphDisplayPanel().addMouseListener(this); 
	//pMainFrame.getGraphDisplayPanel().addMouseMotionListener(this); 
	//register (with myself) to paint nodes/edges and display info panels 
	this.addNodePainter(this);//paint the nodes 
	this.addEdgePainter(this);//paint the edges

	//register myself to receive NodeClickedEvents and EdgeClickedEvents 
	this.addNodeDialogContributor(this); 
	this.addGroupDialogContributor(this);
	this.addEdgeDialogContributor(this); 

	//register to be notified of nodes and edges being created or deleted 
	//this is done in MainClass constructor because this object is
	//instantiated before the Object Maintainer 
	
	//register to contribute a panel to the SonarOptions panel 
        //MainClass.addOptionsPanelContributor(this); 
	mainFrame.addOptionsPanelContributor(this);
	menuManager = new MenuManager(); 
		 
	selectMode=true; 
	handMode=false; 
	zoomMode=false; 

	//this is the number of milliseconds that it waits before redrawing the
	//screen again 
	refreshRate = 500;
	//this thread runs continually in the background to redraw the screen 

    } 
    //CONSTRUCTOR 
    //------------------------------------------------------------------ 
 
    //------------------------------------------------------------------ 
    //MOUSE CLICKED 
    public void mouseClicked(MouseEvent e){} 
    public void mouseClickedCustom(MouseEvent e) 
    { 
	GraphDisplayPanel display = mainFrame.getGraphDisplayPanel(); 
	double x = (double) e.getX();
	double y = (double) e.getY();
	if(mainFrame.getGraphDisplayPanel().planetlab && locationAnalyzer.getLayoutMode() == GridLocationAnalyzer.MATRIX ) {
	    x = x / display.scaleX();
	    y = y / display.scaleY();
	} else {
	    x = display.scaleScreenXCoordToNodeCoord(e.getX()).doubleValue(); 
	    y = display.scaleScreenYCoordToNodeCoord(e.getY()).doubleValue(); 
	}

	if (javax.swing.SwingUtilities.isLeftMouseButton(e)) { 
	    NodeId clickedNodeId = 
	    	locationAnalyzer.findNearestNode(x, y); 
	    if(clickedNodeId != null) { 
		triggerNodeClickedEvent(clickedNodeId); 
		displayNodePropertyDialog(clickedNodeId); 
	    } 
	    //debugging
	    //if(DEBUG) ostore.util.Debug.printtagln("DisplayManager.mouseClicked", e.getX() + ", "+ e.getY() + " "+display.scaleX()+" ("+x+","+y+")");
	} 
	else {	 
	    //TODO:
	    //MainClass.out("DEBUG : triggerEdgeClickedEvent");
	    
	    Vector edgeID = locationAnalyzer.findNearestEdge(
		    x, y); 
	    if((edgeID != null) && (edgeID.size() == 2)) { 
		triggerEdgeClickedEvent((NodeId)edgeID.elementAt(0),
			(NodeId)edgeID.elementAt(1)); 
		displayEdgePropertyDialog((NodeId)edgeID.elementAt(0),
			(NodeId)edgeID.elementAt(1)); 
	    } 
	}	 
    } 
    //MOUSE CLICKED 
    //------------------------------------------------------------------ 
 
 
    //------------------------------------------------------------------ 
    //MOUSE DRAGGED 
    public void mouseDragged(MouseEvent e)  
    { 
	drawCoords(e); 
    }
    //even though using this method instead of my custom method would make the
    //graphics look better I use a custom mouse dragged method because
    //otherwise handling all the events overwhelms the system and all the
    //nodes might time-out 
	 
    public void mouseDraggedCustom(int startX, int startY, MouseEvent e) 
    { 
	if (javax.swing.SwingUtilities.isLeftMouseButton(e)) { 
	    if(selectMode) { 
		triggerNodeDraggedEvent(startX, startY, e); 
	    } else if(handMode) { 
		//scrollWithMouseDrag(startX, startY, e); 
	        triggerNodesSelectedEvent(startX, startY, e);
	    } else if(zoomMode) { 
		zoomToMouseDragRectangle(startX, startY, e); 
	    } 
	} else if (javax.swing.SwingUtilities.isMiddleMouseButton(e)) { 
	    //scrollWithMouseDrag(startX, startY, e); 
	    //highlight those nodes
	    triggerNodesSelectedEvent(startX, startY, e);
	} else if (javax.swing.SwingUtilities.isRightMouseButton(e)) { 
	    zoomToMouseDragRectangle(startX, startY, e); 
	} 
    } 
    //MOUSE DRAGGED 
    //------------------------------------------------------------------ 
 
    public void mouseEntered(MouseEvent e) { } 
    public void mouseExited(MouseEvent e) 
    {         
	//leave these lines if you don't want the custom mouse click to fire
	//after the mouse has left and re-entered the screen 
	//pressedXCoord = -1; 
	//pressedYCoord = -1; 
    } 
		 
    //this function triggers an event for a mouse press 
    public void mousePressed(MouseEvent e) 
    { 
	pressedXCoord = e.getX(); 
	pressedYCoord = e.getY(); 
    } 
		 
    public void mouseReleased(MouseEvent e) 
    { 
	if((pressedXCoord == -1) || (pressedYCoord == -1)) { 
	    return; 
	} 
	int x = e.getX(); 
	int y = e.getY(); 
 
	if((pressedXCoord == x) && (pressedYCoord == y)) { //if a mouse click 
	    mouseClickedCustom(e); 
	} else {
	    //if it was a drag, pass the original coords and the final coords 
	    mouseDraggedCustom(pressedXCoord, pressedYCoord, e); 
	} 
	pressedXCoord = -1;//reset 
	pressedYCoord = -1; 
    } 
		 
    public void mouseMoved(MouseEvent e) 
    { 
	drawCoords(e); 
    } 
 
 
    //this function adds the node dialog contributors 
    public  void addNodeDialogContributor(NodeDialogContributor pContributor) 
    { 
	nodeDialogContributors.add(pContributor); 
    } 
 
    public  void removeNodeDialogContributor(NodeDialogContributor pContributor) 
    { 
	nodeDialogContributors.remove(pContributor); 
    } 
 
    public  void addGroupDialogContributor(GroupDialogContributor pContributor) 
    { 
	groupDialogContributors.add(pContributor); 
    } 
 
    public  void removeGroupDialogContributor(GroupDialogContributor pContributor) 
    { 
	groupDialogContributors.remove(pContributor); 
    } 
 
    //this function adds the edge dialog contributors 
    public  void addEdgeDialogContributor(EdgeDialogContributor pContributor) 
    { 
	edgeDialogContributors.add(pContributor); 
    } 
 
    public  void removeEdgeDialogContributor(EdgeDialogContributor pContributor) 
    { 
	edgeDialogContributors.remove(pContributor); 
    } 
 
    //this function adds the nodeclicked Listeners 
    public  void addNodeClickedEventListener(NodeClickedEventListener pListener) 
    { 
	nodeClickedEventListeners.add(pListener); 
    } 
 
    public  void removeNodeClickedEventListener(NodeClickedEventListener pListener) 
    { 
	nodeClickedEventListeners.remove(pListener); 
    } 
 
    //this function adds the edge clicked Listeners 
    public  void addEdgeClickedEventListener(EdgeClickedEventListener pListener) 
    { 
	edgeClickedEventListeners.add(pListener); 
    } 
		 
    public void removeEdgeClickedEventListener(EdgeClickedEventListener pListener) 
    { 
	edgeClickedEventListeners.remove(pListener); 
    } 
		 
		 
    //************************************************************* 
    //************************************************************* 
    //this is where we register the node, edge and screen painters 
    //************************************************************* 
    //************************************************************* 
	           
    public void addNodePainter(NodePainter painter) 
    { 
	nodePainters.add(painter);//add the painters to the painter list 
    } 
  
    public void removeNodePainter(NodePainter painter) 
    { 
	nodePainters.remove(painter);//add the painters to the painter list 
    } 
  
    public void addEdgePainter(EdgePainter painter) 
    { 
	edgePainters.add(painter);//add the painters to the painter list 
    } 
  
    public void removeEdgePainter(EdgePainter painter) 
    { 
	edgePainters.remove(painter);//add the painters to the painter list 
    } 
                             
    public void addScreenPainter(ScreenPainter painter) 
    { 
	screenPainters.add(painter);//add the painters to the painter list 
    } 
 
    public void removeScreenPainter(ScreenPainter painter) 
    { 
	screenPainters.remove(painter);//add the painters to the painter list 
    } 
	 
 
    //------------------------------------------------------------------------ 
    //*****---TRIGGER NODE CLICKED EVENT---******// 
    //this function sends an event to all node clicked listeners 
    protected void triggerNodeClickedEvent(NodeId pNodeId) 
    { 
	NodeClickedEvent e = new NodeClickedEvent(this, pNodeId, Calendar.getInstance().getTime()); 
	for(int index = 0; index < nodeClickedEventListeners.size(); index++) { 
	    ((NodeClickedEventListener)nodeClickedEventListeners.get(index)).nodeClicked(e); 
	} 
    } 
    //*****---TRIGGER NODE CLICKED EVENT---******// 
    //------------------------------------------------------------------------ 
		 
		 
    //------------------------------------------------------------------------ 
    //*****---TRIGGER EDGE CLICKED EVENT---******// 
    //this function sends an event to all edge-clicked listeners 
    protected void triggerEdgeClickedEvent(NodeId pSourceNodeId, NodeId pDestinationNodeId) 
    { 
	EdgeClickedEvent e = new EdgeClickedEvent(this, pSourceNodeId, pDestinationNodeId, Calendar.getInstance().getTime()); 
	for(int index = 0; index < edgeClickedEventListeners.size(); index++) { 
	    ((EdgeClickedEventListener)edgeClickedEventListeners.get(index)).edgeClicked(e); 
	} 
    } 
    //*****---TRIGGER EDGE CLICKED EVENT---******// 
    //------------------------------------------------------------------------ 
 
	           
    //------------------------------------------------------------------------ 
    //*****---DRAG NEAREST NODE---******// 
    //when in select mode, the mouse draw will drag the nearest node with it 
    public void triggerNodeDraggedEvent(int startX, int startY, MouseEvent e) 
    { 
	GraphDisplayPanel display = mainFrame.getGraphDisplayPanel(); 
	Double x1 = display.scaleScreenXCoordToNodeCoord(startX); 
	Double y1 = display.scaleScreenYCoordToNodeCoord(startY); 
	Double x2 = display.scaleScreenXCoordToNodeCoord(e.getX()); 
	Double y2 = display.scaleScreenYCoordToNodeCoord(e.getY()); 
 
	NodeId pNodeId = locationAnalyzer.findNearestNode(x1.doubleValue(), y1.doubleValue());//this is why we need custom mouse drag (startX andstartY) 
	NodeDraggedEvent event = new NodeDraggedEvent(this, pNodeId, Calendar.getInstance().getTime(), x2.doubleValue(), y2.doubleValue()); 
	for(int index = 0; index < nodeClickedEventListeners.size(); index++) { 
	    ((NodeClickedEventListener)nodeClickedEventListeners.get(index)).nodeDragged(event); 
	} 
 
    } 
    //*****---DRAG NEAREST NODE---******// 
    //------------------------------------------------------------------------ 
    
    public void triggerNodesSelectedEvent(int startX, int startY, MouseEvent e) {
	GraphDisplayPanel display = mainFrame.getGraphDisplayPanel();
	Double x1 = display.scaleScreenXCoordToNodeCoord(startX); 
	Double y1 = display.scaleScreenYCoordToNodeCoord(startY); 
	Double x2 = display.scaleScreenXCoordToNodeCoord(e.getX()); 
	Double y2 = display.scaleScreenYCoordToNodeCoord(e.getY()); 

	Vector nodes = locationAnalyzer.findAllNodesWithinRectangle(x1.doubleValue(), y1.doubleValue(), x2.doubleValue(), y2.doubleValue());

	if(nodes!=null) {
	    NodesSelectedEvent event = new NodesSelectedEvent(this, nodes, Calendar.getInstance().getTime(), x1.doubleValue(), y1.doubleValue(), x2.doubleValue(), y2.doubleValue()); 
	    /*
	     //Instead of immediately grouping nodes, a dialog box comes up
	     //first
	    for(int index = 0; index < nodeClickedEventListeners.size(); index++) { 
		((NodeClickedEventListener)nodeClickedEventListeners.get(index)).nodesSelected(event); 
	    } 
	    */
	    displayGroupPropertyDialog(event); 
	}

    }
 
    //------------------------------------------------------------------------ 
    //*****---Draw Coords with Mouse Event---******// 
    //this is the function that changes the display of the mouse coords on the toolbar 
    public void drawCoords(MouseEvent e) 
    {         //this method should be changed to draw the coords at the current mouse position 
	//it currently draws it on the toolbar 
	String temp = String.valueOf(mainFrame.getGraphDisplayPanel().scaleScreenXCoordToNodeCoord(e.getX())); 
	String text = temp.substring(0,Math.min(4, temp.length())); 
	text = text.concat(","); 
	temp = String.valueOf(mainFrame.getGraphDisplayPanel().scaleScreenYCoordToNodeCoord(e.getY())); 
	text = text.concat(temp.substring(0,Math.min(4, temp.length()))); 
	mainFrame.getCoordLabel().setText(text); 
    } 
    //*****---SCROLL WITH MOUSE DRAG---******// 
    //------------------------------------------------------------------------ 
  
 
 
    //------------------------------------------------------------------------ 
    //*****---SCROLL WITH MOUSE DRAG---******// 
    //when in hand mode, the mouse draw will scroll the screen 
    public void scrollWithMouseDrag(int startX, int startY, MouseEvent e) 
    { 
	int endX = e.getX(); 
	int endY = e.getY(); 
	Point currentPosition = mainFrame.getMainScrollPane().getViewport().getViewPosition(); 
	Point newPosition = new Point(currentPosition.x +(endX-startX), currentPosition.y +(endY-startY)); 
	mainFrame.getMainScrollPane().getViewport().setViewPosition(newPosition); 
    } 
    //*****---SCROLL WITH MOUSE DRAG---******// 
    //------------------------------------------------------------------------ 


    //------------------------------------------------------------------------ 
    //*****---ZOOM TO MOUSE DRAG RECTANGLE---******// 
    //when in zoom mode, the mouse drag will zoom to the selected rectangle 
    public void zoomToMouseDragRectangle(int startX, int startY, MouseEvent e) 
    { 
	int x,y, width, height; 
	double scale; 
	Dimension graphPanelSize = mainFrame.getGraphDisplayPanel().getSize(); 
	Dimension scrollPaneSize = mainFrame.getMainScrollPane().getSize(); 
	int endX = e.getX(); 
	int endY = e.getY(); 
	if(startX < endX)//use the edge of the rectangle that is longer in proportion that the proportion of the window size 
	{ 
	    x = startX; 
	    width = endX-startX; 
	} 
	else 
	{ 
	    x = endX; 
	    width = startX - endX; 
	} 
	if(startY < endY) 
	{ 
	    y = startY; 
	    height = endY-startY; 
	} 
	else 
	{ 
	    y = endY; 
	    height = startY - endY; 
	} 
	if( (height==0) || (width==0)) return; 
	if(width/height > graphPanelSize.getWidth()/graphPanelSize.getHeight())//zoom so that the longer of the width/height of the new rectangle is in view (Since the rect will not be the same proportion as the window) 
	{ 
	    scale = scrollPaneSize.getWidth()/width; 
	} 
	else 
	{ 
	    scale = scrollPaneSize.getHeight()/height; 
	} 
	scale = scale*getScaleFactor(); 
	this.multiplyGraphDisplayPanelSize(scale); 
	mainFrame.getMainScrollPane().getViewport().setViewPosition(new Point(x,y)); 
    }                                             
    //*****---ZOOM TO MOUSE DRAG RECTANGLE---******// 
    //------------------------------------------------------------------------ 



    //------------------------------------------------------------------------ 
    //*****---DISPLAY NODE PROPERTY DIALOG---******// 
    //this function displays the dialog showing all node properties 
    protected void displayNodePropertyDialog(NodeId pNodeId) { 
	TabbedDialog nodeDialog = new TabbedDialog("Node Properties");		 
	ActivePanel currentPanel; 
	NodeDialogContributor listener; 

	for(Enumeration e = nodeDialogContributors.elements(); e.hasMoreElements();)
	{ 
	    listener = ((NodeDialogContributor)e.nextElement()); 
	    currentPanel = listener.getProprietaryNodeInfoPanel(pNodeId); 
	    if(currentPanel != null) { 
		if(currentPanel.getCancelInfoDialog()) {
		    //if you don't want a node dialog to show up, return an Active Component with this set to true 
		    return; 
		} 
		nodeDialog.addActivePanel(currentPanel.getTabTitle(), currentPanel); 
	    } 
	} 
	nodeDialog.setModal(false); 
	nodeDialog.show(); 
    } 
    //*****---DISPLAY NODE PROPERTY DIALOG---******// 
    //------------------------------------------------------------------------ 

    protected void displayGroupPropertyDialog(NodesSelectedEvent event) {
	TabbedDialog groupDialog = new TabbedDialog("Group Properties");		 
	ActivePanel currentPanel; 
	GroupDialogContributor listener; 

	for(Enumeration e = groupDialogContributors.elements(); e.hasMoreElements();)
	{ 
	    listener = ((GroupDialogContributor)e.nextElement()); 
	    currentPanel = listener.getProprietaryGroupInfoPanel(event.getNodes()); 
	    if(currentPanel != null) { 
		if(currentPanel.getCancelInfoDialog()) {
		    //if you don't want a node dialog to show up, return an Active Component with this set to true 
		    return; 
		} 
		groupDialog.addActivePanel(currentPanel.getTabTitle(), currentPanel); 
	    } 
	} 
	groupDialog.setModal(false); 
	groupDialog.show(); 
    }

  
    //------------------------------------------------------------------------ 
    //*****---DISPLAY EDGEPROPERTY DIALOG---******// 
    //this function displays the dialog showing all edge properties 
    protected void displayEdgePropertyDialog(NodeId pSourceNodeId, NodeId pDestinationNodeId) 
    { 
	TabbedDialog edgeDialog = new TabbedDialog("Edge Properties");		 
	ActivePanel currentPanel; 
		 
	for(Enumeration e = edgeDialogContributors.elements(); e.hasMoreElements();) { 
	    currentPanel = ((EdgeDialogContributor)e.nextElement()).getProprietaryEdgeInfoPanel(pSourceNodeId, pDestinationNodeId); 
	    if(currentPanel != null) {
		//if you don't have proprietary info, return a null panel 
		     
		if(currentPanel.getCancelInfoDialog()) {
		    //if you don't want a node dialog to show up, return an Active Component with this set to true 
		    return; 
		} 
		edgeDialog.addActivePanel(currentPanel.getTabTitle(),currentPanel); 
	    } 
	} 
	edgeDialog.setModal(false); 
	edgeDialog.show(); 
    } 
    //*****---DISPLAY EDGEPROPERTY DIALOG---******// 
    //------------------------------------------------------------------------ 
 
 
    //*****---refresh screen NOW---******// 
    //this function will redraw the main screen with all nodes and edges in current positions 
    //public static void refreshScreenNow() 
    public void refreshScreenNow() 
    { 
	mainFrame.paint(mainFrame.getGraphics()); 
    } 
    //*****---refresh screen NOW---******// 
 
	 
    //----------------------------------------------------------------------- 
    //*****---multiplyGraphDisplayPanelSize---******// 
    //public static void multiplyGraphDisplayPanelSize(double factor) 
    public void multiplyGraphDisplayPanelSize(double factor) 
    { 
	Insets inset = mainFrame.getMainScrollPane().getInsets();	  
	Dimension d = mainFrame.getMainScrollPane().getSize();	 
	int x = (int)( (d.width-(inset.left+inset.right))*factor); 
	int y = (int)( (d.height-(inset.top+inset.bottom))*factor); 
	mainFrame.getGraphDisplayPanel().setSize(new Dimension(x, y));//this line makes the scroll pane put scroll bars if necessary 
	mainFrame.getGraphDisplayPanel().setPreferredSize(new Dimension(x, y));//this line changes the size, and the node coordinates are automatically rescaled 
	mainFrame.getMainScrollPane().setViewportView(mainFrame.getGraphDisplayPanel());//this line allows the scroll pane to reevaluate whether scroll bars are needed 
    } 
    //*****---multiplyGraphDisplayPanelSize---******// 
    //----------------------------------------------------------------------- 
 
    public void paintUnderScreen(Graphics g) 
    { 
	ScreenPainter screenPainter; 
	for(Enumeration painters = screenPainters.elements(); painters.hasMoreElements();)  
	    { 
		screenPainter = (ScreenPainter)painters.nextElement(); 
		screenPainter.paintScreenBefore(g); 
	    } 
    } 
 
    public void paintOverScreen(Graphics g) 
    { 
	ScreenPainter screenPainter; 
	for(Enumeration painters = screenPainters.elements(); painters.hasMoreElements();)  
	    { 
		screenPainter = (ScreenPainter)painters.nextElement(); 
		screenPainter.paintScreenAfter(g); 
	    } 
    } 
     
    public void paintAllNodes(NodeId pNodeId, int x1, int y1, int x2, int y2, Graphics g) 
    { 
	NodePainter nodePainter; 
	NodeInfo displayInfo = (NodeInfo)proprietaryNodeInfo.get(pNodeId); 
	if( (displayInfo == null) || (displayInfo.getDisplayThisNode() == false) ) return; 
	for(Enumeration painters = nodePainters.elements(); painters.hasMoreElements();)  
	    { 
		nodePainter = (NodePainter)painters.nextElement(); 
		nodePainter.paintNode(pNodeId, x1, y1, x2, y2, g); 
	    } 
    } 
 
    public void paintAllEdges(NodeId pSourceNodeId, NodeId pDestinationNodeId, int x1, int y1, int x2, int y2, Graphics g) 
    { 
	EdgePainter edgePainter; 
	NodeInfo sourceDisplayInfo = (NodeInfo)proprietaryNodeInfo.get(pSourceNodeId); 
	NodeInfo destinationDisplayInfo= (NodeInfo)proprietaryNodeInfo.get(pDestinationNodeId); 
	if( (sourceDisplayInfo == null) || (destinationDisplayInfo == null)
	    || (sourceDisplayInfo.getDisplayThisNode() == false)
	    || (destinationDisplayInfo.getDisplayThisNode() == false)) return; 
	for(Enumeration painters = edgePainters.elements(); painters.hasMoreElements();)  
	    { 
		edgePainter = (EdgePainter)painters.nextElement(); 
		edgePainter.paintEdge(pSourceNodeId, pDestinationNodeId, x1, y1, x2, y2, g); 
	    } 
    }

    public void paintAllTypedEdges(NodeId source, NodeId dest, int x1, int y1, int x2, int y2, Graphics g)  {
	//EdgePainter edgePainter; 
	NodeInfo sourceDisplayInfo = (NodeInfo)proprietaryNodeInfo.get(source); 
	NodeInfo destinationDisplayInfo= (NodeInfo)proprietaryNodeInfo.get(dest); 
	
	if( (sourceDisplayInfo == null) || (destinationDisplayInfo == null)
	    || (sourceDisplayInfo.getDisplayThisNode() == false)
	    || (destinationDisplayInfo.getDisplayThisNode() == false)) return; 

	/*	    
	 * To be uncommented after implementing more typed edge painters. 	
	 * Currently, only displayManager handles this function. 
	for(Enumeration painters = edgePainters.elements(); painters.hasMoreElements();)  
	{ 	
	typedEdgePainter = (EdgePainter)painters.nextElement(); 	
	typedEdgePainter.paintEdge(pSourceNodeId, pDestinationNodeId, x1, y1, x2, y2, g); 
	} 
	*/
	Hashtable typeToEdge;
	if((typeToEdge = (Hashtable) extraEdgeInfo.get(source, dest)) != null) {
	    paintTypedEdge(typeToEdge, x1, y1, x2, y2, g); 
	}
    }

    /*
     * Sets all the types back to the default icon.
     */
    public void resetAllTypes() {
	for(Enumeration e = proprietaryNodeInfo.elements(); e.hasMoreElements();) {
	    identifyNode(((NodeInfo) e.nextElement()).getNodeId(), NEVER_SET);
	}
    }

    /**
     * Associates a given node with an image. This is primarily for identifying
     * specific functions in an oceanstore archive/dissemination experiment.
     */
    public synchronized void identifyNode(NodeId node, int type) {
    	if(!rolePlaying) return;

	NodeInfo nodeInfo = (NodeInfo) proprietaryNodeInfo.get(node);
	if(nodeInfo == null) {
	    //MainClass.out("DM> Node identified does not exist yet");
	    return;
	}
	ImageIcon ih = null;
	switch(type) {
	    case NEVER_SET:
		ih =  new ImageIcon("images/newimac.gif", "images/newimac.gif");
		nodeInfo.setImageHelper(ih);
		nodeInfo.setImage(ih.getImage());
		nodeInfo.setNodeType(NEVER_SET);
		break;
	    case NON_RESPONSIVE:
		ih = new ImageIcon("images_os/marmoset.jpg",
				"images_os/marmoset.jpg");
		nodeInfo.setImageHelper(ih);
		nodeInfo.setImage(ih.getImage());
		nodeInfo.setNodeType(NON_RESPONSIVE);
		break;
	    case CLIENT:
		ih =  new ImageIcon("images_os/mandrill.jpg",
				"images_os/mandrill.jpg");
		nodeInfo.setImageHelper(ih);
		nodeInfo.setImage(ih.getImage());

		nodeInfo.setNodeType(CLIENT);
		break;
	    case INNER_RING:
		ih = new ImageIcon("images_os/macaq.jpg",
				"images_os/macaq.jpg");
		nodeInfo.setImageHelper(ih);
		nodeInfo.setImage(ih.getImage());

		nodeInfo.setNodeType(INNER_RING);
		break;
	    case FILE_SERVER:
		ih = new ImageIcon("images_os/lemur.jpg",
				"images_os/lemur.jpg");
		nodeInfo.setImageHelper(ih);
		nodeInfo.setImage(ih.getImage());

		nodeInfo.setNodeType(FILE_SERVER);
		break;
	    case BOTH_IR_FS:
		   ih = new ImageIcon("images/oceanblue.gif",
				   "images/oceanblue.gif");
		   nodeInfo.setImageHelper(ih);
		   nodeInfo.setImage(ih.getImage());		    
		   nodeInfo.setNodeType(BOTH_IR_FS);
		   break;
	    case RESP_PARTY:
		   ih = new ImageIcon("images_os/pygmarm.jpg",
				   "images_os/pygmarm.jpg");
		   nodeInfo.setImageHelper(ih);
		   nodeInfo.setImage(ih.getImage());
		   nodeInfo.setNodeType(RESP_PARTY);
		   break;
	    case SET_CREATOR:
		   ih = new ImageIcon("images_os/gorilla.jpg",
				   "images_os/gorilla.jpg");
		   nodeInfo.setImageHelper(ih);
		   nodeInfo.setImage(ih.getImage());
		   nodeInfo.setNodeType(SET_CREATOR); 
		   break;
	    case REPLICA:
		   nodeInfo.setNodeType(REPLICA);
	}	
    }

    public void handleAnalyzerMenuEvent( AnalyzerMenuEvent item ) {
	show_traffic = item.show_traffic;
	show_ICMP = item.show_ICMP;
	show_create_obj = item.show_create_obj;
	show_ir_create = item.show_ir_create;
	show_disseminate_frag = item.show_disseminate_frag;
	show_update = item.show_update;
	show_heartbeat = item.show_heartbeat;
	show_datadrivenhb = item.show_datadrivenhb;
	show_patchwork = item.show_patchwork;
	show_imap = item.show_imap;
	show_req_frag = item.show_req_frag;
    }

    //------------------------------------------------------------------------ 
    //*****---PAINT---******// 
  public void paintNode(NodeId pNodeId, int x1, int y1, int x2, int y2, Graphics g) 
    { 
	NodeInfo nodeInfo = (NodeInfo)proprietaryNodeInfo.get(pNodeId); 

	switch(nodeInfo.getNodeType()) {
	    case CLIENT:
		g.setColor(Color.pink);
		break;
	    case INNER_RING:
		g.setColor(Color.lightGray);
		break;
	    case FILE_SERVER:
		g.setColor(Color.red);
		break;
	    case BOTH_IR_FS:
		g.setColor(Color.green);
		break;
	    case RESP_PARTY:
		g.setColor(Color.yellow);
		break;
	    case SET_CREATOR:
		g.setColor(Color.cyan);
		break;
	    case REPLICA:
		g.setColor(Color.orange);
		break;
	    case NON_RESPONSIVE:
	    case NEVER_SET:
	    default:
		g.setColor(Color.black);
	}
	g.fillOval(x1, y1, x2-x1, x2-x1);
	//g.drawImage(nodeInfo.getImage(), x1, y1, x2-x1, y2-y1, null); 
		 
	if(!nodeInfo.getDisplayNodeId()) return;

	String descrip = "";
	/*
	if(nodeInfo.getGroupLeader() != null) {
	    g.setColor(Color.red);
	    String leader = "<"
		+nodeInfo.getGroupLeader().getNodeId().address().getHostName() + "> ";
	    g.drawString(leader,(x1+x2)/2, y2+(y2-y1)/2);  

	}
	*/
	g.setColor(Color.black); 
	//ostore.util.SecureHash guid = MainClass.objectMaintainer.node_to_guid(pNodeId);
	/*
	if(guid != null && showGuid) {
	    descrip = descrip.concat(guid.toString());
	} else if (showGuid) {
	    descrip = descrip.concat("unknown ");
	}
	if(showPort)
	*/
	//descrip = descrip.concat(pNodeId.address().getHostName());
	//g.drawString(descrip,  (x1+x2)/2, y2+(y2-y1)/4); 
	g.drawString(pNodeId.address().getHostName(), (x1+x2)/2, y2+(y2-y1)/4); 

    } 
    //*****---PAINT---******// 
    //-------------------------------------------------------------------------- 
	 
    public void paintEdge(NodeId source, NodeId dest, int x1, int y1, int x2, int y2, Graphics g) {
	if(showTapestryEdges) {
	    g.setColor(Color.black);
	    g.drawLine(x1, y1, x2, y2); 	
	    if(showTapestryArrows) {
		paintArrow(x1, y1, x2, y2, g, getNodeInfo(source).getArrowValue());
		EdgeInfo edgeInfo = (EdgeInfo) proprietaryEdgeInfo.get(source,dest);
		if(edgeInfo.isBidirectional()) {
		    paintArrow(x2, y2, x1, y1, g, getNodeInfo(dest).getArrowValue());
		}
	    }
	}
    }
 
    //------------------------------------------------------------------------ 
    //*****---GET/SET FUNCTIONS---******// 
    //public static double getScaleFactor() 
    public double getScaleFactor() 
    { 
	Dimension d1 = mainFrame.getMainScrollPane().getSize();	
	Insets inset = mainFrame.getMainScrollPane().getInsets();  
	Dimension d2 = mainFrame.getGraphDisplayPanel().getSize(); 
        return (d2.getHeight()+(inset.top+inset.bottom))/d1.getHeight(); 
    } 
	 
    public static long getRefreshRate() { return refreshRate; } 
    public static void setRefreshRate(long pRefreshRate)
    {
	refreshRate = pRefreshRate;
    } 
 
    public boolean getSelectMode(){return selectMode;} 
    public boolean getHandMode(){return handMode;} 
    public boolean getZoomMode(){return zoomMode;} 
	 
    public void setSelectMode(boolean b){selectMode=b;} 
    public void setHandMode(boolean b){handMode=b;} 
    public void setZoomMode(boolean b){zoomMode=b;} 

    public Hashtable getProprietaryNodeInfo() {
	return proprietaryNodeInfo;
    }
	 
    public NodeInfo getNodeInfo(NodeId nodeNumber)
    {
	return (NodeInfo)proprietaryNodeInfo.get(nodeNumber);
    } 
    public EdgeInfo getEdgeInfo(NodeId sourceNumber, NodeId destinationNumber)
    {
	return (EdgeInfo)proprietaryEdgeInfo.get(sourceNumber,
		destinationNumber);
    } 
    public EdgeInfo getTypedEdgeInfo(NodeId source, NodeId dest) {
        return (EdgeInfo) extraEdgeInfo.get(source, dest);
    }

    public Enumeration getNodeInfo(){return proprietaryNodeInfo.elements();} 
    public Enumeration getEdgeInfo(){return proprietaryEdgeInfo.elements();}
    public Enumeration getTypedEdgeInfo() { return extraEdgeInfo.elements(); }
 
    //*****---GET/SET FUNCTIONS---******// 
    //------------------------------------------------------------------------ 
 
    //------------------------------------------------------------------------ 
    //*****---Node Created---******// 
    //this function defines what you do when a new node is created 
    //It is called by Sonar.PacketAnalyzers.ObjectMainter 
    public synchronized void nodeCreated(NodeEvent e) 
    { 
    	NodeId newNodeId = e.getNodeId();
	//you probably want to create a new info pbject to track the data of 
	//this new node 
    	if(!proprietaryNodeInfo.containsKey(newNodeId)) {
	    //unless it already exists (it might exist if you don't delete it
	    //in NodeDeleted() 
	    proprietaryNodeInfo.put(newNodeId, e.getNodeInfo()); //new NodeInfo(newNodeId));
	} else {
	    NodeInfo nodeInfo = 
	    	(NodeInfo)proprietaryNodeInfo.get(newNodeId);
	    //but you might also want to leave it there but disable it, 
	    //unless this node reappears and you want to use the same info 
	    nodeInfo.setDisplayThisNode(true); 
	} 
    } 
    //*****---Node Created---******// 
    //------------------------------------------------------------------------ 
 
     
    //------------------------------------------------------------------------ 
    //*****---Node Deleted---******// 
    //this function defines what you do when a new node is deleted 
    //It is called by Sonar.PacketAnalyzers.ObjectMainter 
    public synchronized void nodeDeleted(NodeEvent e) 
    { 
    	NodeId deletedNodeId = e.getNodeId();
	//you probably want to delete the info pbject to track the 
	//data of this new node 
    	NodeInfo nodeInfo = 
		(NodeInfo)proprietaryNodeInfo.get(deletedNodeId);
	//but you might also want to leave it there but disable it, unless
	//this node reappears and you want to use the same info 
    	if(nodeInfo!=null) {
	    nodeInfo.setDisplayThisNode(false); 
	} 
    } 
    //*****---Node Deleted---******// 
    //------------------------------------------------------------------------ 
 
 
    //------------------------------------------------------------------------ 
    //*****---Edge Created---******// 
    //this function defines what you do when a new edge is created 
    //It is called by Sonar.PacketAnalyzers.ObjectMainter 
    public synchronized void edgeCreated(EdgeEvent e) 
    { 
    	NodeId source = e.getSourceNodeId(); 
    	NodeId dest = e.getDestinationNodeId();

        //Routing Table edge
	if(!e.hasType()) {
	    EdgeInfo edgeInfo = (EdgeInfo) proprietaryEdgeInfo.get(source,dest); 
	    if(edgeInfo == null) {
		edgeInfo = e.getEdgeInfo(); //new EdgeInfo(source, dest); 
		proprietaryEdgeInfo.put(source, dest, edgeInfo); 
	    } else if(dest.equals(edgeInfo.getSource())) {
		edgeInfo.setBidirectional(true); //Might not be needed
	    }
		    
	} else { //Event edge
	    EdgeInfo edge = e.getEdgeInfo();
	    Hashtable typeToEdge = (Hashtable) extraEdgeInfo.get(source,dest);
	    if(typeToEdge != null) {
	        if(!typeToEdge.contains(edge)) 
		    typeToEdge.put(edge.getType(), edge);
		//} else { //modifyEdgeWidth should be called instead
		//    typeEdgeInfo.instances++;
	    } else {
		typeToEdge = new Hashtable();
		typeToEdge.put(edge.getType(), edge);
		extraEdgeInfo.put(source, dest, typeToEdge);
	    }
	}
    }

    public synchronized void modifyEdgeWidth(EdgeInfo edge) {
	Hashtable typeToEdge = (Hashtable) extraEdgeInfo.get(edge.getSource(), edge.getDest());
	if(typeToEdge != null) {
	    EdgeInfo typedEdgeInfo = null;
	    if((typedEdgeInfo = (EdgeInfo) typeToEdge.get(edge.getType())) != null) {
		if(!typedEdgeInfo.equals(edge)) {
		    typedEdgeInfo.instances = edge.getInstances();
		}
	    }
	}
    }
    
    //*****---Edge Created---******// 
    //------------------------------------------------------------------------ 
 
 
    //------------------------------------------------------------------------ 
    //*****---Edge Deleted---******// 
    //this function defines what you do when a new edge is deleted 
    //It is called by Sonar.PacketAnalyzers.ObjectMainter 
    public synchronized void edgeDeleted(EdgeEvent e) 
    { 
    	NodeId source = e.getSourceNodeId(); 
	NodeId dest = e.getDestinationNodeId();//you probably want to create a new info pbject to track the data of this new node 
	EdgeInfo edgeInfo = null;
	    
    	if(!e.hasType()) {
	    edgeInfo = (EdgeInfo) proprietaryEdgeInfo.get(source, dest);
	    if(edgeInfo.isBidirectional()) {
	        if(source.equals(edgeInfo.getSource()) && dest.equals(edgeInfo.getDest())) {
		    edgeInfo.setSource(dest);
		    edgeInfo.setDest(source);
		    edgeInfo.setBidirectional(false);
	        } else if(dest.equals(edgeInfo.getSource()) && source.equals(edgeInfo.getDest())) {
		    edgeInfo.setBidirectional(false);
		} else {
		    //MainClass.out("Error: sources/dest do not match exisiting bi edge! "+source.port()+"/"+dest.port() + " vs " +edgeInfo.getSource().port() +"/"+ edgeInfo.getDest().port());
		}
	    } else {
	        proprietaryEdgeInfo.remove(source, dest);
	    }
	} else {
	    Hashtable typeToEdge = (Hashtable) extraEdgeInfo.get(source, dest);
	    if(typeToEdge != null) {
		if((edgeInfo = (EdgeInfo) typeToEdge.get(e.getEdgeInfo().getType())) != null) {
		    typeToEdge.remove(e.getEdgeInfo().getType());
		}
	    } // else edge does not exist!
	}
    } 
    //*****---edgeDeleted---******// 
    //------------------------------------------------------------------------ 
 
 
 
    //------------------------------------------------------------------------ 
    //*****---GRAPHICS COMMANDS---******// 
    public void nodeClicked(NodeClickedEvent e){} 
    public void nodeDragged(NodeDraggedEvent e){} 
    
    public void nodesSelected(NodesSelectedEvent e) {}

    public void edgeClicked(EdgeClickedEvent e){} 
	 
    public ActivePanel getProprietaryNodeInfoPanel(NodeId pNodeId) 
    { 
	ProprietaryNodeInfoPanel panel = new ProprietaryNodeInfoPanel((NodeInfo)proprietaryNodeInfo.get(pNodeId)); 
	return (ActivePanel)panel; 
    } 

    public ActivePanel getProprietaryGroupInfoPanel(Vector nodes) {
	ProprietaryGroupInfoPanel panel = new ProprietaryGroupInfoPanel(nodes);
	return (ActivePanel)panel;
    }
 
    public ActivePanel getProprietaryEdgeInfoPanel(NodeId pSourceNodeId, NodeId pDestinationNodeId)  
    { 
	return null; 
	//ProprietaryNodeInfoPanel panel = new ProprietaryNodeInfoPanel((NodeInfo)proprietaryNodeInfo.get(pEdge.getSource().getNodeId())); 
	//return (ActivePanel)panel; 
    } 
 
    public void showOptionsDialog()  
    { 
	StandardDialog newDialog = new StandardDialog(new OptionsPanel()); 
	newDialog.show(); 
    }

    public void showLegend()  
    { 
	StandardDialog newDialog = new StandardDialog(new Legend()); 
	newDialog.show(); 
    } 
 
    public void showEdgeLegend()  
    { 
	StandardDialog newDialog = new StandardDialog(new EdgeLegend()); 
	newDialog.show(); 
    }

    public ActivePanel getOptionsPanel()  
    { 
	return new OptionsPanel(); 
    } 
    public ActivePanel getLegend()  
    { 
	return new Legend(); 
    } 
    //*****---GRAPHICS COMMANDS---******// 
    //------------------------------------------------------------------------ 
 
    //handles setting the icons for a particular group
    //gridlocation analyzer's tab elects the leader
    public class ProprietaryGroupInfoPanel extends Sonar.Dialog.ActivePanel
    {
	Vector group;
	javax.swing.JButton iconButton = new javax.swing.JButton(); 
	javax.swing.JTextField textField1 = new javax.swing.JTextField();
	javax.swing.JTextField textField2 = new javax.swing.JTextField();
	javax.swing.JTextField textField3 = new javax.swing.JTextField();
	javax.swing.JLabel label1 = new javax.swing.JLabel();
	javax.swing.JLabel label2 = new javax.swing.JLabel();
	javax.swing.JLabel label3 = new javax.swing.JLabel();
	javax.swing.JPanel panel1 = new javax.swing.JPanel();

	public ProprietaryGroupInfoPanel(Vector nodes) {
	    group = nodes;

	    tabTitle = "Display";
	    setLayout(null);
	    Insets ins = getInsets();
	    setSize(259,279); 
	    
	    label1.setText("Width"); 
	    add(label1); 
	    label1.setBounds(36,48,84,24); 

	    label2.setText("Height"); 
	    add(label2); 
	    label2.setBounds(36,72,75,24); 

	    textField1.setNextFocusableComponent(textField2); 
	    textField1.setToolTipText("The scale of the coordinate system is determined by the user, and scaled automatically by the system to fit to the screen"); 
	    textField1.setText("0.2"); 
	    add(textField1); 
	    textField1.setBounds(120,48,87,18); 

	    textField2.setNextFocusableComponent(textField3); 
	    textField2.setToolTipText("The scale of the coordinate system is determined by the user, and scaled automatically by the system to fit to the screen"); 
	    textField2.setText("0.2"); 
	    add(textField2); 
	    textField2.setBounds(120,72,87,18); 

	    label3.setText("Image"); 
	    add(label3); 
	    label3.setBounds(36,24,84,24); 
	    textField3.setText("images/newimac.gif");
	    textField3.setNextFocusableComponent(textField1);
	    add(textField3);
	    textField3.setBounds(94, 25, 162, 18);

	    iconButton.setToolTipText("Click this button to see the image that is typed above");
	    iconButton.setText("Preview");
	    add(iconButton);
	    iconButton.setBounds(168, 108, 84, 27);

	    panel1.setLayout(null);
	    add(panel1);
	    panel1.setBounds(36, 165, 153, 126);

	    SymAction lSymAction = new SymAction();
	    iconButton.addActionListener(lSymAction);
	}

	public void applyChanges() {
	    NodeInfo node_info;
	    for(Enumeration e = group.elements(); e.hasMoreElements();) {
		node_info = (NodeInfo) e.nextElement();
		node_info.setImageWidth(Double.parseDouble(textField1.getText())); 
		node_info.setImageHeight(Double.parseDouble(textField2.getText())); 
		node_info.setImageHelper(new ImageIcon(textField3.getText(), textField3.getText()));
		node_info.setImage(node_info.getImageHelper().getImage()); 
	    }
	}
	
	public void initializeDisplayValues() 
	{ 
	    //textField.setText(); //but group assignments
	} 

	class SymAction implements java.awt.event.ActionListener 
	{ 
	    public void actionPerformed(java.awt.event.ActionEvent event) 
	    { 
		Object object = event.getSource(); 
		if (object == iconButton) 
		    iconButton_actionPerformed(event); 
	    } 
	} 

	void iconButton_actionPerformed(java.awt.event.ActionEvent event) 
	{ 
	    // to do: code goes here. 
	    try { 
		ImageIcon imageIcon = new ImageIcon(textField3.getText()); 
		Image tempImage = imageIcon.getImage(); 
		Graphics g = this.getGraphics(); 
		Dimension d = panel1.getSize(); 
		int width = (int)(mainFrame.getGraphDisplayPanel().getXScale()*Double.parseDouble(textField1.getText())); 
		int height = (int)(mainFrame.getGraphDisplayPanel().getYScale()*Double.parseDouble(textField2.getText())); 
		g.drawImage(tempImage, 36,144, width, height, null); 
	    } catch (java.lang.Exception e) {
	    }
	}  
    }
 
    //********************************************************* 
    //********************************************************* 
    //********************************************************* 
    //PROPRIETARY NODE INFO DISPLAY PANEL 
    //This class is an ActivePanel and should have all the information 
    //in GUI form that this class stores with respect to nodes 
    //It should be return with getProprietaryNodeInfoPanel and it will be displayed 
    //with all the other packet analyzer proprietary info when a node is clicked. 
    public class ProprietaryNodeInfoPanel extends Sonar.Dialog.ActivePanel 
    { 
	NodeInfo nodeInfo; 
		 
	public ProprietaryNodeInfoPanel(NodeInfo pNodeInfo) 
	{ 
	    tabTitle = "Display"; 
	    nodeInfo = pNodeInfo; 
	    //{{INIT_CONTROLS 
	    setLayout(null); 
	    Insets ins = getInsets(); 
	    setSize(259,279); 
	    JLabel3.setText("Image Width"); 
	    add(JLabel3); 
	    JLabel3.setBounds(36,48,84,24); 
	    JLabel4.setText("Image Height"); 
	    add(JLabel4); 
	    JLabel4.setBounds(36,72,75,24); 
	    JTextField1.setNextFocusableComponent(JTextField2); 
	    JTextField1.setToolTipText("The scale of the coordinate system is determined by the user, and scaled automatically by the system to fit to the screen"); 
	    JTextField1.setText("0.2"); //1.5
	    add(JTextField1); 
	    JTextField1.setBounds(120,48,87,18); 
	    JTextField2.setNextFocusableComponent(JCheckBox1); 
	    JTextField2.setToolTipText("The scale of the coordinate system is determined by the user, and scaled automatically by the system to fit to the screen"); 
	    JTextField2.setText("0.2");  //3.2
	    add(JTextField2); 
	    JTextField2.setBounds(120,72,87,18); 
	    JLabel1.setText("Image"); 
	    add(JLabel1); 
	    JLabel1.setBounds(36,24,84,24); 
	    JTextField3.setNextFocusableComponent(JTextField1); 
	    JTextField3.setText("images/newimac.gif"); 
	    add(JTextField3); 
	    JTextField3.setBounds(84,24,162,18); 
	    JCheckBox1.setNextFocusableComponent(JCheckBox2); 
	    JCheckBox1.setSelected(true); 
	    JCheckBox1.setToolTipText("Check this if you want this node to appear on the screen"); 
	    JCheckBox1.setText("Display This Node"); 
	    add(JCheckBox1); 
	    JCheckBox1.setBounds(36,96,130,21); 
	    //JCheckBox1.setBounds(36,96,123,21); 
	    JCheckBox2.setNextFocusableComponent(JCheckBox3); 
	    JCheckBox2.setSelected(true); 
	    JCheckBox2.setToolTipText("This should be checked if you want this node to be fit onto the screen"); 
	    JCheckBox2.setText("Fit To Screen"); 
	    add(JCheckBox2); 
	    JCheckBox2.setBounds(36,120,123,21); 
	    JCheckBox3.setNextFocusableComponent(JTextField3); 
	    JCheckBox3.setSelected(true); 
	    //JCheckBox3.setToolTipText("This should be checked if you want the SHA1Hash/Port of the node to be drawn on the screen"); 
	    JCheckBox3.setToolTipText("This should be checked if you want the address of the node to be drawn on the screen"); 
	    //JCheckBox3.setText("Display " + MainClass.objectMaintainer.node_to_guid(nodeInfo.getNodeId()).toString() ); 
	    JCheckBox3.setText("Display " + nodeInfo.getNodeId().address().getHostName());
	    add(JCheckBox3); 
	    //JCheckBox3.setBounds(36,144,123,21); 
	    JCheckBox3.setBounds(36,144,240,21); 
	    JButton1.setToolTipText("Click this button to see the image that is typed above"); 
	    JButton1.setText("Preview"); 
	    add(JButton1); 
	    JButton1.setBounds(168,108,84,27); 
	    JPanel1.setLayout(null); 
	    add(JPanel1); 
	    JPanel1.setBounds(36,165,153,126); 
	    //}} 
 
			//{{REGISTER_LISTENERS 
	    SymAction lSymAction = new SymAction(); 
	    JButton1.addActionListener(lSymAction); 
	    //}} 
	} 
 
	//{{DECLARE_CONTROLS 
	javax.swing.JLabel JLabel3 = new javax.swing.JLabel(); 
	javax.swing.JLabel JLabel4 = new javax.swing.JLabel(); 
	javax.swing.JTextField JTextField1 = new javax.swing.JTextField(); 
	javax.swing.JTextField JTextField2 = new javax.swing.JTextField(); 
	javax.swing.JLabel JLabel1 = new javax.swing.JLabel(); 
	javax.swing.JTextField JTextField3 = new javax.swing.JTextField(); 
	javax.swing.JCheckBox JCheckBox1 = new javax.swing.JCheckBox(); 
	javax.swing.JCheckBox JCheckBox2 = new javax.swing.JCheckBox(); 
	javax.swing.JCheckBox JCheckBox3 = new javax.swing.JCheckBox(); 
	javax.swing.JButton JButton1 = new javax.swing.JButton(); 
	javax.swing.JPanel JPanel1 = new javax.swing.JPanel(); 

	//}} 
 
	public void applyChanges() 
	{ 
	    nodeInfo.setImageWidth(Double.parseDouble(JTextField1.getText())); 
	    nodeInfo.setImageHeight(Double.parseDouble(JTextField2.getText())); 
	    nodeInfo.setDisplayThisNode(JCheckBox1.isSelected()); 
	    nodeInfo.setFitOnScreen(JCheckBox2.isSelected()); 
	    nodeInfo.setDisplayNodeId(JCheckBox3.isSelected()); 
	    nodeInfo.setImageHelper(new ImageIcon(JTextField3.getText(),JTextField3.getText())); 
	    nodeInfo.setImage(nodeInfo.getImageHelper().getImage()); 
	} 
 
	public void initializeDisplayValues() 
	{ 
	    JTextField3.setText(nodeInfo.getImageHelper().getDescription()); 
	    JTextField1.setText(String.valueOf(nodeInfo.getImageWidth())); 
	    JTextField2.setText(String.valueOf(nodeInfo.getImageHeight())); 
	    JCheckBox1.setSelected(nodeInfo.getDisplayThisNode()); 
	    JCheckBox2.setSelected(nodeInfo.getFitOnScreen()); 
	    JCheckBox3.setSelected(nodeInfo.getDisplayNodeId()); 
	} 

	class SymAction implements java.awt.event.ActionListener 
	{ 
	    public void actionPerformed(java.awt.event.ActionEvent event) 
	    { 
		Object object = event.getSource(); 
		if (object == JButton1) 
		    JButton1_actionPerformed(event); 
	    } 
	} 

	void JButton1_actionPerformed(java.awt.event.ActionEvent event) 
	{ 
	    // to do: code goes here. 

	    JButton1_actionPerformed_Interaction1(event); 
	} 

	void JButton1_actionPerformed_Interaction1(java.awt.event.ActionEvent event) 
	{ 
	    try { 
		ImageIcon imageIcon = new ImageIcon(JTextField3.getText()); 
		Image tempImage = imageIcon.getImage(); 
		Graphics g = this.getGraphics(); 
		Dimension d = JPanel1.getSize(); 
		int width = (int)(mainFrame.getGraphDisplayPanel().getXScale()*Double.parseDouble(JTextField1.getText())); 
		int height = (int)(mainFrame.getGraphDisplayPanel().getYScale()*Double.parseDouble(JTextField2.getText())); 
		g.drawImage(tempImage, 36,165, width, height, null); 
	    } catch (java.lang.Exception e) { 
	    } 
	} 
    }
    //PROPRIETARY NODE INFO DISPLAY PANEL 
    //********************************************************* 
    //********************************************************* 
    //********************************************************* 
    
    public class Legend extends ActivePanel implements ItemListener {
	public Legend() {
	    tabTitle = "Node Legend";
	    modal = true;
	    setLayout(null);
	    setSize(280, 225);
	    
	    j_label1.setText("client - pink");
	    add(j_label1);
	    j_label1.setForeground(Color.pink);
	    j_label1.setBounds(20, 10, 150, 24);

	    j_label2.setText("inner ring - light gray");
	    add(j_label2);
	    //j_label2.setBackground(Color.black);
	    //j_label2.setForeground(Color.lightGray);
	    j_label2.setBounds(20, 34, 150, 24);

	    j_label3.setText("file server - red");
	    add(j_label3);
	    j_label3.setForeground(Color.red);
	    j_label3.setBounds(20, 58, 150, 24);

	    j_label4.setText("rp - yellow");
	    add(j_label4);
	    j_label4.setForeground(Color.yellow);
	    j_label4.setBounds(20, 82, 150, 24);

	    j_label5.setText("set creator - cyan");
	    add(j_label5);
	    j_label5.setForeground(Color.cyan);
	    j_label5.setBounds(20,106, 150, 24);

	    j_label6.setText("replica - orange");
	    add(j_label6);
	    j_label6.setForeground(Color.orange);
	    j_label6.setBounds(20,130, 150, 24);

	    j_label0.setText("default/not responding - black");
	    add(j_label0);
	    j_label0.setForeground(Color.black);
	    j_label0.setBounds(20,154, 200, 24);
	}

	javax.swing.JLabel j_label1 = new javax.swing.JLabel();
	javax.swing.JLabel j_label2 = new javax.swing.JLabel();
	javax.swing.JLabel j_label3 = new javax.swing.JLabel();
	javax.swing.JLabel j_label4 = new javax.swing.JLabel();
	javax.swing.JLabel j_label5 = new javax.swing.JLabel();
	javax.swing.JLabel j_label6 = new javax.swing.JLabel();

	javax.swing.JLabel j_label0 = new javax.swing.JLabel();


    }
     
    //TODO change information for edges instead of nodes (as copied above)
    public class EdgeLegend extends ActivePanel implements ItemListener {
	public EdgeLegend() {
	    tabTitle = "Edge Legend";
	    modal = true;
	    setLayout(null);
	    setSize(280, 225);
	    
	    j_label1.setText("client - yellow");
	    add(j_label1);
	    j_label1.setForeground(Color.pink);
	    j_label1.setBounds(20, 10, 150, 24);

	    j_label2.setText("inner ring - light gray");
	    add(j_label2);
	    j_label2.setBounds(20, 34, 150, 24);

	    j_label3.setText("file server - red");
	    add(j_label3);
	    j_label3.setForeground(Color.red);
	    j_label3.setBounds(20, 58, 150, 24);

	    j_label4.setText("rp - yellow");
	    add(j_label4);
	    j_label4.setForeground(Color.yellow);
	    j_label4.setBounds(20, 82, 150, 24);

	    j_label5.setText("set creator - cyan");
	    add(j_label5);
	    j_label5.setForeground(Color.cyan);
	    j_label5.setBounds(20,106, 150, 24);

	    j_label6.setText("replica - orange");
	    add(j_label6);
	    j_label6.setForeground(Color.orange);
	    j_label6.setBounds(20,130, 150, 24);

	    j_label0.setText("default/not responding - black");
	    add(j_label0);
	    j_label0.setForeground(Color.black);
	    j_label0.setBounds(20,154, 200, 24);
	}

	javax.swing.JLabel j_label1 = new javax.swing.JLabel();
	javax.swing.JLabel j_label2 = new javax.swing.JLabel();
	javax.swing.JLabel j_label3 = new javax.swing.JLabel();
	javax.swing.JLabel j_label4 = new javax.swing.JLabel();
	javax.swing.JLabel j_label5 = new javax.swing.JLabel();
	javax.swing.JLabel j_label6 = new javax.swing.JLabel();

	javax.swing.JLabel j_label0 = new javax.swing.JLabel();
    }
 
 
    //*********************************************************************** 
    //*********************************************************************** 
    //Class OptionsPanel  
    //This panel will be shown in a dialog when the users clicks the right menu 
    public class OptionsPanel extends ActivePanel implements ItemListener 
    { 
	public OptionsPanel() 
	{ 
	    tabTitle = "Display Properties"; 
	    modal=true; 
	    //{{INIT_CONTROLS 
	    setLayout(null); 
	    //			Insets ins = getInsets(); 
	    setSize(280,225); 
	    JLabel1.setToolTipText("This is the number of times the screen will be magnified"); 
	    JLabel1.setText("ZoomX:"); 
	    add(JLabel1); 
	    JLabel1.setBounds(24,12,48,24); 
	    JComboBox1.setEditable(true); 
	    JComboBox1.setToolTipText("This is the number of times the screen will be magnified"); 
	    JComboBox1.addItem(new Double(1.0)); 
	    JComboBox1.addItem(new Double(1.5)); 
	    JComboBox1.addItem(new Double(2.0)); 
	    JComboBox1.addItem(new Double(2.5)); 
	    JComboBox1.addItem(new Double(3.0)); 
	    add(JComboBox1); 
	    JComboBox1.setBounds(144,12,108,24); 
	    JLabel2.setToolTipText("Times are listed in milliseconds"); 
	    JLabel2.setText("Screen Refresh Rate:"); 
	    add(JLabel2); 
	    JLabel2.setBounds(24,48,132,24); 
	    JLabel4.setText("msec"); 
	    add(JLabel4); 
	    JLabel4.setFont(new Font("Dialog", Font.BOLD, 9)); 
	    JLabel4.setBounds(204,48,24,24); 
	    JSlider1.setMinimum(100); 
	    JSlider1.setMaximum(10000); 
	    JSlider1.setToolTipText("Slide this to change the refresh rate"); 
	    //JSlider1.setBorder(bevelBorder1); 
	    JSlider1.setValue(1500); 
	    add(JSlider1); 
	    JSlider1.setBounds(60,84,216,21); 
	    //$$ bevelBorder1.move(0,115); 
	    JLabel3.setText("jlabel"); 
	    add(JLabel3); 
	    JLabel3.setFont(new Font("Dialog", Font.BOLD, 16)); 
	    JLabel3.setBounds(156,48,51,27); 
	    JComboBox1.setSelectedIndex(0);

	    JCheckBox3.setSelected(false);
	    JCheckBox3.setToolTipText("This should be checked if you want to display the host address");
	    //JCheckBox3.setToolTipText("This should be checked if you want the SHA1Hash/Port of all the nodes to be drawn on the screen"); 
	    JCheckBox3.setText("Display Address");
	    //JCheckBox3.setText("Display Port & SHA1Hash");
	    add(JCheckBox3); 
	    JCheckBox3.setBounds(36,115,100,21); 

	    JCheckBox4.setSelected(rolePlaying);
	    JCheckBox4.setToolTipText("This should be checked if you'd like to display the appropriate image for a node's 'role'");
	    JCheckBox4.setText("Identify nodes");
	    JCheckBox4.setBounds(36, 136, 100, 21);
	    add(JCheckBox4);

	    JCheckBox5.setSelected(false);
	    JCheckBox5.setToolTipText("This will reset all the nodes to the default icon");
	    JCheckBox5.setText("Reset id's");
	    JCheckBox5.setBounds(36, 157, 100, 21);
	    add(JCheckBox5);

	    /** Routing Table Edges, Arrows **/
	    JCheckBox6.setSelected(showTapestryEdges);
	    JCheckBox6.setToolTipText("This should be checked if you want to display tapestry edges");
	    JCheckBox6.setText("Show Edges");
	    add(JCheckBox6); 
	    JCheckBox6.setBounds(180,115,200,21); 

	    JCheckBox7.setSelected(showTapestryArrows);
	    JCheckBox7.setToolTipText("This should be checked if you'd like to display arrows denoting the direction of edges");
	    JCheckBox7.setText("Show Arrows");
	    JCheckBox7.setBounds(180, 136, 200, 21);
	    add(JCheckBox7);
	    /** Edges, Arrows **/

	    JLabel5.setText("edge time:");
	    add(JLabel5);
	    JLabel5.setBounds(24,178,100,21);

	    JComboBox2.addItem(new String("1 sec"));
	    JComboBox2.addItem(new String("2 secs"));
	    JComboBox2.addItem(new String("3 secs"));
	    JComboBox2.addItem(new String("10 secs"));
	    JComboBox2.addItem(new String("20 secs"));
	    add(JComboBox2);
	    JComboBox2.setBounds(100, 178, 108, 24);

	    JLabel6.setText("node time:");
	    add(JLabel6);
	    JLabel6.setBounds(24,202,100,21);

	    JComboBox3.addItem(new String("20 secs"));
	    JComboBox3.addItem(new String("40 secs"));
	    JComboBox3.addItem(new String("1 min"));
	    JComboBox3.addItem(new String("2 min"));
	    JComboBox3.addItem(new String("3 min"));
	    add(JComboBox3);
	    JComboBox3.setBounds(100, 202, 108, 24);


	    //}} 
 
	    //{{REGISTER_LISTENERS 
	    SymChange lSymChange = new SymChange(); 
	    JSlider1.addChangeListener(lSymChange); 
	    SymAction lSymAction = new SymAction(); 
	    JCheckBox3.addChangeListener(lSymChange); 
	    //JCheckBox4.addChangeListener(lSymChange);
	    //JCheckBox5.addChangeListener(lSymChange);
	    //}} 
	} 
 
	//{{DECLARE_CONTROLS 
	javax.swing.JLabel JLabel1 = new javax.swing.JLabel(); 
	javax.swing.JComboBox JComboBox1 = new javax.swing.JComboBox(); 
	javax.swing.JLabel JLabel2 = new javax.swing.JLabel(); 
	javax.swing.JLabel JLabel4 = new javax.swing.JLabel(); 
	javax.swing.JSlider JSlider1 = new javax.swing.JSlider(); 
	//com.symantec.itools.javax.swing.borders.BevelBorder bevelBorder1 = new com.symantec.itools.javax.swing.borders.BevelBorder(); 
	javax.swing.JLabel JLabel3 = new javax.swing.JLabel(); 

	javax.swing.JCheckBox JCheckBox3 = new javax.swing.JCheckBox(); 
	javax.swing.JCheckBox JCheckBox4 = new javax.swing.JCheckBox();
	javax.swing.JCheckBox JCheckBox5 = new javax.swing.JCheckBox();
	javax.swing.JCheckBox JCheckBox6 = new javax.swing.JCheckBox();
	javax.swing.JCheckBox JCheckBox7 = new javax.swing.JCheckBox();

	javax.swing.JLabel JLabel5 = new javax.swing.JLabel();
	javax.swing.JLabel JLabel6 = new javax.swing.JLabel();
	javax.swing.JComboBox JComboBox2 = new javax.swing.JComboBox();
	javax.swing.JComboBox JComboBox3 = new javax.swing.JComboBox();
	//}} 
	
	
	public void applyChanges() 
	{ 
	    multiplyGraphDisplayPanelSize( Double.parseDouble(JComboBox1.getSelectedItem().toString())); 
	    setRefreshRate(JSlider1.getValue()); 

	    rolePlaying = JCheckBox4.isSelected();
	    if(JCheckBox5.isSelected()) resetAllTypes();

	    showTapestryEdges = JCheckBox6.isSelected();
	    showTapestryArrows = JCheckBox7.isSelected();

	    if(JComboBox2.getSelectedItem().equals("1 sec"))
		edgeEventLongevity = 1000;
	    else if (JComboBox2.getSelectedItem().equals("2 secs"))
		edgeEventLongevity = 2000;
	    else if(JComboBox2.getSelectedItem().equals("3 secs"))
		edgeEventLongevity = 3000;
	    else if(JComboBox2.getSelectedItem().equals("10 secs"))
		edgeEventLongevity = 10000;
	    else if(JComboBox2.getSelectedItem().equals("20 secs"))
		edgeEventLongevity = 20000;

	    if(JComboBox3.getSelectedItem().equals("20 secs"))
		nodeEventLongevity = 20 * 1000;
	    else if (JComboBox3.getSelectedItem().equals("40 secs"))
		nodeEventLongevity = 40 * 1000;
	    else if(JComboBox3.getSelectedItem().equals("1 min"))
		nodeEventLongevity = 60 * 1000;
	    else if(JComboBox3.getSelectedItem().equals("2 min"))
		nodeEventLongevity = 2 * 60 * 1000;
	    else if(JComboBox3.getSelectedItem().equals("3 min"))
		nodeEventLongevity = 3 * 60 * 1000;
	} 
		 
	public void initializeDisplayValues() 
	{ 
	    //double factor = MainClass.displayManager.getScaleFactor(); 
	    double factor = getScaleFactor(); 
	    JComboBox1.setSelectedItem(new Double(factor)); 
	    JSlider1.setValue((int)getRefreshRate()); 
	    JLabel3.setText(String.valueOf(getRefreshRate()));

	    if(edgeEventLongevity == 1000) {
		JComboBox2.setSelectedItem(new String("1 sec"));
	    } else if (edgeEventLongevity == 2000) {
		JComboBox2.setSelectedItem(new String("2 secs"));
	    } else if (edgeEventLongevity == 3000) {
		JComboBox2.setSelectedItem(new String("3 secs"));
	    } else if (edgeEventLongevity == 10000) {
		JComboBox2.setSelectedItem(new String("10 secs"));
	    } else if (edgeEventLongevity == 20000) {
		JComboBox2.setSelectedItem(new String("20 secs"));
	    }
	} 
 
 
	class SymChange implements javax.swing.event.ChangeListener 
	{ 
	    public void stateChanged(javax.swing.event.ChangeEvent event) 
	    { 
		Object object = event.getSource(); 
		if (object == JSlider1) 
		    JSlider1_stateChanged(event); 
		else if (object == JCheckBox3) { 
		    NodeInfo currentNodeInfo; 
		    for(Enumeration e = proprietaryNodeInfo.elements();
			    e.hasMoreElements();) { 
			currentNodeInfo = (NodeInfo)e.nextElement(); 
			currentNodeInfo.setDisplayNodeId(JCheckBox3.isSelected()); 
		    } 
		}
	    } 
	} 
 
	void JSlider1_stateChanged(javax.swing.event.ChangeEvent event) 
	{ 
	    // to do: code goes here. 
	    JSlider1_stateChanged_Interaction1(event); 
	} 
 
	class SymAction implements java.awt.event.ActionListener 
	{ 
	    public void actionPerformed(java.awt.event.ActionEvent event) 
	    { 
	    } 
	} 
 
	void JSlider1_stateChanged_Interaction1(javax.swing.event.ChangeEvent event) 
	{ 
	    try { 
		// convert int->class java.lang.String 
		JLabel3.setText(java.lang.String.valueOf(JSlider1.getValue())); 
	    } catch (java.lang.Exception e) { 
	    } 
	} 
		 
	public void ItemStateChanged(ItemEvent event) 
	{ 
	    Object object = event.getSource(); 
	    //	if(object == JCheckBox3) 
	} 
    } 
    //Class OptionsPanel  
    //*********************************************************************** 
    //*********************************************************************** 
	 
	 
    //********************************************************* 
    //********************************************************* 
    //********************************************************* 
    //MENU MANAGER 
    //This class creates and holds the menu that controls this 
    //DisplayManager.  It returns the menu to whoever wants 
    //to display it and it also handles all events on the menu 
    protected class MenuManager implements /*Serializable,*/ ActionListener, ItemListener 
    { 
	//{{DECLARE_CONTROLS 
	JSeparator separator1 = new JSeparator(); 
	JMenuItem propertiesItem = new JMenuItem(); 
	JSeparator separator2 = new JSeparator(); 

	JMenu legendMenu = new JMenu();
	JMenuItem legendItem = new JMenuItem();
	JMenuItem edgeLegendItem = new JMenuItem();

	//JMenu optionsMenu = new JMenu();

	//JMenu serializeMenu = new JMenu(); 
	//JMenuItem saveNodesItem = new JMenuItem(); 
	//JMenuItem loadNodesItem = new JMenuItem(); 
	//JMenuItem saveEdgesItem = new JMenuItem(); 
	//JMenuItem loadEdgesItem = new JMenuItem(); 

	//}} 
	 
	public MenuManager() 
	{ 
				//{{INIT_CONTROLS 
	
	    mainFrame.DisplayMenu.add(separator1); 
	    //optionsMenu.setText("Options");
	    //optionsMenu.setActionCommand("Options"); 
	    
	    propertiesItem.setText("Display Options"); 
	    propertiesItem.setActionCommand("Display Options"); 
	    mainFrame.DisplayMenu.add(propertiesItem); 
	    //optionsMenu.add(propertiesItem);
	    
	    legendMenu.setText("Keys");
	    legendMenu.setActionCommand("Keys");
	    
	    legendItem.setText("Node Legend"); 
	    legendItem.setActionCommand("Node Legend"); 
	    legendMenu.add(legendItem);
	    //optionsMenu.add(legendItem);
	    
	    edgeLegendItem.setText("Edge Legend"); 
	    edgeLegendItem.setActionCommand("Edge Legend"); 
	    legendMenu.add(edgeLegendItem);
	    
	    mainFrame.DisplayMenu.add(legendMenu); 

	    //mainFrame.DisplayMenu.add(optionsMenu);
	    /*
	    mainFrame.DisplayMenu.add(separator2);     
	    serializeMenu.setText("Serialize"); 
	    serializeMenu.setActionCommand("Serialize"); 
	    saveNodesItem.setText("Save Node Display Info"); 
	    saveNodesItem.setActionCommand("Save Node Display Info"); 
	    serializeMenu.add(saveNodesItem); 
	    loadNodesItem.setText("Load Node Display Info"); 
	    loadNodesItem.setActionCommand("Load Node Display Info"); 
	    serializeMenu.add(loadNodesItem); 
	    saveEdgesItem.setText("Save Edge Display Info"); 
	    saveEdgesItem.setActionCommand("Save Edge Display Info"); 
	    serializeMenu.add(saveEdgesItem); 
	    loadEdgesItem.setText("Load Edge Display Info"); 
	    loadEdgesItem.setActionCommand("Load Edge Display Info"); 
	    serializeMenu.add(loadEdgesItem); 
	    mainFrame.DisplayMenu.add(serializeMenu); 
	    */
				//}} 
 
				//{{REGISTER_LISTENERS 
	    propertiesItem.addActionListener(this); 
	    legendItem.addActionListener(this);
	    edgeLegendItem.addActionListener(this);
	    /*
	    saveNodesItem.addActionListener(this); 
	    loadNodesItem.addActionListener(this); 
	    saveEdgesItem.addActionListener(this); 
	    loadEdgesItem.addActionListener(this); 
	    */
				//}} 
	} 
 
	//---------------------------------------------------------------------- 
	//EVENT HANDLERS 
	//The following two functions handle menu events 
	//The functions following this are the event handling functions 
	public void actionPerformed(ActionEvent e) 
	{ 
	    Object object = e.getSource(); 
	    /*
	    if (object == saveNodesItem) 
		SaveNodes(); 
	    else if (object == loadNodesItem) 
		LoadNodes(); 
	    else if (object == saveEdgesItem) 
		SaveEdges(); 
	    else if (object == loadEdgesItem) 
		LoadEdges(); 
		else */
	    if (object == propertiesItem) {
		showOptionsDialog(); 
	    } else if(object == legendItem) {
		showLegend();
	    } else if(object == edgeLegendItem) {
		showEdgeLegend();
	    }
	} 
 
	public void itemStateChanged(ItemEvent e) 
	{ 
	    Object object = e.getSource(); 
	}		 
	//EVENT HANDLERS 
	//---------------------------------------------------------------------- 
 
	//------------------------------------------------------------------------ 
	//****---SAVE NODES---**** 
       	//takes the node hashtable and saves it to a file 
	public void SaveNodes() 
	{ 
	    FileDialog dialog = new FileDialog(mainFrame, "Save Nodes", FileDialog.SAVE); 
	    dialog.show(); 
	    String filename = dialog.getFile(); 
	    if( (filename != null) && (proprietaryNodeInfo!=null)) 
	    { 
		try { 
		    FileOutputStream fos = new FileOutputStream(filename); 
		    GZIPOutputStream gos = new GZIPOutputStream(fos); 
		    ObjectOutputStream out = new ObjectOutputStream(gos); 
		    out.writeObject(proprietaryNodeInfo); 
		    out.flush(); 
		    out.close(); 
		} 
		catch(Exception e){e.printStackTrace();} 
	    } 
	} 
	//****---SAVE NODES---**** 
	//------------------------------------------------------------------------ 
		  
 
	//------------------------------------------------------------------------ 
	//****---LOAD NODES---**** 
	//takes a file and loads the nodes into proprietaryNodeInfo hashtable 
	public void LoadNodes() 
	{        //in the future, it should prompt the user for which node should be kept 
	    FileDialog dialog = new FileDialog(mainFrame, "Load Nodes", FileDialog.LOAD); 
	    dialog.show(); 
	    String filename = dialog.getFile(); 
	    if(filename != null) { 
		NodeInfo currentNodeInfo; 
		Hashtable newNodes; 
		try { 
		    FileInputStream fis = new FileInputStream(filename); 
		    GZIPInputStream gis = new GZIPInputStream(fis); 
		    ObjectInputStream in = new ObjectInputStream(gis); 
		    newNodes = (Hashtable)in.readObject(); 
		    in.close(); 
		} 
		catch(Exception e){e.printStackTrace(); return;} 
				 
		if((proprietaryNodeInfo == null) || (proprietaryNodeInfo.isEmpty())) {

		    //if there are no nodes yet, just assign the new nodes to the entire vector 
		    proprietaryNodeInfo = newNodes; 
		} 
		else { //otherwise take the new nodes and add them to the vector (first eliminating repeat nodes)  ...in the future, we should ask the user which node to keep, in the case of repeat nodes 
		    for(Enumeration e = newNodes.elements();e.hasMoreElements();) { 
			currentNodeInfo = (NodeInfo)e.nextElement(); 
			proprietaryNodeInfo.remove(currentNodeInfo.getNodeId()); 
			proprietaryNodeInfo.put(currentNodeInfo.getNodeId(), currentNodeInfo); 
		    } 
		} 
	    } 
	} 
	//****---LOAD NODES---**** 
	//------------------------------------------------------------------------ 
	  
 
	//------------------------------------------------------------------------ 
	//****---SAVE EDGES---**** 
	//takes the node hashtable and saves it to a file 
	public void SaveEdges() 
	{ 
	    FileDialog dialog = new FileDialog(mainFrame, "Save Edges", FileDialog.SAVE); 
	    dialog.show(); 
	    String filename = dialog.getFile(); 
	    if( (filename != null) && (proprietaryEdgeInfo!=null)) 
	    { 
		try 
		{ 
		    FileOutputStream fos = new FileOutputStream(filename);
		    GZIPOutputStream gos = new GZIPOutputStream(fos); 
		    ObjectOutputStream out = new ObjectOutputStream(gos); 
		    out.writeObject(proprietaryEdgeInfo); 
		    out.flush(); 
		    out.close(); 
		} 
		catch(Exception e){e.printStackTrace();} 
	    } 
	} 
	//****---SAVE EDGES---**** 
	//------------------------------------------------------------------------ 
		  
 
	//------------------------------------------------------------------------ 
	//****---LOAD EDGES---**** 
	//takes a file and loads the nodes into proprietaryNodeInfo hashtable 
	public void LoadEdges() 
	{      
	    //in the future, it should prompt the user for which node should be kept 
	    FileDialog dialog = new FileDialog(mainFrame, "Load Edges", FileDialog.LOAD); 
	    dialog.show(); 
	    String filename = dialog.getFile(); 
	    if(filename != null) 
	    { 
		EdgeInfo currentEdgeInfo; 
		TwoKeyHashtable newEdges; 
		try 
		{ 
		    FileInputStream fis = new FileInputStream(filename);
		    GZIPInputStream gis = new GZIPInputStream(fis); 
		    ObjectInputStream in = new ObjectInputStream(gis); 
		    newEdges = (TwoKeyHashtable)in.readObject(); 
		    in.close(); 
		} 
		catch(Exception e){e.printStackTrace(); return;} 
		
		if((proprietaryEdgeInfo == null) || (proprietaryEdgeInfo.isEmpty())) //if there are no Edges yet, just assign the new Edges to the entire vector 
		{ 
		    proprietaryEdgeInfo = newEdges; 
		} 
		else//otherwise take the new Edges and add them to the vector (first eliminating repeat Edges)  ...in the future, we should ask the user which node to keep, in the case of repeat Edges 
		{ 
		    for(Enumeration e = newEdges.elements();e.hasMoreElements();)
		    { 
			currentEdgeInfo = (EdgeInfo)e.nextElement(); 
			proprietaryEdgeInfo.remove(currentEdgeInfo.getSource(),currentEdgeInfo.getDest()); 
			proprietaryEdgeInfo.put(currentEdgeInfo.getSource(),currentEdgeInfo.getDest(), currentEdgeInfo); 
		    } 
		} 				 
	    } 
	} 
	//****---LOAD EDGES---**** 
	//------------------------------------------------------------------------ 
    }	           
    //MENU MANAGER 
    //********************************************************* 
    //********************************************************* 
    //********************************************************* 
 
} 
